package com.alibaba.fastsql.interpreter;

import com.alibaba.fastsql.interpreter.expr.*;
import com.alibaba.fastsql.sql.SQLUtils;
import com.alibaba.fastsql.sql.ast.SQLExpr;
import com.alibaba.fastsql.sql.ast.expr.*;
import com.alibaba.fastsql.util.FnvHash;

import java.util.List;
import java.util.Map;

public class TypeCalculation {
    public static Long calculateLiteralValue(String sql,
                                             Map<String, Long> inputs) {
        SQLExpr sqlExpr = SQLUtils.toSQLExpr(sql);
        Expr<Long> expr = toExpr(sqlExpr, Long.class);

        InterpreterContext context = new MapInterpreterContext(inputs);
        Long result = expr.eval(context);
        if (result == null) {
            return Long.valueOf(0);
        }
        return result;
    }

    private static <T> Expr<T> toExpr(SQLExpr x, Class<T> type) {
        if (x instanceof SQLMethodInvokeExpr) {
            return toFunctionCall((SQLMethodInvokeExpr) x, type);
        }

        if (x instanceof SQLIntegerExpr) {
            Number number = ((SQLIntegerExpr) x).getNumber();
            if (number instanceof Long) {
                return (Expr<T>) new ConstLong((Long) number);
            } else  if (number instanceof Integer || number instanceof Short || number instanceof Byte) {
                return (Expr<T>) new ConstLong(number.longValue());
            }
        }

        if (x instanceof SQLBinaryOpExpr) {
            return toExpr((SQLBinaryOpExpr) x, type);
        }

        if (x instanceof SQLIdentifierExpr) {
            SQLIdentifierExpr ident = (SQLIdentifierExpr) x;
            if (type == Long.class) {
                return (Expr<T>) new FieldLong(ident.normalizedName());
            }
        }

        if (x instanceof SQLNullExpr) {
            return LiteralNull.instance;
        }

        throw new UnsupportedOperationException(x.getClass().getName());
    }

    private static <T> Expr<T> toExpr(SQLBinaryOpExpr x, Class<T> type) {
        Expr<T> left = toExpr(x.getLeft(), type);
        Expr<T> right = toExpr(x.getRight(), type);

        switch (x.getOperator()) {
            case Add:
                if (type == Long.class) {
                    return (Expr<T>) new AddLong((Expr<Long>) left, (Expr<Long>) right);
                }
                break;
            case Subtract:
                if (type == Long.class) {
                    return (Expr<T>) new SubLong((Expr<Long>) left, (Expr<Long>) right);
                }
                break;
            case Multiply:
                if (type == Long.class) {
                    return (Expr<T>) new MultipleLong((Expr<Long>) left, (Expr<Long>) right);
                }
                break;
            case Divide:
                if (type == Long.class) {
                    return (Expr<T>) new DivideLong((Expr<Long>) left, (Expr<Long>) right);
                }
                break;
            case GreaterThan:
                if (type == Long.class) {
                    return (Expr<T>) new GTLong((Expr<Long>) left, (Expr<Long>) right);
                }
                break;
            case GreaterThanOrEqual:
                if (type == Long.class) {
                    return (Expr<T>) new GELong((Expr<Long>) left, (Expr<Long>) right);
                }
                break;
            case LessThan:
                if (type == Long.class) {
                    return (Expr<T>) new LTLong((Expr<Long>) left, (Expr<Long>) right);
                }
                break;
            case LessThanOrEqual:
                if (type == Long.class) {
                    return (Expr<T>) new LELong((Expr<Long>) left, (Expr<Long>) right);
                }
                break;
            case Equality:
                if (type == Long.class) {
                    return (Expr<T>) new EQLong((Expr<Long>) left, (Expr<Long>) right);
                }
                break;
            case LessThanOrGreater:
            case NotEqual:
                if (type == Long.class) {
                    return (Expr<T>) new NELong((Expr<Long>) left, (Expr<Long>) right);
                }
                break;
            default:
                throw new UnsupportedOperationException(x.getOperator().name);
        }

        throw new UnsupportedOperationException(x.getClass().getName());
    }

    private static <T> Expr<T> toFunctionCall(SQLMethodInvokeExpr x, Class<T> type) {
        List<SQLExpr> arguments = x.getArguments();
        Expr<T>[] exprList = new Expr[arguments.size()];
        for (int i = 0; i < arguments.size(); i++) {
            exprList[i] = toExpr(arguments.get(i), type);
        }

        long nameHashCode64 = x.methodNameHashCode64();
        if (nameHashCode64 == FnvHash.Constants.MIN) {
            return (Function<T>) new MinLong((Expr<Long>[]) exprList);
        } else if (nameHashCode64 == FnvHash.Constants.MAX) {
            return (Function<T>) new MaxLong((Expr<Long>[]) exprList);
        } else if (nameHashCode64 == FnvHash.Constants.IF && exprList.length == 3) {
            return (Expr<T>) new IFLong((Expr<Boolean>) exprList[0]
                    , (Expr<Long>) exprList[1]
                    , (Expr<Long>) exprList[2]
            );
        }

        throw new UnsupportedOperationException();
    }

}
