package com.alibaba.fastsql.sql.optimizer;

import com.alibaba.fastsql.DbType;
import com.alibaba.fastsql.sql.SQLUtils;
import com.alibaba.fastsql.sql.ast.SQLStatement;
import com.alibaba.fastsql.sql.ast.statement.SQLBlockStatement;
import com.alibaba.fastsql.sql.ast.statement.SQLExprTableSource;
import com.alibaba.fastsql.sql.ast.statement.SQLSelectStatement;
import com.alibaba.fastsql.sql.ast.statement.SQLWithSubqueryClause;
import com.alibaba.fastsql.sql.optimizer.rules.*;
import com.alibaba.fastsql.sql.parser.SQLParserFeature;
import com.alibaba.fastsql.sql.parser.SQLParserUtils;
import com.alibaba.fastsql.sql.parser.SQLStatementParser;
import com.alibaba.fastsql.sql.repository.SchemaRepository;
import com.alibaba.fastsql.sql.repository.SchemaResolveVisitor;
import com.alibaba.fastsql.sql.visitor.SQLASTVisitor;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.TimeZone;

public class Optimizer {
    static long DEFAULT_FEATURES;

    static {
        DEFAULT_FEATURES |= OptimizerFeature.ConstFolding.mask;
        DEFAULT_FEATURES |= OptimizerFeature.PushUp.mask;
        DEFAULT_FEATURES |= OptimizerFeature.PushDown.mask;
        DEFAULT_FEATURES |= OptimizerFeature.PushDownCount.mask;
        DEFAULT_FEATURES |= OptimizerFeature.MergeUnion.mask;
        DEFAULT_FEATURES |= OptimizerFeature.Ordinal.mask;
        DEFAULT_FEATURES |= OptimizerFeature.TypeInference.mask;

        DEFAULT_FEATURES |= OptimizerFeature.ImplicitJoinRewrite.mask;
        DEFAULT_FEATURES |= OptimizerFeature.ClearSelectHint.mask;
    }

    private DbType           dbType;
    private SchemaRepository repository;
    private TimeZone         timeZone;
    private long             features                      = Optimizer.DEFAULT_FEATURES;
    private boolean          rewriteView                   = false;
    private long             defaultLimit                  = -1;
    private String           defaultSchema                 = null;
    private boolean          constFoldingNowFun            = true;
    private int              distinctEagerAggregationCount = 0;

    public Optimizer(DbType dbType, OptimizerFeature... features) {
        this.dbType = dbType;
        if (features != null) {
            for (OptimizerFeature feature : features) {
                this.features |= feature.mask;
            }
        }
    }

    public Optimizer(DbType dbType, SchemaRepository repository, OptimizerFeature... features) {
        this.dbType = dbType;
        this.repository = repository;
        if (features != null) {
            for (OptimizerFeature feature : features) {
                this.features |= feature.mask;
            }
        }
    }

    public Optimizer(DbType dbType, SchemaRepository repository, TimeZone timeZone, OptimizerFeature... features) {
        this.dbType = dbType;
        this.repository = repository;
        this.timeZone = timeZone;
        if (features != null) {
            for (OptimizerFeature feature : features) {
                this.features |= feature.mask;
            }
        }
    }

    public Optimizer(DbType dbType, SchemaRepository repository, TimeZone timeZone, long defaultFeatures, OptimizerFeature... features) {
        this.dbType = dbType;
        this.repository = repository;
        this.timeZone = timeZone;
        this.features = defaultFeatures;
        if (features != null) {
            for (OptimizerFeature feature : features) {
                this.features |= feature.mask;
            }
        }
    }

    public boolean isClearSelectHint() {
        return (this.features & OptimizerFeature.ClearSelectHint.mask) != 0;
    }

    public void setClearSelectHint(boolean value) {
        if (value) {
            this.features |= OptimizerFeature.ClearSelectHint.mask;
        } else {
            this.features &= ~OptimizerFeature.ClearSelectHint.mask;
        }
    }

    public String optimize(String sql) {
        SQLStatementParser parser = SQLParserUtils.createSQLStatementParser(sql, dbType);
        parser.config(SQLParserFeature.EnableSQLBinaryOpExprGroup, true);

        if (OptimizerFeature.isEnabled(features, OptimizerFeature.PipesAsConcat)) {
            parser.config(SQLParserFeature.PipesAsConcat, true);
        }

        List<SQLStatement> stmtList = new ArrayList<SQLStatement>();
        parser.parseStatementList(stmtList, -1, null);

        boolean adsRewrite = OptimizerFeature.isEnabled(features, OptimizerFeature.ADSRewrite);
        for (int i = 0; i < stmtList.size(); i++) {
            SQLStatement stmt = stmtList.get(i);
            if (repository != null) {
                if (adsRewrite) {
                    repository.resolve(stmt, SchemaResolveVisitor.Option.ResolveAllColumn);
                } else {
                    repository.resolve(stmt);
                }
            }
            optimize(stmt);
        }

        if (adsRewrite) {
            if (stmtList.size() == 1
                    && stmtList.get(0) instanceof SQLBlockStatement
                    && ((SQLBlockStatement) stmtList.get(0)).getStatementList().size() == 1) {
                SQLStatement subStmt = ((SQLBlockStatement) stmtList.get(0)).getStatementList().get(0);
                stmtList.set(0, subStmt);
            }
        }

        return SQLUtils.toSQLString(stmtList, dbType);
    }

    public void optimize(SQLStatement stmt) {
        List<SQLASTVisitor> visitors = createVisitors();
        for (int i = 0; i < visitors.size(); i++) {
            SQLASTVisitor v = visitors.get(i);

            stmt.accept(v);

            if (v instanceof DistinctEagerAggregation) {
                distinctEagerAggregationCount += ((DistinctEagerAggregation) v).getReplaceCount();
            }
            if (v instanceof ViewRewriteWithSchema && stmt instanceof SQLSelectStatement) {
                SQLWithSubqueryClause withSubQuery = ((SQLSelectStatement) stmt).getSelect().getWithSubQuery();

                if (withSubQuery != null && withSubQuery.getEntries().size() > 1) {
                    List<SQLWithSubqueryClause.Entry> entries = withSubQuery.getEntries();

                    redo:
                    for (int sourceIndex = 0; sourceIndex < entries.size(); sourceIndex++) {
                        SQLWithSubqueryClause.Entry source = entries.get(sourceIndex);

                        for (int targetIndex = sourceIndex + 1;
                             targetIndex < entries.size() && targetIndex != sourceIndex;
                             targetIndex++) {

                            SQLWithSubqueryClause.Entry target = entries.get(targetIndex);

                            boolean dependency = dependCTE(source, target);

                            if (dependency) {
                                entries.remove(targetIndex);
                                entries.add(sourceIndex, target);
                                sourceIndex = -1;
                                continue redo;
                            }
                        }
                    }
                }
            }
        }
    }

    private boolean dependCTE(SQLWithSubqueryClause.Entry source, SQLWithSubqueryClause.Entry target) {
        TableSourceExtractor tableSourceExtractor = new TableSourceExtractor();
        source.getSubQuery().getQuery().accept(tableSourceExtractor);

        for (SQLExprTableSource tableSource : tableSourceExtractor.getTableSources()) {
            if (tableSource.getTableName().equalsIgnoreCase(target.getAlias())) {
                return true;
            }
        }

        return false;
    }

    private List<SQLASTVisitor> createVisitors() {

        List<SQLASTVisitor> visitors = new ArrayList<SQLASTVisitor>();

        if ((features & OptimizerFeature.ADSRewrite.mask) != 0) {
            visitors.add(new ADSRewrite.TypeResolve());
        }

        if ((features & OptimizerFeature.ViewRewrite.mask) != 0) {
            ViewRewrite v = new ViewRewrite(repository);

            if ((features & OptimizerFeature.BackQuotedNames.mask) != 0) {
                v.setBackQuotedNames(true);
            }

            visitors.add(v);
        }

        if ((features & OptimizerFeature.ViewRewriteWithSchema.mask) != 0) {
            ViewRewriteWithSchema v = new ViewRewriteWithSchema(repository, defaultSchema);

            if ((features & OptimizerFeature.BackQuotedNames.mask) != 0) {
                v.setBackQuotedNames(true);
            }

            visitors.add(v);
        }

        if ((features & OptimizerFeature.PipesAsConcat.mask) != 0) {
            visitors.add(new PipesAsConcat());
        }

        if ((features & OptimizerFeature.ImplicitJoinRewrite.mask) != 0) {
            visitors.add(new ImplicitJoinRewrite());
        }

        if ((features & OptimizerFeature.TypeInference.mask) != 0) {
            visitors.add(new TypeInference(dbType, timeZone));
        }

        if ((features & OptimizerFeature.ConstFolding.mask) != 0) {
            visitors.add(new ConstFolding(dbType, timeZone, constFoldingNowFun));
        }

        if ((features & OptimizerFeature.MergeUnion.mask) != 0) {
            visitors.add(new MergeUnion());
        }

        if ((features & OptimizerFeature.PushDownCount.mask) != 0) {
            visitors.add(new PushDownCount());
        }

        if ((features & OptimizerFeature.PushDown.mask) != 0) {
            visitors.add(new PushDown());
            visitors.add(new PushDownQueryExpr());
        }

        if ((features & OptimizerFeature.PushUp.mask) != 0) {
            visitors.add(new PushUp());
        }

        if ((features & OptimizerFeature.Ordinal.mask) != 0) {
            visitors.add(new Ordinal());
        }

        if ((features & OptimizerFeature.CastFloatToReal.mask) != 0) {
            visitors.add(new CastFloatToReal());
        }

        boolean mergeLimit = (features & OptimizerFeature.NoLimitMerged.mask) != 0;

        if ((features & OptimizerFeature.ADSRewrite.mask) != 0) {
            ADSRewrite v = new ADSRewrite(repository, timeZone, defaultLimit, mergeLimit);

            if ((features & OptimizerFeature.BackQuotedNames.mask) != 0) {
                v.setBackQuotedNames(true);
            }

            v.setClearSelectHint(isClearSelectHint());
            visitors.add(v);
        }

        if ((features & OptimizerFeature.DistinctEagerAggregation.mask) != 0) {
            visitors.add(new DistinctEagerAggregation());
        }

        if ((features & OptimizerFeature.DecorticateExistsWithAGG.mask) != 0) {
            visitors.add(new DecorticateExistsWithAGG());
        }

        if ((features & OptimizerFeature.MultiDistinctEagerAggregation.mask) != 0) {
            visitors.add(new MultiDistinctEagerAggregation());
        }

        return visitors;
    }

    public String getDefaultSchema() {
        return defaultSchema;
    }

    public void setDefaultSchema(String defaultSchema) {
        this.defaultSchema = defaultSchema;
    }

    public DbType getDbType() {
        return dbType;
    }

    public void setDbType(DbType dbType) {
        this.dbType = dbType;
    }

    public SchemaRepository getRepository() {
        return repository;
    }

    public void setRepository(SchemaRepository repository) {
        this.repository = repository;
    }

    public TimeZone getTimeZone() {
        return timeZone;
    }

    public void setTimeZone(TimeZone timeZone) {
        this.timeZone = timeZone;
    }

    public long getFeatures() {
        return features;
    }

    public void setFeatures(long features) {
        this.features = features;
    }

    public boolean isRewriteView() {
        return (features & OptimizerFeature.ViewRewrite.mask) != 0;
    }

    public void setRewriteView(boolean rewriteView) {
        this.features |= OptimizerFeature.ViewRewrite.mask;
    }

    public long getDefaultLimit() {
        return defaultLimit;
    }

    public void setDefaultLimit(long defaultLimit) {
        this.defaultLimit = defaultLimit;
    }

    public boolean isConstFoldingNowFun() {
        return constFoldingNowFun;
    }

    public void setConstFoldingNowFun(boolean constFoldingNowFun) {
        this.constFoldingNowFun = constFoldingNowFun;
    }

    public int getDistinctEagerAggregationCount() {
        return distinctEagerAggregationCount;
    }
}
