package com.alibaba.fastsql.sql.optimizer.rules;

import com.alibaba.fastsql.sql.SQLUtils;
import com.alibaba.fastsql.sql.ast.*;
import com.alibaba.fastsql.sql.ast.expr.*;
import com.alibaba.fastsql.sql.ast.statement.*;
import com.alibaba.fastsql.sql.dialect.mysql.ast.statement.MySqlSelectQueryBlock;
import com.alibaba.fastsql.sql.dialect.mysql.visitor.MySqlSchemaStatVisitor;
import com.alibaba.fastsql.sql.repository.SchemaObject;
import com.alibaba.fastsql.sql.repository.SchemaRepository;
import com.alibaba.fastsql.sql.visitor.SQLASTVisitorAdapter;
import com.alibaba.fastsql.sql.visitor.functions.IfNull;
import com.alibaba.fastsql.stat.TableStat;
import com.alibaba.fastsql.util.FnvHash;
import com.alibaba.fastsql.util.MySqlUtils;
import com.alibaba.fastsql.util.StringUtils;

import java.math.BigDecimal;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.*;

import java.util.ArrayList;
import java.util.List;
import java.util.TimeZone;
import java.util.TreeSet;

import static com.alibaba.fastsql.sql.ast.expr.SQLBinaryOperator.Add;
import static com.alibaba.fastsql.util.StringUtils.quoteAlias;

public class ADSRewrite extends OptimizerVisitor {
    private final static long INFORMATION_SCHEMA_VIEWS = FnvHash.hashCode64("information_schema.views");

    private SchemaRepository repository;
    private boolean noLimitMerged = false;
    private long defaultLimit = -1L;
    private boolean clearSelectHint = true;
    private boolean backQuotedNames = false;

    public ADSRewrite() {

    }

    public ADSRewrite(TimeZone timeZone) {
        this.timeZone = timeZone;
    }

    public ADSRewrite(SchemaRepository repository, TimeZone timeZone, long defaultLimit, boolean noLimitMerged) {
        this.repository = repository;
        this.timeZone = timeZone;
        this.defaultLimit = defaultLimit;
        this.noLimitMerged = noLimitMerged;
    }

    public boolean isClearSelectHint() {
        return clearSelectHint;
    }

    public void setClearSelectHint(boolean clearSelectHint) {
        this.clearSelectHint = clearSelectHint;
    }

    public void preVisit(SQLObject x) {
        /**
         * clear comments
         */

        final List<String> beforeComments = x.getBeforeCommentsDirect();
        if (beforeComments != null && beforeComments.size() > 0) {
            beforeComments.clear();
        }

        final List<String> afterComments = x.getAfterCommentsDirect();
        if (afterComments != null && afterComments.size() > 0) {
            afterComments.clear();
        }

        if (x instanceof SQLStatement) {
            ((SQLStatement) x).setAfterSemi(false);
        }
    }

    public boolean visit(SQLSelectGroupByClause x) {
        if (x.isWithRollUp() || x.isWithCube()) {
            x.setParen(true);
        }

        SQLObject parent = x.getParent();
        if (parent instanceof SQLSelectQueryBlock) {
            SQLSelectQueryBlock queryBlock = (SQLSelectQueryBlock) parent;
            List<SQLExpr> items = x.getItems();
            for (int i = items.size() - 1; i >= 0; i--) {
                SQLExpr item = items.get(i);
                if (item instanceof SQLCharExpr) {
                    SQLIdentifierExpr ident = new SQLIdentifierExpr(((SQLCharExpr) item).getText());
                    if (queryBlock.findSelectItem(ident.nameHashCode64()) != null) {
                        items.set(i, ident);
                        ident.setParent(x);
                    }
                }
            }
        }

        return true;
    }

    public boolean visit(SQLCurrentTimeExpr x) {
        SQLMethodInvokeExpr funtionCall = new SQLMethodInvokeExpr(x.getType().name);
        if (SQLUtils.replaceInParent(x, funtionCall)) {
            funtionCall.accept(this);
            optimizedCount++;
        }
        return false;
    }

    public boolean visit(SQLSelectQueryBlock x) {
        // select * expand
        if (x.getHintsSize() > 0 && clearSelectHint) {
            x.getHints().clear();
        }

        SQLExpr having = null;
        SQLSelectGroupByClause groupBy = x.getGroupBy();
        if (groupBy != null) {
            having = groupBy.getHaving();
        }

        if (having != null) {
            HavingRewrite havingRewrite = new HavingRewrite(x);
            having.accept(havingRewrite);

        }

        // select a from ( select b from t1 limit 0, 10)
        if (x.getLimit() != null) {
            SQLExpr offset = x.getLimit().getOffset();
            if(offset != null && ((SQLNumericLiteralExpr) offset).getNumber().intValue() == 0) {
                x.getLimit().setOffset(null);
            }
        }

        if (x.getParent().getParent() instanceof SQLStatement && x.getFrom() != null) {
            SQLLimit limit = x.getLimit();
            if (limit == null) {
                if(defaultLimit != -1) {
                    x.setLimit(new SQLLimit(Long.valueOf(defaultLimit).intValue()));
                }
            } else {
                if(!noLimitMerged) {
                    SQLExpr rowCountExpr = limit.getRowCount();
                    SQLExpr offsetExpr = limit.getOffset();

                    if (rowCountExpr instanceof SQLNumericLiteralExpr && offsetExpr instanceof SQLNumericLiteralExpr) {
                        int rowCount = ((SQLNumericLiteralExpr) rowCountExpr).getNumber().intValue()
                                + ((SQLNumericLiteralExpr) offsetExpr).getNumber().intValue();

                        x.setLimit(new SQLLimit(rowCount));
                    }
                }
            }
        }
        return super.visit(x);
    }

    void handLargeInlist(SQLSelectQueryBlock queryBlock, SQLInListExpr inListExpr) {
        if (inListExpr.isNot()) {
            return;
        }

        SQLExpr expr = inListExpr.getExpr();
        SQLName column = null;
        if (expr instanceof SQLName) {
            column = (SQLName) expr;
        } else {
            return;
        }

        List<SQLExpr> list = inListExpr.getTargetList();
        if (list.size() < 500) {
            return;
        }

        TreeSet<SQLExpr> set = new TreeSet<SQLExpr>();
        for (SQLExpr item : list) {
            if (item instanceof SQLValuableExpr) {
                set.add(item);
            } else {
                return;
            }
        }

        SQLValuesTableSource values = new SQLValuesTableSource();

        for (SQLExpr item : set) {
            SQLListExpr row = new SQLListExpr(item);
            row.setParent(values);
            values.getValues().add(row);
        }
        SQLTableSource from = queryBlock.getFrom();
        if (from == null) {
            return;
        }

        String alias = "in_0";
        for (int i = 0; i < 100; ++i) {
            alias = "in_" + i;
            if (from.findTableSource(alias) == null) {
                break;
            }
        }
        values.setAlias(alias);

        SQLIdentifierExpr in_column = new SQLIdentifierExpr(alias + "_col");
        values.addColumn(in_column);

        SQLBinaryOpExpr joinCondition = new SQLBinaryOpExpr(column.clone(), SQLBinaryOperator.Equality, new SQLPropertyExpr(alias, in_column.getName()));

        SQLJoinTableSource in_join = new SQLJoinTableSource(from, SQLJoinTableSource.JoinType.INNER_JOIN, values, joinCondition);

        if (replaceInParent(inListExpr, null)) {
            queryBlock.setFrom(in_join);
        }
    }

    public boolean visit(SQLIntegerExpr x) {
        return visitNum(x);
    }

    public boolean visit(SQLNumberExpr x) {
        return visitNum(x);
    }

    public boolean visitNum(SQLNumericLiteralExpr x) {
        SQLObject parent = x.getParent();

        boolean decimal = false;
        if (parent instanceof SQLBinaryOpExpr) {
            SQLBinaryOpExpr binaryOpExpr = (SQLBinaryOpExpr) parent;
            SQLDataType leftType = binaryOpExpr.getLeft().computeDataType();
            SQLDataType rightType = binaryOpExpr.getRight().computeDataType();
            if (leftType == null || rightType == null) {
                return false;
            }

            decimal = leftType.nameHashCode64() == FnvHash.Constants.DECIMAL
                    || rightType.nameHashCode64() == FnvHash.Constants.DECIMAL;
        } else if (parent instanceof SQLInListExpr) {
            SQLExpr expr = ((SQLInListExpr) parent).getExpr();
            SQLDataType type = expr.computeDataType();
            decimal = type != null && type.nameHashCode64() == FnvHash.Constants.DECIMAL;
        } else if (parent instanceof SQLBetweenExpr) {
            SQLExpr expr = ((SQLBetweenExpr) parent).getTestExpr();
            SQLDataType type = expr.computeDataType();
            decimal = type != null && type.nameHashCode64() == FnvHash.Constants.DECIMAL;
        }

        final SQLDecimalExpr decimalExpr;

        Number number = x.getNumber();
        if (number instanceof Integer) {
            decimalExpr = new SQLDecimalExpr(new BigDecimal(number.intValue()));
        } else if (number instanceof Long) {
            decimalExpr = new SQLDecimalExpr(new BigDecimal(number.longValue()));
        } else if (number instanceof Float) {
            decimalExpr = new SQLDecimalExpr(new BigDecimal(number.floatValue()));
        } else if (number instanceof Double) {
            decimalExpr = new SQLDecimalExpr(new BigDecimal(number.doubleValue()));
        } else if (number instanceof BigDecimal) {
            decimalExpr = new SQLDecimalExpr((BigDecimal) number);
        } else {
            decimalExpr = new SQLDecimalExpr(new BigDecimal(number.toString()));
        }

        if (decimal && SQLUtils.replaceInParent(x, decimalExpr)) {
            optimizedCount++;
        }

        return false;
    }

    public boolean visit(SQLCastExpr x) {
        SQLDataType dataType = x.getDataType();
        if (dataType == null) {
            return true;
        }

        long hashCode64 = dataType.nameHashCode64();
        if (hashCode64 == FnvHash.Constants.CHAR || hashCode64 == FnvHash.Constants.STRING) {
            x.setDataType(new SQLCharacterDataType("VARCHAR"));
        } else if (hashCode64 == FnvHash.Constants.INT) {
            x.setDataType(new SQLCharacterDataType("INTEGER"));
        } else if (hashCode64 == FnvHash.Constants.FLOAT) {
            x.setDataType(new SQLCharacterDataType("DOUBLE"));
        } else if (hashCode64 == FnvHash.Constants.LONG) {
            x.setDataType(new SQLCharacterDataType("BIGINT"));
        } else if (hashCode64 == FnvHash.Constants.NUMERIC || hashCode64 == FnvHash.Constants.NUMBER) {
            SQLDataType dataType1 = dataType.clone();
            dataType1.setName("DECIMAL");
            x.setDataType(dataType1);
        }

        hashCode64 = x.getDataType().nameHashCode64();
        if (hashCode64 == FnvHash.Constants.VARCHAR) {
            SQLExpr expr = x.getExpr();
            if (expr instanceof SQLMethodInvokeExpr) {
                SQLMethodInvokeExpr func = (SQLMethodInvokeExpr) expr;
                if (func.methodNameHashCode64() == FnvHash.Constants.JSON_EXTRACT) {
                    func.setMethodName("JSON_EXTRACT_SCALAR");
                    replaceInParent(x, func);
                }
            }
        }

        return true;
    }

    public boolean visit(SQLIdentifierExpr x) {
        long nameHashCode64 = x.nameHashCode64();
        if (nameHashCode64 == FnvHash.Constants.SYSTIMESTAMP
                || nameHashCode64 == FnvHash.Constants.LOCALTIMESTAMP)
        {
            final SQLMethodInvokeExpr target = new SQLMethodInvokeExpr(x.getName());
            if (SQLUtils.replaceInParent(x, target)) {
                target.accept(this);
                optimizedCount++;
                return false;
            }
        }

        String name = x.getName();
        String name2 = quoteAlias(name);
        if (name != name2) {
            x.setName(name2);
        }
        return true;
    }

    public String quoteAlias(String name) {
        if (backQuotedNames) {
            return StringUtils.backquoteAlias(name);
        }
        return StringUtils.quoteAlias(name);
    }

    public boolean visit(SQLWithSubqueryClause x) {
        List<SQLWithSubqueryClause.Entry> entries = x.getEntries();
        for (SQLWithSubqueryClause.Entry entry : entries) {
            entry.setAlias(SQLUtils.normalize(entry.getAlias()));
            entry.getSubQuery().accept(this);
        }
        return false;
    }

    public boolean visit(SQLPropertyExpr x) {
        String name = x.getName();
        String name2 = quoteAlias(name);
        if (name != name2) {
            x.setName(name2);
        }

        x.getOwner().accept(this);

        if (x.hashCode64() == INFORMATION_SCHEMA_VIEWS) {
            x.setName("view");
        }

        return false;
    }

    public boolean visit(SQLBinaryOpExpr x) {
        if (isGroup(x)) {
            return super.visit(x);
        }
        SQLBinaryOperator operator = x.getOperator();
        if (operator == SQLBinaryOperator.DIV) {
            operator = SQLBinaryOperator.Divide;
            x.setOperator(operator);
        }

        SQLExpr left = x.getLeft();
        SQLExpr right = x.getRight();

        switch (operator) {
            case Equality:
                if (right instanceof SQLNullExpr) {
                    x.setOperator(operator = SQLBinaryOperator.Is);
                } else if (right instanceof SQLCharExpr) {
                    SQLDataType dataType = left.computeDataType();
                    if (dataType != null && dataType.nameHashCode64() == FnvHash.Constants.BOOLEAN) {
                        String chars = ((SQLCharExpr) right).getText();
                        if ("Y".equals(chars)) {
                            x.setRight(right = new SQLBooleanExpr(true));
                        } else if ("N".equals(chars)) {
                            x.setRight(right = new SQLBooleanExpr(false));
                        }
                    }
                }
                break;
            case NotEqual:
            case LessThanOrGreater:
                if (right instanceof SQLNullExpr) {
                    x.setOperator(operator = SQLBinaryOperator.IsNot);
                }
                break;

            case BitwiseAnd:
            case BitwiseOr:
            case BitwiseXor:
            case LeftShift:
            case RightShift:
                SQLDataType leftDataType = left.computeDataType();
                if (leftDataType == null || !leftDataType.isInt()) {
                    left = new SQLCastExpr(left, new SQLDataTypeImpl("BIGINT"));
                    x.setLeft(left);
                }

                SQLDataType rightDataType = right.computeDataType();
                if (rightDataType == null || !rightDataType.isInt()) {
                    right = new SQLCastExpr(right, new SQLDataTypeImpl("BIGINT"));
                    x.setRight(right);
                }
                break;
            case LessThanOrEqualOrGreaterThan: {
                SQLMethodInvokeExpr ifExpr =
                    new SQLMethodInvokeExpr("if"
                            , null
                            , SQLBinaryOpExpr.and(
                                    SQLBinaryOpExpr.isNull(left.clone()),
                                    SQLBinaryOpExpr.isNull(right.clone())
                            )
                            , new SQLBooleanExpr(true)
                            , new SQLMethodInvokeExpr("if"
                                , null
                                , SQLBinaryOpExpr.and(
                                        SQLBinaryOpExpr.isNotNull(left.clone()),
                                        SQLBinaryOpExpr.isNotNull(right.clone())
                                )
                                , SQLBinaryOpExpr.eq(left.clone(), right.clone())
                                , new SQLBooleanExpr(false)
                            )
                    );

                replaceInParent(x, ifExpr);
                ifExpr.accept(this);
                return false;

            }
            default:
                break;
        }

        if (operator == SQLBinaryOperator.BitwiseAnd) {
            final SQLMethodInvokeExpr bitwise_and = new SQLMethodInvokeExpr("BITWISE_AND", null, left.clone(), right.clone());
            if (SQLUtils.replaceInParent(x, bitwise_and)) {
                optimizedCount++;
                bitwise_and.accept(this);
                return false;
            }
        } else if (operator == SQLBinaryOperator.BitwiseOr) {
            final SQLMethodInvokeExpr bitwise_and = new SQLMethodInvokeExpr("BITWISE_OR", null, left.clone(), right.clone());
            if (SQLUtils.replaceInParent(x, bitwise_and)) {
                optimizedCount++;
                bitwise_and.accept(this);
                return false;
            }
        } else if (operator == SQLBinaryOperator.LeftShift) {
            final SQLMethodInvokeExpr bitwise_and = new SQLMethodInvokeExpr("BITWISE_LEFT_SHIFT", null, left.clone(), right.clone());
            if (SQLUtils.replaceInParent(x, bitwise_and)) {
                optimizedCount++;
                bitwise_and.accept(this);
                return false;
            }
        } else if (operator == SQLBinaryOperator.RightShift) {
            final SQLMethodInvokeExpr bitwise_and = new SQLMethodInvokeExpr("BITWISE_RIGHT_SHIFT", null, left.clone(), right.clone());
            if (SQLUtils.replaceInParent(x, bitwise_and)) {
                optimizedCount++;
                bitwise_and.accept(this);
                return false;
            }
        } else if (operator == SQLBinaryOperator.BitwiseXor) {
            final SQLMethodInvokeExpr bitwise_and = new SQLMethodInvokeExpr("BITWISE_XOR", null, left.clone(), right.clone());
            if (SQLUtils.replaceInParent(x, bitwise_and)) {
                optimizedCount++;
                bitwise_and.accept(this);
                return false;
            }

        }

        if (left instanceof SQLBinaryOpExpr) {
            SQLBinaryOpExpr binaryOpExprLeft = (SQLBinaryOpExpr) left;
            if (binaryOpExprLeft.getOperator().isArithmetic()) {
                visit(binaryOpExprLeft);
            } else {
                binaryOpExprLeft.accept(this);
            }
        } else {
            left.accept(this);
        }
        right.accept(this);

        if (left instanceof SQLName && right instanceof SQLIntegerExpr) {
            handleNameAndLiteral((SQLName) left, (SQLIntegerExpr) right);
            return false;
        }

        if (left instanceof SQLMethodInvokeExpr
                && ((SQLMethodInvokeExpr) left).methodNameHashCode64() == FnvHash.Constants.JSON_EXTRACT) {
            if (right instanceof SQLCharExpr) {
                ((SQLMethodInvokeExpr) left).setMethodName("JSON_EXTRACT_SCALAR");
            } else if (right instanceof SQLIntegerExpr) {
                SQLCastExpr cast = new SQLCastExpr(left.clone(), SQLIntegerExpr.DATA_TYPE.clone());
                x.setLeft(cast);
                optimizedCount++;
                return false;
            } else if (right instanceof SQLNumberExpr) {
                SQLCastExpr cast = new SQLCastExpr(left.clone(), SQLNumberExpr.DATA_TYPE_DOUBLE.clone());
                x.setLeft(cast);
                optimizedCount++;
                return false;
            } else if (right instanceof SQLBooleanExpr) {
                SQLCastExpr cast = new SQLCastExpr(left.clone(), SQLBooleanExpr.DATA_TYPE.clone());
                x.setLeft(cast);
                optimizedCount++;
                return false;
            }
        }

        final SQLDataType leftDataType = left.computeDataType();
        final SQLDataType rightDataType = right.computeDataType();

        if (leftDataType == null) {
            return true;
        }

        final long leftTypeHash = leftDataType.nameHashCode64();
        switch (x.getOperator()) {
            case GreaterThan:
            case GreaterThanOrEqual:
            case LessThan:
            case LessThanOrEqual:
            case Equality:
            case NotEqual:
            case LessThanOrGreater:
                if(rightDataType != null) {
                    final long rightTypeHash = rightDataType.nameHashCode64();

                    if (leftTypeHash == FnvHash.Constants.BIGINT && (rightTypeHash == FnvHash.Constants.TIMESTAMP
                            || rightTypeHash == FnvHash.Constants.DATE
                            || rightTypeHash == FnvHash.Constants.DATETIME)) {

                        SQLMethodInvokeExpr methodInvokeExpr = new SQLMethodInvokeExpr("to_unixtime");
                        methodInvokeExpr.getArguments().add(right);
                        SQLUtils.replaceInParent(right, methodInvokeExpr);

                    } else if (rightTypeHash == FnvHash.Constants.BIGINT && (leftTypeHash == FnvHash.Constants.TIMESTAMP
                            || leftTypeHash == FnvHash.Constants.DATE
                            || leftTypeHash == FnvHash.Constants.DATETIME)) {
                        SQLMethodInvokeExpr methodInvokeExpr = new SQLMethodInvokeExpr("to_unixtime");
                        methodInvokeExpr.getArguments().add(left);
                        SQLUtils.replaceInParent(left, methodInvokeExpr);
                    } else if (rightTypeHash == FnvHash.Constants.CHAR
                            && right instanceof SQLCharExpr
                            && leftTypeHash == FnvHash.Constants.TIMESTAMP) {
                            if (ConstFolding.repalceToTimestamp((SQLCharExpr) right, timeZone)) {
                                optimizedCount++;
                            }
                    } else if (rightTypeHash == FnvHash.Constants.DATE && leftTypeHash == FnvHash.Constants.TIMESTAMP) {
                        if(x.getRight() instanceof SQLDateExpr) {
                            String text = ((SQLDateExpr) x.getRight()).getLiteral();
                            Date date = MySqlUtils.parseDate(text, timeZone);
                            if (date != null) {
                                SQLTimestampExpr replaceTo = new SQLTimestampExpr(date, timeZone);
                                SQLUtils.replaceInParent(right, replaceTo);
                            }
                        }
                    } else if (leftTypeHash == FnvHash.Constants.BIGINT
                            && (rightTypeHash == FnvHash.Constants.VARCHAR
                                || rightTypeHash == FnvHash.Constants.CHAR)
                            && x.isBothName()
                    ) {

                        SQLCastExpr try_cast = new SQLCastExpr();
                        try_cast.setExpr(right);
                        try_cast.setDataType(SQLIntegerExpr.DATA_TYPE);
                        try_cast.setTry(true);
                        SQLUtils.replaceInParent(right, try_cast);
                    } else if (leftTypeHash == FnvHash.Constants.DATE && right instanceof SQLCharExpr) {

                        String text = ((SQLCharExpr) right).getText();

                        SQLExpr repalceTo;
                        if (text.indexOf('-') < 0) {
                            SQLCastExpr cast = new SQLCastExpr(right.clone(), new SQLDataTypeImpl("DATE"));
                            cast.setTry(true);
                            repalceTo = cast;
                        } else {
                            repalceTo = new SQLDateExpr(text);
                        }

                        SQLUtils.replaceInParent(right, repalceTo);
                    } else if (leftTypeHash == FnvHash.Constants.TIME && right instanceof SQLCharExpr) {

                        String text = ((SQLCharExpr) right).getText();

                        SQLExpr repalceTo;
                        if (text.indexOf(':') < 0) {
                            SQLCastExpr cast = new SQLCastExpr(right.clone(), new SQLDataTypeImpl("TIME"));
                            cast.setTry(true);
                            repalceTo = cast;
                        } else {
                            repalceTo = new SQLTimeExpr(text);
                        }

                        SQLUtils.replaceInParent(right, repalceTo);
                    } else if (rightTypeHash == FnvHash.Constants.BIGINT
                            && (leftTypeHash == FnvHash.Constants.VARCHAR
                                || leftTypeHash == FnvHash.Constants.CHAR)
                            && x.isBothName()
                    ) {

                        SQLCastExpr try_cast = new SQLCastExpr();
                        try_cast.setExpr(left.clone());
                        try_cast.setDataType(SQLIntegerExpr.DATA_TYPE);
                        try_cast.setTry(true);
                        SQLUtils.replaceInParent(left, try_cast);
                    } else if ((leftTypeHash == FnvHash.Constants.CHAR
                            || leftTypeHash == FnvHash.Constants.VARCHAR)
                            && right instanceof SQLNumericLiteralExpr) {
                        String str = right.toString();
                        SQLCharExpr charExpr = new SQLCharExpr(str);
                        x.setRight(charExpr);
                        optimizedCount++;
                    } else if ((leftTypeHash == FnvHash.Constants.INT
                            || leftTypeHash == FnvHash.Constants.BIGINT)) {
                        if(right instanceof SQLCharExpr) {
                            SQLCastExpr cast = new SQLCastExpr(right.clone(), new SQLDataTypeImpl("BIGINT"));
                            x.setRight(cast);
                            optimizedCount++;
                        }
                    }
                }
                break;
            case Subtract: {
                if(rightDataType != null) {
                    final long rightTypeHash = rightDataType.nameHashCode64();
                    if(rightTypeHash == FnvHash.Constants.DATE) {
                        if (leftTypeHash == FnvHash.Constants.DATE) {
                            SQLMethodInvokeExpr dateDiff = new SQLMethodInvokeExpr("DATE_DIFF", null, new SQLCharExpr("DAY"),
                                                                                   right.clone(), left.clone());
                            replaceInParent(x, dateDiff);
                        } else if (leftTypeHash == FnvHash.Constants.CHAR || leftTypeHash == FnvHash.Constants.VARCHAR) {

                            SQLTextLiteralExpr leftTextLiteral = (SQLTextLiteralExpr) left;

                            SQLMethodInvokeExpr dateDiff = new SQLMethodInvokeExpr("DATE_DIFF", null,
                                                                                   new SQLCharExpr("DAY"),
                                                                                   right.clone(), new SQLDateExpr(
                                    leftTextLiteral.getText()));
                            replaceInParent(x, dateDiff);
                        }
                    }
                }
                break;
            }
            case Divide:
            case Multiply:
            case Modulus:
                if (rightDataType != null) {
                    final long rightTypeHash = rightDataType.nameHashCode64();
                    if (leftTypeHash == FnvHash.Constants.VARCHAR) {
                        if(rightTypeHash == FnvHash.Constants.DOUBLE) {
                            SQLCastExpr castExpr = new SQLCastExpr(left, new SQLDataTypeImpl("DOUBLE"));
                            castExpr.setTry(true);
                            x.setLeft(castExpr);
                        } else if (rightTypeHash == FnvHash.Constants.FLOAT) {
                            SQLCastExpr castExpr = new SQLCastExpr(left, new SQLDataTypeImpl("FLOAT"));
                            castExpr.setTry(true);
                            x.setLeft(castExpr);
                        } else if (rightTypeHash == FnvHash.Constants.INT
                                   || rightTypeHash == FnvHash.Constants.BIGINT
                                   || rightTypeHash == FnvHash.Constants.INTEGER
                                   || rightTypeHash == FnvHash.Constants.TINYINT
                                   || rightTypeHash == FnvHash.Constants.SMALLINT) {
                            SQLCastExpr castExpr = new SQLCastExpr(left, new SQLDataTypeImpl("BIGINT"));
                            castExpr.setTry(true);
                            x.setLeft(castExpr);
                        }
                    } else if (rightTypeHash == FnvHash.Constants.VARCHAR) {
                        if (leftTypeHash == FnvHash.Constants.DOUBLE) {
                            SQLCastExpr castExpr = new SQLCastExpr(right, new SQLDataTypeImpl("DOUBLE"));
                            castExpr.setTry(true);
                            x.setRight(castExpr);
                        } else if (leftTypeHash == FnvHash.Constants.FLOAT) {
                            SQLCastExpr castExpr = new SQLCastExpr(right, new SQLDataTypeImpl("FLOAT"));
                            castExpr.setTry(true);
                            x.setRight(castExpr);
                        } else if (leftTypeHash == FnvHash.Constants.INT || leftTypeHash == FnvHash.Constants.BIGINT
                                   || leftTypeHash == FnvHash.Constants.INTEGER
                                   || leftTypeHash == FnvHash.Constants.TINYINT
                                   || leftTypeHash == FnvHash.Constants.SMALLINT) {
                            SQLCastExpr castExpr = new SQLCastExpr(left, new SQLDataTypeImpl("BIGINT"));
                            castExpr.setTry(true);
                            x.setRight(castExpr);
                        }
                    }
                }
                break;

            case Add: {
                if(rightDataType != null) {
                    if ((leftTypeHash == FnvHash.Constants.DATE || leftTypeHash == FnvHash.Constants.TIMESTAMP)
                            && rightDataType.nameHashCode64() == FnvHash.Constants.INTERVAL
                    ) {
                        SQLExpr deltaExpr = right.clone();
                        if (deltaExpr instanceof SQLBinaryOpExpr) {
                            String unit = null;

                            SQLBinaryOpExpr binaryOpRight = (SQLBinaryOpExpr) deltaExpr;
                            SQLExpr rightLeft = binaryOpRight.getLeft();
                            SQLExpr rightRight = binaryOpRight.getRight();
                            if (rightRight instanceof SQLIntervalExpr) {
                                SQLIntervalExpr intervalExpr = (SQLIntervalExpr) rightRight;
                                SQLIntervalUnit intervalUnit = intervalExpr.getUnit();
                                if (intervalUnit != null) {
                                    SQLExpr value = intervalExpr.getValue();
                                    if (value instanceof SQLCharExpr) {
                                        String str = ((SQLCharExpr) value).getText();
                                        try {
                                            long longValue = Long.parseLong(str);
                                            value = SQLIntegerExpr.ofIntOrLong(longValue);
                                        } catch (NumberFormatException ex) {
                                            break;
                                        }
                                    } else if (value instanceof SQLMethodInvokeExpr
                                            && value.computeDataType() != null
                                            && value.computeDataType().isInt()
                                    ) {
                                        SQLIntervalExpr interval2 = intervalExpr.clone();
                                        SQLBinaryOpExpr intervalValue2 = binaryOpRight.clone();
                                        intervalValue2.setRight(value.clone());
                                        interval2.setValue(intervalValue2);

                                        SQLMethodInvokeExpr dateAdd = new SQLMethodInvokeExpr("DATE_ADD", null, left.clone(), interval2);
                                        replaceInParent(x, dateAdd);
                                        return true;
                                    }
                                    binaryOpRight.setRight(value.clone());
                                    unit = intervalUnit.toString();
                                }
                            }

                            if (unit != null) {
                                SQLMethodInvokeExpr dateAdd = new SQLMethodInvokeExpr("DATE_ADD", null, new SQLCharExpr(unit), deltaExpr, left.clone());
                                if (replaceInParent(x, dateAdd)) {
                                    dateAdd.accept(this);
                                }
                                return true;
                            }
                        } else if (deltaExpr instanceof SQLIntervalExpr) {
                            SQLMethodInvokeExpr dateAdd;
                            SQLIntervalExpr intervalExpr = (SQLIntervalExpr) deltaExpr;
                            if (intervalExpr.getUnit() != null) {
                                String unit = intervalExpr.getUnit().name;
                                dateAdd = new SQLMethodInvokeExpr("DATE_ADD", null, new SQLCharExpr(unit), intervalExpr.getValue().clone(), left.clone());
                            } else {
                                dateAdd = new SQLMethodInvokeExpr("DATE_ADD", null, left.clone(), deltaExpr);
                            }

                            if (replaceInParent(x, dateAdd)) {
                                dateAdd.accept(this);
                            }
                            return true;
                        }

                    }
                }
                break;
            }
            default:
                break;
        }

        return true;
    }

    private void handleNameAndLiteral(SQLName name, SQLLiteralExpr literal) {
        final SQLDataType dataType = name.computeDataType();
        if (dataType == null) {
            return;
        }

        final long hashCode64 = dataType.nameHashCode64();
        if (hashCode64 == FnvHash.Constants.SMALLINT) {
            if (literal instanceof SQLIntegerExpr) {
                if (SQLUtils.replaceInParent(literal
                        , new SQLSmallIntExpr(
                                ((SQLIntegerExpr) literal)
                                        .getNumber()
                                        .shortValue())))
                {
                    optimizedCount++;
                }
            }
        } else  if (hashCode64 == FnvHash.Constants.TINYINT) {
            if (literal instanceof SQLIntegerExpr) {
                if (SQLUtils.replaceInParent(literal
                        , new SQLTinyIntExpr(
                                ((SQLIntegerExpr) literal)
                                        .getNumber()
                                        .byteValue())))
                {
                    optimizedCount++;
                }
            }
        }


    }

    @Override
    public boolean visit(SQLMatchAgainstExpr x) {

        for (SQLExpr column : x.getColumns()) {
            column.accept(this);
        }

        x.getAgainst().accept(this);

        List<SQLExpr> params = new ArrayList<SQLExpr>();
        params.addAll(x.getColumns());
        params.add(x.getAgainst());

        SQLMethodInvokeExpr fun = new SQLMethodInvokeExpr("MATCH_AGAINST", null, params);

        // select * from t1 where match(col1) against('1')  --> select * from t1 where match(col1) against('1') > 0
        final SQLObject parent = x.getParent();

        SQLBinaryOperator operator = null;
        if (parent instanceof SQLBinaryOpExpr) {
            operator = ((SQLBinaryOpExpr) parent).getOperator();
        } else if (parent instanceof SQLBinaryOpExprGroup) {
            operator = ((SQLBinaryOpExprGroup) parent).getOperator();
        }

        SQLExpr replaceTo = null;
        if (parent instanceof SQLSelectQueryBlock || (operator != null && operator.isLogical())) {
            replaceTo = new SQLBinaryOpExpr(fun, SQLBinaryOperator.GreaterThan, new SQLIntegerExpr(0));
        } else {
            replaceTo = fun;
        }

        SQLUtils.replaceInParent(x, replaceTo);
        return false;
    }

    public boolean visit(SQLMethodInvokeExpr x) {
        long hash = x.methodNameHashCode64();

        String methodName = x.getMethodName();
        x.setMethodName(SQLUtils.normalize(methodName));

        List<SQLExpr> arguments = x.getArguments();
        for (SQLExpr argument : arguments) {
            argument.accept(this);
        }

        if (hash == FnvHash.Constants.DATEDIFF) {
            func_datediff(x);
        } else if (hash == FnvHash.Constants.ADD_MONTHS && arguments.size() == 2) {
            func_add_months(x);

        } else if (hash == FnvHash.Constants.DAYOFMONTH && arguments.size() == 1) {
            func_dayofmonth(x);

        } else if (hash == FnvHash.Constants.FROM_UNIXTIME) {
            func_from_unixtime(x);

        } else if (hash == FnvHash.Constants.TO_UNIXTIME) {
            func_to_unixtime(x);

        } else if (hash == FnvHash.Constants.DAYOFYEAR && arguments.size() == 1) {
            func_dayofyear(x);

        } else if (hash == FnvHash.Constants.DAYOFWEEK && arguments.size() == 1) {
            x.setMethodName("DAY_OF_WEEK");

        } else if (hash == FnvHash.Constants.TO_BASE64 && arguments.size() == 1) {
            SQLExpr arg0 = arguments.get(0);
            SQLCastExpr castExpr = new SQLCastExpr(arg0, new SQLDataTypeImpl("VARBINARY"));

            arguments.set(0, castExpr);

            optimizedCount++;

        } else if (hash == FnvHash.Constants.DAYNAME && arguments.size() == 1) {
            x.setMethodName("DATE_FORMAT");

            arguments.add(1, new SQLCharExpr("%W"));

        } else if (hash == FnvHash.Constants.MONTH_BETWEEN && arguments.size() == 2) {
            func_month_between(x);
        } else if (hash == FnvHash.Constants.UNIX_TIMESTAMP) {
            func_unix_timestamp(x);
        } else if ((hash == FnvHash.Constants.HOUR
                    || hash == FnvHash.Constants.MINUTE || hash == FnvHash.Constants.SECOND) && arguments.size() == 1) {
            SQLExpr arg0 = x.getArguments().get(0);
            if (arg0 instanceof SQLCharExpr || arg0 instanceof SQLIdentifierExpr) {
                x.getArguments().set(0, new SQLCastExpr(arg0, new SQLDataTypeImpl("TIME")));
            }

        } else if (hash == FnvHash.Constants.TIMESTAMPADD && arguments.size() == 3) {
            func_timestampadd(x);

        } else if (hash == FnvHash.Constants.WEEK) {
            func_week(x);

        } else if (hash == FnvHash.Constants.DATE_TRUNC && arguments.size() == 2) {
            func_date_trunc(x);

        } else if (hash == FnvHash.Constants.TIMESTAMPDIFF && arguments.size() == 3) {
            func_timestampdiff(x);

        } else if (hash == FnvHash.Constants.MONTHNAME && arguments.size() == 1) {
            func_monthname(x);

        } else if (hash == FnvHash.Constants.DECODE) {
            func_decode(x);

        } else if (hash == FnvHash.Constants.CONCAT) {
            func_concat(x);

        } else if (hash == FnvHash.Constants.WEEKOFYEAR) {
            func_weekofyear(x);

        } else if (hash == FnvHash.Constants.POSITION) {
            func_position(x);

//        } else if (hash == FnvHash.Constants.YEARWEEK || hash == FnvHash.Constants.YEAR_OF_WEEK || hash == FnvHash.Constants.YOW) {
//            func_yearofweek(x);

        } else if (hash == FnvHash.Constants.YEARMONTH && arguments.size() == 1) {
            func_yearmonth(x);

        } else if (hash == FnvHash.Constants.TIMESTAMP) {
            func_timestamp(x);

        } else if (hash == FnvHash.Constants.TO_TIMESTAMP) {
            func_to_timestamp(x);

        } else if (hash == FnvHash.Constants.TO_CHAR) {
            func_to_char(x);

        } else if (hash == FnvHash.Constants.DATE_SUB || hash == FnvHash.Constants.SUBDATE) {
            func_date_sub(x);

        } else if (hash == FnvHash.Constants.DATE_ADD
                || hash == FnvHash.Constants.ADDDATE
                || hash == FnvHash.Constants.DATEADD) {
            if (argsAllConst(x)) { // 特例：莲花池需求，date_add 强制折叠
                func_str_date_add(x);
            } else {
                func_date_add(x);
            }

        } else if (hash == FnvHash.Constants.CURDATE) {
            x.setMethodName("CURDATE");

        } else if (hash == FnvHash.Constants.SYSTIMESTAMP) {
            x.setMethodName("CURRENT_TIMESTAMP");

        } else if (hash == FnvHash.Constants.TIMEDIFF) {
            List<SQLExpr> args = x.getArguments();
            SQLExpr arg0 = args.get(0);
            SQLExpr arg1 = args.get(1);

            if (arg0 instanceof SQLIdentifierExpr) {
                arguments.set(0, new SQLCastExpr(arg0, new SQLCharacterDataType("VARCHAR")));
            }

            if (arg1 instanceof SQLIdentifierExpr) {
                arguments.set(1, new SQLCastExpr(arg1, new SQLCharacterDataType("VARCHAR")));
            }

        } else if (hash == FnvHash.Constants.ADDTIME) {
            func_addtime(x);
        } else if (hash == FnvHash.Constants.NVL2 && arguments.size() == 3) {
            func_nvl2(x);

        } else if (hash == FnvHash.Constants.LOCATE) {
            func_locate(x);

        } else if (hash == FnvHash.Constants.GREATEST) {
            func_greatest(x);

        } else if (hash == FnvHash.Constants.LEAST) {
            func_least(x);

        } else if (hash == FnvHash.Constants.TRIM) {
            func_trim(x);

        } else if (hash == FnvHash.Constants.UDF_SYS_ROWCOUNT) {
            func_udf_sys_rowcount(x);

        } else if (hash == FnvHash.Constants.TRUNCATE) {
            func_truncate(x);

        } else if (hash == FnvHash.Constants.TRUNC) {
            func_trunc(x);

        } else if (hash == FnvHash.Constants.LEFT) {
            x.setMethodName("PRESTO_LEFT");

        } else if (hash == FnvHash.Constants.RIGHT) {
            x.setMethodName("PRESTO_RIGHT");

        } else if (hash == FnvHash.Constants.REGEXP_SUBSTR) {
            x.setMethodName("REGEXP_EXTRACT");

        } else if (hash == FnvHash.Constants.LCASE) {
            x.setMethodName("LOWER");

        } else if (hash == FnvHash.Constants.UCASE) {
            x.setMethodName("UPPER");

        } else if (hash == FnvHash.Constants.SUBSTR || hash == FnvHash.Constants.SUBSTRING) {
            func_substr(x);
        } else if (hash == FnvHash.Constants.LOCALTIMESTAMP) {
            x.setMethodName("LOCALTIMESTAMPFUNC");

        } else if (hash == FnvHash.Constants.SPACE) {
            x.setMethodName("rpad");
            SQLExpr arg0 = x.getArguments().get(0);

            x.getArguments().set(0, new SQLCharExpr(""));
            x.getArguments().add(arg0);
            x.getArguments().add(new SQLCharExpr(" "));

        } else if (hash == FnvHash.Constants.LOCALTIME) {
            x.setMethodName("LOCALTIMEFUNC");

        } else if (hash == FnvHash.Constants.SESSIONTIMEZONE) {
            x.setMethodName("CURRENT_TIMEZONE");

        } else if (hash == FnvHash.Constants.DBTIMEZONE) {
            x.setMethodName("CURRENT_TIMEZONE");

        } else if (hash == FnvHash.Constants.BITAND) {
            x.setMethodName("BITWISE_AND");

        } else if (hash == FnvHash.Constants.CURRENT_TIME) {
            x.setMethodName("CURRENTTIMEFUNC");

        } else if (hash == FnvHash.Constants.CURRENT_DATE) {
            x.setMethodName("curdate");

        } else if (hash == FnvHash.Constants.CURRENT_TIMESTAMP) {
            x.setMethodName("now");

        } else if (hash == FnvHash.Constants.JSON_EXTRACT) {
            func_json_extract(x);

        } else if (hash == FnvHash.Constants.JSON_EXTRACT_SCALAR) {
            func_json_extract_scalar(x);

        } else if (hash == FnvHash.Constants.CHAR_LENGTH || hash == FnvHash.Constants.CHARACTER_LENGTH) {
            x.setMethodName("LENGTH");

        } else if (hash == FnvHash.Constants.CHAR) {
            List<SQLExpr> args = x.getArguments();
            for (int i = 0; i < args.size(); i++) {
                SQLExpr arg = args.get(i);

                if (arg instanceof SQLCharExpr) {
                    String text = ((SQLCharExpr) arg).getText();
                    args.set(i, new SQLIntegerExpr(Integer.valueOf(text)));
                } else if (arg instanceof SQLIdentifierExpr) {
                    SQLCastExpr castExpr = new SQLCastExpr(arg, new SQLDataTypeImpl("BIGINT"));
                    args.set(i, castExpr);
                }
            }
        } else if (hash == FnvHash.Constants.NULLIF) {
            func_nullif(x);
        } else if (hash == FnvHash.Constants.COALESCE
                || hash == FnvHash.Constants.IFNULL
                || hash == FnvHash.Constants.NVL) {
            func_ifnull(x);

        } else if (hash == FnvHash.Constants.ISNULL) {
            func_isnull(x);

        } else if (hash == FnvHash.Constants.CONVERT_TZ) {
            func_convert_tz(x);
        }

        final SQLExpr owner = x.getOwner();
        if (owner != null) {
            owner.accept(this);
        }

        for (SQLExpr arg : x.getArguments()) {
            arg.accept(this);
        }

        return false;
    }


    private void func_substr(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (x.getFrom() != null) {
            if (x.getFor() != null) {
                x.getArguments().add(x.getFrom());
                x.getArguments().add(x.getFor());

                x.setFrom(null);
                x.setFor(null);
                x.setMethodName("SUBSTR");
            } else {
                x.setMethodName("SUBSTRING");
            }
        } else {
            SQLExpr arg0 = x.getArguments().get(0);

            SQLDataType arg0DateType = arg0.computeDataType();
            if(arg0DateType != null) {
                long bash64 = arg0DateType.nameHashCode64();
                if (arg0DateType != null &&
                        (bash64 != FnvHash.Constants.CHAR && bash64 != FnvHash.Constants.VARCHAR)) {
                    SQLCastExpr replace = new SQLCastExpr(arg0.clone(), new SQLDataTypeImpl(SQLDataTypeImpl.Constants.VARCHAR));
                    arguments.set(0, replace);
                }
            }

            x.setMethodName("SUBSTR");
        }
    }

    private boolean argsAllConst(SQLMethodInvokeExpr x) {
        final long hash = x.methodNameHashCode64();

        boolean allConst = true;
        for (int i = 0; i < x.getArguments().size(); i++) {
            SQLExpr arg = x.getArguments().get(i);

            if (arg instanceof SQLValuableExpr) {

            } else if (arg instanceof SQLIntervalExpr && ((SQLIntervalExpr) arg).getValue() instanceof SQLValuableExpr) {

            } else if (hash == FnvHash.Constants.TIMESTAMPDIFF && i == 0 && arg instanceof SQLIdentifierExpr) {
            } else if (hash == FnvHash.Constants.GET_FORMAT && i == 0 && arg instanceof SQLIdentifierExpr) {
            } else if (hash == FnvHash.Constants.IF || hash == FnvHash.Constants.JSON_EXTRACT
                       || hash == FnvHash.Constants.JSON_ARRAY_GET || hash == FnvHash.Constants.DATE_FORMAT) {
                // continue
            } else {
                allConst = false;
            }
        }

        return allConst;
    }

    private void dateAdd(SQLMethodInvokeExpr x, SQLExpr arg0, SQLExpr arg1, boolean negative) {
        String text;
        boolean timestamp = false;
        if (arg0 instanceof SQLCharExpr) {
            text = ((SQLCharExpr) arg0).getText();
            if (text.length() > 10) {
                timestamp = true;
            }
        } else if (arg0 instanceof SQLDateExpr) {
            text = ((SQLDateExpr) arg0).getLiteral();
        } else {
            text = ((SQLTimestampExpr) arg0).getLiteral();
            timestamp = true;
        }

        SQLIntervalUnit unit = null;
        SQLExpr valueExpr = null;
        if (arg1 instanceof SQLIntegerExpr) {
            valueExpr = arg1;
            unit = SQLIntervalUnit.DAY;
        } else {
            SQLIntervalExpr interval = (SQLIntervalExpr) arg1;
            unit = interval.getUnit();
            valueExpr = interval.getValue();
        }

        Date date = MySqlUtils.parseDate(text, timeZone);
        if (date != null && timeZone != null && valueExpr instanceof SQLIntegerExpr && unit != null) {
            int delta = ((SQLIntegerExpr) valueExpr).getNumber().intValue();
            if (negative) {
                delta = -delta;
            }

            Calendar calendar = Calendar.getInstance(timeZone);
            calendar.setTime(date);
            switch (unit) {
                case YEAR:
                    calendar.add(Calendar.YEAR, delta);
                    break;
                case MONTH:
                    calendar.add(Calendar.MONTH, delta);
                    break;
                case DAY:
                    calendar.add(Calendar.DAY_OF_YEAR, delta);
                    break;
                case HOUR:
                    calendar.add(Calendar.HOUR_OF_DAY, delta);
                    break;
                case MINUTE:
                    calendar.add(Calendar.MINUTE, delta);
                    break;
                case SECOND:
                    calendar.add(Calendar.SECOND, delta);
                    break;
                default:
                    return;
            }

            DateFormat tzFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
            tzFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
            tzFormat.setTimeZone(timeZone);

            String result = tzFormat.format(calendar.getTime());

            SQLExpr replaceTo;
            if (result.endsWith(" 00:00:00") && !timestamp) {
                replaceTo = new SQLDateExpr(result.substring(0, result.length() - 9));
            } else {
                replaceTo = new SQLTimestampExpr(result);
            }
            if (replaceInParent(x, replaceTo)) {
                optimizedCount++;
            }

        }
    }

    private void func_str_date_add(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();

        if (arguments.size() == 2 && (arguments.get(0) instanceof SQLCharExpr || (arguments.get(0) instanceof SQLDateExpr)
                                      || arguments.get(0) instanceof SQLTimestampExpr) && (arguments.get(1) instanceof SQLIntervalExpr || arguments.get(1) instanceof SQLIntegerExpr)) {
            dateAdd(x, arguments.get(0), arguments.get(1), false);
            return;
        }

        if (arguments.size() == 3 && arguments.get(0) instanceof SQLDateExpr && arguments.get(1) instanceof SQLIntegerExpr
            && arguments.get(2) instanceof SQLCharExpr) {
            SQLDateExpr dateExpr = (SQLDateExpr) arguments.get(0);
            int delta = ((SQLIntegerExpr) arguments.get(1)).getNumber().intValue();
            String type = ((SQLCharExpr) arguments.get(2)).getText();

            if ("day".equalsIgnoreCase(type)) {
                SQLDateExpr dateCloned = dateExpr.clone();
                if (dateCloned.addDay(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            } else if ("month".equalsIgnoreCase(type)) {
                SQLDateExpr dateCloned = dateExpr.clone();
                if (dateCloned.addMonth(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            }
        } else if (arguments.size() == 3 && arguments.get(2) instanceof SQLDateExpr && arguments.get(1) instanceof SQLIntegerExpr
                   && arguments.get(0) instanceof SQLCharExpr) {
            SQLDateExpr dateExpr = (SQLDateExpr) arguments.get(2);
            int delta = ((SQLIntegerExpr) arguments.get(1)).getNumber().intValue();
            String type = ((SQLCharExpr) arguments.get(0)).getText();

            if ("day".equalsIgnoreCase(type)) {
                SQLDateExpr dateCloned = dateExpr.clone();
                if (dateCloned.addDay(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            } else if ("month".equalsIgnoreCase(type)) {
                SQLDateExpr dateCloned = dateExpr.clone();
                if (dateCloned.addMonth(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            }

        } else if (arguments.size() == 3 && arguments.get(0) instanceof SQLTimestampExpr && arguments.get(1) instanceof SQLIntegerExpr
                   && arguments.get(2) instanceof SQLCharExpr) {
            SQLTimestampExpr dateExpr = (SQLTimestampExpr) arguments.get(0);
            int delta = ((SQLIntegerExpr) arguments.get(1)).getNumber().intValue();
            String type = ((SQLCharExpr) arguments.get(2)).getText();

            if ("day".equalsIgnoreCase(type)) {
                SQLTimestampExpr dateCloned = dateExpr.clone();
                if (dateCloned.addDay(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            } else if ("month".equalsIgnoreCase(type)) {
                SQLTimestampExpr dateCloned = dateExpr.clone();
                if (dateCloned.addMonth(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            }

        } else if (arguments.size() == 3 && arguments.get(2) instanceof SQLTimestampExpr && arguments.get(1) instanceof SQLIntegerExpr
                   && arguments.get(0) instanceof SQLCharExpr) {
            SQLTimestampExpr dateExpr = (SQLTimestampExpr) arguments.get(2);
            int delta = ((SQLIntegerExpr) arguments.get(1)).getNumber().intValue();
            String type = ((SQLCharExpr) arguments.get(0)).getText();

            if ("day".equalsIgnoreCase(type)) {
                SQLTimestampExpr dateCloned = dateExpr.clone();
                if (dateCloned.addDay(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            } else if ("month".equalsIgnoreCase(type)) {
                SQLTimestampExpr dateCloned = dateExpr.clone();
                if (dateCloned.addMonth(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            }

        } else if (arguments.size() == 3 && arguments.get(2) instanceof SQLCharExpr && arguments.get(1) instanceof SQLIntegerExpr
                   && arguments.get(0) instanceof SQLCharExpr) {
            String dateText = ((SQLCharExpr) arguments.get(2)).getText();
            int delta = ((SQLIntegerExpr) arguments.get(1)).getNumber().intValue();
            String type = ((SQLCharExpr) arguments.get(0)).getText();

            Date date = MySqlUtils.parseDate(dateText, timeZone);
            if (date == null) {
                return;
            }

            Calendar calendar = timeZone != null ? Calendar.getInstance(timeZone) : Calendar.getInstance();
            calendar.setTime(date);

            boolean isDate = calendar.get(Calendar.MILLISECOND) == 0 && calendar.get(Calendar.SECOND) == 0
                             && calendar.get(Calendar.MINUTE) == 0 && calendar.get(Calendar.HOUR_OF_DAY) == 0;


            boolean isTimestampType = false;
            if (x.getParent() instanceof SQLBinaryOpExpr) {
                SQLDataType dataType = ((SQLBinaryOpExpr) x.getParent()).getLeft().computeDataType();

                if (dataType != null && dataType.nameHashCode64() == FnvHash.Constants.TIMESTAMP) {
                    isTimestampType = true;
                }
            }

            SQLExpr replaceTo = null;
            if ("day".equalsIgnoreCase(type)) {
                calendar.add(Calendar.DAY_OF_MONTH, delta);
                if(isTimestampType) {
                    replaceTo = new SQLTimestampExpr(calendar.getTime(), timeZone);
                } else if (isDate) {
                    replaceTo = new SQLDateExpr(calendar.getTime(), timeZone);
                } else {
                    replaceTo = new SQLTimestampExpr(calendar.getTime(), timeZone);
                }
            } else if ("month".equalsIgnoreCase(type)) {
                calendar.add(Calendar.MONTH, delta);
                if (isTimestampType) {
                    replaceTo = new SQLTimestampExpr(calendar.getTime(), timeZone);
                } else if (isDate) {
                    replaceTo = new SQLDateExpr(calendar.getTime(), timeZone);
                } else {
                    replaceTo = new SQLTimestampExpr(calendar.getTime(), timeZone);
                }
            } else {
                func_date_add(x);
            }

            if (replaceTo != null && replaceInParent(x, replaceTo)) {
                optimizedCount++;
            }
        }
    }

    private void func_truncate(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 2) {
            return;
        }
        SQLExpr arg0 = arguments.get(0);

        if (arg0 instanceof SQLCastExpr) {
            return;
        }

        if (arg0 instanceof SQLDecimalExpr) {
            return;
        }

        if (arg0 instanceof SQLNumericLiteralExpr) {
            SQLDecimalExpr decimalExpr = new SQLDecimalExpr(arg0.toString());
            arguments.set(0, decimalExpr);
        }
    }

    private void func_date_trunc(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        SQLExpr arg1 = arguments.get(1);

    }

    private void func_week(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 1) {
            SQLExpr arg0 = arguments.get(0);
            SQLDataType arg0Type = arg0.computeDataType();

            if (arg0Type != null && arg0Type.isInt()) {
                SQLCastExpr charCast = new SQLCastExpr(arg0.clone(), SQLCharExpr.DATA_TYPE.clone());
                SQLCastExpr cast = new SQLCastExpr(charCast, SQLDateExpr.DATA_TYPE.clone());
                x.setArgument(0, cast);
                optimizedCount++;
            }

            arguments.add(new SQLIntegerExpr(0));
        } else if (arguments.size() == 2) {
            SQLExpr arg0 = arguments.get(0);
            SQLDataType arg0Type = arg0.computeDataType();
            if (arg0Type != null && arg0Type.isString()) {
                SQLCastExpr cast = new SQLCastExpr(arg0.clone(), SQLDateExpr.DATA_TYPE.clone());
                x.setArgument(0, cast);
                optimizedCount++;
            }
        }
    }

    private void func_position(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 2) {
            x.setMethodName("STRPOS");
        } else if (arguments.size() == 3) {
            SQLMethodInvokeExpr subStrFun = new SQLMethodInvokeExpr("SUBSTR", null, arguments.get(1), arguments.get(2));

            SQLMethodInvokeExpr posFun = new SQLMethodInvokeExpr("STRPOS", null, subStrFun, arguments.get(0));
            if (SQLUtils.replaceInParent(x, posFun)) {
                optimizedCount++;
            }
        }
    }

    private void func_group_concat(SQLMethodInvokeExpr x) {
        if (x.getArguments().size() > 0) {
            SQLExpr arg0 = x.getArguments().get(0);
            if (!(arg0 instanceof SQLCastExpr) && !(arg0 instanceof SQLCharExpr)) {
                SQLCastExpr cast = new SQLCastExpr(arg0.clone(), new SQLCharacterDataType("VARCHAR"));
                x.setArgument(0, cast);
            }
        }

        if (x.getArguments().size() > 1) {
            SQLMethodInvokeExpr concat = new SQLMethodInvokeExpr("CONCAT", null, x.getArguments());

            x.getArguments().clear();
            x.getArguments().add(concat);

            optimizedCount++;
        }
    }

    private void func_convert_tz(SQLMethodInvokeExpr x) {
        if (x.getArguments().size() == 3) {
            SQLExpr arg0 = x.getArguments().get(0);
            if (!(arg0 instanceof SQLCastExpr) && !(arg0 instanceof SQLCharExpr)) {
                SQLCastExpr cast = new SQLCastExpr(arg0.clone(), new SQLCharacterDataType("VARCHAR"));
                x.setArgument(0, cast);
            }
        }
    }

    private void func_concat(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        final int size = arguments.size();

        if (size == 1) {
            SQLExpr arg = arguments.get(0);

            if (arg instanceof SQLCastExpr) {

            } else {
                SQLExpr replaceTo = null;
                SQLDataType dataType = arg.computeDataType();
                if (dataType == null) {
                    arguments.set(0, new SQLCastExpr(arg, new SQLDataTypeImpl("VARCHAR")));
                } else if (dataType != null
                        && (dataType.nameHashCode64() != FnvHash.Constants.CHAR
                        && dataType.nameHashCode64() != FnvHash.Constants.VARCHAR))
                {
                    arguments.set(0, new SQLCastExpr(arg, new SQLDataTypeImpl("VARCHAR")));
                }
            }

            arguments.add(new SQLCharExpr(""));

        } else {
            for (int i = 0; i < arguments.size(); i++) {
                SQLExpr arg = arguments.get(i);

                if (arg instanceof SQLCastExpr) {
                    SQLExpr c0 = ((SQLCastExpr) arg).getExpr();
                    if (c0 instanceof SQLIdentifierExpr) {
                        SQLDataType sqlDataType = c0.computeDataType();
                        if (sqlDataType != null && sqlDataType.isString()) {
                            replaceInParent(arg, c0);
                        }
                    }
                    continue;
                }

                if (arg instanceof SQLMethodInvokeExpr) {
                    SQLMethodInvokeExpr func = (SQLMethodInvokeExpr) arg;
                    if (func.methodNameHashCode64() == FnvHash.Constants.JSON_EXTRACT) {
                        func.setMethodName("JSON_EXTRACT_SCALAR");
                        continue;
                    }
                }

                SQLDataType dataType = arg.computeDataType();
                if (dataType == null) {
                    arguments.set(i, new SQLCastExpr(arg, new SQLDataTypeImpl("VARCHAR")));
                } else if (dataType != null && (dataType.nameHashCode64() != FnvHash.Constants.CHAR
                                         && dataType.nameHashCode64() != FnvHash.Constants.VARCHAR)) {
                    arguments.set(i, new SQLCastExpr(arg, new SQLDataTypeImpl("VARCHAR")));
                }
            }
        }
    }

    private void func_decode(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        final int size = arguments.size();
        if (size < 3) {
            return;
        }

        final SQLCaseExpr caseExpr = new SQLCaseExpr();
        caseExpr.setValueExpr(arguments.get(0).clone());

        int end = size % 2 == 1 ? size : size - 1;
        for (int i = 1; i < end; i+= 2) {
            caseExpr.addItem(arguments.get(i), arguments.get(i + 1));
        }
        if (end != size) {
            caseExpr.setElseExpr(arguments.get(end));
        }
        if (SQLUtils.replaceInParent(x, caseExpr)) {
            optimizedCount++;
        }
    }

    private void func_monthname(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        SQLExpr arg0 = arguments.get(0);

        SQLMethodInvokeExpr replaceTo = new SQLMethodInvokeExpr("DATE_FORMAT", null, arg0, new SQLCharExpr("%M"));
        if (SQLUtils.replaceInParent(x, replaceTo)) {
            optimizedCount++;
        }
    }

    private void func_yearmonth(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();

        SQLExpr arg0 = arguments.get(0);
        if(!(arg0 instanceof SQLCastExpr)) {
            arg0 = new SQLCastExpr(arg0, new SQLDataTypeImpl("TIMESTAMP"));
        }

        SQLMethodInvokeExpr replaceTo = new SQLMethodInvokeExpr("DATE_FORMAT", null, arg0, new SQLCharExpr("%Y-%m"));
        if (SQLUtils.replaceInParent(x, replaceTo)) {
            optimizedCount++;
        }
    }

    private void func_timestampdiff(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);
        SQLExpr arg2 = arguments.get(2);

        if (arg0 instanceof SQLIdentifierExpr) {
            arg0 = new SQLCharExpr(((SQLIdentifierExpr) arg0).getName());
        }

        if (arg1 instanceof SQLIdentifierExpr) {
            arg1 = new SQLCastExpr(arg1, new SQLDataTypeImpl("TIMESTAMP"));
        } else if (arg1 instanceof SQLCharExpr) {
            String text = ((SQLCharExpr) arg1).getText();
            if(text.length() == 8 && text.charAt(2) == ':' && text.charAt(5) == ':') {
                arg1 = new SQLTimeExpr(((SQLCharExpr) arg1).getText());
            } else {
                arg1 = new SQLTimestampExpr(((SQLCharExpr) arg1).getText());
            }
        }

        if (arg2 instanceof SQLIdentifierExpr) {
            arg2 = new SQLCastExpr(arg2, new SQLDataTypeImpl("TIMESTAMP"));
        } else if (arg2 instanceof SQLCharExpr) {
            String text = ((SQLCharExpr) arg2).getText();
            if (text.length() == 8 && text.charAt(2) == ':' && text.charAt(5) == ':') {
                arg2 = new SQLTimeExpr(((SQLCharExpr) arg2).getText());
            } else {
                arg2 = new SQLTimestampExpr(((SQLCharExpr) arg2).getText());
            }
        }

        SQLMethodInvokeExpr replaceTo = new SQLMethodInvokeExpr("DATE_DIFF", null, arg0, arg1, arg2);
        if (SQLUtils.replaceInParent(x, replaceTo)) {
            optimizedCount++;
        }
    }

    private void func_timestampadd(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);
        SQLExpr arg2 = arguments.get(2);

        if (arg0 instanceof SQLIdentifierExpr) {
            arg0 = new SQLCharExpr(((SQLIdentifierExpr) arg0).getName());
        }

        if (arg2 instanceof SQLCharExpr) {
            String text = ((SQLCharExpr) arg2).getText();
            if (text.length() == 10) {
                arg2 = new SQLDateExpr(text);
            } else if (text.length() == 19) {
                arg2 = new SQLTimestampExpr(text);
            }
        }

        SQLMethodInvokeExpr replaceTo = new SQLMethodInvokeExpr("DATE_ADD", null, arg0, arg1, arg2);
        if (SQLUtils.replaceInParent(x, replaceTo)) {
            optimizedCount++;
        }
    }

    private void func_month_between(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);

        SQLMethodInvokeExpr replaceTo = new SQLMethodInvokeExpr("DATE_DIFF", null, new SQLCharExpr("MONTH"), arg0, arg1);
        if (SQLUtils.replaceInParent(x, replaceTo)) {
            optimizedCount++;
        }
    }

    private void func_unix_timestamp(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 1) {
            return;
        }
        SQLExpr arg0 = arguments.get(0);

        SQLDataType sqlDataType = arg0.computeDataType();
        if (sqlDataType != null && sqlDataType.isString()) {
            arguments.set(0, new SQLCastExpr(arg0, new SQLDataTypeImpl("TIMESTAMP")));
        }
    }

    private void func_add_months(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);

        SQLDataType dataType = arg0.computeDataType();
        if (dataType != null
                && (dataType.nameHashCode64() == FnvHash.Constants.CHAR || dataType.nameHashCode64() == FnvHash.Constants.VARCHAR)) {
            arg0 = new SQLCastExpr(arg0, new SQLDataTypeImpl("DATE"));
        }

        SQLMethodInvokeExpr replaceTo = new SQLMethodInvokeExpr("DATE_ADD", null, new SQLCharExpr("MONTH"), arg1, arg0);
        if (SQLUtils.replaceInParent(x, replaceTo)) {
            optimizedCount++;
        }
    }

    private void func_from_unixtime(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();

        if (arguments.size() != 2) {
            return;
        }

        SQLExpr arg1 = arguments.get(1);
        if (arg1 instanceof SQLCharExpr) {
            String format = ((SQLCharExpr) arg1).getText();
            if (format.indexOf('%') >= 0) {
                SimpleDateFormat javaFormat = (SimpleDateFormat) MySqlUtils.toJavaFormat(format);
                String pattern = javaFormat.toPattern();
                SQLCharExpr patternExpr = new SQLCharExpr(pattern);
                x.setArgument(1, patternExpr);
            }
        }
    }

    private void func_to_unixtime(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 1) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        if (arg0 instanceof SQLBinaryOpExpr) {
            SQLBinaryOpExpr binaryOpExpr = (SQLBinaryOpExpr) arg0;
            if (!binaryOpExpr.getOperator().isArithmetic()) {
                return;
            }
            SQLDataType leftDataType = binaryOpExpr.getLeft().computeDataType();
            SQLDataType rightDataType = binaryOpExpr.getRight().computeDataType();

            if (leftDataType != null && leftDataType.nameHashCode64() == FnvHash.Constants.TIMESTAMP
                    && rightDataType != null && rightDataType.nameHashCode64() == FnvHash.Constants.TIMESTAMP) {
                replaceInParent(x, arg0.clone());
            }
        }
    }

    private void func_trim(SQLMethodInvokeExpr x) {
        if (x.getFrom() != null) {
            x.getArguments().add(0, x.getFrom());
            x.setFrom(null);
            x.setTrimOption(null);
        }
    }

    private void func_greatest(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 0) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);

        SQLDataType arg0DataType = arg0.computeDataType();
        if (arg0DataType != null) {

            boolean isDateType = false;
            SQLDataTypeImpl castType = null;

            if (arg0DataType.nameHashCode64() == FnvHash.Constants.TIMESTAMP) {
                isDateType = true;
                castType = new SQLDataTypeImpl(SQLDataType.Constants.TIMESTAMP);
            } else if (arg0DataType.nameHashCode64() == FnvHash.Constants.DATE) {
                isDateType = true;
                castType = new SQLDataTypeImpl(SQLDataType.Constants.DATE);
            }

            if (isDateType) {
                for (int i = 1; i < arguments.size(); i++) {
                    SQLCastExpr sqlCastExpr = new SQLCastExpr(arguments.get(i), castType);
                    arguments.set(i, sqlCastExpr);
                }
            }
        }
    }

    private void func_least(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 0) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);

        SQLDataType arg0DataType = arg0.computeDataType();
        if (arg0DataType != null) {

            boolean isDateType = false;
            SQLDataTypeImpl castType = null;

            if (arg0DataType.nameHashCode64() == FnvHash.Constants.TIMESTAMP) {
                isDateType = true;
                castType = new SQLDataTypeImpl(SQLDataType.Constants.TIMESTAMP);
            } else if (arg0DataType.nameHashCode64() == FnvHash.Constants.DATE) {
                isDateType = true;
                castType = new SQLDataTypeImpl(SQLDataType.Constants.DATE);
            }

            if (isDateType) {
                for (int i = 1; i < arguments.size(); i++) {
                    SQLCastExpr sqlCastExpr = new SQLCastExpr(arguments.get(i), castType);
                    arguments.set(i, sqlCastExpr);
                }
            }
        }
    }

    private void func_locate(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 2 && arguments.size() != 3) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);

        if (arguments.size() == 2) {
            SQLMethodInvokeExpr strpos
                    = new SQLMethodInvokeExpr("STRPOS", null, arg1.clone(), arg0.clone());

            if (SQLUtils.replaceInParent(x, strpos)) {
                optimizedCount++;
            }
            return;
        }

        SQLExpr arg2 = arguments.get(2);

        SQLMethodInvokeExpr substr
                = new SQLMethodInvokeExpr("SUBSTR", null
                    , arg1.clone()
                    , arg2.clone());

        SQLMethodInvokeExpr strpos
                = new SQLMethodInvokeExpr("STRPOS", null, substr, arg0.clone());

        SQLExpr delta;
        SQLBinaryOperator op = Add;
        if (arg2 instanceof SQLIntegerExpr) {
            int deltaVal = ((SQLIntegerExpr) arg2).getNumber().intValue() - 1;
            if (deltaVal < 0) {
                deltaVal = -deltaVal;
                op = SQLBinaryOperator.Subtract;
            }
            delta = new SQLIntegerExpr(deltaVal);
        } else {
            delta = new SQLBinaryOpExpr(arg2.clone(), SQLBinaryOperator.Subtract, new SQLIntegerExpr(1));
        }

        SQLBinaryOpExpr gt = new SQLBinaryOpExpr(strpos, SQLBinaryOperator.GreaterThan, new SQLIntegerExpr(0), dbType);

        SQLExpr replaceTo = new SQLMethodInvokeExpr("IF", null, gt
                , new SQLBinaryOpExpr(strpos, op, delta, dbType), new SQLIntegerExpr(0));

        if (SQLUtils.replaceInParent(x, replaceTo)) {
            optimizedCount++;
        }
    }

    private void func_json_extract(SQLMethodInvokeExpr x) {

    }

    private void func_json_extract_scalar(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 2) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);

        if (arg1 instanceof SQLCharExpr) {
            String path = ((SQLCharExpr) arg1).getText();
            if (path.startsWith("$.")
                    && arg0 instanceof SQLMethodInvokeExpr
                    && ((SQLMethodInvokeExpr) arg0).methodNameHashCode64() == FnvHash.Constants.JSON_EXTRACT) {
                SQLMethodInvokeExpr json_extract = (SQLMethodInvokeExpr) arg0;
                if (json_extract.getArguments().size() == 2
                        && json_extract.getArguments().get(1) instanceof SQLCharExpr)
                {
                    String path2 = ((SQLCharExpr) json_extract.getArguments().get(1)).getText();
                    SQLExpr r_arg0 = json_extract.getArguments().get(0).clone();
                    String r_path = path2 + path.substring(1);
                    SQLExpr r_arg1 = new SQLCharExpr(r_path);
                    x.setArgument(0, r_arg0);
                    x.setArgument(1, r_arg1);
                }
            }
        }
    }

    private void func_addtime(SQLMethodInvokeExpr x) {
        List<SQLExpr> args = x.getArguments();
        SQLExpr arg0 = args.get(0);

        if (arg0 instanceof SQLName) {
            x.setArgument(0, new SQLCastExpr(arg0.clone(), new SQLCharacterDataType("VARCHAR")));
        }
    }

    private void func_nvl2(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 3) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);
        SQLExpr arg2 = arguments.get(2);

        SQLMethodInvokeExpr ifExpr = new SQLMethodInvokeExpr("IF", null, SQLBinaryOpExpr.isNull(arg0.clone()), arg2.clone(), arg1.clone());

        if (SQLUtils.replaceInParent(x, ifExpr)) {
            optimizedCount++;
        }
    }

    private void func_nullif(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 2) {
            SQLExpr arg0 = arguments.get(0);
            SQLExpr arg1 = arguments.get(1);

            if (arg0 instanceof SQLName && arg1 instanceof SQLLiteralExpr) {
                handleNameLiteral(x, arg0, (SQLLiteralExpr) arg1);
            } else if(arg1 instanceof SQLName && arg0 instanceof SQLLiteralExpr){
                handleNameLiteral(x, arg1, (SQLLiteralExpr) arg0);
            }
            return;
        }
    }


    private void func_isnull(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 1) {
            SQLBinaryOpExpr isNullExpr = SQLBinaryOpExpr.isNull(arguments.get(0).clone());
            replaceInParent(x, isNullExpr);
            isNullExpr.accept(this);
            optimizedCount++;
            return;
        } else if(arguments.size() == 2){
            x.setMethodName("COALESCE");
            func_ifnullCast(x);
            optimizedCount++;
            return;
        } else {
            x.setMethodName("COALESCE");
            optimizedCount++;
        }
    }

    private void func_ifnull(SQLMethodInvokeExpr x) {
        x.setMethodName("COALESCE");
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 2) {
            return;
        }
        func_ifnullCast(x);
        optimizedCount++;
    }

    private void func_ifnullCast(SQLMethodInvokeExpr x) {
        List<SQLExpr> args = x.getArguments();

        if (args.size() == 2) {

            SQLExpr arg0 = args.get(0);
            SQLExpr arg1 = args.get(1);

            if (arg0 instanceof SQLValuableExpr && arg1 instanceof SQLValuableExpr) {
                Object val = IfNull.instance.eval(x);
                if (val instanceof Integer) {
                    if (replaceInParent(x, new SQLIntegerExpr((Integer) val))) {
                        optimizedCount++;
                    }
                    return;
                }
                if (val instanceof String) {
                    if (replaceInParent(x, new SQLCharExpr((String) val))) {
                        optimizedCount++;
                    }
                    return;
                }

            } else {
                SQLDataType arg0DateType = arg0.computeDataType();
                SQLDataType arg1DateType = arg1.computeDataType();

                if (arg0 instanceof SQLName && arg0DateType != null && (arg0DateType.isInt() || arg0DateType.isNumberic())
                        && (arg1 instanceof SQLLiteralExpr && arg1DateType != null && (arg1DateType.isInt() || arg1DateType.isNumberic()))) {
                    return;
                }

                if (arg0DateType == null && arg1DateType != null) {
                    if(arg1DateType.nameHashCode64() == FnvHash.Constants.NUMBER
                            || arg1DateType.nameHashCode64() == FnvHash.Constants.NUMERIC) {
                        SQLCastExpr castExpr = new SQLCastExpr(arg0, new SQLDataTypeImpl("DOUBLE"));
                        castExpr.setTry(true);
                        args.set(0, castExpr);
                    } else {
                        SQLCastExpr castExpr = new SQLCastExpr(arg0, new SQLDataTypeImpl(arg1DateType.getName()));
                        castExpr.setTry(true);
                        args.set(0, castExpr);
                    }
                    return;
                } else if (arg1DateType != null) {
                    long arg0Hash = arg0DateType.nameHashCode64();
                    long arg1Hash = arg1DateType.nameHashCode64();

                    if ((arg0Hash == FnvHash.Constants.VARCHAR
                            || arg0Hash == FnvHash.Constants.CHAR)
                            && (arg1Hash == FnvHash.Constants.VARCHAR
                            || arg1Hash == FnvHash.Constants.CHAR)) {
                        return;
                    }

                    if ((arg0Hash == FnvHash.Constants.INT
                            || arg0Hash == FnvHash.Constants.BIGINT
                            || arg0Hash == FnvHash.Constants.INTEGER)
                            && (arg1Hash == FnvHash.Constants.INT
                            || arg1Hash == FnvHash.Constants.BIGINT
                            || arg1Hash == FnvHash.Constants.INTEGER)) {
                        return;
                    }

                    if (arg0 instanceof SQLAggregateExpr
                            && (arg1DateType.isInt()
                            || arg1Hash == FnvHash.Constants.DOUBLE
                            || arg1Hash == FnvHash.Constants.FLOAT
                            || arg1Hash == FnvHash.Constants.DECIMAL)) {
                        return;
                    }

                    if (x.getParent() instanceof SQLMethodInvokeExpr) {
                        if ((x.getParent() instanceof SQLAggregateExpr)
                                && arg1 instanceof SQLNumericLiteralExpr
                                && !(arg0DateType.isInt()
                                || arg0Hash == FnvHash.Constants.DOUBLE
                                || arg0Hash == FnvHash.Constants.FLOAT
                                || arg0Hash == FnvHash.Constants.DECIMAL
                                )) {
                            SQLCastExpr castExpr = new SQLCastExpr(arg0, new SQLDataTypeImpl("BIGINT"));
                            castExpr.setTry(true);
                            args.set(0, castExpr);
                            return;
                        }
                    }

                    if(arg1Hash == FnvHash.Constants.VARCHAR
                            || arg1Hash == FnvHash.Constants.CHAR
                            || arg1Hash == FnvHash.Constants.TEXT
                            || arg1Hash == FnvHash.Constants.CLOB) {
                        SQLCastExpr castExpr = new SQLCastExpr(arg0, new SQLDataTypeImpl("VARCHAR"));
                        castExpr.setTry(true);
                        args.set(0, castExpr);
                    } else if ((arg1DateType.isInt())
                            && (arg0Hash == FnvHash.Constants.VARCHAR
                            || arg0Hash == FnvHash.Constants.CHAR
                            || arg0Hash == FnvHash.Constants.STRING)) {
                        SQLCastExpr castExpr = new SQLCastExpr(arg1, new SQLDataTypeImpl("VARCHAR"));
                        args.set(1, castExpr);
                    } else if ((arg1DateType.isInt())
                            && (arg0Hash == FnvHash.Constants.NUMBER
                            || arg0Hash == FnvHash.Constants.FLOAT
                            || arg0Hash == FnvHash.Constants.DOUBLE
                            || arg0Hash == FnvHash.Constants.DECIMAL)) {
                        SQLCastExpr castExpr = new SQLCastExpr(arg1, new SQLDataTypeImpl("DOUBLE"));
                        args.set(1, castExpr);
                    } else {
                        SQLCastExpr castExpr;
                        if (arg1DateType.isInt()) {
                            castExpr = new SQLCastExpr(arg0, new SQLDataTypeImpl(SQLDataType.Constants.BIGINT));
                        } else if (arg1Hash == FnvHash.Constants.NUMBER
                                || arg1Hash == FnvHash.Constants.FLOAT
                                || arg1Hash == FnvHash.Constants.DOUBLE
                                || arg1Hash == FnvHash.Constants.DECIMAL) {
                            castExpr = new SQLCastExpr(arg0, new SQLDataTypeImpl("DOUBLE"));
                        } else if (arg1Hash == FnvHash.Constants.CHAR) {
                            castExpr = new SQLCastExpr(arg0, new SQLDataTypeImpl("VARCHAR"));
                        } else {
                            castExpr = new SQLCastExpr(arg0, new SQLDataTypeImpl(arg1DateType.getName()));
                        }
                        castExpr.setTry(true);
                        args.set(0, castExpr);
                    }

                }
            }
        }
    }

    private void func_to_char(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 2) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);

        x.setMethodName("DATE_FORMAT");
        SQLDataType type0 = arg0.computeDataType();
        if (type0 == null) {
            return;
        }

        long hash = type0.nameHashCode64();
        if (hash == FnvHash.Constants.CHAR || hash == FnvHash.Constants.VARCHAR) {
            SQLExpr replaceTo = null;
            if (arg0 instanceof SQLCharExpr) {
                String text = ((SQLCharExpr) arg0).getText();
                replaceTo = new SQLTimestampExpr(text);
            } else {
                replaceTo = new SQLCastExpr(arg0.clone(), new SQLDataTypeImpl("TIMESTAMP"));
            }
            if (replaceTo != null) {
                arguments.set(0, replaceTo);
            }
        }
    }

    private void func_to_timestamp(SQLMethodInvokeExpr x) {
        x.setMethodName("DATE_PARSE");
        optimizedCount++;
    }

    private void func_timestamp(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 2) {
            return;
        }

        SQLExpr arg1 = arguments.get(1);
        if (arg1 instanceof SQLCharExpr) {
            SQLCharExpr charExpr = (SQLCharExpr) arg1;
            String fmt = charExpr.getText();
            if ("yyyy-MM-dd HH:mm:ss".equals(fmt)) {
                charExpr.setText("%Y-%m-%d %k:%i:%s");
            }
        }
        optimizedCount++;
    }

    private void func_weekofyear(SQLMethodInvokeExpr x) {
        x.setMethodName("WEEK");

        final List<SQLExpr> arguments = x.getArguments();

        if(arguments.size() == 1) {
            arguments.add(1, new SQLIntegerExpr(3));
            optimizedCount++;
        }

    }

    private void func_dayofmonth(SQLMethodInvokeExpr x) {
        x.setMethodName("DAY_OF_MONTH");
        optimizedCount++;
    }

    private void func_dayofyear(SQLMethodInvokeExpr x) {
        x.setMethodName("DAY_OF_YEAR");

        SQLExpr arg0 = x.getArguments().get(0);

        if (arg0 instanceof SQLIdentifierExpr) {
            x.getArguments().set(0, new SQLCastExpr(arg0, new SQLDataTypeImpl("TIMESTAMP")));
        } else if (arg0 instanceof SQLCharExpr) {
            x.getArguments().set(0, new SQLTimestampExpr(((SQLCharExpr) arg0).getText()));
        }

        optimizedCount++;
    }

    private void func_udf_sys_rowcount(SQLMethodInvokeExpr x) {
        SQLAggregateExpr count = new SQLAggregateExpr("COUNT");
        for (SQLExpr arg : x.getArguments()) {
            count.addArgument(arg.clone());
        }

        if (SQLUtils.replaceInParent(x, count)) {
            optimizedCount++;
        }
    }

    private void func_trunc(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 1 && arguments.size() != 2) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        SQLDataType type0 = arg0.computeDataType();

        if (type0 == null) {
            return;
        }

        if (arguments.size() == 1) {
            long typeHash = type0.nameHashCode64();
            if (typeHash == FnvHash.Constants.TINYINT || typeHash == FnvHash.Constants.SMALLINT
                || typeHash == FnvHash.Constants.INT || typeHash == FnvHash.Constants.INTEGER
                || typeHash == FnvHash.Constants.BIGINT || typeHash == FnvHash.Constants.FLOAT
                || typeHash == FnvHash.Constants.DOUBLE || typeHash == FnvHash.Constants.NUMBER
                || typeHash == FnvHash.Constants.DECIMAL) {
                x.setMethodName("TRUNCATE");
            }
        }
    }

    private void func_date_sub(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 2) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);

        SQLDataType type0 = arg0.computeDataType();
        SQLDataType type1 = arg1.computeDataType();

        SQLExpr intervalValue;
        String unit;
        if (arg1 instanceof SQLIntervalExpr) {
            SQLIntervalExpr interval = (SQLIntervalExpr) arg1;
            intervalValue = interval.getValue();
            unit = interval.getUnit().name();
        } else {
            intervalValue = arg1;
            unit = "Day";
        }

        SQLMethodInvokeExpr dateAdd = new SQLMethodInvokeExpr("DATE_ADD");
        dateAdd.addArgument(new SQLCharExpr(unit));
        dateAdd.addArgument(new SQLUnaryExpr(SQLUnaryOperator.Negative, intervalValue.clone()));
        dateAdd.addArgument(arg0.clone());

        if (SQLUtils.replaceInParent(x, dateAdd)) {
            dateAdd.accept(this);
            optimizedCount++;
        }
    }

    private void func_date_add(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 2) {

            SQLExpr arg0 = arguments.get(0);
            SQLExpr arg1 = arguments.get(1);

            SQLExpr intervalValue;
            String unit;
            if (arg1 instanceof SQLIntervalExpr) {
                SQLIntervalExpr interval = (SQLIntervalExpr) arg1;
                intervalValue = interval.getValue();
                unit = interval.getUnit().name();
            } else {
                intervalValue = arg1;
                unit = "Day";
            }

            SQLMethodInvokeExpr dateAdd = new SQLMethodInvokeExpr("DATE_ADD");
            dateAdd.addArgument(new SQLCharExpr(unit));
            dateAdd.addArgument(intervalValue.clone());
            SQLExpr arg2 = arg0.clone();

            SQLDataType arg2DataType = arg2.computeDataType();
            if (arg2DataType != null && arg2DataType.nameHashCode64() == FnvHash.Constants.DATE) {
                long unitHash = FnvHash.hashCode64(unit);
                if (unitHash == FnvHash.Constants.SECOND
                        || unitHash == FnvHash.Constants.MINUTE
                        || unitHash == FnvHash.Constants.HOUR) {
                    arg2 = new SQLCastExpr(arg2, new SQLDataTypeImpl("TIMESTAMP"));
                }
            }

            dateAdd.addArgument(arg2);

            if (SQLUtils.replaceInParent(x, dateAdd)) {
                optimizedCount++;
                dateAdd.accept(this);
                return;
            }
        } else if (arguments.size() == 3) {
            SQLExpr arg0 = arguments.get(0);
            SQLExpr arg2 = arguments.get(2);

            SQLDataType type2 = arg2.computeDataType();

            long arg2Hash = type2 != null ? type2.nameHashCode64() : 0L;

            if (arg2Hash == FnvHash.Constants.TIMESTAMP
                || arg2Hash == FnvHash.Constants.DATETIME) {
                return;
            }

            if (arg2Hash == FnvHash.Constants.DATE) {
                long unitHash = FnvHash.hashCode64(arg0.toString());
                if (unitHash == FnvHash.Constants.SECOND
                        || unitHash == FnvHash.Constants.MINUTE
                        || unitHash == FnvHash.Constants.HOUR) {
                    arg2 = new SQLCastExpr(arg2, new SQLDataTypeImpl("TIMESTAMP"));
                    arg2.setParent(x);
                    arguments.set(2, arg2);
                    return;
                }
            }

            if (arg2 instanceof SQLTextLiteralExpr) {
                int length = ((SQLTextLiteralExpr)arg2).getText().length();
                if (length == 10) {
                    arguments.set(2, new SQLCastExpr(arg2, new SQLDataTypeImpl("DATE")));
                } else if (length >= 19) {
                    arguments.set(2, new SQLCastExpr(arg2, new SQLDataTypeImpl("TIMESTAMP")));
                }
            } else {
                if (arg2Hash == FnvHash.Constants.CHAR || arg2Hash == FnvHash.Constants.VARCHAR) {
                    arguments.set(2, new SQLCastExpr(arg2, new SQLDataTypeImpl("TIMESTAMP")));
                }
            }
        }
    }

    private void func_datediff(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 2 && arguments.size() != 3) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);
        SQLExpr arg2 = arguments.size() == 3 ? arguments.get(2) : null;

        SQLDataType type0 = arg0.computeDataType();
        SQLDataType type1 = arg1.computeDataType();

        if (type0 == null || type1 == null) {
            return;
        }

        if (arguments.size() == 2) {
            if (arg0 instanceof SQLCharExpr) {
                arguments.set(1, new SQLDateExpr(((SQLCharExpr) arg0).getText()));
            } else if (arg0 instanceof SQLIntegerExpr) {
                arguments.set(1, new SQLDateExpr((arg0.toString())));
            } else {
                arguments.set(1, new SQLCastExpr(arg0, new SQLCharacterDataType("DATE")));
            }

            if (arg1 instanceof SQLCharExpr) {
                arguments.set(0, new SQLDateExpr(((SQLCharExpr) arg1).getText()));
            } else if(arg1 instanceof SQLIntegerExpr) {
                arguments.set(0, new SQLDateExpr((arg1.toString())));
            } else {
                arguments.set(0, new SQLCastExpr(arg1, new SQLCharacterDataType("DATE")));
            }
            arguments.add(0, new SQLCharExpr("DAY"));

            for (SQLExpr arg : arguments) {
                arg.setParent(x);
            }
            x.setMethodName("DATE_DIFF");
        }
    }

    public boolean visit(SQLSelectItem x) {
        String alias = x.getAlias();

        String alias2 = quoteAlias(alias);
        if (alias2 != alias) {
            x.setAlias(alias2);
        }

        final SQLExpr expr = x.getExpr();
        if (expr instanceof SQLIdentifierExpr) {
            visit((SQLIdentifierExpr) expr);
        } else if (expr instanceof SQLPropertyExpr) {
            visit((SQLPropertyExpr) expr);
        } else {
            expr.accept(this);
        }

        return false;
    }

    public boolean visit(SQLBlockStatement x) {
        List<SQLStatement> stmtList = x.getStatementList();
        if (stmtList.size() == 2 && stmtList.get(1) instanceof SQLCommitStatement) {
            stmtList.remove(1);
        }
        return true;
    }

    public boolean visit(SQLForStatement x) {
        SQLExpr range = x.getRange();

        if (!(range instanceof SQLQueryExpr)) {
            return true;
        }

        SQLSelect rangeSelect = ((SQLQueryExpr) range).getSubQuery();
        SQLSelectQueryBlock rangeQuery = rangeSelect.getQueryBlock();
        if (rangeQuery == null) {
            return true;
        }

        if (!(rangeQuery.getFrom() instanceof SQLExprTableSource)) {
            return true;
        }
        if (rangeQuery.getWhere() != null) {
            return true;
        }

        SQLExprTableSource rangeQueryFrom = (SQLExprTableSource) rangeQuery.getFrom();

        List<SQLStatement> stmtList = x.getStatements();
        if (stmtList.size() != 1 || !(stmtList.get(0) instanceof SQLUpdateStatement)) {
            return true;
        }

        if (!(x.getParent() instanceof SQLBlockStatement)) {
            return true;
        }
        SQLBlockStatement parent = (SQLBlockStatement) x.getParent();
        int index = parent.getStatementList().indexOf(x);
        if (index == -1) {
            return true;
        }

        SQLUpdateStatement update  = (SQLUpdateStatement) stmtList.get(0);

        SQLExprTableSource joinRight = rangeQueryFrom.clone();
        joinRight.setAlias(x.getIndex().toString());
        SQLJoinTableSource join = new SQLJoinTableSource(update.getTableSource(), SQLJoinTableSource.JoinType.COMMA, joinRight, null);
        update.setTableSource(join);

        update.accept(this);

        parent.getStatementList().set(index, update);

        return false;
    }

    public boolean visit(SQLExprTableSource x) {
        String alias = x.getAlias();
        String alias2 = quoteAlias(alias);
        if (alias2 != alias) {
            x.setAlias(alias2);
        }

        final SQLExpr expr = x.getExpr();

        if (expr instanceof SQLMethodInvokeExpr
                && ((SQLMethodInvokeExpr) expr).methodNameHashCode64() == FnvHash.Constants.ANN)
        {
            ann(x, (SQLMethodInvokeExpr) expr);
        }

        return super.visit(x);
    }

    private static class ANNCheckDistanceVisitor extends OptimizerVisitor {
        private boolean hasAnnDistance;
        private SQLExprTableSource x;

        public ANNCheckDistanceVisitor(SQLExprTableSource x) {
            this.x = x;
        }

        public boolean visit(SQLIdentifierExpr x) {
            if (x.nameHashCode64() == FnvHash.Constants.ANN_DISTANCE) {
                hasAnnDistance = true;
            }
            return false;
        }

        public boolean visit(SQLExprTableSource x1) {
            if (this.x == x1) {
                return false;
            }
            return true;
        }
    }

    private void ann(final SQLExprTableSource x, SQLMethodInvokeExpr ann) {
        SQLObject parent = x.getParent();

        SQLSelectStatement stmt = null;
        SQLSelectQueryBlock parentQuery = null;
        boolean hasAnnDistance = false;
        for (SQLObject p = parent;p != null;p = p.getParent()) {
            if (p instanceof SQLSelectQueryBlock && parentQuery == null) {
                parentQuery = (SQLSelectQueryBlock) p;
                continue;
            }

            if (p instanceof SQLSelectStatement) {
                stmt = (SQLSelectStatement) p;
                break;
            }
        }

        if (stmt != null) {
            ANNCheckDistanceVisitor v = new ANNCheckDistanceVisitor(x);
            stmt.accept(v);
            hasAnnDistance = v.hasAnnDistance;
        }

        final List<SQLExpr> arguments = ann.getArguments();

        SQLExpr query = arguments.get(0);
        SQLExpr column = arguments.get(1);
        SQLExpr input = arguments.get(2);
        SQLExpr topN = arguments.get(3);
        SQLExpr param = null;
        if (arguments.size() > 4) {
            param = arguments.get(4);
        }

        SQLOrderingSpecification orderBySpec = SQLOrderingSpecification.ASC;

        SQLAnnIndex.Distance distance = null;
        if (column instanceof SQLIdentifierExpr) {
            final SQLIdentifierExpr identColumn = (SQLIdentifierExpr) column;
            SQLColumnDefinition resolvedColumn = identColumn.getResolvedColumn();
            if (resolvedColumn != null
                    && resolvedColumn.getAnnIndex() != null) {
                distance = resolvedColumn.getAnnIndex().getDistance();
            }
        }

        if (param instanceof SQLCharExpr) {
            String paramText = ((SQLCharExpr) param).getText();
            if (paramText.toLowerCase().indexOf("distance_measure=dotproduct") != -1) {
                distance = SQLAnnIndex.Distance.DotProduct;
            }
        }

        if (distance == SQLAnnIndex.Distance.DotProduct) {
            orderBySpec = SQLOrderingSpecification.DESC;
        }

        SQLExpr ann2;
        if (arguments.size() == 4) {
            ann2 = new SQLMethodInvokeExpr("ANN", null, column, input, topN);
        } else {
            ann2 = new SQLMethodInvokeExpr("ANN", null, column, input, topN, arguments.get(4));
        }
        SQLExpr where = new SQLBinaryOpExpr(ann2, SQLBinaryOperator.GreaterThanOrEqual, new SQLRealExpr(0.0F));


        SQLName table = null;

        String ann_distance = "ann_distance";
//        SQLIdentifierExpr ann_distanceExpr = new SQLIdentifierExpr("ann_distance");

        SQLTableSource tableSource;
        if (query instanceof SQLName) {
            table = (SQLName) query;
            tableSource = new SQLExprTableSource(query.clone());
        } else {
            SQLQueryExpr queryExpr = (SQLQueryExpr) query;
            tableSource = new SQLSubqueryTableSource(queryExpr.getSubQuery().clone());
        }

        SchemaRepository repository = this.repository;
        if (repository == null) {
            repository = new SchemaRepository(dbType);
        }
        MySqlSchemaStatVisitor statVisitor = new MySqlSchemaStatVisitor(repository) {
            public boolean visit(SQLExprTableSource x1) {
                if (x == x1) {
                    return false;
                }
                return super.visit(x1);
            }
        };

        if (parentQuery != null) {
            if (this.repository == null) {
                repository.resolve(parentQuery);
            }

            statVisitor.visit(parentQuery);
            if (parentQuery.getWhere() == null && !parentQuery.hasSelectAggregation()) {
                parentQuery.setWhere(where);
                parentQuery.setFrom(tableSource);
                parentQuery.setLimit(new SQLLimit(topN));
                Integer selectIndex = null;
                for (int i = 0; i < parentQuery.getSelectList().size(); i++) {
                    SQLSelectItem selectItem = parentQuery.getSelectList().get(i);
                    SQLExpr selectItemExpr = selectItem.getExpr();
                    if (selectItemExpr instanceof SQLName
                            && ((SQLName) selectItemExpr).getSimpleName().equalsIgnoreCase(ann_distance)) {
                        selectItem.setExpr(ann2.clone());
                        if (selectItem.getAlias() == null) {
                            selectItem.setAlias(ann_distance);
                        }
                        selectIndex = i + 1;
                    }
                }
//                    parentQuery.addSelectItem(ann2.clone(), "ann_distance");
                parentQuery.addOrderBy(
                        selectIndex != null
                                ? new SQLIntegerExpr(selectIndex)
                                : ann2.clone()
                        , orderBySpec);
                return;
            }
        }

        SQLSubqueryTableSource replaceTo = null;
        SQLSelectQueryBlock queryBlock = new MySqlSelectQueryBlock();

        if (table != null && statVisitor.getColumns().size() > 0) {
            for (TableStat.Column statColumn : statVisitor.getColumns()) {
                if (FnvHash.hashCode64(statColumn.getTable()) == table.hashCode64()) {
                    if (statColumn.getName().equalsIgnoreCase(ann_distance)) {
                        queryBlock.addSelectItem(ann2.clone(), ann_distance);
                    } else {
                        queryBlock.addSelectItem(new SQLIdentifierExpr(statColumn.getName()));
                    }
                }
            }
        } else {
            queryBlock.addSelectItem(new SQLAllColumnExpr());
        }

        queryBlock.setFrom(tableSource);

        queryBlock.addWhere(where);
        queryBlock.setLimit(new SQLLimit(topN));
        if (hasAnnDistance && queryBlock.findSelectItem("ann_distance") == null) {
            queryBlock.addSelectItem(ann2, "ann_distance");
        }
        queryBlock.addOrderBy(ann2.clone(), orderBySpec);
        replaceTo = new SQLSubqueryTableSource(new SQLSelect(queryBlock), x.getAlias());

        if (SQLUtils.replaceInParent(x, replaceTo)) {
            optimizedCount++;
        }
    }

    public boolean visit(SQLUnionQueryTableSource x) {
        String alias = x.getAlias();

        String alias2 = quoteAlias(alias);
        if (alias2 != alias) {
            x.setAlias(alias2);
        }

        return true;
    }

    public boolean visit(SQLSubqueryTableSource x) {
        String alias = x.getAlias();

        String alias2 = quoteAlias(alias);
        if (alias2 != alias) {
            x.setAlias(alias2);
        }

        visit(x.getSelect());

        return false;
    }

    public boolean visit(SQLAggregateExpr x) {
        List<SQLExpr> arguments = x.getArguments();

        String methodName = x.getMethodName();
        x.setMethodName(SQLUtils.normalize(methodName));

        long hash = x.methodNameHashCode64();
        if (hash == FnvHash.Constants.COUNT
            && x.getOption() == SQLAggregateOption.DISTINCT
            && arguments.size() > 1
            && x.getOver() == null) {
            
            SQLListExpr listExpr = new SQLListExpr();
            for (SQLExpr arg : arguments) {
                listExpr.addItem(arg.clone());
            }
            arguments.clear();
            arguments.add(listExpr);

            optimizedCount++;
        } else if (hash == FnvHash.Constants.GROUP_CONCAT) {
            func_group_concat(x);
        }
//        else if (hash == FnvHash.Constants.SUM && arguments.size() == 1) {
//            SQLExpr arg0 = arguments.get(0);
//            SQLDataType sqlDataType = arg0.computeDataType();
//            if(sqlDataType != null && sqlDataType.isString()) {
//                SQLCastExpr castExpr = new SQLCastExpr(arg0, new SQLDataTypeImpl(SQLDataTypeImpl.Constants.DOUBLE));
//                castExpr.setTry(true);
//                arguments.set(0, castExpr);
//                optimizedCount++;
//            }
//        }

        return true;
    }

    public boolean visit(SQLJoinTableSource x) {

//        SQLExpr on = x.getCondition();

//        if (on instanceof SQLBooleanExpr) {
//            boolean value = ((SQLBooleanExpr) on).getBooleanValue();
//            if (value) {
//                SQLJoinTableSource.JoinType joinType = x.getJoinType();
//                if (joinType == SQLJoinTableSource.JoinType.LEFT_OUTER_JOIN
//                        || joinType == SQLJoinTableSource.JoinType.INNER_JOIN)
//                {
//                    x.setCondition(null);
//                    x.setJoinType(SQLJoinTableSource.JoinType.COMMA);
//                }
//            }
//        }

        SQLJoinTableSource.JoinType joinType = x.getJoinType();
        if (joinType == SQLJoinTableSource.JoinType.COMMA
                && x.getLeft() instanceof SQLExprTableSource
                && x.getRight() instanceof SQLExprTableSource) {
            x.putAttribute("ads.comma_join", true);
        }

        return true;
    }

    private void visitUnion(SQLUnionQuery x) {
        SQLUnionOperator operator = x.getOperator();
        SQLSelectQuery left = x.getLeft();
        SQLSelectQuery right = x.getRight();


        SQLOrderBy orderBy = x.getOrderBy();
        if (left instanceof SQLUnionQuery
                && ((SQLUnionQuery) left).getOperator() == operator
                && !right.isBracket()
                && orderBy == null) {

            SQLUnionQuery leftUnion = (SQLUnionQuery) left;

            if (right instanceof SQLSelectQueryBlock) {
                if (x.getOperator() == SQLUnionOperator.UNION_ALL) {
                    right.setBracket(true);
                } else {
                    SQLSelectQueryBlock rightQueryBlock = (SQLSelectQueryBlock) right;
                    if (rightQueryBlock.getOrderBy() != null || rightQueryBlock.getLimit() != null) {
                        rightQueryBlock.setBracket(true);
                    }
                }
            }

            List<SQLSelectQuery> rights = new ArrayList<SQLSelectQuery>();
            rights.add(right);

            for (;;) {
                SQLSelectQuery leftLeft = leftUnion.getLeft();
                SQLSelectQuery leftRight = leftUnion.getRight();

                if (leftRight instanceof SQLSelectQueryBlock) {
                    if (x.getOperator() == SQLUnionOperator.UNION_ALL) {
                        leftRight.setBracket(true);
                    } else {
                        SQLSelectQueryBlock leftRightQueryBlock = (SQLSelectQueryBlock) leftRight;
                        if (leftRightQueryBlock.getOrderBy() != null || leftRightQueryBlock.getLimit() != null) {
                            leftRightQueryBlock.setBracket(true);
                        }
                    }
                }

                if ((!leftUnion.isBracket())
                        && leftUnion.getOrderBy() == null
                        && (!leftLeft.isBracket())
                        && (!leftRight.isBracket())
                        && leftLeft instanceof SQLUnionQuery
                        && ((SQLUnionQuery) leftLeft).getOperator() == operator) {
                    rights.add(leftRight);
                    leftUnion = (SQLUnionQuery) leftLeft;
                    continue;
                } else {
                    rights.add(leftRight);
                    rights.add(leftLeft);
                }
                break;
            }

            for (int i = rights.size() - 1; i >= 0; i--) {
                SQLSelectQuery item = rights.get(i);
                item.accept(this);

            }
            return;
        }

        if (left != null) {
            for (;;) {
                if (left.getClass() == SQLUnionQuery.class) {
                    SQLUnionQuery leftUnion = (SQLUnionQuery) left;
                    SQLSelectQuery leftLeft = leftUnion.getLeft();
                    SQLSelectQuery leftRigt = leftUnion.getRight();
                    if ((!leftUnion.isBracket())
                            && leftUnion.getRight() instanceof SQLSelectQueryBlock
                            && leftUnion.getLeft() != null
                            && leftUnion.getOrderBy() == null)
                    {
                        if (leftLeft.getClass() == SQLUnionQuery.class) {
                            visit((SQLUnionQuery) leftLeft);
                        } else {
                            leftLeft.accept(this);
                        }
                        leftRigt.accept(this);
                    } else {
                        visit(leftUnion);
                    }
                } else {
                    left.accept(this);
                }
                break;
            }
        }

        if (right == null) {
            return;
        }

        right.accept(this);

        if (orderBy != null) {
            orderBy.accept(this);
        }

        SQLLimit limit = x.getLimit();
        if (limit != null) {
            limit.accept(this);
        }

    }

    public boolean visit(SQLUnionQuery x) {
        if (x.getRelations().size() > 2) {
            for (SQLSelectQuery relation : x.getRelations()) {
                relation.accept(this);
            }
            return false;
        }

        if (x.getLimit() == null
                && x.getOrderBy() == null
                && x.getLeft() instanceof SQLUnionQuery
                && ((SQLUnionQuery) x.getLeft()).getLeft() instanceof SQLUnionQuery
        ) {
            visitUnion(x);
            return false;
        }

        SQLSelectQuery left = x.getLeft();
        if (left instanceof SQLSelectQueryBlock) {
            if (x.getOperator() == SQLUnionOperator.UNION_ALL) {
                left.setBracket(true);
            } else {
                SQLSelectQueryBlock leftQueryBlock = (SQLSelectQueryBlock) left;
                if (leftQueryBlock.getLimit() != null || leftQueryBlock.getOrderBy() != null) {
                    left.setBracket(true);
                }
            }
        }
        if (left instanceof SQLUnionQuery) {
            visit((SQLUnionQuery) left);
        } else {
            left.accept(this);
        }

        SQLSelectQuery right = x.getRight();
        if (right instanceof SQLSelectQueryBlock) {
            if (x.getOperator() == SQLUnionOperator.UNION_ALL) {
                right.setBracket(true);
            } else {
                SQLSelectQueryBlock rightQueryBlock = (SQLSelectQueryBlock) right;
                if (rightQueryBlock.getLimit() != null || rightQueryBlock.getOrderBy() != null) {
                    right.setBracket(true);
                }
            }
        }
        if (right instanceof SQLUnionQuery) {
            visit((SQLUnionQuery) right);
        } else {
            right.accept(this);
        }

        if (x.getOperator() == SQLUnionOperator.MINUS) {
            x.setOperator(SQLUnionOperator.EXCEPT);
        }

        boolean hasLimitOrderBy = x.getLimit() != null || x.getOrderBy() != null;
        SQLObject parent = x.getParent();
        if (parent instanceof SQLUnionQuery && hasLimitOrderBy) {
            x.setBracket(true);
        }

        return true;
    }

    public boolean visit(SQLExtractExpr x) {
        SQLIntervalUnit unit = x.getUnit();
        SQLExpr value = x.getValue();

        switch (unit) {
            case YEAR:
            case QUARTER:
            case MONTH:
            case DAY:
            case HOUR:
            case MINUTE:
            case SECOND:
                SQLUtils.replaceInParent(x, new SQLMethodInvokeExpr(unit.name(), null, value));
                break;
            case WEEK:
                SQLMethodInvokeExpr weekFun = new SQLMethodInvokeExpr("WEEK");
                weekFun.addArgument(new SQLCastExpr(value, new SQLDataTypeImpl("TIMESTAMP")));
                SQLUtils.replaceInParent(x, weekFun);
                break;
            case DAY_OF_MONTH:
                SQLUtils.replaceInParent(x, new SQLMethodInvokeExpr("DAY", null, value));
                break;

            case DAY_OF_WEEK:
            case DOW:
                SQLUtils.replaceInParent(x, new SQLMethodInvokeExpr("DAY_OF_WEEK", null, value));
                break;

            case DAY_OF_YEAR:
            case DOY:
                SQLMethodInvokeExpr invokeExpr = new SQLMethodInvokeExpr(null, null, value);
                func_dayofyear(invokeExpr);
                SQLUtils.replaceInParent(x, invokeExpr);
                break;

            case YEAR_OF_WEEK:
            case YOW:
                if (value instanceof SQLCharExpr) {
                    SQLExpr replaceTo = new SQLTimestampExpr(((SQLCharExpr) value).getText());
                    SQLUtils.replaceInParent(x, new SQLMethodInvokeExpr("YEAR_OF_WEEK", null, replaceTo));
                } else if (value instanceof SQLIdentifierExpr) {
                    SQLExpr replaceTo = new SQLCastExpr(value, new SQLDataTypeImpl("TIMESTAMP"));
                    SQLUtils.replaceInParent(x, new SQLMethodInvokeExpr("YEAR_OF_WEEK", null, replaceTo));
                }
                break;

            case TIMEZONE_HOUR:
            case TIMEZONE_MINUTE:
                SQLUtils.replaceInParent(x, new SQLMethodInvokeExpr(unit.name(), null, new SQLCastExpr(value, new SQLDataTypeImpl("TIMESTAMP"))));
                break;
        }


        return true;
    }

    class HavingRewrite extends SQLASTVisitorAdapter {
        private final SQLSelectQueryBlock queryBlock;

        public HavingRewrite(SQLSelectQueryBlock queryBlock) {
            this.queryBlock = queryBlock;
        }

        public boolean visit(SQLIdentifierExpr x) {
            if ( x.getParent() instanceof SQLPropertyExpr) {
                return false;
            }
            SQLSelectItem selectItem = x.getResolvedSelectItem();
            if (selectItem == null) {
                List<SQLSelectItem> selectList = queryBlock.getSelectList();
                for (int i = 0; i < selectList.size(); i++) {
                    SQLSelectItem item = selectList.get(i);

                    if (x.nameHashCode64() == item.alias_hash()) {
                        selectItem = item;
                        break;
                    }
                }
            }

            if (selectItem == null) {
                return false;
            }

            boolean isAgg = false;

            SQLObject item = selectItem.getExpr();
            while (!(item instanceof SQLSelectQueryBlock) && !(item instanceof SQLSelect)) {
                if(item instanceof SQLAggregateExpr) {
                    isAgg = true;
                    break;
                }
                item = item.getParent();
            }

            SQLExpr clonedItemExpr = selectItem.getExpr().clone();

            boolean allColumn = false;
            if (clonedItemExpr instanceof SQLAllColumnExpr
                    || (clonedItemExpr instanceof SQLPropertyExpr
                        && ((SQLPropertyExpr) clonedItemExpr).getName().equals("*"))
            ) {
                allColumn = true;
            }
            if ((!allColumn) && !(isAgg && x.getParent() instanceof SQLAggregateExpr)) {
                SQLUtils.replaceInParent(x, clonedItemExpr);
            }

            return true;
        }
    }

    public boolean visit(SQLLimit x) {
        final SQLExpr offset = x.getOffset();
        if (offset != null) {
            if (offset.equals(new SQLIntegerExpr(0))) {
                x.setOffset(null);
            } else {
                offset.accept(this);
            }
        }

        final SQLExpr rowCount = x.getRowCount();
        if (rowCount != null) {
            rowCount.accept(this);
        }
        return false;
    }

    class ViewRewrite extends SQLASTVisitorAdapter {

        public boolean visit(SQLExprTableSource x) {
            SQLExpr expr = x.getExpr();

            if (expr instanceof SQLName) {
                String name = ((SQLName) expr).getSimpleName();

                SQLTableSource resolvedTableSource = null;
                if (expr instanceof SQLIdentifierExpr) {
                    resolvedTableSource = ((SQLIdentifierExpr) expr).getResolvedTableSource();
                }

                if (resolvedTableSource == null) {
                    SchemaObject table = repository.findTable(name);
                    if (table != null) {
                        return true;
                    }
                    SchemaObject view = repository.findView(name);
                    if (view != null) {
                        SQLCreateViewStatement statement = (SQLCreateViewStatement) view.getStatement();
                        List<SQLTableElement> columns = statement.getColumns();

                        String subQuerySQL = statement.getSubQuery().toString();
                    }
                }
            }

            return true;
        }
    }

    public boolean visit(SQLSelectOrderByItem x) {
        SQLObject parent = x.getParent();
        if (parent != null && parent.getParent() instanceof SQLSelectQueryBlock) {
            SQLExpr expr = x.getExpr();

            SQLSelectQueryBlock queryBlock = (SQLSelectQueryBlock) parent.getParent();
            List<SQLSelectItem> selectList = queryBlock.getSelectList();

            boolean found = false;
            for (SQLSelectItem item : selectList) {
                if (item.getExpr().equals(expr)) {
                    found = true;
                    break;
                }
            }

            if (!found) {
                if (expr instanceof SQLIdentifierExpr) {
                    long nameHashCode64 = ((SQLIdentifierExpr) expr).nameHashCode64();
                    for (SQLSelectItem item : selectList) {
                        if (item.alias_hash() == nameHashCode64) {
                            break;
                        }
                    }
                } else if (expr instanceof SQLAggregateExpr) {
                    SQLAggregateExpr agg = (SQLAggregateExpr) expr;
                    if (agg.getArguments().size() == 1 && agg.getArguments().get(0) instanceof SQLIdentifierExpr) {
                        SQLIdentifierExpr ident = (SQLIdentifierExpr) agg.getArguments().get(0);
                        for (SQLSelectItem item : selectList) {
                            SQLExpr itemExpr = item.getExpr();
                            if (itemExpr instanceof SQLAggregateExpr
                                    && ((SQLAggregateExpr) itemExpr).getArguments().size() == 1
                                    && ((SQLAggregateExpr) itemExpr).getArguments().get(0) instanceof SQLPropertyExpr
                            ) {
                                SQLPropertyExpr propertyExpr = (SQLPropertyExpr) ((SQLAggregateExpr) itemExpr).getArguments().get(0);
                                if (propertyExpr.nameHashCode64() == ident.nameHashCode64()) {
                                    SQLAggregateExpr agg2 = agg.clone();
                                    agg2.setArgument(0, propertyExpr.clone());
                                    if (agg2.equals(itemExpr)) {
                                        x.setExpr(agg2);
                                    }
                                }
                            }
                        }
                    }

                }
            }
        }

        return super.visit(x);
    }

    public boolean visit(SQLBetweenExpr x) {
        x.getTestExpr().accept(this);
        x.getBeginExpr().accept(this);
        x.getEndExpr().accept(this);

        SQLExpr expr = x.getTestExpr();
        SQLExpr begin = x.getBeginExpr();
        SQLExpr end = x.getEndExpr();
        if (expr instanceof SQLMethodInvokeExpr
                && ((SQLMethodInvokeExpr) expr).methodNameHashCode64() == FnvHash.Constants.JSON_EXTRACT) {
            if (begin instanceof SQLCharExpr) {
                ((SQLMethodInvokeExpr) expr).setMethodName("JSON_EXTRACT_SCALAR");
            } else if (end instanceof SQLNumberExpr || begin instanceof SQLNumberExpr) {
                SQLCastExpr cast = new SQLCastExpr(expr.clone(), SQLNumberExpr.DATA_TYPE_DOUBLE.clone());
                x.setTestExpr(cast);
                optimizedCount++;
                return false;
            } else if (begin instanceof SQLIntegerExpr || end instanceof SQLIntegerExpr) {
                SQLCastExpr cast = new SQLCastExpr(expr.clone(), SQLIntegerExpr.DATA_TYPE.clone());
                x.setTestExpr(cast);
                optimizedCount++;
                return false;
            } else if (begin instanceof SQLBooleanExpr) {
                SQLCastExpr cast = new SQLCastExpr(expr.clone(), SQLBooleanExpr.DATA_TYPE.clone());
                x.setTestExpr(cast);
                optimizedCount++;
                return false;
            }
        }

        return true;
    }

    public boolean visit(SQLInListExpr x) {
        SQLExpr expr = x.getExpr();
        expr.accept(this);
        if (x.getTargetList().size() == 0) {
            return true;
        }

        SQLExpr right = x.getTargetList().get(0);
        if (expr instanceof SQLMethodInvokeExpr) {
            SQLMethodInvokeExpr methodInvokeExpr = (SQLMethodInvokeExpr) expr;
            long methodHash = methodInvokeExpr.methodNameHashCode64();
            if (methodHash == FnvHash.Constants.JSON_EXTRACT) {
                if (right instanceof SQLCharExpr) {
                    methodInvokeExpr.setMethodName("JSON_EXTRACT_SCALAR");
                } else if (right instanceof SQLIntegerExpr) {
                    SQLCastExpr cast = new SQLCastExpr(expr.clone(), SQLIntegerExpr.DATA_TYPE.clone());
                    x.setExpr(cast);
                    optimizedCount++;
                    return false;
                } else if (right instanceof SQLNumberExpr) {
                    SQLCastExpr cast = new SQLCastExpr(expr.clone(), SQLNumberExpr.DATA_TYPE_DOUBLE.clone());
                    x.setExpr(cast);
                    optimizedCount++;
                    return false;
                } else if (right instanceof SQLBooleanExpr) {
                    SQLCastExpr cast = new SQLCastExpr(expr.clone(), SQLBooleanExpr.DATA_TYPE.clone());
                    x.setExpr(cast);
                    optimizedCount++;
                    return false;
                }
            } else {
                for (SQLExpr target : x.getTargetList()) {
                    if (target instanceof SQLLiteralExpr) {
                        SQLLiteralExpr literal = (SQLLiteralExpr) target;
                        handleNameLiteral(x, expr, literal);
                    }
                }
            }
        }

        return true;
    }

    public static class TypeResolve extends OptimizerVisitor {
        public boolean visit(SQLMethodInvokeExpr x) {
            List<SQLExpr> arguments = x.getArguments();
            if (x.methodNameHashCode64() == FnvHash.Constants.FROM_UNIXTIME) {
                if (arguments.size() == 1) {
                    x.setResolvedReturnDataType(new SQLDataTypeImpl("TIMESTAMP"));
                } else if (arguments.size() == 2) {
                    x.setResolvedReturnDataType(new SQLCharacterDataType("VARCHAR"));
                }
                return true;
            }

            if (x.methodNameHashCode64() == FnvHash.Constants.STR_TO_DATE) {
                if (arguments.size() == 2 && arguments.get(1) instanceof SQLCharExpr) {
                    String fmt = ((SQLCharExpr) arguments.get(1)).getText();
                    if ("%Y-%m-%d".equals(fmt)) {
                        x.setResolvedReturnDataType(new SQLDataTypeImpl("DATE"));
                    }

                    if (fmt.indexOf("%i") >= 0 || fmt.indexOf("%H") >= 0 || fmt.indexOf("%s") >= 0) {
                        x.setResolvedReturnDataType(new SQLDataTypeImpl("TIMESTAMP"));
                    }
                }
            }

            return true;
        }
    }

    public boolean visit(SQLSelect x) {
        if (x.getHintsSize() > 0) {
            x.getHints().clear();
        }
        return super.visit(x);
    }

    public boolean isBackQuotedNames()
    {
        return backQuotedNames;
    }

    public void setBackQuotedNames(boolean backQuotedNames)
    {
        this.backQuotedNames = backQuotedNames;
    }
}
