package com.alibaba.fastsql.sql.optimizer.rules;

import com.alibaba.fastsql.sql.SQLUtils;
import com.alibaba.fastsql.sql.ast.SQLCommentHint;
import com.alibaba.fastsql.sql.ast.SQLExpr;
import com.alibaba.fastsql.sql.ast.SQLName;
import com.alibaba.fastsql.sql.ast.SQLObject;
import com.alibaba.fastsql.sql.ast.expr.*;
import com.alibaba.fastsql.sql.ast.statement.*;
import com.alibaba.fastsql.sql.dialect.mysql.ast.statement.MySqlSelectQueryBlock;
import com.alibaba.fastsql.sql.visitor.SQLASTVisitorAdapter;
import com.alibaba.fastsql.util.FnvHash;

import java.math.BigDecimal;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

public class DynamicFilterVisitor extends OptimizerVisitor {
    private DynamicQueryHandler dynamicQueryHandler;
    protected Exception error = null;

    private int tableALiasSeed = 1;

    protected String genTableAlias() {
        return "DFP_T_" + (tableALiasSeed++) + "_";
    }

    public DynamicQueryHandler getDynamicQueryHandler() {
        return dynamicQueryHandler;
    }

    public void setDynamicQueryHandler(DynamicQueryHandler dynamicQueryHandler) {
        this.dynamicQueryHandler = dynamicQueryHandler;
    }

    static class Plan {
        MySqlSelectQueryBlock queryBlock;
        List<SQLInListExpr> ins = new ArrayList<SQLInListExpr>();
        List<Integer> inSelectItemIndexes = new ArrayList<Integer>();
        SQLValuesTableSource values;
        SQLTableSource mapTableSource;
        List<SQLExpr> mapTableSourceFilters = new ArrayList<SQLExpr>();

        List<SQLName> columns = new ArrayList<SQLName>();
        List<SQLName> notResolvedcolumns = new ArrayList<SQLName>();

        public boolean isIncludeAllColumns() {
            if (notResolvedcolumns.size() > 0) {
                return false;
            }

            for (SQLName column : columns) {
                if (!values.getColumns().contains(column)) {
                    return false;
                }
            }

            return true;
        }
    }

    public Exception getError() {
        return error;
    }

    public boolean visit(MySqlSelectQueryBlock x) {
        if (dynamicQueryHandler == null) {
            return super.visit(x);
        }

        List<SQLTableSource> mappJoinTableSources = x.getMappJoinTableSources();
        if (mappJoinTableSources.size() == 0) {
            return super.visit(x);
        }

        for (int i = mappJoinTableSources.size() - 1; i >= 0; i--) {
            SQLTableSource tableSource = mappJoinTableSources.get(i);
            SQLObject parent = tableSource.getParent();

            if (!(parent instanceof SQLJoinTableSource)) {
                mappJoinTableSources.remove(i);
                continue;
            }

            SQLJoinTableSource join = (SQLJoinTableSource) parent;
            SQLJoinTableSource.JoinType joinType = join.getJoinType();
            if (joinType == null) {
                continue;
            }

            switch (joinType) {
                case JOIN:
                case INNER_JOIN:
                    break;
                case LEFT_OUTER_JOIN:
                    if (tableSource == join.getRight()) {
                        mappJoinTableSources.remove(i);
                        continue;
                    }
                    break;
                case RIGHT_OUTER_JOIN:
                    if (tableSource == join.getLeft()) {
                        mappJoinTableSources.remove(i);
                        continue;
                    }
                    break;
                default:
                    mappJoinTableSources.remove(i);
                    continue;
            }
        }

        List<SQLExpr> queryConditions = null;
        SQLExpr where = x.getWhere();
        if (where != null) {
            queryConditions = SQLBinaryOpExpr.split(where, SQLBinaryOperator.BooleanAnd);
        }

        List<Plan> planList = new ArrayList<Plan>();

        for (SQLTableSource tableSource : mappJoinTableSources) {
            Plan plan = new Plan();

            long aliasHashCode64 = tableSource.aliasHashCode64();

            if (aliasHashCode64 == 0 && tableSource instanceof SQLExprTableSource) {
                SQLName tableName = (SQLName) ((SQLExprTableSource) tableSource).getExpr();
                aliasHashCode64 = tableName.nameHashCode64();
            }

            plan.queryBlock = new MySqlSelectQueryBlock();

            plan.queryBlock.setFrom(tableSource.clone());
            if (tableSource instanceof SQLExprTableSource) {
                SQLExpr tableSourceExpr = ((SQLExprTableSource) tableSource).getExpr();
                if (tableSourceExpr instanceof SQLIdentifierExpr) {
                    SQLIdentifierExpr identTableSourceExpr = (SQLIdentifierExpr) tableSourceExpr;
                    SQLTableSource resolvedTableSource = identTableSourceExpr.getResolvedTableSource();
                    if (resolvedTableSource instanceof SQLWithSubqueryClause.Entry) {
                        SQLSelect subQuery = ((SQLWithSubqueryClause.Entry) resolvedTableSource).getSubQuery().clone();

                        plan.queryBlock.setFrom(subQuery, identTableSourceExpr.normalizedName());
                    }
                }
            }

            plan.values = new SQLValuesTableSource();
            plan.values.setAlias(tableSource.computeAlias());
            plan.mapTableSource = tableSource;

            SQLObject parent = tableSource.getParent();
            if (parent instanceof SQLJoinTableSource) {
                SQLJoinTableSource join = (SQLJoinTableSource) parent;
                SQLExpr on = join.getCondition();
                if (on == null) {
                    continue;
                }

                List<SQLExpr> joinConditions = SQLBinaryOpExpr.split(on, SQLBinaryOperator.BooleanAnd);
                for (SQLExpr joinCondition : joinConditions) {
                    if (joinCondition instanceof SQLBinaryOpExpr) {
                        SQLBinaryOpExpr binaryOpExpr = (SQLBinaryOpExpr) joinCondition;
                        if (binaryOpExpr.isBothName() && binaryOpExpr.getOperator() != SQLBinaryOperator.Equality) {
                            return false;
                        }
                    }
                }

                buildQuery(aliasHashCode64, plan, joinConditions);
                buildQuery(aliasHashCode64, plan, queryConditions);
            }

            planList.add(plan);
        }

        boolean hasNotSupportType = false;
        for (int i = 0; i < planList.size(); i++) {
            Plan plan = planList.get(i);
            MySqlSelectQueryBlock dynamicQuery = plan.queryBlock;

            String dynamicQuerySql = dynamicQuery.toString();
            List<Object[]> rows;
            try {
                rows = dynamicQueryHandler.executeQuery(dynamicQuerySql);
            } catch (Exception ex) {
                error = ex;
                // skip
                return false;
            }

            for (Object[] row : rows) {
                SQLListExpr rowExpr = new SQLListExpr();

                for (int selectItemIndex : plan.inSelectItemIndexes) {
                    Object colVal = row[selectItemIndex];

                    SQLExpr expr = null;
                    if (colVal instanceof Byte
                            || colVal instanceof Short
                            || colVal instanceof Integer
                            || colVal instanceof Long)
                    {
                        expr = new SQLIntegerExpr((Number) colVal);
                    } else if (colVal instanceof BigDecimal) {
                        expr = new SQLDecimalExpr((BigDecimal) colVal);
                    } else if (colVal instanceof String) {
                        expr = new SQLCharExpr((String) colVal);
                    } else if (colVal instanceof Character) {
                        expr = new SQLCharExpr(colVal.toString());
                    } else if (colVal instanceof Boolean) {
                        expr = new SQLBooleanExpr((Boolean) colVal);
                    } else if (colVal instanceof java.sql.Date) {
                        expr = new SQLDateExpr((java.sql.Date) colVal);
                    } else if (colVal instanceof java.sql.Timestamp) {
                        expr = new SQLTimestampExpr((java.sql.Timestamp) colVal);
                    } else {
                        hasNotSupportType = true;
                        if (colVal == null) {
                            error = new SQLException("not support null value.");
                        } else {
                            error = new SQLException("not support Type : " + colVal.getClass());
                        }
                        break;
                    }

                    plan.ins.get(selectItemIndex).addTarget(expr);

                    rowExpr.addItem(expr.clone());
                }

                plan.values.addValue(rowExpr);
            }
        }

        if (hasNotSupportType) {
            return false;
        }

        for (Plan plan : planList) {
            for (SQLInListExpr in : plan.ins) {
                if (in.getTargetList().size() == 0) {
                    error = new SQLException("in targetList is empty");
                    return false;
                }
            }
        }

        for (Plan plan : planList) {
            PlanVisitor v = new PlanVisitor(plan);
            x.accept(v);
        }

        int tableSourceReplacedCount = 0;
        for (Plan plan : planList) {
            for (SQLInListExpr in : plan.ins) {
                x.addCondition(in);
            }

            if (!plan.isIncludeAllColumns()) {
                continue;
            }

            boolean tableSourceReplaced = false;
            if (plan.values.getColumns().size() == plan.ins.size()) {
                if (SQLUtils.replaceInParent(plan.mapTableSource, null)) {
                    tableSourceReplaced = true;
                }
            } else if (SQLUtils.replaceInParent(plan.mapTableSource, plan.values)) {
                tableSourceReplaced = true;
            }

            if (tableSourceReplaced) {
                tableSourceReplacedCount++;
                for (SQLExpr filter : plan.mapTableSourceFilters) {
                    SQLUtils.replaceInParent(filter, null);
                }
            }

        }

        if (planList.size() > 0
                && planList.size() == mappJoinTableSources.size()
                && tableSourceReplacedCount == planList.size()
        ) {
            x.clearMapJoinHint();
        }

        return false;
    }

    private class PlanVisitor extends SQLASTVisitorAdapter {
        Plan plan;

        public PlanVisitor(Plan plan) {
            this.plan = plan;
        }

        public boolean visit(SQLPropertyExpr x) {
            SQLExpr owner = x.getOwner();
            if (owner instanceof SQLIdentifierExpr && ((SQLIdentifierExpr) owner).nameHashCode64() == plan.mapTableSource.aliasHashCode64()) {
                plan.columns.add(new SQLIdentifierExpr(x.getName()));
            }
            return false;
        }

        public boolean visit(SQLExprTableSource x) {
            return false;
        }

        public boolean visit(SQLAggregateExpr x) {
            if (x.methodNameHashCode64() == FnvHash.Constants.COUNT
                    && x.getArguments().size() == 1) {
                SQLExpr arg = x.getArguments().get(0);
                if (arg instanceof SQLAllColumnExpr) {
                    return false;
                }
            }
            return true;
        }

        public boolean visit(SQLIdentifierExpr x) {
            SQLTableSource resolvedTableSource = x.getResolvedTableSource();
            if (resolvedTableSource == null) {
                plan.notResolvedcolumns.add(x);
            } else if (resolvedTableSource == plan.mapTableSource) {
                plan.columns.add(x);
            }
            return false;
        }

        public boolean visit(SQLAllColumnExpr x) {
            plan.notResolvedcolumns.add(new SQLIdentifierExpr("*"));
            return false;
        }

        public boolean visit(SQLBinaryOpExpr x) {
            if (plan.mapTableSourceFilters.contains(x)) {
                return false;
            }
            return true;
        }
    }

    private void buildQuery(long aliasHashCode64
            , Plan plan
            , List<SQLExpr> joinConditions
    ) {
        if (joinConditions == null) {
            return;
        }

        SQLTableSource from = plan.queryBlock.getFrom();
        if (from.getAlias() == null
                && from instanceof SQLExprTableSource
                && ((SQLExprTableSource) from).getExpr() instanceof SQLPropertyExpr) {
            SQLPropertyExpr propertyExpr = (SQLPropertyExpr) ((SQLExprTableSource) from).getExpr();
            from.setAlias(propertyExpr.getName());
        }

        for (SQLExpr condition : joinConditions) {
            if (condition instanceof SQLBinaryOpExpr) {

                SQLBinaryOpExpr binaryOpCond = (SQLBinaryOpExpr) condition;

                if (binaryOpCond.isBothName()) {
                    SQLName leftName = (SQLName) binaryOpCond.getLeft();
                    SQLName rightName = (SQLName) binaryOpCond.getRight();

                    if (leftName instanceof SQLPropertyExpr
                            && ((SQLPropertyExpr) leftName).getOwner() instanceof SQLIdentifierExpr
                            && ((SQLIdentifierExpr) ((SQLPropertyExpr) leftName).getOwner()).nameHashCode64() == aliasHashCode64)
                    {
                        plan.queryBlock.addSelectItem(leftName.clone());
                        plan.values.addColumn(rightName.getSimpleName());

                        SQLInListExpr in = new SQLInListExpr();
                        in.setExpr(rightName.clone());
                        plan.ins.add(in);
                        plan.inSelectItemIndexes.add(plan.queryBlock.getSelectList().size() - 1);
                    } else if (rightName instanceof SQLPropertyExpr
                            && ((SQLPropertyExpr) rightName).getOwner() instanceof SQLIdentifierExpr
                            && ((SQLIdentifierExpr) ((SQLPropertyExpr) rightName).getOwner()).nameHashCode64() == aliasHashCode64)
                    {
                        plan.queryBlock.addSelectItem(rightName.clone());
                        plan.values.addColumn(rightName.getSimpleName());

                        SQLInListExpr in = new SQLInListExpr();
                        in.setExpr(leftName.clone());
                        plan.ins.add(in);
                        plan.inSelectItemIndexes.add(plan.queryBlock.getSelectList().size() - 1);
                    }
                }

                SQLName name = null;
                SQLLiteralExpr literal = null;
                if (binaryOpCond.isLeftNameAndRightLiteral()) {
                    name = (SQLName) binaryOpCond.getLeft();
                    literal = (SQLLiteralExpr) binaryOpCond.getRight();
                } else if (binaryOpCond.isLeftLiteralAndRightName()) {
                    name = (SQLName) binaryOpCond.getRight();
                    literal = (SQLLiteralExpr) binaryOpCond.getLeft();
                }

                if (name == null) {
                    continue;
                }

                if (name instanceof SQLPropertyExpr
                        && ((SQLPropertyExpr) name).getOwner() instanceof SQLIdentifierExpr
                        && ((SQLIdentifierExpr) ((SQLPropertyExpr) name).getOwner()).nameHashCode64() == aliasHashCode64)
                {
                    plan.queryBlock.addCondition(condition.clone());
                    plan.mapTableSourceFilters.add(condition);
                } else if (name instanceof SQLIdentifierExpr
                        && ((SQLIdentifierExpr) name).getResolvedTableSource() != null
                        && ((SQLIdentifierExpr) name).getResolvedTableSource() == plan.mapTableSource
                ) {
                    plan.queryBlock.addCondition(condition.clone());
                    plan.mapTableSourceFilters.add(condition);
                }
            } else if (condition instanceof SQLInListExpr) {
                SQLInListExpr inListExpr = (SQLInListExpr) condition;
                if (inListExpr.getExpr() instanceof SQLName) {
                    SQLName name = (SQLName) inListExpr.getExpr();

                    boolean allLiteral = true;
                    for (SQLExpr item : inListExpr.getTargetList()) {
                        if (!(item instanceof SQLLiteralExpr)) {
                            allLiteral = false;
                            break;
                        }
                    }


                    if (allLiteral
                            && name instanceof SQLPropertyExpr
                            && ((SQLPropertyExpr) name).getOwner() instanceof SQLIdentifierExpr
                            && ((SQLIdentifierExpr) ((SQLPropertyExpr) name).getOwner()).nameHashCode64() == aliasHashCode64)
                    {
                        plan.queryBlock.addCondition(condition.clone());
                        plan.mapTableSourceFilters.add(condition);
                    }
                }
            } else if (condition instanceof SQLBetweenExpr) {
                SQLBetweenExpr betweenExpr = (SQLBetweenExpr) condition;
                if (betweenExpr.getTestExpr() instanceof SQLName) {
                    SQLName name = (SQLName) betweenExpr.getTestExpr();

                    boolean allLiteral = betweenExpr.getBeginExpr() instanceof SQLLiteralExpr && betweenExpr.getEndExpr() instanceof SQLLiteralExpr;

                    if (allLiteral
                            && name instanceof SQLPropertyExpr
                            && ((SQLPropertyExpr) name).getOwner() instanceof SQLIdentifierExpr
                            && ((SQLIdentifierExpr) ((SQLPropertyExpr) name).getOwner()).nameHashCode64() == aliasHashCode64)
                    {
                        plan.queryBlock.addCondition(condition.clone());
                        plan.mapTableSourceFilters.add(condition);
                    }
                }
            }
        }
    }

    public static interface DynamicQueryHandler {
        List<Object[]> executeQuery(String sql);
    }
}
