package com.alibaba.fastsql.sql.optimizer.rules;

import com.alibaba.fastsql.sql.SQLUtils;
import com.alibaba.fastsql.sql.ast.*;
import com.alibaba.fastsql.sql.ast.expr.*;
import com.alibaba.fastsql.sql.ast.statement.*;
import com.alibaba.fastsql.sql.dialect.mysql.ast.statement.MySqlSelectQueryBlock;
import com.alibaba.fastsql.sql.dialect.mysql.visitor.MySqlASTVisitorAdapter;
import com.alibaba.fastsql.util.FnvHash;
import com.alibaba.fastsql.util.MySqlUtils;
import com.alibaba.fastsql.util.StringUtils;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.TimeZone;

import static com.alibaba.fastsql.sql.ast.expr.SQLBinaryOperator.BooleanAnd;

public class OptimizerVisitor extends MySqlASTVisitorAdapter {

    protected TimeZone timeZone;
    protected int optimizedCount = 0;
    protected int repalceFailCount = 0;

    public TimeZone getTimeZone() {
        return timeZone;
    }

    public void setTimeZone(TimeZone timeZone) {
        this.timeZone = timeZone;
    }

    public boolean visit(SQLSelectStatement x) {
        x.getSelect()
                .accept(this);
        return false;
    }

    public boolean visit(SQLSelect x) {
        final SQLWithSubqueryClause with = x.getWithSubQuery();
        if (with != null)  {
            with.accept(this);
        }

        x.getQuery()
                .accept(this);
        return false;
    }

    public boolean visit(SQLSelectItem x) {
        final SQLExpr expr = x.getExpr();

        if (expr instanceof SQLIdentifierExpr) {
            visit((SQLIdentifierExpr) expr);
        } else if (expr instanceof SQLPropertyExpr) {
            visit((SQLPropertyExpr) expr);
        } else {
            expr.accept(this);
        }

        return false;
    }

    public boolean visit(SQLWithSubqueryClause x) {
        List<SQLWithSubqueryClause.Entry> entries = x.getEntries();
        for (int i = 0; i < entries.size(); i++) {
            entries.get(i).accept(this);
        }
        return false;
    }

    public boolean visit(SQLSubqueryTableSource x) {
        visit(x.getSelect());
        return false;
    }

    public boolean visit(SQLPropertyExpr x) {
        x.getOwner()
                .accept(this);
        return false;
    }

    public boolean visit(SQLExprTableSource x) {
        final SQLExpr expr = x.getExpr();
        if (expr instanceof SQLIdentifierExpr) {
            visit((SQLIdentifierExpr) expr);
        } else if (expr instanceof SQLMethodInvokeExpr) {
            visit((SQLMethodInvokeExpr) expr);
        } else
            {
            expr.accept(this);
        }
        return false;
    }

    public boolean visit(SQLMethodInvokeExpr x) {
        for (SQLExpr arg : x.getArguments()) {
            arg.accept(this);
        }
        return false;
    }

    public boolean visit(SQLQueryExpr x) {
        visit(x.getSubQuery());
        return false;
    }

    public boolean visit(MySqlSelectQueryBlock x) {
        return visit((SQLSelectQueryBlock) x);
    }

    public boolean visit(SQLSelectQueryBlock x) {
        for (SQLSelectItem selectItem : x.getSelectList()) {
            visit(selectItem);
        }

        final SQLTableSource from = x.getFrom();
        if (from != null) {
            from.accept(this);
        }

        final SQLExprTableSource into = x.getInto();
        if (into != null) {
            into.accept(this);
        }

        final SQLExpr where = x.getWhere();
        if (where != null) {
            where.accept(this);
        }

        final SQLExpr startWith = x.getStartWith();
        if (startWith != null) {
            startWith.accept(this);
        }

        final SQLExpr connectBy = x.getConnectBy();
        if (connectBy != null) {
            connectBy.accept(this);
        }

        final SQLSelectGroupByClause groupBy = x.getGroupBy();
        if (groupBy != null) {
            groupBy.accept(this);
        }

        final SQLOrderBy orderBy = x.getOrderBy();
        if (orderBy != null) {
            orderBy.accept(this);
        }

        final List<SQLSelectOrderByItem> distributeBy = x.getDistributeByDirect();
        if (distributeBy != null) {
            for (SQLSelectOrderByItem item : distributeBy) {
                item.accept(this);
            }
        }

        final List<SQLSelectOrderByItem> sortBy = x.getSortByDirect();
        if (sortBy != null) {
            for (SQLSelectOrderByItem item : sortBy) {
                item.accept(this);
            }
        }

        final SQLExpr waitTime = x.getWaitTime();
        if (waitTime != null) {
            waitTime.accept(this);
        }

        final SQLLimit limit = x.getLimit();
        if (limit != null) {
            limit.accept(this);
        }

        return false;
    }

    protected boolean isGroup(SQLBinaryOpExpr x) {
        final SQLBinaryOperator operator = x.getOperator();
        if (operator == SQLBinaryOperator.BooleanAnd || operator == SQLBinaryOperator.BooleanOr) {
            SQLExpr left = x.getLeft();
            if (left instanceof SQLBinaryOpExpr
                    && ((SQLBinaryOpExpr) left).getOperator() == operator) {
                return true;
            }

            SQLExpr right = x.getRight();
            if (right instanceof SQLBinaryOpExpr
                    && ((SQLBinaryOpExpr) right).getOperator() == operator) {
                return true;
            }
        }

        return false;
    }

    public boolean visit(SQLBinaryOpExpr x) {
        if (isGroup(x)) {
            SQLBinaryOpExprGroup group = new SQLBinaryOpExprGroup(x.getOperator());
            SQLBinaryOpExpr.split(group.getItems(), x, x.getOperator());
            for (SQLExpr item : group.getItems()) {
                item.setParent(group);
            }
            if (SQLUtils.replaceInParent(x, group)) {
                group.accept(this);
                return false;
            }
        }

        x.getLeft().accept(this);
        x.getRight().accept(this);
        return false;
    }

    public boolean visit(SQLBinaryOpExprGroup x) {
        for (int i = 0; i < x.getItems().size(); i++) {
            x.getItems().get(i).accept(this);
        }

        return false;
    }

    public boolean visit(SQLOrderBy x) {
        for (int i = x.getItems().size() - 1; i >= 0; i--) {
            SQLSelectOrderByItem item = x.getItems().get(i);
            item.accept(this);
        }
        return false;
    }

    public boolean visit(SQLSelectOrderByItem x) {
        x.getExpr().accept(this);
        return false;
    }

    public boolean visit(SQLLimit x) {
        final SQLExpr offset = x.getOffset();
        if (offset != null) {
            offset.accept(this);
        }

        final SQLExpr rowCount = x.getRowCount();
        if (rowCount != null) {
            rowCount.accept(this);
        }
        return false;
    }

    public int getReplaceCount() {
        return optimizedCount;
    }

    public void handleNameLiteral(SQLExpr x, SQLExpr name, SQLLiteralExpr literal) {
        SQLDataType dataType = name.computeDataType();
        if (dataType == null) {
            return;
        }

        final long dataTypeNameHash = dataType.nameHashCode64();

        if (dataTypeNameHash == FnvHash.Constants.TINYINT) {
            if (literal instanceof SQLCharExpr) {
                String text = ((SQLCharExpr) literal).getText();
                SQLExpr replaceTo = null;
                try {
                    text = text.trim();
                    byte val = Byte.parseByte(text);
                    replaceTo = new SQLTinyIntExpr(val);
                } catch (NumberFormatException ex) {
                }

                if (replaceTo != null && replaceInParent(literal, replaceTo)) {
                    optimizedCount++;
                    return;
                }
            } else if (literal instanceof SQLIntegerExpr) {
                SQLExpr replaceTo = null;
                try {
                    int intValue = ((SQLIntegerExpr) literal).getNumber().intValue();

                    if (intValue >= Byte.MIN_VALUE && intValue <= Byte.MAX_VALUE) {
                        replaceTo = new SQLTinyIntExpr((byte) intValue);
                    }
                } catch (NumberFormatException ex) {
                }

                if (replaceTo != null && replaceInParent(literal, replaceTo)) {
                    optimizedCount++;
                    return;
                }
            }
        }
        if (dataType.isInt()) {
            if (literal instanceof SQLCharExpr) {
                String text = ((SQLCharExpr) literal).getText();
                SQLExpr replaceTo = null;
                try {
                    text = text.trim();
                    // id = '123' -> id = 123
                    long val = Long.parseLong(text);
                    replaceTo = SQLIntegerExpr.ofIntOrLong(val);

                    if (dataTypeNameHash == FnvHash.Constants.INT && (val < Integer.MIN_VALUE || val > Integer.MAX_VALUE)) {
                        if (x instanceof SQLInListExpr) {
                            ((SQLInListExpr) x).getTargetList().remove(literal);
                        }
                    }
                } catch (NumberFormatException ex) {
                    // id = '123' -> false
                    if (x instanceof SQLBinaryOpExpr) {
                        switch (((SQLBinaryOpExpr) x).getOperator()) {
                            case Equality:
                                if (replaceInParent(x, new SQLBooleanExpr(false))) {
                                    optimizedCount++;
                                }
                                break;
                            default:
                                if (StringUtils.isEmpty(text) || "null".equalsIgnoreCase(text)) {
                                    ((SQLBinaryOpExpr) x).setRight(new SQLNullExpr());
                                    optimizedCount++;
                                } else {
                                    SQLCastExpr literalCast = new SQLCastExpr(literal, new SQLDataTypeImpl(dataType.getName()));
                                    literalCast.setTry(true);
                                    ((SQLBinaryOpExpr) x).setRight(literalCast);
                                    optimizedCount++;
                                }
                                break;
                        }
                        return;
                    } else if (literal.getParent() instanceof SQLInListExpr) {
                        replaceInParent(literal, null);
                        return;
                    } else if ((!(literal.getParent() instanceof SQLBetweenExpr))) {
                        SQLCastExpr nameCast = new SQLCastExpr(name.clone(), new SQLDataTypeImpl("VARCHAR"));
                        replaceInParent(name, nameCast);
                        return;
                    } else if (literal.getParent() instanceof SQLBetweenExpr) {
                        if ("null".equalsIgnoreCase(text)) {
                            replaceInParent(literal, new SQLNullExpr());
                            optimizedCount++;
                        }
                    }
                }

                if (replaceTo != null && replaceInParent(literal, replaceTo)) {
                    optimizedCount++;
                    return;
                }
            } else if (literal instanceof SQLNumberExpr) {
                SQLNumberExpr numberExpr = (SQLNumberExpr) literal;
                Number number = numberExpr.getNumber();
                if (number instanceof BigDecimal) {
                    BigDecimal decimal = (BigDecimal) number;

                    if (decimal.doubleValue() == decimal.longValue()) {
                        BigInteger integer = decimal.toBigInteger();
                        numberExpr.setNumber(integer);
                    }
                }
            } else if (literal instanceof SQLBooleanExpr) {
                SQLBooleanExpr booleanExpr = (SQLBooleanExpr) literal;
                boolean booleanValue = booleanExpr.getBooleanValue();

                SQLIntegerExpr intExpr;

                if (booleanValue) {
                    intExpr = new SQLIntegerExpr(1);
                } else {
                    intExpr = new SQLIntegerExpr(0);
                }

                if (replaceInParent(literal, intExpr)) {
                    optimizedCount++;
                }
            }
        } else if (dataType.isString()) {
            if (literal instanceof SQLIntegerExpr) {
                boolean cmp = false;
                if (x instanceof SQLBinaryOpExpr) {
                    SQLBinaryOperator op = ((SQLBinaryOpExpr) x).getOperator();
                    switch (op) {
                        case Equality:
                        case NotEqual:
                        case LessThanOrGreater:
                        case LessThan:
                        case LessThanOrEqual:
                        case GreaterThan:
                        case GreaterThanOrEqual:
                            cmp = true;
                            break;
                        default:
                            break;
                    }

                } else if (x instanceof SQLInListExpr) {
                    cmp = true;
                } else if (x instanceof SQLBetweenExpr) {
                    cmp = true;
                } else if (x instanceof SQLMethodInvokeExpr
                           && ((SQLMethodInvokeExpr) x).methodNameHashCode64() == FnvHash.Constants.COALESCE) {
                    cmp = true;
                }

                if (cmp) {
                    String text = literal.toString();
                    SQLCharExpr charExpr = new SQLCharExpr(text);
                    if (replaceInParent(literal, charExpr)) {
                        optimizedCount++;
                    }
                } else {
                    SQLCastExpr cast = new SQLCastExpr(name.clone(), SQLIntegerExpr.DATA_TYPE.clone());
                    if (replaceInParent(name, cast)) {
                        optimizedCount++;
                    }
                }
            } else if (literal instanceof SQLBooleanExpr) {
                String text = "false";

                if (((SQLBooleanExpr) literal).getBooleanValue()) {
                    text = "true";
                }

                SQLCharExpr charExpr = new SQLCharExpr(text);
                if (replaceInParent(literal, charExpr)) {
                    optimizedCount++;
                }
            } else if (literal instanceof SQLNumberExpr) {
                if (((SQLNumberExpr) literal).getNumber() instanceof BigDecimal) {

                    SQLCharExpr charExpr = new SQLCharExpr(literal.toString());
                    if (replaceInParent(literal, charExpr)) {
                        optimizedCount++;
                    }
                }

            }
        } else if (dataTypeNameHash == FnvHash.Constants.DATE) {
            String text = null;
            if (literal instanceof SQLCharExpr) {
                text = ((SQLCharExpr) literal).getText();
            } else if (literal instanceof SQLIntegerExpr) {
                text = literal.toString();
            }

            if (text != null) {
                if (text.length() == 0) {
                    replaceInParent(literal, new SQLNullExpr());
                    return;
                }

                if (text.length() == 21 && text.endsWith(" 00:00:00.0")) {
                    text = text.substring(0, 10);
                } else if (text.length() == 19 && text.endsWith(" 00:00:00")) {
                    text = text.substring(0, 10);
                } else if (text.length() == 7 && text.charAt(4) == '-') {
                    replaceInParent(x, new SQLBooleanExpr(false));
                    return;
                }

                SQLExpr replaceTo = null;
                if (text.length() == 19 || text.length() == 21) {
                    replaceTo = new SQLTimestampExpr(text);
                } else {
                    replaceTo = new SQLDateExpr(text);
                }

                replaceInParent(literal, replaceTo);
            }
        } else if (dataTypeNameHash == FnvHash.Constants.TIMESTAMP) {
            if (repalceToTimestamp(literal, timeZone)) {
                optimizedCount++;
            }
        } else if (dataTypeNameHash == FnvHash.Constants.TIME) {
            if (literal instanceof SQLCharExpr) {
                String text = ((SQLCharExpr) literal).getText();
                SQLTimeExpr timeExpr = new SQLTimeExpr(text);
                if (replaceInParent(literal, timeExpr)) {
                    optimizedCount++;
                }
            }
        } else if (dataTypeNameHash == FnvHash.Constants.DOUBLE) {
            if (literal instanceof SQLNumberExpr) {
                double value = ((SQLNumberExpr) literal).getNumber().doubleValue();
                SQLNumberExpr numberExpr = new SQLNumberExpr(value);
                if (replaceInParent(literal, numberExpr)) {
                    optimizedCount++;
                }
            } else if (literal instanceof SQLCharExpr) {
                String chars = ((SQLCharExpr) literal).getText();

                try {
                    double value = Double.parseDouble(chars);
                    SQLNumberExpr numberExpr = new SQLNumberExpr(value);
                    if (replaceInParent(literal, numberExpr)) {
                        optimizedCount++;
                    }
                } catch (NumberFormatException e) {
                    if (x instanceof SQLBinaryOpExpr && "null".equalsIgnoreCase(chars) && (
                            ((SQLBinaryOpExpr) x).getOperator() == SQLBinaryOperator.NotEqual
                            || ((SQLBinaryOpExpr) x).getOperator() == SQLBinaryOperator.LessThanOrGreater)) {
                        replaceInParent(literal, new SQLNullExpr());
                    } else {
                        SQLBooleanExpr booleanExpr = new SQLBooleanExpr(false);
                        replaceInParent(x, booleanExpr);
                    }
                }

            } else if (literal instanceof SQLBooleanExpr) {
                boolean booleanValue = ((SQLBooleanExpr) literal).getBooleanValue();

                SQLNumberExpr numberExpr = new SQLNumberExpr(0.0);

                if (booleanValue) {
                    numberExpr = new SQLNumberExpr(1.0);
                }
                replaceInParent(literal, numberExpr);
            }
        } else if (dataTypeNameHash == FnvHash.Constants.FLOAT) {
            if (literal instanceof SQLNumberExpr) {
                float value = ((SQLNumberExpr) literal).getNumber().floatValue();
                SQLNumberExpr numberExpr = new SQLNumberExpr(value);
                if (replaceInParent(literal, numberExpr)) {
                    optimizedCount++;
                }
            } else if (literal instanceof SQLCharExpr) {
                String chars = ((SQLCharExpr) literal).getText();

                try {
                    float value = Float.parseFloat(chars);
                    SQLNumberExpr numberExpr = new SQLNumberExpr(value);
                    if (replaceInParent(literal, numberExpr)) {
                        optimizedCount++;
                    }
                } catch (NumberFormatException e) {
                    if (x instanceof SQLBinaryOpExpr && "null".equalsIgnoreCase(chars) && (
                            ((SQLBinaryOpExpr) x).getOperator() == SQLBinaryOperator.NotEqual
                            || ((SQLBinaryOpExpr) x).getOperator() == SQLBinaryOperator.LessThanOrGreater)) {
                        replaceInParent(literal, new SQLNullExpr());
                    } else {
                        SQLBooleanExpr booleanExpr = new SQLBooleanExpr(false);
                        if (replaceInParent(literal, booleanExpr)) {
                            optimizedCount++;
                        }
                    }
                }
            } else if (literal instanceof SQLBooleanExpr) {
                boolean booleanValue = ((SQLBooleanExpr) literal).getBooleanValue();

                SQLNumberExpr numberExpr = new SQLNumberExpr(0.0);

                if (booleanValue) {
                    numberExpr = new SQLNumberExpr(1.0);
                }

                if (replaceInParent(literal, numberExpr)) {
                    optimizedCount++;
                }
            }
        } else if (dataTypeNameHash == FnvHash.Constants.DECIMAL) {
            if (literal instanceof SQLCharExpr) {
                String chars = ((SQLCharExpr) literal).getText();

                try {
                    BigDecimal value = new BigDecimal(chars);
                    SQLNumberExpr numberExpr = new SQLNumberExpr(value);
                    if (replaceInParent(literal, numberExpr)) {
                        optimizedCount++;
                    }
                } catch (NumberFormatException e) {
                    if (x instanceof SQLBinaryOpExpr && "null".equalsIgnoreCase(chars) && (
                            ((SQLBinaryOpExpr) x).getOperator() == SQLBinaryOperator.NotEqual
                            || ((SQLBinaryOpExpr) x).getOperator() == SQLBinaryOperator.LessThanOrGreater)) {
                        replaceInParent(literal, new SQLNullExpr());
                    } else {
                        SQLBooleanExpr booleanExpr = new SQLBooleanExpr(false);
                        if (replaceInParent(literal, booleanExpr)) {
                            optimizedCount++;
                        }
                    }
                }
            } else if (literal instanceof SQLBooleanExpr) {
                boolean booleanValue = ((SQLBooleanExpr) literal).getBooleanValue();

                SQLNumberExpr numberExpr = new SQLNumberExpr(0.0);

                if (booleanValue) {
                    numberExpr = new SQLNumberExpr(1.0);
                }

                if (replaceInParent(literal, numberExpr)) {
                    optimizedCount++;
                }
            }
        } else if (dataTypeNameHash == FnvHash.Constants.BOOLEAN) {
            if (literal instanceof SQLIntegerExpr) {
                int val = ((SQLIntegerExpr) literal).getNumber().intValue();
                SQLBooleanExpr bool = null;
                if (val == 1) {
                    bool = new SQLBooleanExpr(true);
                } else if (val == 0) {
                    bool = new SQLBooleanExpr(false);
                }

                if (bool != null && replaceInParent(literal, bool)) {
                    optimizedCount++;
                }
            }
        }

    }

    public static boolean repalceToTimestamp(SQLLiteralExpr literal, TimeZone timeZone) {
        String text = null;
        if (literal instanceof SQLDateExpr) {
            text = ((SQLDateExpr) literal).getLiteral();
        } else if (literal instanceof SQLCharExpr) {
            text = ((SQLCharExpr) literal).getText();
        } else if (literal instanceof SQLTimestampExpr) {
            String str = ((SQLTimestampExpr) literal).getLiteral();
            if (str.length() < 19) {
                text = str;
            }
        }

        if (text == null || "".equals(text.trim())) {
            return false;
        }

        text = text.trim();

        if (text.length() == 19) {
            SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
            try {
                if (timeZone != null) {
                    dateFormat.setTimeZone(timeZone);
                }
                Date date = dateFormat.parse(text);
                text = dateFormat.format(date);
                return SQLUtils.replaceInParent(literal, new SQLTimestampExpr(text));
            } catch (ParseException e) {
                // skip
            }
        } else if (text.length() == 20 && text.indexOf("  ") > 0) {
            SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd  HH:mm:ss");
            try {
                if (timeZone != null) {
                    dateFormat.setTimeZone(timeZone);
                }
                Date date = dateFormat.parse(text);
                text = dateFormat.format(date);
                SQLTimestampExpr replaceTo = new SQLTimestampExpr(date, timeZone);
                return SQLUtils.replaceInParent(literal, replaceTo);
            } catch (ParseException e) {
                // skip
            }
        }

        if (text.length() < 19) {
            Date date = MySqlUtils.parseDate(text, timeZone);
            if (date == null && text.length() == 16 && text.charAt(13) == ':') {
                date = MySqlUtils.parseDate(text + ":00", timeZone);
            }
            if (date != null) {
                SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
                if (timeZone != null) {
                    dateFormat.setTimeZone(timeZone);
                }
                text = dateFormat.format(date);
                return SQLUtils.replaceInParent(literal, new SQLTimestampExpr(text));
            }
        }

        int dotIndex = text.indexOf('.');
        if (dotIndex == -1) {
            Date date = MySqlUtils.parseDate(text, timeZone);
            if (date != null) {
                SQLTimestampExpr replaceTo = new SQLTimestampExpr(date, timeZone);
                return SQLUtils.replaceInParent(literal, replaceTo);
            }
        } else if (dotIndex == 19 && text.length() > 23) {
            // 2019-02-09 01:00:00.783820, presto 不支持 6位精度
            text = text.substring(0,23);

            SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSS");
            if (timeZone != null) {
                dateFormat.setTimeZone(timeZone);
            }

            Date date = MySqlUtils.parseDate(text,timeZone);
            text = dateFormat.format(date);
        }

        if (text.length() == 20 && text.endsWith(":000")) {
            text = text.substring(0, text.length() - 1);
        } else if (text.length() == 22 && text.endsWith(":00:00:00")) {
            text = text.substring(0, text.length() - 3);
        }

        return SQLUtils.replaceInParent(literal, new SQLTimestampExpr(text));
    }

    public boolean replaceInParent(SQLExpr x, SQLExpr target) {
        boolean result = SQLUtils.replaceInParent(x, target);
        if (!result) {
            repalceFailCount++;
        } else {
            optimizedCount++;
        }
        return result;
    }
}
