package com.alibaba.fastsql.sql.optimizer.rules;

import com.alibaba.fastsql.DbType;
import com.alibaba.fastsql.funtions.FunctionDef;
import com.alibaba.fastsql.funtions.FunctionRepository;
import com.alibaba.fastsql.sql.SQLUtils;
import com.alibaba.fastsql.sql.ast.*;
import com.alibaba.fastsql.sql.ast.expr.*;
import com.alibaba.fastsql.sql.ast.statement.*;
import com.alibaba.fastsql.sql.dialect.mysql.ast.statement.MySqlInsertStatement;
import com.alibaba.fastsql.sql.parser.SQLExprParser;
import com.alibaba.fastsql.sql.parser.SQLParser;
import com.alibaba.fastsql.sql.repository.SchemaObject;
import com.alibaba.fastsql.util.FnvHash;
import com.alibaba.fastsql.util.MySqlUtils;

import java.math.BigDecimal;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.TimeZone;

import static com.alibaba.fastsql.sql.ast.expr.SQLBinaryOperator.Equality;
import static com.alibaba.fastsql.sql.ast.expr.SQLBinaryOperator.NotEqual;

public class TypeInference extends OptimizerVisitor {
    private FunctionRepository functionRepository = FunctionRepository.MY_SQL;

    public TypeInference() {

    }

    public TypeInference(DbType dbType, TimeZone timeZone) {
        this.dbType = dbType;
        this.timeZone = timeZone;
    }


    public TypeInference(TimeZone timeZone) {
        this.timeZone = timeZone;
    }

    public boolean visit(MySqlInsertStatement x) {
        return visit((SQLInsertStatement) x);
    }

    public boolean visit(SQLInsertStatement x) {
        {
            SQLExprTableSource tableSource = x.getTableSource();
            if (tableSource != null) {
                tableSource.accept(this);
            }

            List<SQLExpr> columns = x.getColumns();
            for (SQLExpr column : columns) {
                column.accept(this);
            }

            List<SQLInsertStatement.ValuesClause> valuesList = x.getValuesList();
            for (SQLInsertStatement.ValuesClause item : valuesList) {
                item.accept(this);
            }

            SQLSelect query = x.getQuery();
            if (query != null) {
                query.accept(this);
            }
        }

        SQLSelect select = x.getQuery();
        if (select == null) {
            return false;
        }

        List<SQLDataType> dataTypes = new ArrayList<SQLDataType>();
        List<SQLExpr> columns = x.getColumns();
        if (columns.size() > 0) {
            for (SQLExpr column : columns) {
                dataTypes.add(column.computeDataType());
            }
        } else {
            SchemaObject schemaObject = x.getTableSource().getSchemaObject();
            if (schemaObject != null && schemaObject.getStatement() instanceof SQLCreateTableStatement) {
                SQLCreateTableStatement ddl = (SQLCreateTableStatement) schemaObject.getStatement();
                List<SQLColumnDefinition> columnDefinitions = ddl.getColumnDefinitions();
                for (SQLColumnDefinition columnDefinition : columnDefinitions) {
                    dataTypes.add(columnDefinition.getDataType());
                }
            }
        }

        inference(dataTypes, select.getQuery());

        return false;
    }

    private void inference(List<SQLDataType> dataTypes, SQLSelectQuery query) {
        if (query instanceof SQLSelectQueryBlock) {
            inference(dataTypes, (SQLSelectQueryBlock) query);
        } else if (query instanceof SQLUnionQuery) {
            // TODO
        }
    }

    private void inference(List<SQLDataType> dataTypes, SQLSelectQueryBlock queryBlock) {
        List<SQLSelectItem> selectList = queryBlock.getSelectList();

        for (int i = 0; i < selectList.size() && i < dataTypes.size(); i++) {
            SQLDataType dataType = dataTypes.get(i);
            if (dataType == null) {
                continue;
            }

            SQLSelectItem selectItem = selectList.get(i);
            SQLExpr expr = selectItem.getExpr();
            SQLDataType exprDataType = expr.computeDataType();
            if (exprDataType == null) {
                continue;
            }
            final long exprDataTypeHash = exprDataType.nameHashCode64();

            SQLExpr replaceTo = null;
            long dataTypeHash = dataType.nameHashCode64();

            if (exprDataTypeHash == dataTypeHash) {
                continue;
            }


            if (dataTypeHash == FnvHash.Constants.DATETIME) {
                if (expr instanceof SQLCharExpr) {
                    replaceTo = new SQLTimestampExpr(((SQLCharExpr) expr).getText());
                } else if (exprDataType.isInt() || exprDataType.isString()) {
                    replaceTo = new SQLCastExpr(expr.clone(), dataType.clone());
                }
            } else if (dataTypeHash == FnvHash.Constants.BIGINT
                    || dataTypeHash == FnvHash.Constants.INT
                    || dataTypeHash == FnvHash.Constants.INTEGER
                    || dataTypeHash == FnvHash.Constants.SMALLINT
                    || dataTypeHash == FnvHash.Constants.TINYINT) {
                if (expr instanceof SQLCharExpr) {
                    String text = ((SQLCharExpr) expr).getText();
                    try {
                        long value = Long.parseLong(text);
                        replaceTo = SQLIntegerExpr.ofIntOrLong(value);
                    } catch (NumberFormatException x) {
                        // skip
                    }
                } else if (exprDataTypeHash == FnvHash.Constants.VARCHAR
                        || exprDataTypeHash == FnvHash.Constants.CHAR
                        || exprDataTypeHash == FnvHash.Constants.FLOAT
                        || exprDataTypeHash == FnvHash.Constants.DOUBLE
                        || exprDataTypeHash == FnvHash.Constants.NUMBER) {
                    replaceTo = new SQLCastExpr(expr.clone(), dataType.clone());
                } else if (exprDataTypeHash == FnvHash.Constants.BIGINT
                        && (dataTypeHash == FnvHash.Constants.INT
                            || dataTypeHash == FnvHash.Constants.INTEGER
                            || dataTypeHash == FnvHash.Constants.SMALLINT
                            || dataTypeHash == FnvHash.Constants.TINYINT)) {
                    replaceTo = new SQLCastExpr(expr.clone(), dataType.clone());
                }
            } else if (dataTypeHash == FnvHash.Constants.CHAR || dataTypeHash == FnvHash.Constants.VARCHAR) {
                if (expr instanceof SQLNumericLiteralExpr) {
                    replaceTo = new SQLCharExpr(expr.toString());
                } else if (exprDataType.isInt()) {
                    replaceTo = new SQLCastExpr(expr.clone(), dataType.clone());
                }
            }

            if (replaceTo != null) {
                selectItem.setExpr(replaceTo);
            }
        }
    }

    public boolean visit(SQLBinaryOpExpr x) {
        if (isGroup(x)) {
            return super.visit(x);
        }

        final SQLExpr left = x.getLeft();
        final SQLExpr right = x.getRight();

        left.accept(this);
        right.accept(this);

        SQLDataType leftDataType = left.computeDataType();
        final SQLDataType rightDataType = right.computeDataType();

        if (leftDataType == null) {
            return false;
        }

        final long leftTypeHash = leftDataType.nameHashCode64();
        long rightTypeHash = rightDataType != null
                ? rightDataType.nameHashCode64()
                : 0;

        switch (x.getOperator()) {
            case Like:
            case NotLike:
                if (leftDataType.isInt()
                        || leftTypeHash == FnvHash.Constants.FLOAT
                        || leftTypeHash == FnvHash.Constants.DOUBLE
                        || leftTypeHash == FnvHash.Constants.NUMBER
                        || leftTypeHash == FnvHash.Constants.DECIMAL
                        || leftTypeHash == FnvHash.Constants.DATE
                        || leftTypeHash == FnvHash.Constants.DATETIME
                        || leftTypeHash == FnvHash.Constants.TIMESTAMP
                        || leftTypeHash == FnvHash.Constants.TIME) {
                    x.setLeft(new SQLCastExpr(left, new SQLCharacterDataType("CHAR")));
                }
                break;
            case Add:
            case Subtract:
                if (leftTypeHash == FnvHash.Constants.CHAR || leftTypeHash == FnvHash.Constants.VARCHAR) {
                    if (right instanceof SQLIntervalExpr) {
                        if (left instanceof SQLCharExpr) {
                            String text = ((SQLCharExpr) left).getText();
                            Date date = MySqlUtils.parseDate(text, timeZone);
                            if (date == null) {
                                return false;
                            }
                            SQLTimestampExpr ts = new SQLTimestampExpr(date, timeZone);
                            x.setLeft(ts);
                        } else {
                            SQLCastExpr castExpr = new SQLCastExpr(left.clone(), new SQLDataTypeImpl("TIMESTAMP"));
                            x.setLeft(castExpr);
                        }
                        break;
                    } else if (rightDataType != null && rightDataType.isInt()) {
                        SQLCastExpr castExpr = new SQLCastExpr(left.clone(), new SQLDataTypeImpl("BIGINT"));
                        x.setLeft(castExpr);
                    }
                } else if (leftDataType != null && leftDataType.isInt()) {
                    if (rightTypeHash == FnvHash.Constants.TIMESTAMP && dbType == DbType.mysql) {
                        SQLExpr replaceTo = null;
                        if (right instanceof SQLTimestampExpr) {
                            Date date = ((SQLTimestampExpr) right).getDate(timeZone);
                            if (date != null) {
                                SimpleDateFormat format = new SimpleDateFormat("yyyyMMddHHmmss");
                                if (timeZone != null) {
                                    format.setTimeZone(timeZone);
                                }
                                String dateStr = format.format(date);
                                long intVal = Long.parseLong(dateStr);


                                if (left instanceof SQLIntegerExpr) {
                                    Number number = ((SQLIntegerExpr) left).getNumber();
                                    if (number instanceof Integer || number instanceof Long) {
                                        long leftValue = number.longValue();

                                        long result;
                                        if (x.getOperator() == SQLBinaryOperator.Add) {
                                            result = leftValue + intVal;
                                        } else {
                                            result = leftValue - intVal;
                                        }
                                        if (SQLUtils.replaceInParent(x, new SQLIntegerExpr(result))) {
                                            return false;
                                        }
                                    }
                                }

                                replaceTo = new SQLIntegerExpr(intVal);
                            }
                        } else {
                            SQLMethodInvokeExpr chars = new SQLMethodInvokeExpr("date_format", null, right.clone(), new SQLCharExpr("%Y%m%d%H%i%s"));
                            replaceTo = new SQLCastExpr(chars, new SQLDataTypeImpl("BIGINT"));
                        }

                        if (replaceTo != null) {
                            x.setRight(replaceTo);
                        }
                    }
                }

                if (leftDataType.isString()
                        && (!(left instanceof SQLTextLiteralExpr))
                        && rightTypeHash == FnvHash.Constants.NUMBER) {
                    SQLCastExpr castExpr = new SQLCastExpr(left.clone(), SQLNumberExpr.DATA_TYPE_DOUBLE.clone());
                    x.setLeft(castExpr);
                    break;
                }

                break;
            case Multiply:
            case Divide:
            case DIV: {
                if (leftDataType.isString()
                        && (!(left instanceof SQLTextLiteralExpr))) {
                    if (rightTypeHash == FnvHash.Constants.NUMBER) {
                        SQLCastExpr castExpr = new SQLCastExpr(left.clone(), SQLNumberExpr.DATA_TYPE_DOUBLE.clone());
                        x.setLeft(castExpr);
                        optimizedCount++;
                    } else if (right instanceof SQLIntegerExpr) {
                        SQLCastExpr castExpr = new SQLCastExpr(left.clone(), SQLIntegerExpr.DATA_TYPE.clone());
                        x.setLeft(castExpr);
                        optimizedCount++;
                    }
                    break;
                } else if (leftDataType.isString()
                        && left instanceof SQLTextLiteralExpr) {
                    if (rightDataType.isInt() || rightTypeHash == FnvHash.Constants.NUMBER) {
                        SQLCastExpr castExpr = new SQLCastExpr(left.clone(), SQLNumberExpr.DATA_TYPE_DOUBLE.clone());
                        x.setLeft(castExpr);
                        optimizedCount++;
                    }
                }
            }
                break;
            case LessThan:
            case LessThanOrEqual:
            case GreaterThanOrEqual:
            case GreaterThan:
            case Equality:
            case NotEqual:
            case LessThanOrGreater:
                if (leftDataType != null) {
                    long leftDateTypeHash = leftDataType.nameHashCode64();
                    if (leftDateTypeHash == FnvHash.Constants.INT
                            || leftDateTypeHash == FnvHash.Constants.INTEGER
                            || leftDateTypeHash == FnvHash.Constants.BIGINT) {
                        if (right instanceof SQLCharExpr) {
                            String text = ((SQLCharExpr) right).getText();
                            try {
                                long val = Long.parseLong(text);
                                x.setRight(new SQLIntegerExpr(val));
                                optimizedCount++;
                            } catch (NumberFormatException ex) {
                                // skip
                            }
                        } else if (rightDataType != null && rightDataType.isString() && right instanceof SQLName) {
                            SQLCastExpr cast = new SQLCastExpr(right.clone(), leftDataType.clone());
                            cast.setTry(true);
                            x.setRight(cast);
                        }
                    } else if (leftDateTypeHash == FnvHash.Constants.DOUBLE
                            || leftDateTypeHash == FnvHash.Constants.FLOAT
                            || leftDateTypeHash == FnvHash.Constants.DECIMAL
                            || leftDateTypeHash == FnvHash.Constants.NUMBER
                            || leftDateTypeHash == FnvHash.Constants.NUMERIC) {
                        if (right instanceof SQLCharExpr) {
                            String text = ((SQLCharExpr) right).getText();
                            try {
                                if (text.indexOf('.') == -1 && text.indexOf('e') == -1 && text.indexOf('E') == -1) {
                                    long val = Long.parseLong(text);
                                    x.setRight(new SQLNumberExpr((double) val));
                                    optimizedCount++;
                                }
                            } catch (NumberFormatException ex) {
                                // skip
                            }
                        }
                    } else if (leftDateTypeHash == FnvHash.Constants.CHAR
                            || leftDateTypeHash == FnvHash.Constants.VARCHAR) {
                        if (right instanceof SQLNumericLiteralExpr) {
                            if (x.getOperator() != NotEqual && x.getOperator() != Equality) {
                                Number number = ((SQLNumericLiteralExpr) right).getNumber();
                                SQLCastExpr cast = null;
                                if (number instanceof Integer
                                        || number instanceof Long) {
                                    cast = new SQLCastExpr(left.clone(), new SQLDataTypeImpl(SQLDataTypeImpl.Constants.BIGINT));
                                } else if (number instanceof Float
                                        || number instanceof Double
                                        || number instanceof BigDecimal) {
                                    cast = new SQLCastExpr(left.clone(), new SQLDataTypeImpl(SQLDataTypeImpl.Constants.DOUBLE));
                                }
                                if (cast != null) {
                                    cast.setTry(true);
                                    x.setLeft(cast);
                                    optimizedCount++;
                                }
                            } else if(left instanceof SQLAggregateExpr) {

                            } else {
                                String str = right.toString();
                                SQLCharExpr charExpr = new SQLCharExpr(str);
                                x.setRight(charExpr);
                                optimizedCount++;
                            }
                        }
                    } else if (leftDateTypeHash == FnvHash.Constants.BOOLEAN) {
                        if (right instanceof SQLCharExpr) {
                            String str = ((SQLCharExpr) right).getText();
                            if ("true".equalsIgnoreCase(str) || "1".equals(str)) {
                                SQLBooleanExpr booleanExpr = new SQLBooleanExpr(true);
                                x.setRight(booleanExpr);
                                optimizedCount++;
                            } else if ("false".equalsIgnoreCase(str) || "0".equals(str)) {
                                SQLBooleanExpr booleanExpr = new SQLBooleanExpr(false);
                                x.setRight(booleanExpr);
                                optimizedCount++;
                            }
                        } else if (right instanceof SQLIntegerExpr) {
                            Number num = ((SQLIntegerExpr) right).getNumber();
                            if (num.equals(1)) {
                                SQLBooleanExpr booleanExpr = new SQLBooleanExpr(true);
                                x.setRight(booleanExpr);
                                optimizedCount++;
                            } else if (num.equals(0)) {
                                SQLBooleanExpr booleanExpr = new SQLBooleanExpr(false);
                                x.setRight(booleanExpr);
                                optimizedCount++;
                            }
                        }
                    } else if (leftDateTypeHash == FnvHash.Constants.TIME) {
                        if (right instanceof SQLCharExpr) {
                            String str = ((SQLCharExpr) right).getText();
                            x.setRight(new SQLTimeExpr(str));
                            optimizedCount++;
                        }
                    }
                }
                break;
            default:
                break;
        }


        return false;
    }

    public boolean visit(SQLAggregateExpr x) {
        return visit((SQLMethodInvokeExpr) x);
    }

    public boolean visit(SQLMethodInvokeExpr x) {
        final long hash = x.methodNameHashCode64();

        if (x.getResolvedReturnDataType() == null && (hash == FnvHash.Constants.DATEDIFF || hash == FnvHash.Constants.DATE_DIFF)) {
            x.setResolvedReturnDataType(SQLIntegerExpr.DATA_TYPE);
        }

        if (x.methodNameHashCode64() == FnvHash.Constants.IF) {
            func_if(x);
            return true;
        }

        FunctionDef functionDef = findFunction(hash);

        final List<SQLExpr> arguments = x.getArguments();
        if (functionDef != null) {
            if ((!functionDef.isVarArgs()) && arguments.size() > functionDef.getParameterTypes().size()) {
                return true;
            }

            for (int i = 0; i < arguments.size(); i++) {
                SQLExpr replaceTo = null;

                final SQLDataType type = functionDef.getParemeterType(i);
                final SQLExpr arg = arguments.get(i);
                SQLDataType argDataType = arg.computeDataType();

                if (type == null) {
                    continue;
                }



                if (FnvHash.Constants.CHAR == type.nameHashCode64()) {
                    if (arg instanceof SQLIntegerExpr) {
                        String chars = ((SQLIntegerExpr) arg).getNumber().toString();
                        SQLCharExpr charExpr = new SQLCharExpr(chars);
                        charExpr.setParent(x);
                        arguments.set(i, charExpr);
                        optimizedCount++;
                        continue;
                    } else if (arg instanceof SQLTimestampExpr) {
                        String chars = ((SQLTimestampExpr) arg).getLiteral();
                        SQLCharExpr charExpr = new SQLCharExpr(chars);
                        charExpr.setParent(x);
                        arguments.set(i, charExpr);
                        optimizedCount++;
                        continue;
                    }

                    if (argDataType == null) {
                        if (arg instanceof SQLMethodInvokeExpr) {
                            SQLMethodInvokeExpr argFunction = (SQLMethodInvokeExpr) arg;
                            FunctionDef paramFunctionDef = findFunction(argFunction.methodNameHashCode64());
                            if (paramFunctionDef != null) {
                                argDataType = paramFunctionDef.getReturnType();
                            } else if (argFunction.methodNameHashCode64() == FnvHash.Constants.ROUND) {
                                argDataType = SQLNumberExpr.DATA_TYPE_NUMBER; // number or date or timestamp
                            }
                        }
                    }

                    if (argDataType != null) {
                        long argTypeHash = argDataType.nameHashCode64();
                        if (argTypeHash == FnvHash.Constants.TINYINT
                                || argTypeHash == FnvHash.Constants.SMALLINT
                                || argTypeHash == FnvHash.Constants.INT
                                || argTypeHash == FnvHash.Constants.INTEGER
                                || argTypeHash == FnvHash.Constants.BIGINT
                                || argTypeHash == FnvHash.Constants.FLOAT
                                || argTypeHash == FnvHash.Constants.DOUBLE
                                || argTypeHash == FnvHash.Constants.NUMBER
                                || argTypeHash == FnvHash.Constants.DECIMAL
                                || argTypeHash == FnvHash.Constants.DATE
                                || argTypeHash == FnvHash.Constants.DATETIME
                                || argTypeHash == FnvHash.Constants.TIMESTAMP
                                || argTypeHash == FnvHash.Constants.TIME) {
                            replaceTo = new SQLCastExpr(arg, new SQLCharacterDataType("VARCHAR"));
                        }
                    }
                } else if (FnvHash.Constants.TIMESTAMP == type.nameHashCode64()) {

                    if (arg instanceof SQLCharExpr) {
                        final String text = ((SQLCharExpr) arg).getText();
                        if (text.length() == 10) {
                            Date date = MySqlUtils.parseDate(text, timeZone);
                            if(date != null) {
                                replaceTo = new SQLDateExpr(text);
                            }
                        } else if (text.length() > 10) {
                            Date date = MySqlUtils.parseDate(text, timeZone);
                            if (date != null) {
                                replaceTo = new SQLTimestampExpr(text);
                            }
                        }
                    } else if (argDataType != null) {
                        long argTypeHash = argDataType.nameHashCode64();
                        if (argTypeHash == FnvHash.Constants.VARCHAR
                                || argTypeHash == FnvHash.Constants.CHAR) {
                            SQLCastExpr castExpr = new SQLCastExpr(arg, new SQLCharacterDataType("TIMESTAMP"));
                            castExpr.setParent(x);
                            arguments.set(i, castExpr);
                            optimizedCount++;
                        }
                    }

                } else if (FnvHash.Constants.DATE == type.nameHashCode64()) {

                    if (arg instanceof SQLCharExpr && ((SQLCharExpr) arg).getText().length() > 10) {
                        replaceTo = new SQLTimestampExpr(((SQLCharExpr) arg).getText());
                    } else if (argDataType != null) {
                        long argTypeHash = argDataType.nameHashCode64();
                        if (argTypeHash == FnvHash.Constants.VARCHAR
                                || argTypeHash == FnvHash.Constants.CHAR) {
                            SQLCastExpr castExpr = new SQLCastExpr(arg, new SQLCharacterDataType("DATE"));
                            castExpr.setParent(x);
                            arguments.set(i, castExpr);
                            optimizedCount++;
                        }
                    }

                } else if (FnvHash.Constants.BIGINT == type.nameHashCode64()) {
                    if (arg instanceof SQLCharExpr) {
                        String text = ((SQLCharExpr) arg).getText();
                        try {
                            long intVal = Long.parseLong(text);
                            replaceTo = new SQLIntegerExpr(intVal);
                        } catch (NumberFormatException ex) {
                            // skip
                        }

                    }

                } else if (FnvHash.Constants.VARBINARY == type.nameHashCode64()) {
                    if (argDataType != null) {
                        long argTypeHash = argDataType.nameHashCode64();
                        if (argTypeHash == FnvHash.Constants.VARCHAR
                                || argTypeHash == FnvHash.Constants.CHAR) {
                            SQLCastExpr castExpr = new SQLCastExpr(arg, new SQLCharacterDataType("VARBINARY"));
                            castExpr.setParent(x);
                            arguments.set(i, castExpr);
                            optimizedCount++;
                        }
                    }
                }

                if (replaceTo != null) {
                    replaceTo.setParent(x);
                    arguments.set(i, replaceTo);
                    optimizedCount++;
                }
            }

            return true;
        }

        return true;
    }

    public FunctionDef findFunction(long nameHash) {
        return functionRepository.find(nameHash);
    }

    public void func_if(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 3) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);
        SQLExpr arg2 = arguments.get(2);

        SQLDataType trueExprType = arg1.computeDataType();
        if (trueExprType == null) {
            return;
        }

        SQLDataType falseExprType = arg2.computeDataType();
        if (falseExprType == null) {
            return;
        }

        if (trueExprType.isInt()) {
            if (falseExprType.isString()) {
                SQLCastExpr cast = new SQLCastExpr(arg2.clone(), trueExprType.clone());
                cast.setTry(true);
                x.setArgument(2, cast);
            }
        }
    }

    public boolean visit(SQLCaseExpr x) {
        List<SQLCaseExpr.Item> items = x.getItems();
        if (items.size() == 0) {
            return true;
        }

        if (x.getValueExpr() == null) {
            for (SQLCaseExpr.Item item : items) {
                SQLExpr conditionExpr = item.getConditionExpr();

                if (conditionExpr instanceof SQLIntegerExpr) {
                    Number num = ((SQLIntegerExpr) conditionExpr).getNumber();
                    if (num instanceof Integer && num.intValue() == 1) {
                        item.setConditionExpr(new SQLBooleanExpr(true));
                        continue;
                    } else if (num instanceof Integer && num.intValue() == 0) {
                        item.setConditionExpr(new SQLBooleanExpr(false));
                        continue;
                    }
                }

                SQLDataType conditionDataType = conditionExpr.computeDataType();
                if (conditionDataType != null && conditionDataType.isInt()) {
                    SQLCastExpr cast = new SQLCastExpr(conditionExpr.clone(), conditionDataType.clone());
                    item.setConditionExpr(cast);
                }
            }
        }

        SQLExpr firstValueExpr = items.get(0).getValueExpr();
        if (firstValueExpr == null) {
            return true;
        }

        SQLDataType firstValueDataType = firstValueExpr.computeDataType();
        if (firstValueDataType == null) {
            return true;
        }

        for (int i = 1; i < items.size(); i++) {
            SQLCaseExpr.Item item = items.get(i);
            SQLExpr valueExpr = item.getValueExpr();
            if (valueExpr == null) {
                continue;
            }

            SQLDataType valueDataType = valueExpr.computeDataType();
            if (valueDataType == null) {
                continue;
            }

            if (firstValueDataType.isString()) {
                if (valueDataType.isString()) {
                    continue;
                }

                if (valueDataType.nameHashCode64() == FnvHash.Constants.TIMESTAMP) {
                    SQLCastExpr castExpr = new SQLCastExpr(valueExpr.clone(), firstValueDataType.clone());
                    items.get(i).setValueExpr(castExpr);
                    optimizedCount++;
                }
            } else if (firstValueDataType.nameHashCode64() == FnvHash.Constants.TIMESTAMP) {
                if (valueExpr instanceof SQLCharExpr) {
                    item.setValueExpr(
                            new SQLTimestampExpr(((SQLCharExpr) valueExpr).getText())
                    );
                    optimizedCount++;
                } else if (valueDataType.isString()) {
                    SQLCastExpr castExpr = new SQLCastExpr(valueExpr.clone(), firstValueDataType.clone());
                    x.setElseExpr(castExpr);
                    optimizedCount++;
                }
            }
        }

        SQLExpr elseExpr = x.getElseExpr();
        if (elseExpr == null) {
            return true;
        }

        SQLDataType elseDataType = elseExpr.computeDataType();
        if (elseDataType == null) {
            return true;
        }

        if (firstValueDataType.isString()) {
            if (elseDataType.nameHashCode64() == FnvHash.Constants.TIMESTAMP) {
                SQLCastExpr castExpr = new SQLCastExpr(elseExpr.clone(), firstValueDataType.clone());
                x.setElseExpr(castExpr);
                optimizedCount++;
            }
        } else if (firstValueDataType.nameHashCode64() == FnvHash.Constants.TIMESTAMP) {
            if (elseExpr instanceof SQLCharExpr) {
                x.setElseExpr(
                        new SQLTimestampExpr(((SQLCharExpr) elseExpr).getText())
                );
                optimizedCount++;
            } else if (elseDataType.isString()) {
                SQLCastExpr castExpr = new SQLCastExpr(elseExpr.clone(), firstValueDataType.clone());
                x.setElseExpr(castExpr);
                optimizedCount++;
            }
        }

        return true;
    }

    public boolean visit(SQLInListExpr x) {
        List<SQLExpr> list = x.getTargetList();

        if (list.size() == 0) {
            return true;
        }

        SQLDataType firstDataType = list.get(0).computeDataType();
        if (firstDataType == null) {
            return true;
        }

        for (int i = 1; i < list.size(); i++) {
            SQLDataType itemDataType = list.get(i).computeDataType();
            if (itemDataType == null || !itemDataType.equals(firstDataType)) {
                return true;
            }
        }

        SQLExpr expr = x.getExpr();

        if (expr instanceof SQLCastExpr) {
            SQLCastExpr cast = (SQLCastExpr) expr;
            SQLDataType castExprDataType = cast.getExpr().computeDataType();
            if (cast.getDataType().isString()
                    && firstDataType.isInt()
                    && castExprDataType.isInt()
            ) {
                x.setExpr(cast.getExpr().clone());
                optimizedCount++;
            }
        }

        SQLDataType dataType = expr.computeDataType();
        if (dataType != null && dataType.isInt()) {
            for (int i = list.size() - 1; i >= 0; i--) {
                SQLExpr item = list.get(i);
                if (item instanceof SQLCharExpr) {
                    String str = ((SQLCharExpr) item).getText();
                    try {
                        long l = Long.parseLong(str);
                        SQLIntegerExpr intExpr = SQLIntegerExpr.ofIntOrLong(l);
                        intExpr.setParent(x);
                        list.set(i, intExpr);
                    } catch (Exception ex) {
                        // skip
                    }
                }
            }
        } else if (dataType != null && (dataType.nameHashCode64() == FnvHash.Constants.DATE
                || dataType.nameHashCode64() == FnvHash.Constants.TIMESTAMP
                || dataType.nameHashCode64() == FnvHash.Constants.TIME)) {
            for (int i = list.size() - 1; i >= 0; i--) {
                SQLExpr item = list.get(i);
                if (item instanceof SQLCharExpr) {
                    handleNameLiteral(x, expr, (SQLCharExpr)item);
                }
            }
        }

        return true;
    }

    public boolean visit(SQLBetweenExpr x) {
        x.getBeginExpr().accept(this);
        x.getEndExpr().accept(this);

        SQLDataType beginType = x.getBeginExpr().computeDataType();
        SQLDataType endType = x.getEndExpr().computeDataType();
        SQLDataType testType = x.getTestExpr().computeDataType();

        if (beginType == null || endType == null) {
            return false;
        }

        if (beginType.isString() && endType.isInt()) {
            SQLCastExpr cast = new SQLCastExpr(x.getBeginExpr().clone(), endType.clone());
            x.setBeginExpr(cast);
            optimizedCount++;
        } else if (endType.isString() && beginType.isInt()) {
            SQLCastExpr cast = new SQLCastExpr(x.getEndExpr().clone(), beginType.clone());
            x.setEndExpr(cast);
            optimizedCount++;
        } else if (testType != null && testType.nameHashCode64() == FnvHash.Constants.TIMESTAMP) {
            if (x.getBeginExpr() instanceof SQLLiteralExpr) {
                if (ConstFolding.repalceToTimestamp((SQLLiteralExpr) x.getBeginExpr(), timeZone)) {
                    optimizedCount++;
                }
            }

            if (x.getEndExpr() instanceof SQLLiteralExpr) {
                if (ConstFolding.repalceToTimestamp((SQLLiteralExpr) x.getEndExpr(), timeZone)) {
                    optimizedCount++;
                }
            }
        } else if (beginType.nameHashCode64() == FnvHash.Constants.DATE && endType.nameHashCode64() == FnvHash.Constants.DATE && x.getTestExpr() instanceof SQLCharExpr) {
            SQLDateExpr sqlDateExpr = new SQLDateExpr(((SQLCharExpr) x.getTestExpr()).getText());
            x.setTestExpr(sqlDateExpr);
            optimizedCount++;
        }

        return false;
    }


    public boolean visit(SQLNotExpr x) {
        SQLExpr expr = x.getExpr();

        if (expr instanceof SQLIntegerExpr) {
            Number number = ((SQLIntegerExpr) expr).getNumber();
            if (number instanceof Integer) {
                if (number.intValue() == 1) {
                    x.setExpr(new SQLBooleanExpr(true));
                } else {
                    if (number.intValue() == 0) {
                        x.setExpr(new SQLBooleanExpr(false));
                    }
                }
            }
        }

        return true;
    }

}
