package com.alibaba.fastsql.sql.optimizer.rules;

import com.alibaba.fastsql.DbType;
import com.alibaba.fastsql.sql.SQLUtils;
import com.alibaba.fastsql.sql.ast.SQLExpr;
import com.alibaba.fastsql.sql.ast.SQLName;
import com.alibaba.fastsql.sql.ast.SQLObject;
import com.alibaba.fastsql.sql.ast.SQLStatement;
import com.alibaba.fastsql.sql.ast.expr.SQLIdentifierExpr;
import com.alibaba.fastsql.sql.ast.expr.SQLMethodInvokeExpr;
import com.alibaba.fastsql.sql.ast.expr.SQLPropertyExpr;
import com.alibaba.fastsql.sql.ast.statement.*;
import com.alibaba.fastsql.sql.dialect.mysql.ast.statement.MySqlInsertStatement;
import com.alibaba.fastsql.sql.optimizer.Optimizer;
import com.alibaba.fastsql.sql.parser.ParserException;
import com.alibaba.fastsql.sql.repository.Schema;
import com.alibaba.fastsql.sql.repository.SchemaObject;
import com.alibaba.fastsql.sql.repository.SchemaRepository;
import com.alibaba.fastsql.sql.visitor.SQLASTVisitorAdapter;
import com.alibaba.fastsql.util.FnvHash;
import com.alibaba.fastsql.util.StringUtils;

import java.util.*;

import static com.alibaba.fastsql.util.StringUtils.backquoteAlias;
import static com.alibaba.fastsql.util.StringUtils.quoteAlias;

/**
 * @author lijun.cailj 2018/9/7
 */
public class ViewRewriteWithSchema extends OptimizerVisitor {

    private List<SchemaObject> views = new ArrayList();
    private SchemaRepository repository;
    private String defaultSchema;
    private boolean backQuotedNames;

    private Map<String, SchemaObject> viewMap         = new HashMap();
    private Map<String, String>       originalViewMap = new HashMap<String, String>();
    private Map<String, String>       newViewMap      = new HashMap<String, String>();

    public ViewRewriteWithSchema(SchemaRepository repository, String defaultSchema) {
        this.repository = repository;
        this.defaultSchema = defaultSchema;
    }

    @Override
    public void endVisit(SQLSelectStatement x) {
        ViewRewriteRefAlias refAlias = new ViewRewriteRefAlias(newViewMap);
        x.getSelect().getQuery().accept(refAlias);
    }

    @Override
    public boolean visit(SQLExprTableSource x) {
        String alias = x.getAlias();
        String alias2 = backQuotedNames
                ? backquoteAlias(alias)
                : quoteAlias(alias);
        if (alias2 != alias) {
            x.setAlias(alias2);
        }

        final SQLExpr expr = x.getExpr();

        if (expr instanceof SQLName && repository != null) {
            String name = ((SQLName) expr).getSimpleName();

            Schema schema = null;

            SQLTableSource resolvedTableSource = null;
            if (expr instanceof SQLIdentifierExpr) {
                resolvedTableSource = ((SQLIdentifierExpr) expr).getResolvedTableSource();
                if (!StringUtils.isEmpty(defaultSchema)) {
                    schema = repository.findSchema(defaultSchema);
                } else {
                    schema = repository.getDefaultSchema();
                }
            } else if (expr instanceof SQLPropertyExpr) {
                resolvedTableSource = ((SQLPropertyExpr) expr).getResolvedTableSource();
                schema = repository.findSchema(((SQLPropertyExpr) expr).getOwnernName());
            }

            if (resolvedTableSource == null && schema != null && x.getSchemaObject() == null) {
                SchemaObject view = schema.findView(name);
                if (view != null) {
                    views.add(view);

                    String viewName = buildViewName(view);
                    viewMap.put(viewName, view);
                    newViewMap.put(SQLUtils.normalize(name.toLowerCase()), viewName);

                    SQLExprTableSource replaceTo = new SQLExprTableSource(viewName);
                    replaceTo.setAlias(x.getAlias());
                    SQLUtils.replaceInParent(x, replaceTo);

                    SQLCreateViewStatement stmt = (SQLCreateViewStatement) view.getStatement();
                    SQLSelect subQuery = stmt.getSubQuery();

                    try {
                        subQuery.getQuery().accept(this);
                        ViewRewriteRefAlias refAlias = new ViewRewriteRefAlias(newViewMap);
                        subQuery.getQuery().accept(refAlias);
                    } catch (StackOverflowError error) {
                        throw new ParserException("view has circular dependencies");
                    }
                }
            }
        }

        return super.visit(x);
    }

    private String buildViewName(SchemaObject view) {
        StringBuilder viewName = new StringBuilder();
        String simpleName = view.getSchema().getSimpleName();
        if (!StringUtils.isEmpty(simpleName)) {
            long fnv32Lower = FnvHash.fnv_32_lower(simpleName);
            if (fnv32Lower < 0) {
                fnv32Lower = -fnv32Lower;
            }
            viewName.append(simpleName);
            viewName.append("_");
            viewName.append(fnv32Lower);
            viewName.append("_");
        }
        viewName.append(view.getName());

        return SQLUtils.normalize(viewName.toString().toLowerCase());
    }

    public boolean visit(SQLSelect x) {
        final SQLWithSubqueryClause with = x.getWithSubQuery();
        if (with != null)  {
            with.accept(this);
        }

        x.getQuery()
                .accept(this);


        SQLStatement stmt = null;
        for (SQLObject parent = x.getParent(); parent != null; parent = parent.getParent()) {
            if (parent instanceof SQLSelectStatement) {
                stmt = (SQLSelectStatement) parent;
                break;
            } else if (parent instanceof MySqlInsertStatement) {
                stmt = (MySqlInsertStatement) parent;
                break;
            }
        }

        if (stmt == null) {
            return false;
        }

        SQLWithSubqueryClause withSubQuery = null;
        SQLSelect select = null;

        if (stmt instanceof SQLSelectStatement) {
            select = ((SQLSelectStatement) stmt).getSelect();
            withSubQuery = select.getWithSubQuery();
        } else if (stmt instanceof MySqlInsertStatement) {
            select = ((MySqlInsertStatement) stmt).getQuery();
            withSubQuery = select.getWithSubQuery();
        }

        if (withSubQuery == null && views != null && views.size() > 0) {
            withSubQuery = new SQLWithSubqueryClause();
            select.setWithSubQuery(withSubQuery);
        }

        for (int i = views.size() - 1; i >= 0; i--) {
            SchemaObject view = views.get(i);

            String alias = newViewMap.get(view.getName().toLowerCase());

            if (withSubQuery.findEntry(FnvHash.fnv1a_64(alias)) != null) {
                continue;
            }

            SQLWithSubqueryClause.Entry entry = new SQLWithSubqueryClause.Entry();

            SQLCreateViewStatement viewStatement = (SQLCreateViewStatement) view.getStatement();
            if (viewStatement.getColumns() != null) {
                for (SQLTableElement element : viewStatement.getColumns()) {
                    if (element instanceof SQLColumnDefinition) {
                        entry.getColumns().add(((SQLColumnDefinition) element).getName());
                    }
                }
            }
            SQLSelect subQuery = viewStatement.getSubQuery().clone();
            if (subQuery.getWithSubQuery() != null && subQuery.getWithSubQuery().getEntries().size() > 0) {
                for (SQLWithSubqueryClause.Entry item
                        : subQuery.getWithSubQuery().getEntries()) {
                    SQLWithSubqueryClause.Entry clonedItem = item.clone();
                    subQuery.setWithSubQuery(null);
                    if (withSubQuery.findEntry(item.aliasHashCode64()) != null) {
                        continue;
                    }

                    withSubQuery.addEntry(clonedItem);
                    visit(
                            clonedItem.getSubQuery()
                    );
                }
            }

            SchemaBuilderVisitor schemaBuilderVisitor = new SchemaBuilderVisitor(view);
            subQuery.accept(schemaBuilderVisitor);

            entry.setSubQuery(subQuery);


            entry.setAlias(alias);
            if (x.getParent() instanceof SQLWithSubqueryClause.Entry) {
                if (withSubQuery.findEntry(entry.aliasHashCode64()) == null) {
                    withSubQuery.getEntries().add(0, entry);
                }
            } else {
                if (withSubQuery.findEntry(entry.aliasHashCode64()) == null) {
                    withSubQuery.getEntries().add(entry);
                }
            }
            entry.setParent(withSubQuery);
        }

        return false;
    }

    class SchemaBuilderVisitor extends OptimizerVisitor {

        private SchemaObject view;

        public SchemaBuilderVisitor(SchemaObject view) {
            this.view = view;
        }

        public boolean visit(SQLUnionQueryTableSource x) {
            x.getUnion().accept(this);
            return false;
        }

        public boolean visit(SQLExprTableSource x) {

            if (!StringUtils.isEmpty(x.getSchema())) {
                return false;
            }

            Schema schema = repository.findSchema(x.getSchema());

            SchemaObject tableOrView = schema.findTableOrView(x.getTableName());

            if (tableOrView != null) {
                String tableSchemaSimpleName = tableOrView.getSchema().getSimpleName();

                if (StringUtils.isEmpty(tableSchemaSimpleName)) {
                    String viewSchemaSimple = view.getSchema().getSimpleName();
                    if (!StringUtils.isEmpty(viewSchemaSimple)) {
                        x.setSchema(viewSchemaSimple);
                    }
                } else {
                    x.setSchema(tableSchemaSimpleName);
                }
            }

            return false;
        }
    }

    public boolean isBackQuotedNames()
    {
        return backQuotedNames;
    }

    public void setBackQuotedNames(boolean backQuotedNames)
    {
        this.backQuotedNames = backQuotedNames;
    }
}
