package com.alibaba.fastsql.support.protobuf;

import com.alibaba.fastsql.DbType;
import com.alibaba.fastsql.sql.ast.expr.SQLDateExpr;
import com.alibaba.fastsql.sql.ast.expr.SQLTimeExpr;
import com.alibaba.fastsql.sql.ast.expr.SQLTimestampExpr;
import com.alibaba.fastsql.sql.parser.SQLInsertValueHandler;
import com.alibaba.fastsql.util.FnvHash;
import com.alibaba.fastsql.util.MySqlUtils;
import com.google.protobuf.Descriptors;
import com.google.protobuf.DynamicMessage;

import java.math.BigDecimal;
import java.sql.SQLException;
import java.text.DateFormat;
import java.text.ParseException;
import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;
import java.util.UUID;

public abstract class ProtobufInsertValueHandler implements SQLInsertValueHandler {
    protected DbType dbType;
    protected final Descriptors.Descriptor descriptor;
    protected TimeZone timeZone;

    public ProtobufInsertValueHandler(Descriptors.Descriptor descriptor) {
        this.descriptor = descriptor;
    }

    public ProtobufInsertValueHandler(Descriptors.Descriptor descriptor, TimeZone timeZone) {
        this.descriptor = descriptor;
        this.timeZone = timeZone;
    }

    public ProtobufInsertValueHandler(DbType dbType, Descriptors.Descriptor descriptor, TimeZone timeZone) {
        this.dbType = dbType;
        this.descriptor = descriptor;
        this.timeZone = timeZone;
    }

    @Override
    public Object newRow() throws SQLException {
        return DynamicMessage.newBuilder(descriptor);
    }

    protected abstract Descriptors.FieldDescriptor getField(int columnIndex) throws SQLException;

    @Override
    public void processInteger(Object row, int index, Number number) throws SQLException {
        DynamicMessage.Builder msgBuilder = (DynamicMessage.Builder) row;

        Descriptors.FieldDescriptor field = getField(index);
        Descriptors.FieldDescriptor.Type fieldType = field.getType();

        switch (fieldType) {
            case INT32: {
                Integer intValue = null;
                if (number instanceof Integer) {
                    intValue = (Integer) number;
                } else {
                    throw new InvalidValueException("int value over flow", field.getName(), number.toString());
                }
                msgBuilder.setFieldDirect(field, intValue);
                break;
            }
            case INT64: {
                Long longValue = null;
                if (number instanceof Integer) {
                    longValue = number.longValue();
                } else if (number instanceof Long) {
                    longValue = (Long) number;
                } else {
                    throw new InvalidValueException("int value over flow", field.getName(), number.toString());
                }
                msgBuilder.setFieldDirect(field, longValue);
                break;
            }
            case FLOAT: {
                msgBuilder.setFieldDirect(field, number.floatValue());
                break;
            }
            case DOUBLE: {
                msgBuilder.setFieldDirect(field, number.doubleValue());
                break;
            }
            case JSON:
            case STRING:
                msgBuilder.setFieldDirect(field, number.toString());
                break;
            case BOOL: {
                if (number instanceof Integer) {
                    int intValue = number.intValue();
                    if (intValue == 1) {
                        msgBuilder.setFieldDirect(field, true);
                    } else if (intValue == 0) {
                        msgBuilder.setFieldDirect(field, false);
                    } else {
                        throw new InvalidValueException("can not cast to boolean", field.getName(), number.toString());
                    }
                } else {
                    throw new InvalidValueException("can not cast to boolean", field.getName(), number.toString());
                }
                break;
            }
            default:
                throw new InvalidValueException("type convert not support : " + fieldType, field.getName(), number.toString());
        }
    }

    @Override
    public void processString(Object row, int index, String strVal) throws SQLException {
        DynamicMessage.Builder msgBuilder = (DynamicMessage.Builder) row;

        Descriptors.FieldDescriptor field = getField(index);
        Descriptors.FieldDescriptor.Type fieldType = field.getType();

        switch (fieldType) {
            case INT32: {
                int intValue;
                try {
                    intValue = Integer.parseInt(strVal);
                } catch (NumberFormatException e) {
                    throw new InvalidValueException("type convert parseLong error", field.getName(), strVal);
                }
                msgBuilder.setFieldDirect(field, intValue);
                break;
            }
            case INT64: {
                long longValue;
                try {
                    longValue = Long.parseLong(strVal);
                } catch (NumberFormatException e) {
                    throw new InvalidValueException("type convert parseLong error", field.getName(), strVal);
                }
                msgBuilder.setFieldDirect(field, longValue);
                break;
            }
            case FLOAT: {
                float floatValue;
                try {
                    floatValue = Float.parseFloat(strVal);
                } catch (NumberFormatException e) {
                    throw new InvalidValueException("type convert parseLong error", field.getName(), strVal);
                }
                msgBuilder.setFieldDirect(field, floatValue);
                break;
            }
            case DOUBLE: {
                double doubleValue;
                try {
                    doubleValue = Double.parseDouble(strVal);
                } catch (NumberFormatException e) {
                    throw new InvalidValueException("type convert parseLong error", field.getName(), strVal);
                }
                msgBuilder.setFieldDirect(field, doubleValue);
                break;
            }
            case JSON:
                msgBuilder.setFieldDirect(field, strVal);
                break;
            case STRING:
                msgBuilder.setFieldDirect(field, strVal);
                break;
            case TIME: {
                Date datetime = MySqlUtils.parseDate(strVal, timeZone);
                if (datetime != null) {
                    processTime(row, index, datetime);
                } else if (strVal.equals("0000-00-00 00:00:00") || strVal.equals("0000-00-00") || strVal.equals("00:00:00")) {
                    processTime(row, index, new Date(0));
                } else {
                    processIllegalTimestamp(field, strVal);
                }
                break;
            }
            case DATE:{
                Date datetime = MySqlUtils.parseDate(strVal, timeZone);
                if (datetime != null) {
                    processDate(row, index, datetime);
                } else if (strVal.equals("0000-00-00 00:00:00") || strVal.equals("0000-00-00") || strVal.equals("00:00:00")) {
                    processDate(row, index, new Date(0));
                } else {
                    processIllegalTimestamp(field, strVal);
                }
                break;
            }
            case TIMESTAMP: {
                Date datetime = MySqlUtils.parseDate(strVal, timeZone);
                if (datetime != null) {
                    msgBuilder.setFieldDirect(field, (int) (datetime.getTime() / 1000));
                } else if (strVal.equals("0000-00-00 00:00:00") || strVal.equals("0000-00-00") || strVal.equals("00:00:00")) {
                    msgBuilder.setFieldDirect(field, 0);
                } else {
                    processIllegalTimestamp(field, strVal);
                }
                break;
            }
            case DATETIME: {
                Date datetime = MySqlUtils.parseDate(strVal, timeZone);
                if (datetime != null) {
                    msgBuilder.setFieldDirect(field, (int) (datetime.getTime() / 1000));
                } else {
                    processIllegalTimestamp(field, strVal);
                }
                break;
            }
            case BOOL:
                if (strVal.equalsIgnoreCase("true")
                        || strVal.equalsIgnoreCase("Y")
                        || strVal.equalsIgnoreCase("1")) {
                    msgBuilder.setFieldDirect(field, true);
                } else if (strVal.equalsIgnoreCase("false")
                        || strVal.equalsIgnoreCase("N")
                        || strVal.equalsIgnoreCase("0")) {
                    msgBuilder.setFieldDirect(field, false);
                } else {
                    throw new InvalidValueException("type convert not support : " + fieldType, field.getName(), strVal);
                }
                break;
            case BYTES: {

            }
                break;
            default:
                throw new InvalidValueException("type convert not support : " + fieldType, field.getName(), strVal);
        }
    }

    protected void processIllegalTimestamp(Descriptors.FieldDescriptor field, String value) throws SQLException {
        throw new InvalidValueException("type convert not support : Timestamp", field.getName(), value);
    }

    @Override
    public void processDate(Object row, int index, String value) throws SQLException {
        processString(row, index, value);
    }

    public void processTime(Object row, int index, Date value) throws SQLException {
        Calendar cale;
        if (timeZone == null) {
            cale = Calendar.getInstance();
        } else {
            cale = Calendar.getInstance(timeZone);
        }

        int hour = cale.get(Calendar.HOUR_OF_DAY);
        int minute = cale.get(Calendar.MINUTE);
        int second = cale.get(Calendar.SECOND);

        final char[] chars = new char[10];
        chars[0] = (char) (hour/10 + '0');
        chars[1] = (char) (hour%10 + '0');
        chars[2] = ':';
        chars[3] = (char) (minute/10 + '0');
        chars[4] = (char) (minute%10 + '0');
        chars[5] = ':';
        chars[6] = (char) (second/10 + '0');
        chars[7] = (char) (second%10 + '0');

        DynamicMessage.Builder msgBuilder = (DynamicMessage.Builder) row;

        Descriptors.FieldDescriptor field = getField(index);
        String strVal = new String(chars);
        msgBuilder.setFieldDirect(field, strVal);
    }

    @Override
    public void processDate(Object row, int index, java.util.Date value) throws SQLException {
        Calendar cale;
        if (timeZone == null) {
            cale = Calendar.getInstance();
        } else {
            cale = Calendar.getInstance(timeZone);
        }
        cale.setTime(value);

        int year = cale.get(Calendar.YEAR);
        int month = cale.get(Calendar.MONTH) + 1;
        int dayOfMonth = cale.get(Calendar.DAY_OF_MONTH);

        final char[] chars = new char[10];
        chars[0] = (char) (year/1000 + '0');
        chars[1] = (char) ((year/100)%10 + '0');
        chars[2] = (char) ((year/10)%10 + '0');
        chars[3] = (char) (year%10 + '0');
        chars[4] = '-';
        chars[5] = (char) (month/10 + '0');
        chars[6] = (char) (month%10 + '0');
        chars[7] = '-';
        chars[8] = (char) (dayOfMonth/10 + '0');
        chars[9] = (char) (dayOfMonth%10 + '0');

        DynamicMessage.Builder msgBuilder = (DynamicMessage.Builder) row;

        Descriptors.FieldDescriptor field = getField(index);
        String strVal = new String(chars);
        msgBuilder.setFieldDirect(field, strVal);
    }

    @Override
    public void processTimestamp(Object row, int index, String value) throws SQLException {
        processString(row, index, value);
    }

    @Override
    public void processTimestamp(Object row, int index, java.util.Date value) throws SQLException {
        Calendar cale;
        if (timeZone == null) {
            cale = Calendar.getInstance();
        } else {
            cale = Calendar.getInstance(timeZone);
        }
        cale.setTime(value);

        int year = cale.get(Calendar.YEAR);
        int month = cale.get(Calendar.MONTH) + 1;
        int dayOfMonth = cale.get(Calendar.DAY_OF_MONTH);
        int hour = cale.get(Calendar.HOUR_OF_DAY);
        int minute = cale.get(Calendar.MINUTE);
        int second = cale.get(Calendar.SECOND);

        final char[] chars = new char[19];
        chars[0] = (char) (year/1000 + '0');
        chars[1] = (char) ((year/100)%10 + '0');
        chars[2] = (char) ((year/10)%10 + '0');
        chars[3] = (char) (year%10 + '0');
        chars[4] = '-';
        chars[5] = (char) (month/10 + '0');
        chars[6] = (char) (month%10 + '0');
        chars[7] = '-';
        chars[8] = (char) (dayOfMonth/10 + '0');
        chars[9] = (char) (dayOfMonth%10 + '0');
        chars[10] = ' ';
        chars[11] = (char) (hour/10 + '0');
        chars[12] = (char) (hour%10 + '0');
        chars[13] = ':';
        chars[14] = (char) (minute/10 + '0');
        chars[15] = (char) (minute%10 + '0');
        chars[16] = ':';
        chars[17] = (char) (second/10 + '0');
        chars[18] = (char) (second%10 + '0');

        String strVal = new String(chars);
        processString(row, index, strVal);
    }

    @Override
    public void processTime(Object row, int index, String value) throws SQLException {
        processString(row, index, value);
    }

    @Override
    public void processDecimal(Object row, int index, BigDecimal value) throws SQLException {
        DynamicMessage.Builder msgBuilder = (DynamicMessage.Builder) row;

        Descriptors.FieldDescriptor field = getField(index);
        Descriptors.FieldDescriptor.Type fieldType = field.getType();

        switch (fieldType) {
            case INT32: {
                int intValue;
                try {
                    intValue = value.intValue();
                } catch (NumberFormatException e) {
                    throw new InvalidValueException("type convert parseInt error", field.getName(), value.toString());
                }
                msgBuilder.setFieldDirect(field, intValue);
                break;
            }
            case INT64: {
                long longValue;
                try {
                    longValue = value.longValue();
                } catch (NumberFormatException e) {
                    throw new InvalidValueException("type convert parseLong error", field.getName(), value.toString());
                }
                msgBuilder.setFieldDirect(field, longValue);
                break;
            }
            case FLOAT: {
                float floatValue;
                try {
                    floatValue = value.floatValue();
                } catch (NumberFormatException e) {
                    throw new InvalidValueException("type convert parseFloat error", field.getName(), value.toString());
                }
                msgBuilder.setFieldDirect(field, floatValue);
                break;
            }
            case DOUBLE: {
                double doubleValue;
                try {
                    doubleValue = value.doubleValue();
                } catch (NumberFormatException e) {
                    throw new InvalidValueException("type convert parseDouble error", field.getName(), value.toString());
                }
                msgBuilder.setFieldDirect(field, doubleValue);
                break;
            }
            case STRING:
                msgBuilder.setFieldDirect(field, value.toString());
                break;
            default:
                throw new InvalidValueException("type convert not support : " + fieldType, field.getName(), value.toString());
        }
    }

    @Override
    public void processBoolean(Object row, int index, boolean value) throws SQLException {
        DynamicMessage.Builder msgBuilder = (DynamicMessage.Builder) row;

        Descriptors.FieldDescriptor field = getField(index);
        Descriptors.FieldDescriptor.Type fieldType = field.getType();

        switch (fieldType) {
            case INT32:
            case INT64:{
                int intValue = value ? 1 : 0;
                msgBuilder.setFieldDirect(field, intValue);
                break;
            }
            case STRING:
                msgBuilder.setFieldDirect(field, Boolean.toString(value));
                break;
            case BOOL:
                msgBuilder.setFieldDirect(field, value);
                break;
            default:
                throw new InvalidValueException("type convert not support : " + fieldType, field.getName(), Boolean.toString(value));
        }
    }

    @Override
    public void processNull(Object row, int index) throws SQLException {

    }

    protected SQLException processValueError(Descriptors.FieldDescriptor field, SQLException ex) {
        return ex;
    }

    @Override
    public void processFunction(Object row
            , int index
            , String funcName
            , long funcNameHashCode64
            , Object... values) throws SQLException
    {
        if (funcNameHashCode64 == FnvHash.Constants.UUID) {
            String uuid = UUID.randomUUID().toString();
            processString(row, index, uuid);
            return;
        }

        if (funcNameHashCode64 == FnvHash.Constants.STR_TO_DATE || funcNameHashCode64 == FnvHash.Constants.DATE_PARSE) {
            if (values.length == 2) {
                String str = (String) values[0];
                String fmt = (String) values[1];
                DateFormat dateFormat = MySqlUtils.toJavaFormat(fmt, timeZone);
                if (dateFormat != null) {
                    try {
                        Date date = dateFormat.parse(str);
                        processDate(row, index, date);
                        return;
                    } catch (ParseException e) {
                        e.printStackTrace();
                    }
                }
            }
        }

        Descriptors.FieldDescriptor field = getField(index);
        throw new InvalidValueException("insert value not support function : " + funcName, field.getName(), funcName);
    }

    @Override
    public void processRow(Object row) throws SQLException {

    }
    @Override
    public void processComplete() throws SQLException {

    }
}
