package com.alibaba.fastsql.sql.optimizer.rules;

import com.alibaba.fastsql.DbType;
import com.alibaba.fastsql.sql.SQLUtils;
import com.alibaba.fastsql.sql.ast.*;
import com.alibaba.fastsql.sql.ast.expr.*;
import com.alibaba.fastsql.sql.ast.statement.*;
import com.alibaba.fastsql.sql.dialect.mysql.ast.statement.MySqlSelectQueryBlock;
import com.alibaba.fastsql.sql.visitor.functions.*;
import com.alibaba.fastsql.util.*;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.RoundingMode;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.*;

import static com.alibaba.fastsql.sql.ast.expr.SQLBinaryOperator.*;

public class ConstFolding extends OptimizerVisitor {
    private boolean constFoldingNowFun;
    private Date now;

    public ConstFolding() {

    }

    public ConstFolding(DbType dbType, TimeZone timeZone, Boolean constFoldingNowFun) {
        this.dbType = dbType;
        this.timeZone = timeZone;
        this.constFoldingNowFun = constFoldingNowFun;
    }

    public ConstFolding(DbType dbType) {
        this.dbType = dbType;
    }

    public void endVisit(MySqlSelectQueryBlock x) {
        endVisit((SQLSelectQueryBlock) x);
    }

    public void endVisit(SQLSelectQueryBlock x) {
        SQLExpr where = x.getWhere();

        if (where instanceof SQLBooleanExpr) {
            boolean value = ((SQLBooleanExpr) where).getValue();
            if (value) {
                x.setWhere(null);
            } else {
                if (x.getGroupBy() == null
                        && x.getSelectList().size() == 1) {
                    SQLSelectItem selectItem = x.getSelectList().get(0);
                    SQLExpr selectItemExpr = selectItem.getExpr();
                    if (selectItemExpr instanceof SQLAggregateExpr) {
                        SQLAggregateExpr aggExpr = (SQLAggregateExpr) selectItemExpr;
                        if (aggExpr.methodNameHashCode64() == FnvHash.Constants.COUNT
                                && aggExpr.getOver() == null) {
                            selectItem.setExpr(new SQLIntegerExpr(0));
                            x.setFrom((SQLTableSource) null);
                            x.setOrderBy(null);
                            x.setWhere(null);
                            optimizedCount++;
                        }
                    }
                }
            }
        }
    }

    public boolean visit(SQLUnaryExpr x) {
        x.getExpr().accept(this);

        SQLExpr expr = x.getExpr();
        SQLExpr replaceTo = null;
        if (x.getOperator() == SQLUnaryOperator.Negative && expr instanceof SQLValuableExpr) {
            Object value = ((SQLValuableExpr) expr).getValue();
            if (value instanceof Long) {
                long longVal = ((Long) value).longValue();
                replaceTo = new SQLIntegerExpr(-longVal);
            } else if (value instanceof Integer) {
                long intValue = ((Integer) value).intValue();
                replaceTo = new SQLIntegerExpr(-intValue);
            }
        }

        if (replaceTo != null) {
            replaceInParent(x, replaceTo);
        }

        return false;
    }

    public boolean visit(SQLNotExpr x) {
        x.getExpr().accept(this);

        SQLExpr expr = x.getExpr();
        SQLExpr replaceTo = null;
        if (expr instanceof SQLValuableExpr) {
            Object value = ((SQLValuableExpr) expr).getValue();
            if (value instanceof Boolean) {
                replaceTo = new SQLBooleanExpr(!((Boolean) value).booleanValue());
            } else if (value instanceof Integer) {
                int intVal = ((Integer) value).intValue();
                if (intVal == 0) {
                    replaceTo = new SQLBooleanExpr(true);
                } else if (intVal == 1) {
                    replaceTo = new SQLBooleanExpr(false);
                }
            }
        }

        if (replaceTo != null) {
            replaceInParent(x, replaceTo);
        }

        return false;
    }

    public boolean visit(SQLCaseExpr x) {
        if (x.getValueExpr() != null) {
            x.getValueExpr().accept(this);
        }

        boolean allChar = true;

        final List<SQLCaseExpr.Item> items = x.getItems();
        for (SQLCaseExpr.Item item : items) {
            item.accept(this);
            if (!(item.getValueExpr() instanceof SQLCharExpr)) {
                allChar = false;
            }
        }
        {
            final SQLExpr elseExpr = x.getElseExpr();
            if (elseExpr != null) {
                SQLDataType sqlDataType = elseExpr.computeDataType();
                if (sqlDataType != null && elseExpr instanceof SQLValuableExpr && (sqlDataType.isNumberic() || sqlDataType.isInt()) && allChar) {
                    replaceInParent(elseExpr, new SQLCharExpr(elseExpr.toString()));
                }
                elseExpr.accept(this);
            }
        }

        if (x.getValueExpr() instanceof SQLName) {
            SQLName valueExpr = (SQLName) x.getValueExpr();
            for (SQLCaseExpr.Item item : items) {
                if (item.getConditionExpr() instanceof SQLLiteralExpr) {
                    handleNameLiteral(null, valueExpr, (SQLLiteralExpr) item.getConditionExpr());
                }
            }
        }

        final SQLExpr valueExpr = x.getValueExpr();
        Object value = null;
        if (valueExpr instanceof SQLValuableExpr) {
            value = ((SQLValuableExpr) valueExpr).getValue();
        }

        if (value != null) {
            for (SQLCaseExpr.Item item : items) {
                if (item.getConditionExpr() instanceof SQLValuableExpr) {
                    Object condition = ((SQLValuableExpr) item.getConditionExpr()).getValue();
                    if (value.equals(condition)) {
                        if (replaceInParent(x, item.getValueExpr())) {
                            return false;
                        }
                    }
                }
            }
        } else {
            final SQLObject parent = x.getParent();
            for (SQLCaseExpr.Item item : items) {
                final SQLExpr itemCondition = item.getConditionExpr();
                Boolean conditionIsTrue = false;

                if (itemCondition instanceof SQLValuableExpr) {
                    Object condition = ((SQLValuableExpr) itemCondition).getValue();
                    if (Boolean.TRUE.equals(condition)) {
                        conditionIsTrue = true;
                    } else {
                        if (Boolean.FALSE.equals(condition)) {
                            continue;
                        }
                    }
                } else if (parent instanceof SQLBinaryOpExpr
                        && ((SQLBinaryOpExpr) parent).getOperator().isRelational()) {
                    Object parentParent = parent.getParent();
                    if (parentParent instanceof SQLBinaryOpExpr) {
                        SQLBinaryOpExpr binaryOpParent = (SQLBinaryOpExpr) parentParent;

                        if (binaryOpParent.getOperator() == BooleanAnd) {
                            if (itemCondition.equals(binaryOpParent.getLeft()) || itemCondition.equals(binaryOpParent.getRight())) {
                                conditionIsTrue = true;
                            }
                        }
                    } else if (parentParent instanceof SQLBinaryOpExprGroup) {
                        SQLBinaryOpExprGroup binaryOpParentGroup = (SQLBinaryOpExprGroup) parentParent;

                        if (binaryOpParentGroup.getOperator() == BooleanAnd) {
                            if (binaryOpParentGroup.getItems().contains(itemCondition)) {
                                conditionIsTrue = true;
                            }
                        }
                    }
                }

                if (conditionIsTrue && replaceInParent(x, item.getValueExpr())) {
                    return false;
                }

                break;
            }
        }

        final SQLExpr elseExpr = x.getElseExpr();

        int falseCount = 0;
        for (SQLCaseExpr.Item item : items) {
            SQLExpr conditionExpr = item.getConditionExpr();
            if (conditionExpr instanceof SQLBooleanExpr
                    && !((SQLBooleanExpr) conditionExpr).getBooleanValue()) {
                falseCount++;
            }
        }
        if (falseCount == items.size() && elseExpr != null) {
            replaceInParent(x, elseExpr.clone());
            return false;
        }


        if (valueExpr == null && elseExpr instanceof SQLQueryExpr) {
            SQLSelect elseSelect = ((SQLQueryExpr) elseExpr)
                    .getSubQuery()
                    .clone();
            if (x.getItems().size() == 1
                    && x.getItems().get(0).getConditionExpr() instanceof SQLBinaryOpExpr
                    && x.getItems().get(0).getValueExpr() instanceof SQLQueryExpr)
            {
                SQLBinaryOpExpr itemConditionExpr = (SQLBinaryOpExpr) x.getItems().get(0).getConditionExpr();
                SQLQueryExpr    itemValueExpr     = (SQLQueryExpr) x.getItems().get(0).getValueExpr();

                if (itemConditionExpr.getLeft() instanceof SQLQueryExpr) {
                    SQLSelect itemConditionSelect = ((SQLQueryExpr) itemConditionExpr.getLeft())
                            .getSubQuery()
                            .clone();
                    SQLSelect itemValueSelect = itemValueExpr.getSubQuery()
                            .clone();

                    SQLSelectQueryBlock elseQueryBlock = elseSelect.getQueryBlock();
                    SQLSelectQueryBlock itemConditionQueryBlock = itemConditionSelect.getQueryBlock();
                    SQLSelectQueryBlock itemValueQueryBlock     = itemValueSelect.getQueryBlock();
                    if (elseQueryBlock != null && itemConditionQueryBlock != null
                            && elseQueryBlock.getSelectList().size() == 1
                            && itemConditionQueryBlock.getSelectList().size() == 1
                            && itemValueQueryBlock.getSelectList().size() == 1)
                    {
                        SQLSelectItem elseSelectItem = elseQueryBlock.getSelectList().get(0);
                        SQLSelectItem itemConditionSelectItem = itemConditionQueryBlock.getSelectList().get(0);
                        SQLSelectItem itemValueSelectItem     = itemValueQueryBlock.getSelectList().get(0);

                        SQLExpr elseSelectItemExpr = elseSelectItem.getExpr();
                        SQLExpr itemSelectItemExpr = itemConditionSelectItem.getExpr();
                        SQLExpr itemValueSelectItemExpr = itemValueSelectItem.getExpr();

                        elseSelectItem.setExpr(new SQLIntegerExpr(1));
                        itemConditionSelectItem.setExpr(new SQLIntegerExpr(1));
                        itemValueSelectItem.setExpr(new SQLIntegerExpr(1));

                        if (elseQueryBlock.equals(itemConditionQueryBlock)
                                && elseQueryBlock.equals(itemValueQueryBlock))
                        {
                            SQLBinaryOpExpr condition2 = itemConditionExpr.clone();
                            condition2.setLeft(itemSelectItemExpr.clone());

                            SQLCaseExpr caseExpr2 = new SQLCaseExpr();
                            caseExpr2.addItem(condition2, itemValueSelectItemExpr);
                            caseExpr2.setElseExpr(elseSelectItemExpr);

                            itemConditionSelectItem.setExpr(caseExpr2);

                            SQLQueryExpr query2 = new SQLQueryExpr(itemConditionSelect);
                            if (replaceInParent(x, query2)) {
                                caseExpr2.accept(this);
                                return false;
                            }
                        }
                    }
                }
            }
        }

        return false;
    }

    public boolean visit(SQLCastExpr x) {
        x.getExpr().accept(this);

        SQLDataType dataType = x.getDataType();
        final SQLExpr expr = x.getExpr();
        long dataTypeHash = dataType.nameHashCode64();

        SQLExpr replaceTo = null;
        if (dataTypeHash == FnvHash.Constants.BIGINT) {
            // '123' -> 123
            if (expr instanceof SQLCharExpr) {
                String str = ((SQLCharExpr) expr).getText();
                try {
                    long val = Long.parseLong(str);
                    replaceTo = new SQLIntegerExpr(val);
                } catch (NumberFormatException ex) {
                    // skip
                }
            }
        } else if (dataTypeHash == FnvHash.Constants.INT || dataTypeHash == FnvHash.Constants.INTEGER) {
            // 123 -> 123
            if (expr instanceof SQLCharExpr) {
                String str = ((SQLCharExpr) expr).getText();
                try {
                    int val = Integer.parseInt(str);
                    replaceTo = new SQLIntegerExpr(val);
                } catch (NumberFormatException ex) {
                    // skip
                }
            }
        } else if (dataTypeHash == FnvHash.Constants.SMALLINT) {
            // 123 -> 123
            if (expr instanceof SQLCharExpr) {
                String str = ((SQLCharExpr) expr).getText();
                try {
                    int val = Short.parseShort(str);
                    replaceTo = new SQLIntegerExpr(val);
                } catch (NumberFormatException ex) {
                    // skip
                }
            } else if (expr instanceof SQLIntegerExpr) {
                Number number = ((SQLIntegerExpr) expr).getNumber();
                if (number instanceof Integer
                        && number.intValue() >= Short.MIN_VALUE
                        && number.intValue() <= Short.MAX_VALUE) {
                    replaceTo = expr.clone();
                }
            }
        } else if (dataTypeHash == FnvHash.Constants.DECIMAL) {
            if (expr instanceof SQLCharExpr && dataType.getArguments().size() == 0) {
                // '123' -> DECIMAL '123'
                String str = ((SQLCharExpr) expr).getText();
                try {
                    replaceTo = new SQLDecimalExpr(str);
                } catch (NumberFormatException ex) {
                    // skip
                }
            }
        } else if (dataTypeHash == FnvHash.Constants.VARCHAR
                || dataTypeHash == FnvHash.Constants.VARCHAR2
                || dataTypeHash == FnvHash.Constants.CHAR
                || dataTypeHash == FnvHash.Constants.TEXT
                || dataTypeHash == FnvHash.Constants.TINYTEXT
                || dataTypeHash == FnvHash.Constants.MEDIUMTEXT
                || dataTypeHash == FnvHash.Constants.LONGTEXT
                || dataTypeHash == FnvHash.Constants.CLOB
                || dataTypeHash == FnvHash.Constants.NCLOB
                || dataTypeHash == FnvHash.Constants.NVARCHAR
                || dataTypeHash == FnvHash.Constants.NVARCHAR2
                || dataTypeHash == FnvHash.Constants.NCHAR_VARYING
                || dataTypeHash == FnvHash.Constants.STRING) {
            // 123 -> '123'
            if (expr instanceof SQLNumericLiteralExpr) {
                String str = expr.toString();
                try {
                    replaceTo = new SQLCharExpr(str);
                } catch (NumberFormatException ex) {
                    // skip
                }
            } else if (expr instanceof SQLDateExpr) {
                replaceTo = new SQLCharExpr(((SQLDateExpr) expr).getLiteral());
            } else {
                final SQLDataType sqlDataType = expr.computeDataType();
                if (sqlDataType == null) {
                    return false;
                }

                final long exprDataTypeHash = sqlDataType.nameHashCode64();
                if (exprDataTypeHash == FnvHash.Constants.VARCHAR
                        || exprDataTypeHash == FnvHash.Constants.CHAR
                        || exprDataTypeHash == FnvHash.Constants.TEXT
                        || exprDataTypeHash == FnvHash.Constants.STRING) {
                    replaceTo = expr.clone();
                }
            }
        } else if (dataTypeHash == FnvHash.Constants.DATE) {
            if (expr instanceof SQLIntegerExpr && ((SQLIntegerExpr) expr).getNumber() instanceof Integer) {
                int intVal = ((SQLIntegerExpr) expr).getNumber().intValue();
                if (intVal >= 10000000 && intVal <= 99999999) {
                    int year = intVal / 10000;
                    int mmdd = intVal - year * 10000;
                    int mm = mmdd / 100;
                    int dd = mmdd % 100;

                    StringBuffer buf = new StringBuffer();
                    buf.append(year);
                    buf.append('-');
                    if (mm < 10) {
                        buf.append('0');
                    }
                    buf.append(mm);
                    buf.append('-');
                    if (dd < 10) {
                        buf.append('0');
                    }
                    buf.append(dd);

                    replaceTo = new SQLDateExpr(buf.toString());
                }
            } else if (expr instanceof SQLCharExpr) {
                String text = ((SQLCharExpr) expr).getText();
                Date date = MySqlUtils.parseDate(text, timeZone);
                if (date != null) {
                    if (text.length() == 8) {
                        SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd");
                        if (timeZone != null) {
                            dateFormat.setTimeZone(timeZone);
                        }
                        text = dateFormat.format(date);
                    }
                    replaceTo = new SQLDateExpr(text);
                }
            } else if (expr instanceof SQLDateExpr) {
                replaceTo = expr.clone();
            }
        } else if (dataTypeHash == FnvHash.Constants.TIMESTAMP) {
            if (expr instanceof SQLCharExpr) {
                String text = ((SQLCharExpr) expr).getText();
                Date date = MySqlUtils.parseDate(text, timeZone);
                if (date != null) {
                    replaceTo = new SQLTimestampExpr(date, timeZone);
                }
            }
        }

        if (replaceTo != null && replaceInParent(x, replaceTo)) {
            optimizedCount++;
        }

        return false;
    }

    public boolean visit(SQLCurrentTimeExpr x) {
        if(!constFoldingNowFun) {
            return false;
        }

        if (now == null) {
            now = new Date();
        }

        SQLExpr replaceTo;
        switch (x.getType()) {
            case CURRENT_DATE:
            case CURDATE:
                replaceTo = new SQLDateExpr(now, timeZone);
                break;
            case LOCALTIME:
            case CURRENT_TIME:
            case CURTIME:
                replaceTo = new SQLTimeExpr(now, timeZone);
                break;
            default:
                replaceTo = new SQLTimestampExpr(now, timeZone);
                break;
        }

        if (replaceInParent(x, replaceTo)) {
            optimizedCount++;
        }

        return false;
    }

    public boolean visit(SQLIdentifierExpr x) {
        long hash = x.hashCode64();

        SQLExpr replaceTo = null;
        if (constFoldingNowFun
                && (hash == FnvHash.Constants.SYSDATE
                || hash == FnvHash.Constants.NOW
                || hash == FnvHash.Constants.CURRENT_TIMESTAMP
                || hash == FnvHash.Constants.LOCALTIMESTAMP
                || hash == FnvHash.Constants.SYSTIMESTAMP
                || hash == FnvHash.Constants.CURRENT_DATE
                || hash == FnvHash.Constants.CURDATE)
                && timeZone != null) {
            if (now == null) {
                now = new Date();
            }

            if (hash == FnvHash.Constants.CURRENT_DATE || hash == FnvHash.Constants.CURDATE) {
                replaceTo = new SQLDateExpr(now, timeZone);
            } else {
                replaceTo = new SQLTimestampExpr(now, timeZone);
            }
        }

        if (replaceTo != null && replaceInParent(x, replaceTo)) {
            optimizedCount++;
            return false;
        }

        if (x.getResolvedColumn() == null
                && (x.getParent() instanceof SQLBinaryOpExpr || x.getParent() instanceof SQLMethodInvokeExpr || x.getParent() instanceof SQLCastExpr)
                && x.getResolvedTableSource() != null
                && x.getResolvedTableSource() instanceof SQLSubqueryTableSource) {
            SQLSubqueryTableSource resolvedTableSource = (SQLSubqueryTableSource) x.getResolvedTableSource();
            SQLSelectQuery query = resolvedTableSource.getSelect().getQuery();

            if(query instanceof MySqlSelectQueryBlock) {
                SQLSelectItem item = ((MySqlSelectQueryBlock) query).findSelectItem(x.getName());

                if (item != null) {
                    if (item.getExpr() instanceof SQLLiteralExpr) {
                        replaceInParent(x, item.getExpr());
                        return true;
                    }
                }
            }
        }

        return false;
    }

    public boolean visit(SQLMethodInvokeExpr x) {
        SQLExpr owner = x.getOwner();
        if (owner != null) {
            owner.accept(this);
        }

        for (SQLExpr arg : x.getArguments()) {
            arg.accept(this);
        }

        SQLExpr for_ = x.getFor();
        if (for_ != null) {
            for_.accept(this);
        }

        final long hash = x.methodNameHashCode64();

        if (x.getArguments().size() == 1
                && x.getArguments().get(0) instanceof SQLMethodInvokeExpr
                && ((SQLMethodInvokeExpr) x.getArguments().get(0)).methodNameHashCode64() == FnvHash.Constants.FROM_UNIXTIME
                && ((SQLMethodInvokeExpr) x.getArguments().get(0)).getArguments().size() == 1
                && ((SQLMethodInvokeExpr) x.getArguments().get(0)).getArguments().get(0) instanceof SQLTimestampExpr
                && (hash == FnvHash.Constants.YEAR
                || hash == FnvHash.Constants.MONTH
                || hash == FnvHash.Constants.DAY
                || hash == FnvHash.Constants.HOUR)
        ) {
            SQLExpr ts = ((SQLMethodInvokeExpr) x.getArguments().get(0)).getArguments().get(0);
            x.setArgument(0, ts.clone());
        }

        if (constFoldingNowFun && timeZone != null
                && (hash == FnvHash.Constants.SYSDATE
                || hash == FnvHash.Constants.NOW
                || hash == FnvHash.Constants.CURRENT_TIMESTAMP
                || hash == FnvHash.Constants.LOCALTIMESTAMP
                || hash == FnvHash.Constants.SYSTIMESTAMP
                || hash == FnvHash.Constants.CURRENT_DATE
                || hash == FnvHash.Constants.CURDATE
        )) {
            final boolean date = hash == FnvHash.Constants.CURRENT_DATE || hash == FnvHash.Constants.CURDATE;

            SimpleDateFormat format;
            format = new SimpleDateFormat(date ? "yyyy-MM-dd" : "yyyy-MM-dd HH:mm:ss");
            format.setTimeZone(timeZone);

            if (now == null) {
                now = new Date();
            }

            String text = format.format(now);

            SQLExpr timestampExpr = date
                    ? new SQLDateExpr(text)
                    : new SQLTimestampExpr(text);
            if (replaceInParent(x, timestampExpr)) {
                optimizedCount++;
                return false;
            }
        }

        if (timeZone != null
                && (hash == FnvHash.Constants.CURRENT_TIME
                || hash == FnvHash.Constants.CURTIME)) {

            SimpleDateFormat format;
            format = new SimpleDateFormat("HH:mm:ss");
            format.setTimeZone(timeZone);

            if (now == null) {
                now = new Date();
            }

            String text = format.format(now);

            SQLExpr timeExpr = new SQLTimeExpr(text);
            if (replaceInParent(x, timeExpr)) {
                optimizedCount++;
                return false;
            }
        }

        if ((hash == FnvHash.Constants.IFNULL
             || hash == FnvHash.Constants.ISNULL
                || hash == FnvHash.Constants.COALESCE)
                && x.getArguments().size() > 1) {
            List<SQLExpr> arguments = x.getArguments();

            SQLExpr arg0 = arguments.get(0);
            SQLDataType arg0DataType = arg0.computeDataType();
            if (arg0DataType == null && !(arg0 instanceof SQLNullExpr)) {
                return false;
            }
        }

        boolean allConst = isAllConst(x);

        if (x.getArguments().size() >= 1) {
            boolean isNull = false;
            SQLExpr arg0 = x.getArguments().get(0);
            if (arg0 instanceof SQLNullExpr) {
                isNull = true;
            } else if (arg0 instanceof SQLCastExpr && ((SQLCastExpr) arg0).getExpr() instanceof SQLNullExpr) {
                isNull = true;
            }

            if (isNull) {
                if (hash == FnvHash.Constants.FLOOR
                        || hash == FnvHash.Constants.FROM_DAYS
                        || hash == FnvHash.Constants.DATE_ADD
                        || hash == FnvHash.Constants.DATEADD
                        || hash == FnvHash.Constants.DATE_SUB
                        || hash == FnvHash.Constants.DATE_DIFF
                        || hash == FnvHash.Constants.DATE_PARSE
                        || hash == FnvHash.Constants.DATE_FORMAT
                        || hash == FnvHash.Constants.YEAR
                        || hash == FnvHash.Constants.MONTH
                        || hash == FnvHash.Constants.DATE
                        || hash == FnvHash.Constants.DAY
                        || hash == FnvHash.Constants.QUARTER
                        || hash == FnvHash.Constants.HOUR
                        || hash == FnvHash.Constants.MINUTE
                        || hash == FnvHash.Constants.CONCAT
                        || hash == FnvHash.Constants.ADDDATE
                        || hash == FnvHash.Constants.TO_DAYS
                ) {
                    if (replaceInParent(x, new SQLNullExpr())) {
                        return false;
                    }
                }
            }

        }

        if (allConst) {
            List<SQLExpr> arguments = x.getArguments();
            if (hash == FnvHash.Constants.SUBSTR || hash == FnvHash.Constants.SUBSTRING) {

                if (x.getFrom() != null && x.getFor() != null) {
                    x.getArguments().add(x.getFrom());
                    x.getArguments().add(x.getFor());

                    x.setFrom(null);
                    x.setFor(null);
                }

                SQLExpr arg0 = arguments.get(0);
                if (arguments.size() == 3 && !(arg0 instanceof SQLCharExpr) && arg0 instanceof SQLValuableExpr) {
                    String s = ((SQLValuableExpr) arg0).getValue().toString();
                    arguments.set(0, new SQLCharExpr(s));
                }

                // fname = substr('123', 1, 2) -> fname = '12'
                Object val = Substring.instance.eval(x);
                if (val instanceof String) {
                    if (replaceInParent(x, new SQLCharExpr((String) val))) {
                        optimizedCount++;
                    }
                    return false;
                }
            } else if (hash == FnvHash.Constants.CONCAT) {
                // concat('1', '2') -> '12'
                Object val = com.alibaba.fastsql.sql.visitor.functions.Concat.instance.eval(x);
                if (val instanceof String) {
                    if (replaceInParent(x, new SQLCharExpr((String) val))) {
                        optimizedCount++;
                    }
                    return false;
                }
            } else if (hash == FnvHash.Constants.LCASE || hash == FnvHash.Constants.LOWER) {
                // lcase('aBc') -> 'abc'
                Object val = Lcase.instance.eval(x);
                if (val instanceof String) {
                    if (replaceInParent(x, new SQLCharExpr((String) val))) {
                        optimizedCount++;
                    }
                    return false;
                }
            } else if (hash == FnvHash.Constants.UCASE || hash == FnvHash.Constants.UPPER) {
                // ucase('aBc') -> 'ABC'
                Object val = Ucase.instance.eval(x);
                if (val instanceof String) {
                    if (replaceInParent(x, new SQLCharExpr((String) val))) {
                        optimizedCount++;
                    }
                    return false;
                }
            } else if (hash == FnvHash.Constants.LENGTH || hash == FnvHash.Constants.LEN) {
                // len('aBc') -> 3
                Object val = Length.instance.eval(x);
                if (val instanceof Integer) {
                    if (replaceInParent(x, new SQLIntegerExpr((Integer) val))) {
                        optimizedCount++;
                    }
                    return false;
                }
            } else if (hash == FnvHash.Constants.LEFT) {
                func_left(x);
            } else if (hash == FnvHash.Constants.HEX) {
                func_hex(x);
            } else if (hash == FnvHash.Constants.UNHEX) {
//                func_unhex(x);
            } else if (hash == FnvHash.Constants.ISNULL) {
                func_isnull(x);
            } else if (hash == FnvHash.Constants.TO_UNIXTIME) {
                func_to_unixtime(x);
            } else if (hash == FnvHash.Constants.LOCATE) {
                func_locate(x);
            } else if (hash == FnvHash.Constants.REVERSE) {
                func_reverse(x);
            } else if (hash == FnvHash.Constants.RIGHT) {
                func_right(x);
            } else if (hash == FnvHash.Constants.RTRIM) {
                func_rtrim(x);
            } else if (hash == FnvHash.Constants.LTRIM) {
                func_ltrim(x);
            } else if (hash == FnvHash.Constants.TRIM) {
                func_trim(x);

            } else if (hash == FnvHash.Constants.GREATEST) {
                // greast（1，2，3) -> 3
                func_greatest(x);
            } else if (hash == FnvHash.Constants.LEAST) {
                // least（1，2，3) -> 1
                func_least(x);
            } else if (hash == FnvHash.Constants.IFNULL) {
                func_ifnull(x);
            } else if (hash == FnvHash.Constants.COALESCE) {
                func_coalesce(x);
            } else if (hash == FnvHash.Constants.IF) {
                func_if(x);

            } else if (hash == FnvHash.Constants.NULLIF) {
                func_nullif(x);
            } else if (hash == FnvHash.Constants.REPLACE) {
                if (arguments.size() == 3) {
                    try {
                        String arg0 = ((SQLCharExpr) arguments.get(0)).getText();
                        String arg1 = ((SQLCharExpr) arguments.get(1)).getText();
                        String arg2 = ((SQLCharExpr) arguments.get(2)).getText();

                        String newStr = arg0.replaceAll(arg1, arg2);
                        replaceInParent(x, new SQLCharExpr(newStr));
                    } catch (Exception e) {
                        //
                        return false;
                    }
                }

            } else if (hash == FnvHash.Constants.TO_DATE) {
                boolean supportDateExpr = dbType == DbType.mysql
                        || dbType == DbType.ads
                        || dbType == DbType.hive;

                if (arguments.size() == 2
                        && arguments.get(0) instanceof SQLCharExpr
                        && arguments.get(1) instanceof SQLCharExpr
                        && supportDateExpr) {
                    String chars = ((SQLCharExpr) arguments.get(0)).getText();
                    String format = ((SQLCharExpr) arguments.get(1)).getText();
                    DateFormat dateFormat = MySqlUtils.toJavaFormat(format);
                    if (dateFormat != null) {
                        try {
                            Date date = dateFormat.parse(chars);
                            String dateLiteral = new SimpleDateFormat("yyyy-MM-dd").format(date);
                            if (replaceInParent(x, new SQLDateExpr(dateLiteral))) {
                                optimizedCount++;
                            }
                        } catch (ParseException e) {
                            // skip
                            return false;
                        }
                    }
                }
            } else if (hash == FnvHash.Constants.STR_TO_DATE
                    || hash == FnvHash.Constants.DATE_PARSE
                    || hash == FnvHash.Constants.TO_DATE) {
                func_str_to_date(x);
            } else if (hash == FnvHash.Constants.JSON_EXTRACT) {
                func_jsonExtract(x);
            } else if (hash == FnvHash.Constants.JSON_ARRAY_GET) {
                func_json_arrayGet(x);
            } else if (hash == FnvHash.Constants.ABS) {
                func_abs(x);
            } else if (hash == FnvHash.Constants.FLOOR) {
                func_floor(x);

            } else if (hash == FnvHash.Constants.INTERVAL) {
                func_interval(x);
            } else if (hash == FnvHash.Constants.DATEADD
                    || hash == FnvHash.Constants.DATE_ADD
                    || hash == FnvHash.Constants.ADDDATE) {
                func_str_date_add(x, arguments);

            } else if (hash == FnvHash.Constants.DATE_SUB || hash == FnvHash.Constants.SUBDATE) {
                func_str_date_sub(x, arguments);
            } else if (hash == FnvHash.Constants.ADDTIME) {
                func_addTime(x, arguments);
            } else if (hash == FnvHash.Constants.SUBTIME) {
                func_subTime(x, arguments);

            } else if (hash == FnvHash.Constants.GET_FORMAT) {
                func_get_format(x);
            } else if (hash == FnvHash.Constants.DAY_OF_WEEK || hash == FnvHash.Constants.DAYOFWEEK) {
                func_day_of_week(x);
            } else if (hash == FnvHash.Constants.DATE_FORMAT) {
                func_date_format(x, arguments);
            } else if (hash == FnvHash.Constants.DATEDIFF || hash == FnvHash.Constants.DATE_DIFF) {
                func_datediff(x);
            } else if (hash == FnvHash.Constants.DATE_TRUNC) {
                func_datetrunc(x);
            } else if (hash == FnvHash.Constants.DATE) {
                func_date(x);
            } else if (hash == FnvHash.Constants.TIME) {
                func_time(x);
            } else if (hash == FnvHash.Constants.TIMESTAMP) {
                func_timestamp(x);
            } else if (hash == FnvHash.Constants.TIMESTAMPDIFF) {
                func_timestamp_diff(x);
            } else if (hash == FnvHash.Constants.MONTHNAME) {
                func_monthname(x);
            } else if (hash == FnvHash.Constants.MONTH_BETWEEN) {
                func_month_between(x);

            } else if (hash == FnvHash.Constants.TRUNC) {
                func_trunc(x);

            } else if (hash == FnvHash.Constants.PERIOD_ADD) {
                func_period_add(x);
            } else if (hash == FnvHash.Constants.YEAR
                    || hash == FnvHash.Constants.QUARTER
                    || hash == FnvHash.Constants.MONTH
                    || hash == FnvHash.Constants.DAY
                    || hash == FnvHash.Constants.WEEK
                    || hash == FnvHash.Constants.WEEKDAY
                    || hash == FnvHash.Constants.WEEKOFYEAR
                    || hash == FnvHash.Constants.DAYOFMONTH
                    || hash == FnvHash.Constants.HOUR
                    || hash == FnvHash.Constants.MINUTE
                    || hash == FnvHash.Constants.SECOND) {
                if (arguments.size() == 1) {
                    SQLExpr arg0 = arguments.get(0);

                    String text = null;
                    if (arg0 instanceof SQLCharExpr) {
                        text = ((SQLCharExpr) arg0).getText();
                    } else if (arg0 instanceof SQLDateExpr) {
                        text = ((SQLDateExpr) arg0).getLiteral();
                    } else if (arg0 instanceof SQLTimestampExpr) {
                        text = ((SQLTimestampExpr) arg0).getLiteral();
                    } else if (arg0 instanceof SQLIntegerExpr) {
                        text = String.valueOf(((SQLIntegerExpr) arg0).getNumber().intValue());
                    }

                    if (text == null) {
                        return false;
                    }

                    if (text.length() == 10) {
                        if (hash == FnvHash.Constants.HOUR || hash == FnvHash.Constants.MINUTE || hash == FnvHash.Constants.SECOND) {
                            return false;
                        }
                    }

                    Date d0 = MySqlUtils.parseDate(text, timeZone);
                    if (d0 == null) {
                        return false;
                    }

                    Calendar calendar;
                    if (timeZone != null) {
                        calendar = Calendar.getInstance(timeZone);
                    } else {
                        calendar = Calendar.getInstance();
                    }
                    calendar.setTime(d0);

                    Integer value = null;
                    if (hash == FnvHash.Constants.YEAR) {
                        value = calendar.get(Calendar.YEAR);
                    } else if (hash == FnvHash.Constants.QUARTER) {
                        value = (calendar.get(Calendar.MONTH)) / 3 + 1;
                    } else if (hash == FnvHash.Constants.MONTH) {
                        value = calendar.get(Calendar.MONTH) + 1;
                    } else if (hash == FnvHash.Constants.DAY || hash == FnvHash.Constants.DAYOFMONTH) {
                        value = calendar.get(Calendar.DAY_OF_MONTH);
                    } else if (hash == FnvHash.Constants.WEEK) {
                        value = calendar.get(Calendar.WEEK_OF_YEAR) - 1;
                    } else if (hash == FnvHash.Constants.WEEKOFYEAR) {
                        calendar.setFirstDayOfWeek(2);
                        value = calendar.get(Calendar.WEEK_OF_YEAR);
                    } else if (hash == FnvHash.Constants.WEEKDAY) {
                        value = (calendar.get(Calendar.DAY_OF_WEEK) + 5) % 7;
                    } else if (hash == FnvHash.Constants.HOUR) {
                        value = calendar.get(Calendar.HOUR_OF_DAY);
                    } else if (hash == FnvHash.Constants.MINUTE) {
                        value = calendar.get(Calendar.MINUTE);
                    } else if (hash == FnvHash.Constants.SECOND) {
                        value = calendar.get(Calendar.SECOND);
                    }

                    if (value != null && replaceInParent(x, new SQLIntegerExpr(value))) {
                        optimizedCount++;
                    }
                }

                if (arguments.size() == 2) {
                    String text = null;
                    SQLExpr arg0 = arguments.get(0);

                    if (arg0 instanceof SQLCharExpr) {
                        text = ((SQLCharExpr) arg0).getText();
                    }

                    if (text != null && replaceInParent(arg0, new SQLDateExpr(text))) {
                        optimizedCount++;
                    }
                }

            } else if (hash == FnvHash.Constants.TO_CHAR) {
                func_to_char(x, arguments);
            } else if (hash == FnvHash.Constants.UNIX_TIMESTAMP) {
                func_unix_timestamp(x);
            } else if (hash == FnvHash.Constants.FROM_UNIXTIME) {
                func_from_unixtime(x);
            } else if (hash == FnvHash.Constants.LAST_DAY) {
                func_last_day(x);
            } else if (hash == FnvHash.Constants.MAKEDATE) {
                func_makedate(x);
            } else if (hash == FnvHash.Constants.ASCII) {
                func_ascii(x);

            } else if (hash == FnvHash.Constants.DAYNAME) {
                func_dayname(x);

            } else if (hash == FnvHash.Constants.ORD) {
                func_ord(x);

            } else if (hash == FnvHash.Constants.REPEAT) {
                func_repeat(x);

            } else if (hash == FnvHash.Constants.SPACE) {
                func_space(x);

            } else if (hash == FnvHash.Constants.SUBSTRING_INDEX) {
                func_substring_index(x);
            }
        }

        return false;
    }

    private void func_to_unixtime(SQLMethodInvokeExpr x) {
        if (x.getArguments().size() == 1) {
            SQLExpr arg0 = x.getArguments().get(0);
            if (arg0 instanceof SQLCharExpr) {
                String text = ((SQLCharExpr) arg0).getText();
                SQLTimestampExpr timestampExpr = new SQLTimestampExpr(text);
                x.getArguments().set(0, timestampExpr);
            }
        }
    }

    private void func_time(SQLMethodInvokeExpr x) {

        if(x.getArguments().size() == 1) {
            SQLExpr arg0 = x.getArguments().get(0);

            SimpleDateFormat format = new SimpleDateFormat("HH:mm:ss");
            if (arg0 instanceof SQLTimestampExpr) {
                String text = format.format(((SQLTimestampExpr) arg0).getDate(timeZone));
                SQLExpr timeExpr = new SQLCharExpr(text);
                replaceInParent(x, timeExpr);
            } else if (arg0 instanceof SQLDateExpr) {
                String text = format.format(((SQLDateExpr) arg0).getDate(timeZone));
                SQLExpr timeExpr = new SQLCharExpr(text);
                replaceInParent(x, timeExpr);
            }
        }
    }

    private void func_to_char(SQLMethodInvokeExpr x, List<SQLExpr> arguments) {
        if (arguments.size() != 2) {
            return;
        }

        final SQLExpr arg0 = arguments.get(0);
        final SQLExpr arg1 = arguments.get(1);

        String format;
        if (arg1 instanceof SQLCharExpr) {
            format = ((SQLCharExpr) arg1).getText();
        } else {
            return;
        }

        Date date;
        if (arg0 instanceof SQLDateExpr) {
            date = ((SQLDateExpr) arg0).getDate(timeZone);
        } else if (arg0 instanceof SQLTimestampExpr) {
            date = ((SQLTimestampExpr) arg0).getDate(timeZone);
        } else {
            return;
        }

        DateFormat javaFormat = MySqlUtils.toJavaFormat(format, timeZone);
        if (javaFormat == null) {
            return;
        }

        String chars = javaFormat.format(date);
        if (replaceInParent(x, new SQLCharExpr(chars))) {
            optimizedCount++;
        }
    }

    private static boolean isAllConst(SQLMethodInvokeExpr x) {
        boolean allConst = true;
        final long hash = x.methodNameHashCode64();
        for (int i = 0; i < x.getArguments().size(); i++) {
            SQLExpr arg = x.getArguments().get(i);

            if (arg instanceof SQLValuableExpr) {

            } else if (arg instanceof SQLIntervalExpr && ((SQLIntervalExpr) arg).getValue() instanceof SQLValuableExpr) {

            } else if (hash == FnvHash.Constants.TIMESTAMPDIFF && i == 0 && arg instanceof SQLIdentifierExpr) {
            } else if (hash == FnvHash.Constants.GET_FORMAT && i == 0 && arg instanceof SQLIdentifierExpr) {
            } else if (hash == FnvHash.Constants.IF || hash == FnvHash.Constants.JSON_EXTRACT
                       || hash == FnvHash.Constants.JSON_ARRAY_GET || hash == FnvHash.Constants.DATE_FORMAT
                       || hash == FnvHash.Constants.DATE_ADD || hash == FnvHash.Constants.ADDDATE) {
                // continue
            } else {
                allConst = false;
            }
        }
        return allConst;
    }

    private void func_substring_index(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 3) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);
        SQLExpr arg2 = arguments.get(2);

        if (arg0 instanceof SQLCharExpr && arg1 instanceof SQLCharExpr && arg2 instanceof SQLIntegerExpr) {
            String str0 = ((SQLCharExpr) arg0).getText();
            String str1 = ((SQLCharExpr) arg1).getText();
            int count = ((SQLIntegerExpr) arg2).getNumber().intValue();


            SQLExpr replaceTo;
            if (count > 0) {
                int index = -1;
                for (int i = 0; i < count; ++i) {
                    index = str0.indexOf(str1, index + 1);
                    if (index == -1) {
                        break;
                    }
                }

                if (index == -1) {
                    replaceTo = arg0.clone();
                } else {
                    String result = str0.substring(0, index);
                    replaceTo = new SQLCharExpr(result);
                }
            } else if (count < 0) {
                int index = str0.length();
                for (int i = 0; i < -count; ++i) {
                    index = str0.lastIndexOf(str1, index - 1);
                    if (index == -1) {
                        break;
                    }
                }

                if (index == -1) {
                    replaceTo = arg0.clone();
                } else {
                    String result = str0.substring(index + 1);
                    replaceTo = new SQLCharExpr(result);
                }
            } else {
                return;
            }

            if (replaceInParent(x, replaceTo)) {
                optimizedCount++;
            }
        }
    }

    private void func_space(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 1) {
            return;
        }
        SQLExpr arg0 = arguments.get(0);
        if (arg0 instanceof SQLIntegerExpr) {
            int val = ((SQLIntegerExpr) arg0).getNumber().intValue();

            char[] buf = new char[val];
            Arrays.fill(buf, ' ');
            if (replaceInParent(x, new SQLCharExpr(new String(buf)))) {
                optimizedCount++;
            }
        }
    }

    private void func_repeat(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 2) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);
        if (arg0 instanceof SQLCharExpr && arg1 instanceof SQLIntegerExpr) {
            String text = ((SQLCharExpr) arg0).getText();
            long val = ((SQLIntegerExpr) arg1).getNumber().longValue();

            if (val <= 0 || val > 100) {
                return;
            }

            SQLExpr replaceTo = null;
            if (val == 1) {
                replaceTo = arg0.clone();
            } else {
                StringBuffer buf = new StringBuffer(text.length() * (int) val);
                for (int i = 0; i < val; ++i) {
                    buf.append(text);
                }
                replaceTo = new SQLCharExpr(buf.toString());
            }
            if (replaceInParent(x, replaceTo)) {
                optimizedCount++;
            }
        }
    }

    private void func_ord(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 1) {
            return;
        }
        SQLExpr arg0 = arguments.get(0);
        if (arg0 instanceof SQLCharExpr) {
            String text = ((SQLCharExpr) arg0).getText();
            if (text.length() == 1) {
                char ch = text.charAt(0);
                if (replaceInParent(x, new SQLIntegerExpr((int) ch))) {
                    optimizedCount++;
                }
            }
        }
    }

    private void func_day_of_week(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 1) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);

        String text;
        if (arg0 instanceof SQLCharExpr) {
            text = ((SQLCharExpr) arg0).getText();
        } else if (arg0 instanceof SQLDateExpr) {
            text = ((SQLDateExpr) arg0).getLiteral();
        } else {
            return;
        }

        Date date = MySqlUtils.parseDate(text, timeZone);
        if (date == null) {
            return;
        }

        Calendar calendar = timeZone != null ? Calendar.getInstance(timeZone) : Calendar.getInstance();
        calendar.setTime(date);

        int dayOfWeek = calendar.get(Calendar.DAY_OF_WEEK);

        if (replaceInParent(x, new SQLIntegerExpr(dayOfWeek))) {
            optimizedCount++;
        }
    }

    private void func_date_format(SQLMethodInvokeExpr x, List<SQLExpr> arguments) {
        if (arguments.size() != 2) {
            return;
        }

        if (arguments.get(0) instanceof SQLCharExpr) {
            String str = ((SQLCharExpr) arguments.get(0)).getText();
            Date date = MySqlUtils.parseDate(str, timeZone);
            if (date == null) {
                return;
            }

            if (str.length() == 19 || str.length() == 21) {
                SQLTimestampExpr ts = new SQLTimestampExpr(date, timeZone);
                ts.setParent(x);
                arguments.set(0, ts);
            } else if (str.length() == 10) {
                SQLDateExpr dateExpr = new SQLDateExpr(date, timeZone);
                dateExpr.setParent(x);
                arguments.set(0, dateExpr);
            }
        }

        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);
        if (arg0 instanceof SQLTimestampExpr
                && arg1 instanceof SQLCharExpr) {
            SQLTimestampExpr timestampExpr = (SQLTimestampExpr) arg0;
            String format = ((SQLCharExpr) arg1).getText();

            DateFormat dateFormat = mysqlDateFormat(format, timeZone);
            if (dateFormat != null) {
                Date date = timestampExpr.getDate(timeZone);
                if (date == null) {
                    return;
                }

                String text = dateFormat.format(date);
                if (replaceInParent(x, new SQLCharExpr(text))) {
                    optimizedCount++;
                }
            }
        } else if (arg0 instanceof SQLDateExpr
                && arg1 instanceof SQLCharExpr) {
            SQLDateExpr timestampExpr = (SQLDateExpr) arg0;
            String format = ((SQLCharExpr) arg1).getText();

            if (format.contains("%u") || format.contains("%U")) {
                replaceInParent(timestampExpr, new SQLCharExpr(timestampExpr.getLiteral()));
            } else {
                DateFormat dateFormat = mysqlDateFormat(format, timeZone);
                if (dateFormat != null) {
                    Date date = timestampExpr.getDate(timeZone);
                    if (date == null) {
                        return;
                    }

                    String text = dateFormat.format(date);
                    if (replaceInParent(x, new SQLCharExpr(text))) {
                        optimizedCount++;
                    }
                }
            }
        } else if (arg1 instanceof SQLCharExpr) {
            String format = ((SQLCharExpr) arg1).getText();
            SQLObject parent = x.getParent();

            if (parent instanceof SQLMethodInvokeExpr) {
                long parentFuncNameHash = ((SQLMethodInvokeExpr) parent).methodNameHashCode64();
                if (parentFuncNameHash == FnvHash.Constants.DATE_DIFF
                        || parentFuncNameHash == FnvHash.Constants.DATEDIFF)
                {
                    if ("%Y%m%d".equals(format) || "%Y-%m-%d".equals(format)) {
                        SQLMethodInvokeExpr date = new SQLMethodInvokeExpr("DATE", null, arg0.clone());
                        replaceInParent(x, date);
                    }
                }
            }
        }
    }

    private void func_subTime(SQLMethodInvokeExpr x, List<SQLExpr> arguments) {
        if (arguments.size() != 2) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);

        if (arg0 instanceof SQLNullExpr) {
            replaceInParent(x, new SQLNullExpr());
            return;
        }
        SQLDataType dataType = arg0.computeDataType();
        if(dataType != null) {
            long nameHashCode64 = dataType.nameHashCode64();
            if (nameHashCode64 == FnvHash.Constants.TIMESTAMP) {
                SQLCastExpr castExpr = new SQLCastExpr(arg0, new SQLDataTypeImpl("VARCHAR"));
                arguments.set(0, castExpr);
                optimizedCount++;
            }
        }
    }

    private void func_addTime(SQLMethodInvokeExpr x, List<SQLExpr> arguments) {
        if (arguments.size() != 2) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);

        if (arg0 instanceof SQLNullExpr) {
            replaceInParent(x, new SQLNullExpr());
            return;
        }

        Date date = null;
        if (arg0 instanceof SQLCharExpr) {
            String strVal = ((SQLCharExpr) arg0).getText();
            date = MySqlUtils.parseDate(strVal, timeZone);
        } else if (arg0 instanceof SQLTimestampExpr) {
            String strVal = ((SQLTimestampExpr) arg0).getLiteral();
            date = MySqlUtils.parseDate(strVal, timeZone);
        }

        if (date == null) {
            return;
        }

        String times;
        if (arg1 instanceof SQLCharExpr) {
            times = ((SQLCharExpr) arg1).getText();
        } else {
            return;
        }

        SQLExpr replaceTo = null;
        // ('2018-07-18 00:00:00', '1 00:00:00'
        if (times.length() == 10) {
            if (times.charAt(1) == ' ' && times.charAt(4) == ':' && times.charAt(7) == ':') {
                char c0 = times.charAt(0);
                char c2 = times.charAt(2);
                char c3 = times.charAt(3);
                char c5 = times.charAt(5);
                char c6 = times.charAt(6);
                char c8 = times.charAt(8);
                char c9 = times.charAt(9);

                if (c0 >= '0' && c0 <= '9'
                        && c2 >= '0' && c2 <= '9'
                        && c3 >= '0' && c3 <= '9'
                        && c5 >= '0' && c5 <= '9'
                        && c6 >= '0' && c6 <= '9'
                        && c8 >= '0' && c8 <= '9'
                        && c9 >= '0' && c9 <= '9') {
                    int days = c0 - '0';
                    int hours = (c2 - '0') * 10 + (c3 - '0');
                    int minutes = (c5 - '0') * 10 + (c6 - '0');
                    int seconds = (c8 - '0') * 10 + (c9 - '0');

                    Calendar calendar = timeZone == null
                            ? Calendar.getInstance()
                            : Calendar.getInstance(timeZone);
                    calendar.setTime(date);

                    if (days != 0) {
                        calendar.add(Calendar.DAY_OF_YEAR, days);
                    }
                    if (hours != 0) {
                        calendar.add(Calendar.HOUR_OF_DAY, hours);
                    }
                    if (minutes != 0) {
                        calendar.add(Calendar.MINUTE, minutes);
                    }
                    if (seconds != 0) {
                        calendar.add(Calendar.SECOND, seconds);
                    }

                    replaceTo = new SQLTimestampExpr(calendar.getTime(), timeZone);
                }
            }

            if (replaceTo != null) {
                replaceInParent(x, replaceTo);
            }
        }
    }

    private void func_str_date_add(SQLMethodInvokeExpr x, List<SQLExpr> arguments) {
        if (arguments.size() == 2 && arguments.get(1) instanceof SQLIntervalExpr) {
            SQLIntervalExpr intervalExpr = (SQLIntervalExpr) arguments.get(1);
            SQLExpr intervalExprValue = intervalExpr.getValue();
            if (intervalExprValue instanceof SQLIntegerExpr
                    && ((SQLIntegerExpr) intervalExprValue).getNumber() instanceof Integer
                    && ((SQLIntegerExpr) intervalExprValue).getNumber().intValue() == 0
            ) {
                SQLExpr arg0 = arguments.get(0);
                SQLDataType dateType = arg0.computeDataType();
                if (dateType == null
                        || (dateType.nameHashCode64() != FnvHash.Constants.TIMESTAMP
                            && dateType.nameHashCode64() != FnvHash.Constants.TIMESTAMP)
                ) {
                    SQLCastExpr cast = new SQLCastExpr(arg0.clone(), new SQLDataTypeImpl("TIMESTAMP"));
                    replaceInParent(x, cast);
                } else {
                    replaceInParent(x, arg0.clone());
                }


                return;
            }
        }

        boolean supportDateAdd = dbType == DbType.mysql
                || dbType == DbType.ads;

        if (arguments.size() == 2
                && (arguments.get(0) instanceof SQLCharExpr
                || (arguments.get(0) instanceof SQLDateExpr)
                || arguments.get(0) instanceof SQLTimestampExpr)
                && (arguments.get(1) instanceof SQLIntervalExpr || arguments.get(1) instanceof SQLIntegerExpr)) {
            dateAdd(x, arguments.get(0), arguments.get(1), false);
            return;
        }

        if (arguments.size() == 3
                && arguments.get(0) instanceof SQLDateExpr
                && arguments.get(1) instanceof SQLIntegerExpr
                && arguments.get(2) instanceof SQLCharExpr
                && supportDateAdd) {
            SQLDateExpr dateExpr = (SQLDateExpr) arguments.get(0);
            int delta = ((SQLIntegerExpr) arguments.get(1)).getNumber().intValue();
            String type = ((SQLCharExpr) arguments.get(2)).getText();

            if ("day".equalsIgnoreCase(type)) {
                SQLDateExpr dateCloned = dateExpr.clone();
                if (dateCloned.addDay(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            } else if ("month".equalsIgnoreCase(type)) {
                SQLDateExpr dateCloned = dateExpr.clone();
                if (dateCloned.addMonth(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            }
        } else if (arguments.size() == 3
                && arguments.get(2) instanceof SQLDateExpr
                && arguments.get(1) instanceof SQLIntegerExpr
                && arguments.get(0) instanceof SQLCharExpr
                && supportDateAdd) {
            SQLDateExpr dateExpr = (SQLDateExpr) arguments.get(2);
            int delta = ((SQLIntegerExpr) arguments.get(1)).getNumber().intValue();
            String type = ((SQLCharExpr) arguments.get(0)).getText();

            if ("day".equalsIgnoreCase(type)) {
                SQLDateExpr dateCloned = dateExpr.clone();
                if (dateCloned.addDay(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            } else if ("month".equalsIgnoreCase(type)) {
                SQLDateExpr dateCloned = dateExpr.clone();
                if (dateCloned.addMonth(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            } else if ("year".equalsIgnoreCase(type)) {
                SQLDateExpr dateCloned = dateExpr.clone();
                if (dateCloned.addYear(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            }

        } else if (arguments.size() == 3
                && arguments.get(0) instanceof SQLTimestampExpr
                && arguments.get(1) instanceof SQLIntegerExpr
                && arguments.get(2) instanceof SQLCharExpr
                && supportDateAdd) {
            SQLTimestampExpr dateExpr = (SQLTimestampExpr) arguments.get(0);
            int delta = ((SQLIntegerExpr) arguments.get(1)).getNumber().intValue();
            String type = ((SQLCharExpr) arguments.get(2)).getText();

            if ("day".equalsIgnoreCase(type)) {
                SQLTimestampExpr dateCloned = dateExpr.clone();
                if (dateCloned.addDay(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            } else if ("month".equalsIgnoreCase(type)) {
                SQLTimestampExpr dateCloned = dateExpr.clone();
                if (dateCloned.addMonth(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            }

        } else if (arguments.size() == 3
                && arguments.get(2) instanceof SQLTimestampExpr
                && arguments.get(1) instanceof SQLIntegerExpr
                && arguments.get(0) instanceof SQLCharExpr
                && supportDateAdd) {
            SQLTimestampExpr dateExpr = (SQLTimestampExpr) arguments.get(2);
            int delta = ((SQLIntegerExpr) arguments.get(1)).getNumber().intValue();
            String type = ((SQLCharExpr) arguments.get(0)).getText();

            if ("day".equalsIgnoreCase(type)) {
                SQLTimestampExpr dateCloned = dateExpr.clone();
                if (dateCloned.addDay(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            } else if ("month".equalsIgnoreCase(type)) {
                SQLTimestampExpr dateCloned = dateExpr.clone();
                if (dateCloned.addMonth(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            } else if ("hour".equalsIgnoreCase(type)) {
                SQLTimestampExpr dateCloned = dateExpr.clone();
                if (dateCloned.addHour(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            } else if ("minute".equalsIgnoreCase(type)) {
                SQLTimestampExpr dateCloned = dateExpr.clone();
                if (dateCloned.addMiniute(delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            }

        } else if (arguments.size() == 3
                && arguments.get(2) instanceof SQLCharExpr
                && arguments.get(1) instanceof SQLIntegerExpr
                && arguments.get(0) instanceof SQLCharExpr
                && supportDateAdd) {
            String dateText = ((SQLCharExpr) arguments.get(2)).getText();
            int delta = ((SQLIntegerExpr) arguments.get(1)).getNumber().intValue();
            String type = ((SQLCharExpr) arguments.get(0)).getText();

            Date date = MySqlUtils.parseDate(dateText, timeZone);
            if (date == null) {
                return;
            }

            Calendar calendar = timeZone != null
                    ? Calendar.getInstance(timeZone)
                    : Calendar.getInstance();
            calendar.setTime(date);

            boolean isDate = calendar.get(Calendar.MILLISECOND) == 0
                    && calendar.get(Calendar.SECOND) == 0
                    && calendar.get(Calendar.MINUTE) == 0
                    && calendar.get(Calendar.HOUR_OF_DAY) == 0;

            SQLExpr replaceTo = null;
            if ("day".equalsIgnoreCase(type)) {
                calendar.add(Calendar.DAY_OF_MONTH, delta);
                if (isDate) {
                    replaceTo = new SQLDateExpr(calendar.getTime(), timeZone);
                } else {
                    replaceTo = new SQLTimestampExpr(calendar.getTime(), timeZone);
                }
            } else if ("month".equalsIgnoreCase(type)) {
                calendar.add(Calendar.MONTH, delta);
                if (isDate) {
                    replaceTo = new SQLDateExpr(calendar.getTime(), timeZone);
                } else {
                    replaceTo = new SQLTimestampExpr(calendar.getTime(), timeZone);
                }
            }

            if (replaceTo != null && replaceInParent(x, replaceTo)) {
                optimizedCount++;
            }
        }
    }

    private void func_str_date_sub(SQLMethodInvokeExpr x, List<SQLExpr> arguments) {
        boolean supportDateAdd = dbType == DbType.mysql
                || dbType == DbType.ads;

        if (arguments.size() == 2
                && (arguments.get(0) instanceof SQLCharExpr
                || (arguments.get(0) instanceof SQLDateExpr)
                || arguments.get(0) instanceof SQLTimestampExpr)
                && (arguments.get(1) instanceof SQLIntervalExpr || arguments.get(1) instanceof SQLIntegerExpr)) {
            dateAdd(x, arguments.get(0), arguments.get(1), true);
            return;
        }

        if (arguments.size() == 3
                && arguments.get(0) instanceof SQLDateExpr
                && arguments.get(1) instanceof SQLIntegerExpr
                && arguments.get(2) instanceof SQLCharExpr
                && supportDateAdd) {
            SQLDateExpr dateExpr = (SQLDateExpr) arguments.get(0);
            int delta = ((SQLIntegerExpr) arguments.get(1)).getNumber().intValue();
            String type = ((SQLCharExpr) arguments.get(2)).getText();

            if ("day".equalsIgnoreCase(type)) {
                SQLDateExpr dateCloned = dateExpr.clone();
                if (dateCloned.addDay(-delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            }
        } else if (arguments.size() == 3
                && arguments.get(2) instanceof SQLDateExpr
                && arguments.get(1) instanceof SQLIntegerExpr
                && arguments.get(0) instanceof SQLCharExpr
                && supportDateAdd) {
            SQLDateExpr dateExpr = (SQLDateExpr) arguments.get(2);
            int delta = ((SQLIntegerExpr) arguments.get(1)).getNumber().intValue();
            String type = ((SQLCharExpr) arguments.get(0)).getText();

            if ("day".equalsIgnoreCase(type)) {
                SQLDateExpr dateCloned = dateExpr.clone();
                if (dateCloned.addDay(-delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            }

        } else if (arguments.size() == 3
                && arguments.get(0) instanceof SQLTimestampExpr
                && arguments.get(1) instanceof SQLIntegerExpr
                && arguments.get(2) instanceof SQLCharExpr
                && supportDateAdd) {
            SQLTimestampExpr dateExpr = (SQLTimestampExpr) arguments.get(0);
            int delta = ((SQLIntegerExpr) arguments.get(1)).getNumber().intValue();
            String type = ((SQLCharExpr) arguments.get(2)).getText();

            if ("day".equalsIgnoreCase(type)) {
                SQLTimestampExpr dateCloned = dateExpr.clone();
                if (dateCloned.addDay(-delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            }

        } else if (arguments.size() == 3
                && arguments.get(2) instanceof SQLTimestampExpr
                && arguments.get(1) instanceof SQLIntegerExpr
                && arguments.get(0) instanceof SQLCharExpr
                && supportDateAdd) {
            SQLTimestampExpr dateExpr = (SQLTimestampExpr) arguments.get(2);
            int delta = ((SQLIntegerExpr) arguments.get(1)).getNumber().intValue();
            String type = ((SQLCharExpr) arguments.get(0)).getText();

            if ("day".equalsIgnoreCase(type)) {
                SQLTimestampExpr dateCloned = dateExpr.clone();
                if (dateCloned.addDay(-delta)) {
                    if (replaceInParent(x, dateCloned)) {
                        optimizedCount++;
                    }
                }
            }
        }
    }

    private void func_dayname(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 1) {
            SQLExpr arg0 = arguments.get(0);
            if (arg0 instanceof SQLCharExpr) {
                String str0 = ((SQLCharExpr) arg0).getText();

                Date date = MySqlUtils.parseDate(str0, timeZone);
                if (date == null) {
                    return;
                }

                if (str0.length() == 19) {
                    SQLTimestampExpr ts = new SQLTimestampExpr(str0);
                    ts.setParent(x);
                    arguments.set(0, ts);
                } else if (str0.length() == 10) {
                    SQLDateExpr dateExpr = new SQLDateExpr(str0);
                    dateExpr.setParent(x);
                    arguments.set(0, dateExpr);
                }
            }
        }

        Date date = null;
        if (arguments.get(0) instanceof SQLDateExpr) {
            date = ((SQLDateExpr) arguments.get(0)).getDate(timeZone);
        } else if (arguments.size() > 1 && arguments.get(1) instanceof SQLTimestampExpr) {
            date = ((SQLTimestampExpr) arguments.get(0)).getDate(timeZone);
        }

        if (date == null) {
            return;
        }

        Calendar calendar = timeZone != null
                ? Calendar.getInstance(timeZone)
                : Calendar.getInstance();
        calendar.setTime(date);
        int dayOfWeek = calendar.get(Calendar.DAY_OF_WEEK);

        String name = null;
        switch (dayOfWeek) {
            case 1:
                name = "Sunday";
                break;
            case 2:
                name = "Monday";
                break;
            case 3:
                name = "Tuesday";
                break;
            case 4:
                name = "Wednesday";
                break;
            case 5:
                name = "Thursday";
                break;
            case 6:
                name = "Friday";
                break;
            case 7:
                name = "Saturday";
                break;
            default:
                break;
        }

        if (name != null && replaceInParent(x, new SQLCharExpr(name))) {
            optimizedCount++;
        }
    }

    private void func_ascii(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 1) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);

        if (arg0 instanceof SQLCharExpr) {
            String text = ((SQLCharExpr) arg0).getText();

            int ascii;
            if (text.length() > 0) {
                ascii = text.charAt(0);
            } else {
                ascii = 0;
            }
            if (replaceInParent(x, new SQLIntegerExpr(ascii))) {
                optimizedCount++;
            }
        }
    }

    private void func_makedate(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 2) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);

        if (arg0 instanceof SQLIntegerExpr && arg1 instanceof SQLIntegerExpr) {
            int year = ((SQLIntegerExpr) arg0).getNumber().intValue();
            int dayOfYear = ((SQLIntegerExpr) arg1).getNumber().intValue();

            if (dayOfYear > 0) {
                Calendar calendar = timeZone != null ? Calendar.getInstance(timeZone) : Calendar.getInstance();
                calendar.set(year, 0, 1, 0, 0, 0);
                calendar.set(Calendar.MILLISECOND, 9);
                calendar.set(Calendar.DAY_OF_YEAR, dayOfYear);

                Date date = calendar.getTime();
                if (replaceInParent(x, new SQLDateExpr(date, timeZone))) {
                    optimizedCount++;
                }
            } else {
                if (replaceInParent(x, new SQLNullExpr())) {
                    optimizedCount++;
                }
            }
        }
    }

    private void func_last_day(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 1) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);

        String text = null;
        if (arg0 instanceof SQLTimestampExpr) {
            text = ((SQLTimestampExpr) arg0).getLiteral();
        } else if (arg0 instanceof SQLCharExpr) {
            text = ((SQLCharExpr) arg0).getText();
        } else if (arg0 instanceof SQLDateExpr) {
            text = ((SQLDateExpr) arg0).getLiteral();
        }

        Date date = MySqlUtils.parseDate(text, timeZone);
        if (date == null) {
            return;
        }


        Calendar calendar = timeZone != null ? Calendar.getInstance(timeZone) : Calendar.getInstance();
        calendar.setTime(date);
        int val = calendar.getActualMaximum(Calendar.DAY_OF_MONTH);
        calendar.set(Calendar.DAY_OF_MONTH, val);

        Date lastDay = calendar.getTime();
        if (replaceInParent(x, new SQLDateExpr(lastDay, timeZone))) {
            optimizedCount++;
        }
    }

    private void func_from_unixtime(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 1) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);

        if (arg0 instanceof SQLIntegerExpr) {
            long seconds = ((SQLIntegerExpr) arg0).getNumber().intValue();
            Date date = new Date(seconds * 1000);
            if (replaceInParent(x, new SQLTimestampExpr(date, timeZone))) {
                optimizedCount++;
            }
        }
    }

    private void func_unix_timestamp(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 1) {
            return;
        }
        SQLExpr arg0 = arguments.get(0);

        String text = null;
        if (arg0 instanceof SQLTimestampExpr) {
            text = ((SQLTimestampExpr) arg0).getLiteral();
        } else if (arg0 instanceof SQLCharExpr) {
            text = ((SQLCharExpr) arg0).getText();
        } else if (arg0 instanceof SQLDateExpr) {
            text = ((SQLDateExpr) arg0).getLiteral();
        }

        if (text == null) {
            return;
        }

        Date date = MySqlUtils.parseDate(text, timeZone);
        if (date == null) {
            return;
        }

        long millis = date.getTime();
        if (millis < 0) {
            return;
        }
        int seconds = (int) (millis / 1000);

//        if (((long) seconds) * 1000 != millis) {
//            return;
//        }

        if (replaceInParent(x, new SQLIntegerExpr(seconds))) {
            optimizedCount++;
        }
    }

    private void func_json_arrayGet(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 2) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);

        if (arg0 instanceof SQLMethodInvokeExpr
                && ((SQLMethodInvokeExpr) arg0).methodNameHashCode64() == FnvHash.Constants.JSON_EXTRACT
                && ((SQLMethodInvokeExpr) arg0).getArguments().size() == 2
                && ((SQLMethodInvokeExpr) arg0).getArguments().get(1) instanceof SQLCharExpr
                && arg1 instanceof SQLIntegerExpr) {
            SQLMethodInvokeExpr jsonExtract = (SQLMethodInvokeExpr) arg0;
            SQLExpr expr = jsonExtract.getArguments().get(0);
            SQLIntegerExpr arrayIndex = (SQLIntegerExpr) arg1; //(SQLIntegerExpr) jsonExtract.getArguments().get(1);
            String path = ((SQLCharExpr) jsonExtract.getArguments().get(1)).getText();

            String path2 = path + "[" + arrayIndex.getNumber() + "]";

            SQLMethodInvokeExpr replaceTo = new SQLMethodInvokeExpr(x.getMethodName(), null, expr.clone(), new SQLCharExpr(path2));
            replaceInParent(x, replaceTo);
        }
    }

    private void func_jsonExtract(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 2) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);

        if (arg0 instanceof SQLMethodInvokeExpr
            && ((SQLMethodInvokeExpr) arg0).methodNameHashCode64() == FnvHash.Constants.JSON_ARRAY_GET
            && ((SQLMethodInvokeExpr) arg0).getArguments().size() == 2
                && ((SQLMethodInvokeExpr) arg0).getArguments().get(1) instanceof SQLIntegerExpr
                && arg1 instanceof SQLCharExpr) {
            SQLMethodInvokeExpr json_array_get = (SQLMethodInvokeExpr) arg0;
            SQLExpr expr = json_array_get.getArguments().get(0);
            SQLIntegerExpr arrayIndex = (SQLIntegerExpr) json_array_get.getArguments().get(1);
            String path = ((SQLCharExpr) arg1).getText();

            if (path.startsWith("$.") && arrayIndex.getNumber() instanceof Integer) {
                String rewritePath = "$[" + arrayIndex.getNumber() + "]." + path.substring(2);
                SQLMethodInvokeExpr replaceTo = new SQLMethodInvokeExpr(x.getMethodName(), null, expr.clone(), new SQLCharExpr(rewritePath));
                replaceInParent(x, replaceTo);
            }
        }
    }

    private void func_trunc(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 2) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);

        Date date0 = null;
        if (arg0 instanceof SQLTimestampExpr) {
            date0 = ((SQLTimestampExpr) arg0).getDate(timeZone);
        } else if (arg0 instanceof SQLCharExpr) {
            String str0 = ((SQLCharExpr) arg1).getText();
            date0 = MySqlUtils.parseDate(str0, timeZone);
        }

        if (date0 == null) {
            return;
        }

        if (!(arg1 instanceof SQLCharExpr)) {
            return;
        }

        Calendar calendar = timeZone == null ? Calendar.getInstance() : Calendar.getInstance(timeZone);
        calendar.setTime(date0);

        long unitHash = FnvHash.hashCode64(
                ((SQLCharExpr) arg1).getText());

        boolean timestamp = false;
        if (unitHash == FnvHash.Constants.YEAR) {
            calendar.set(Calendar.MONTH, 0);
            calendar.set(Calendar.DAY_OF_MONTH, 1);
            calendar.set(Calendar.HOUR_OF_DAY, 0);
            calendar.set(Calendar.MINUTE, 0);
            calendar.set(Calendar.SECOND, 0);
            calendar.set(Calendar.MILLISECOND, 0);
        } else if (unitHash == FnvHash.Constants.MONTH) {
            calendar.set(Calendar.DAY_OF_MONTH, 1);
            calendar.set(Calendar.HOUR_OF_DAY, 0);
            calendar.set(Calendar.MINUTE, 0);
            calendar.set(Calendar.SECOND, 0);
            calendar.set(Calendar.MILLISECOND, 0);
        } else if (unitHash == FnvHash.Constants.DAY) {
            calendar.set(Calendar.HOUR_OF_DAY, 0);
            calendar.set(Calendar.MINUTE, 0);
            calendar.set(Calendar.SECOND, 0);
            calendar.set(Calendar.MILLISECOND, 0);
        } else if (unitHash == FnvHash.Constants.HOUR) {
            calendar.set(Calendar.MINUTE, 0);
            calendar.set(Calendar.SECOND, 0);
            calendar.set(Calendar.MILLISECOND, 0);
            timestamp = true;
        } else if (unitHash == FnvHash.Constants.MINUTE) {
            calendar.set(Calendar.SECOND, 0);
            calendar.set(Calendar.MILLISECOND, 0);
            timestamp = true;
        } else if (unitHash == FnvHash.Constants.SECOND) {
            calendar.set(Calendar.MILLISECOND, 0);
            timestamp = true;
        } else {
            return;
        }

        SQLExpr replaceTo = timestamp
                ? new SQLTimestampExpr(calendar.getTime(), timeZone)
                : new SQLDateExpr(calendar.getTime(), timeZone);
        if (replaceInParent(x, replaceTo)) {
            optimizedCount++;
        }
    }

    private void func_month_between(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 2) {
            return;
        }

        SQLExpr arg1 = arguments.get(0);
        SQLExpr arg2 = arguments.get(1);

        String str1 = null, str2 = null;
        if (arg1 instanceof SQLCharExpr) {
            str1 = ((SQLCharExpr) arg1).getText();
        }
        if (arg2 instanceof SQLCharExpr) {
            str2 = ((SQLCharExpr) arg2).getText();
        }

        Integer delta = Jdk8TimeUtils.monthBetween(str1, str2, timeZone);
        if (delta != null && replaceInParent(x, new SQLIntegerExpr(delta))) {
            optimizedCount++;
        }
    }

    private void func_monthname(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 1) {
            return;
        }
        SQLExpr arg0 = arguments.get(0);

        Date date = null;
        if (arg0 instanceof SQLDateExpr) {
            date = ((SQLDateExpr) arg0).getDate(timeZone);
        } else if (arg0 instanceof SQLCharExpr) {
            String text = ((SQLCharExpr) arg0).getText();
            SimpleDateFormat format
                    = text.length() == 10
                    ? new SimpleDateFormat("yyyy-MM-dd")
                    : new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
            try {
                date = format.parse(text);
            } catch (ParseException e) {
                // skip
            }
        }

        if (date == null) {
            return;
        }

        Calendar calendar = timeZone == null ? Calendar.getInstance() : Calendar.getInstance(timeZone);
        calendar.setTime(date);
        int month = calendar.get(Calendar.MONTH);

        String monthName = null;
        switch (month) {
            case 0:
                monthName = "January";
                break;
            case 1:
                monthName = "February";
                break;
            case 2:
                monthName = "March";
                break;
            case 3:
                monthName = "April";
                break;
            case 4:
                monthName = "May";
                break;
            case 5:
                monthName = "June";
                break;
            case 6:
                monthName = "July";
                break;
            case 7:
                monthName = "August";
                break;
            case 8:
                monthName = "September";
                break;
            case 9:
                monthName = "October";
                break;
            case 10:
                monthName = "November";
                break;
            case 11:
                monthName = "December";
                break;
            default:
                break;
        }

        if (monthName != null && replaceInParent(x, new SQLCharExpr(monthName))) {
            optimizedCount++;
        }
    }

    private void func_date(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 1) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        if (arg0 instanceof SQLCharExpr) {
            String str0 = ((SQLCharExpr) arg0).getText();

            Date date = MySqlUtils.parseDate(str0, timeZone);

            if (date != null && replaceInParent(x, new SQLDateExpr(date, timeZone))) {
                optimizedCount++;
            }
        }

    }

    private void func_timestamp(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 1) {

            SQLExpr arg0 = arguments.get(0);

            String str0;
            if (arg0 instanceof SQLCharExpr) {
                str0 = ((SQLCharExpr) arg0).getText();
                if (str0.indexOf('.') > 0) {
                    replaceInParent(x, new SQLTimestampExpr(str0));
                    return;
                }
            } else if (arg0 instanceof SQLIntegerExpr) {
                str0 = ((SQLIntegerExpr) arg0).getNumber().toString();
            } else {
                return;
            }

            Date date = MySqlUtils.parseDate(str0, timeZone);
            if (date == null) {
                return;
            }

            replaceInParent(x, new SQLTimestampExpr(date, timeZone));
        } else if (arguments.size() == 2) {
            SQLExpr arg0 = arguments.get(0);
            SQLExpr arg1 = arguments.get(1);

            if (arg0 instanceof SQLCharExpr && arg1 instanceof SQLCharExpr) {
                String str = ((SQLCharExpr) arg0).getText();
                String fmt = ((SQLCharExpr) arg1).getText();
                DateFormat dateFormat = MySqlUtils.toJavaFormat(fmt, timeZone);
                if (dateFormat != null) {
                    try {
                        Date date = dateFormat.parse(str);
                        SQLTimestampExpr ts = new SQLTimestampExpr(date, timeZone);
                        replaceInParent(x, ts);
                    } catch (ParseException e) {
                        // skip
                    }
                }
            }
        }
    }

    private void func_datetrunc(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if(arguments.size() != 2) {
            return;
        }

        SQLExpr arg1 = arguments.get(1);
        if (arg1 instanceof SQLCharExpr) {
            SQLTimestampExpr ts = new SQLTimestampExpr(((SQLCharExpr) arg1).getText());
            x.setArgument(1, ts);
        }
    }

    private void func_datediff(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();

        if(arguments.size() == 2 && (arguments.get(0) instanceof SQLNullExpr
                                    || arguments.get(1) instanceof SQLNullExpr)) {
            if (replaceInParent(x, new SQLNullExpr())) {
                optimizedCount++;
            }
        }

        if (arguments.size() == 2
                && arguments.get(0) instanceof SQLCharExpr
                && arguments.get(1) instanceof SQLCharExpr) {
            String str0 = ((SQLCharExpr) arguments.get(0)).getText();
            String str1 = ((SQLCharExpr) arguments.get(1)).getText();

            Integer days = Jdk8TimeUtils.dateDiff(str0, str1, timeZone);
            if (days != null && replaceInParent(x, new SQLIntegerExpr(days))) {
                optimizedCount++;
            }
        }

        if (arguments.size() == 3
                && timeZone != null) {
            SQLExpr arg0 = arguments.get(0);
            SQLExpr arg1 = arguments.get(1);
            SQLExpr arg2 = arguments.get(2);

            long unitHash;
            if (arg0 instanceof SQLIdentifierExpr) {
                unitHash = ((SQLIdentifierExpr) arg0).nameHashCode64();
            } else if (arg0 instanceof SQLCharExpr) {
                String text = ((SQLCharExpr) arg0).getText();
                unitHash = FnvHash.hashCode64(text);
            } else {
                return;
            }

            if (arg1 instanceof SQLCharExpr && arg2 instanceof SQLTimeExpr) {
                arg1 = new SQLTimeExpr(((SQLCharExpr) arg1).getText());
                x.setArgument(1, arg1);
                return;
            }

            String str1 = null, str2 = null;
            if (arg1 instanceof SQLCharExpr) {
                str1 = ((SQLCharExpr) arg1).getText();
            } else if (arg1 instanceof SQLDateExpr) {
                str1 = ((SQLDateExpr) arg1).getLiteral();
            } else if (arg1 instanceof SQLTimestampExpr) {
                str1 = ((SQLTimestampExpr) arg1).getLiteral();
            }

            if (arg2 instanceof SQLCharExpr) {
                str2 = ((SQLCharExpr) arg2).getText();
            } else if (arg2 instanceof SQLDateExpr) {
                str2 = ((SQLDateExpr) arg2).getLiteral();
            } else if (arg2 instanceof SQLTimestampExpr) {
                str2 = ((SQLTimestampExpr) arg2).getLiteral();
            }

            Integer delta = Jdk8TimeUtils.dateDiff1(unitHash, str1, str2, timeZone);
            if (delta != null && replaceInParent(x, new SQLIntegerExpr(delta))) {
                optimizedCount++;
            }
        }
    }

    private void func_timestamp_diff(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 3
                && timeZone != null
                && arguments.get(0) instanceof SQLIdentifierExpr) {
            SQLExpr arg1 = arguments.get(1);
            SQLExpr arg2 = arguments.get(2);

            long unitHash = ((SQLIdentifierExpr) arguments.get(0)).nameHashCode64();

            String str1 = null, str2 = null;
            if (arg1 instanceof SQLCharExpr) {
                str1 = ((SQLCharExpr) arg1).getText();
            }
            if (arg2 instanceof SQLCharExpr) {
                str2 = ((SQLCharExpr) arg2).getText();
            }

            Integer delta = Jdk8TimeUtils.timestampDiff(unitHash, str1, str2, timeZone);
            if (delta == null) {
                return;
            }
            if (replaceInParent(x, new SQLIntegerExpr(delta))) {
                optimizedCount++;
            }
        }
    }

    private void func_get_format(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 2
                && arguments.get(0) instanceof SQLIdentifierExpr
                && arguments.get(1) instanceof SQLCharExpr) {
            long nameHash = ((SQLIdentifierExpr) arguments.get(0)).nameHashCode64();
            String str1 = ((SQLCharExpr) arguments.get(1)).getText();

            SQLExpr replaceTo = null;
            if (nameHash == FnvHash.Constants.DATE) {
                if (str1.equals("USA")) {
                    replaceTo = new SQLCharExpr("%m.%d.%Y");
                } else if (str1.equals("JIS") || str1.equals("ISO")) {
                    replaceTo = new SQLCharExpr("%Y-%m-%d");
                } else if (str1.equals("EUR")) {
                    replaceTo = new SQLCharExpr("%d.%m.%Y");
                } else if (str1.equals("INTERNAL")) {
                    replaceTo = new SQLCharExpr("%Y%m%d");
                }
            } else if (nameHash == FnvHash.Constants.DATETIME) {
                if (str1.equals("USA") || str1.equals("EUR")) {
                    replaceTo = new SQLCharExpr("%Y-%m-%d %H.%i.%s");
                } else if (str1.equals("JIS") || str1.equals("ISO")) {
                    replaceTo = new SQLCharExpr("%Y-%m-%d %H:%i:%s");
                } else if (str1.equals("INTERNAL")) {
                    replaceTo = new SQLCharExpr("%Y%m%d%H%i%s");
                }
            } else if (nameHash == FnvHash.Constants.TIME) {
                if (str1.equals("USA")) {
                    replaceTo = new SQLCharExpr("%h:%i:%s %p");
                } else if (str1.equals("JIS") || str1.equals("ISO")) {
                    replaceTo = new SQLCharExpr("%H:%i:%s");
                } else if (str1.equals("EUR")) {
                    replaceTo = new SQLCharExpr("%H.%i.%s");
                } else if (str1.equals("INTERNAL")) {
                    replaceTo = new SQLCharExpr("%H%i%s");
                }
            }

            if (replaceTo != null && replaceInParent(x, replaceTo)) {
                optimizedCount++;
            }
        }
    }

    private void func_str_to_date(SQLMethodInvokeExpr x) {
        List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 2 && arguments.get(0) instanceof SQLTimestampExpr) {
            SQLTimestampExpr ts = (SQLTimestampExpr) arguments.get(0);
            x.setArgument(0, new SQLCharExpr(ts.getLiteral()));
            optimizedCount++;
        }

        if (arguments.size() == 2
                && arguments.get(0) instanceof SQLCharExpr
                && arguments.get(1) instanceof SQLCharExpr) {
            String str = ((SQLCharExpr) arguments.get(0)).getText();
            String fmt = ((SQLCharExpr) arguments.get(1)).getText();

            if (fmt.indexOf("%Y") == -1 || fmt.indexOf("%m") == -1) {
                return;
            }

            SQLExpr replaceTo = null;
            if (fmt.equals("%Y-%m-%d %H:%i:%s")) {
                replaceTo = new SQLTimestampExpr(str);
            } else if (fmt.equals("%Y-%m-%d")) {
                replaceTo = new SQLDateExpr(str);
//            } else if (fmt.equals("%Y")) {
//                int year = Integer.parseInt(str);
//                replaceTo = new SQLCharExpr(year + "-00-00");
//            } else if (fmt.equals("%m")) {
//                int month = Integer.parseInt(str);
//                String dateStr = (month < 10 ? "0000-0" : "0000-") + month + "-00";
//                replaceTo = new SQLCharExpr(dateStr);
            } else {
                if (str.indexOf("-00") != -1 || str.indexOf("/00") != -1) {
                    return;
                }

                Date date = null;
                DateFormat dateFormat;
                dateFormat = mysqlDateFormat(fmt, timeZone);
                if (dateFormat != null) {
                    try {
                        date = dateFormat.parse(str);
                    } catch (ParseException e) {

                    }
                }

                if (date != null) {
                    Calendar calendar = timeZone == null
                            ? Calendar.getInstance()
                            : Calendar.getInstance(timeZone);
                    calendar.setTime(date);

                    int hour = calendar.get(Calendar.HOUR_OF_DAY);
                    int minute = calendar.get(Calendar.MINUTE);
                    int second = calendar.get(Calendar.SECOND);
                    int millis = calendar.get(Calendar.MILLISECOND);

                    if (hour == 0 && minute == 0 && second == 0 && millis == 0) {
                        replaceTo = new SQLDateExpr(date, timeZone);
                    } else {
                        replaceTo = new SQLTimestampExpr(date, timeZone);
                    }
                }
            }


            if (replaceTo != null
                    && replaceInParent(x, replaceTo)) {
                optimizedCount++;
            }
        }
    }

    private void func_if(SQLMethodInvokeExpr x) {
        for (SQLExpr arg : x.getArguments()) {
            arg.accept(this);
        }

        if (x.getArguments().size() != 3) {
            return;
        }

        SQLExpr itemCondition = x.getArguments().get(0);
        SQLExpr trueValueExpr = x.getArguments().get(1);
        SQLExpr falseValueExpr = x.getArguments().get(2);
        if (itemCondition instanceof SQLValuableExpr) {
            Object condVal = ((SQLValuableExpr) itemCondition).getValue();
            if (condVal instanceof Boolean) {
                if (((Boolean) condVal).booleanValue()) {
                    replaceInParent(x, trueValueExpr);
                } else {
                    replaceInParent(x, falseValueExpr);
                }
            }
            return;
        }

        boolean conditionIsTrue = false;
        final SQLObject parent = x.getParent();
        if (parent instanceof SQLBinaryOpExpr
                && ((SQLBinaryOpExpr) parent).getOperator().isRelational()) {
            Object parentParent = parent.getParent();
            if (parentParent instanceof SQLBinaryOpExpr) {
                SQLBinaryOpExpr binaryOpParent = (SQLBinaryOpExpr) parentParent;

                if (binaryOpParent.getOperator() == BooleanAnd) {
                    if (itemCondition.equals(binaryOpParent.getLeft()) || itemCondition.equals(binaryOpParent.getRight())) {
                        conditionIsTrue = true;
                    }
                }
            } else if (parentParent instanceof SQLBinaryOpExprGroup) {
                SQLBinaryOpExprGroup binaryOpParentGroup = (SQLBinaryOpExprGroup) parentParent;

                if (binaryOpParentGroup.getOperator() == BooleanAnd) {
                    if (binaryOpParentGroup.getItems().contains(itemCondition)) {
                        conditionIsTrue = true;
                    }
                }
            }
        }

        if (conditionIsTrue) {
            replaceInParent(x, trueValueExpr);
        }
    }

    private void func_coalesce(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        int nullCount = 0;
        for (SQLExpr arg : arguments) {
            if (arg instanceof SQLLiteralExpr) {
                if (arg instanceof SQLNullExpr) {
                    nullCount++;
                    continue;
                }
                if (replaceInParent(x, arg.clone())) {
                    optimizedCount++;
                    return;
                }
            }
        }

        if (nullCount == arguments.size()
                && replaceInParent(x, new SQLNullExpr())) {
            optimizedCount++;
        }
    }

    private void func_ifnull(SQLMethodInvokeExpr x) {
        Object val = IfNull.instance.eval(x);
        if (val instanceof Integer) {
            if (replaceInParent(x, new SQLIntegerExpr((Integer) val))) {
                optimizedCount++;
            }
            return;
        }

        if (val instanceof String) {
            if (replaceInParent(x, new SQLCharExpr((String) val))) {
                optimizedCount++;
            }
            return;
        }
    }

    private void func_locate(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 2
                && arguments.get(0) instanceof SQLCharExpr
                && arguments.get(1) instanceof SQLCharExpr) {
            String str0 = ((SQLCharExpr) arguments.get(0)).getText();
            String str1 = ((SQLCharExpr) arguments.get(1)).getText();
            int index = str1.indexOf(str0) + 1;

            if (replaceInParent(x, new SQLIntegerExpr(index))) {
                optimizedCount++;
            }
        } else if (arguments.size() == 3
                && arguments.get(0) instanceof SQLCharExpr
                && arguments.get(1) instanceof SQLCharExpr
                && arguments.get(2) instanceof SQLIntegerExpr) {
            String str0 = ((SQLCharExpr) arguments.get(0)).getText();
            String str1 = ((SQLCharExpr) arguments.get(1)).getText();
            int offset = ((SQLIntegerExpr) arguments.get(2)).getNumber().intValue();
            if (offset <= 0) {
                return;
            }

            int index = str1.indexOf(str0, offset - 1) + 1;

            if (replaceInParent(x, new SQLIntegerExpr(index))) {
                optimizedCount++;
            }
        }
    }

    private void func_left(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 2
                && arguments.get(0) instanceof SQLCharExpr
                && arguments.get(1) instanceof SQLIntegerExpr) {
            String str = ((SQLCharExpr) arguments.get(0)).getText();
            int len = ((SQLIntegerExpr) arguments.get(1)).getNumber().intValue();

            if (len <= 0) {
                len = 0;
            }

            if(str.length() < len) {
                len = str.length();
            }

            String left = str.substring(0, len);

            if (replaceInParent(x, new SQLCharExpr(left))) {
                optimizedCount++;
            }
        }
    }

    private void func_right(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 2
                && arguments.get(0) instanceof SQLCharExpr
                && arguments.get(1) instanceof SQLIntegerExpr) {
            String str = ((SQLCharExpr) arguments.get(0)).getText();
            int len = ((SQLIntegerExpr) arguments.get(1)).getNumber().intValue();

            if (len <= 0) {
                len = 0;
            }

            if (len >= str.length()) {
                len = str.length();
            }
            String right = str.substring(str.length() - len, str.length());

            if (replaceInParent(x, new SQLCharExpr(right))) {
                optimizedCount++;
            }
        }
    }

    private void func_rtrim(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 1
                && arguments.get(0) instanceof SQLCharExpr) {
            String str = ((SQLCharExpr) arguments.get(0)).getText();
            String rtrim = StringUtils.rtrim(str);

            if (rtrim != str && replaceInParent(x, new SQLCharExpr(rtrim))) {
                optimizedCount++;
            }
        }
    }

    private void func_ltrim(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 1
                && arguments.get(0) instanceof SQLCharExpr) {
            String str = ((SQLCharExpr) arguments.get(0)).getText();
            String ltrim = StringUtils.ltrim(str);

            if (ltrim != str && replaceInParent(x, new SQLCharExpr(ltrim))) {
                optimizedCount++;
            }
        }
    }

    private void func_trim(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 1
                && arguments.get(0) instanceof SQLCharExpr) {
            String str = ((SQLCharExpr) arguments.get(0)).getText();
            String trim = str.trim();

            SQLExpr from = x.getFrom();

            if (from != null) {
                return;
            }

            if (trim != str && replaceInParent(x, new SQLCharExpr(trim))) {
                optimizedCount++;
            }
        }
    }

    private void func_reverse(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 1
                && arguments.get(0) instanceof SQLCharExpr) {
            String str = ((SQLCharExpr) arguments.get(0)).getText();
            char[] chars = new char[str.length()];
            for (int i = 0; i < str.length(); ++i) {
                chars[chars.length - i - 1] = str.charAt(i);
            }

            if (replaceInParent(x, new SQLCharExpr(new String(chars)))) {
                optimizedCount++;
            }
        }
    }

    private void func_hex(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 1) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);

        if (arg0 instanceof SQLCharExpr) {
            String text = ((SQLCharExpr) arg0).getText();
            try {
                String str = HexBin.encode(text.getBytes("UTF-8"));
                if (replaceInParent(x, new SQLCharExpr(str))) {
                    optimizedCount++;
                }
            } catch (Exception e) {

            }
        } else if (arg0 instanceof SQLHexExpr) {
            String text = ((SQLHexExpr) arg0).getHex();
            if (replaceInParent(x, new SQLCharExpr(text))) {
                optimizedCount++;
            }
        }
    }

    private void func_isnull(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 1) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        if (arg0 instanceof SQLNullExpr) {
            replaceInParent(x, new SQLBooleanExpr(true));
        } else if (arg0 instanceof SQLLiteralExpr) {
            replaceInParent(x, new SQLBooleanExpr(false));
        }
    }

    private void func_unhex(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 1) {
            return;
        }

        final SQLObject parent = x.getParent();
        if (parent instanceof SQLBinaryOpExpr) {
            final SQLBinaryOperator operator = ((SQLBinaryOpExpr) parent).getOperator();
            switch (operator) {
                case BooleanAnd:
                case BooleanOr:
                case BooleanXor:
                case BitwiseXor:
                case BitwiseAnd:
                case BitwiseOr:
                case Is:
                case IsNot:
                case GreaterThan:
                case GreaterThanOrEqual:
                case LessThan:
                case LessThanOrEqual:
                case LeftShift:
                case RightShift:
                    return;
                default:
                    break;
            }
        } else if (parent instanceof SQLBetweenExpr || parent instanceof SQLInListExpr || parent instanceof SQLUnaryExpr) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        if (arg0 instanceof SQLCharExpr) {
            String text = ((SQLCharExpr) arg0).getText();
            for (int i = 0; i < text.length(); ++i) {
                char ch = text.charAt(i);
                if (ch >= '0' && ch <= '9') {
                    continue;
                }
                if (ch >= 'a' && ch <= 'f') {
                    continue;
                }
                if (ch >= 'A' && ch <= 'F') {
                    continue;
                }
                return;
            }

            SQLExpr replaceTo = new SQLHexExpr(text);
            if (replaceInParent(x, replaceTo)) {
                optimizedCount++;
                replaceTo.accept(this);
            }
        }
    }

    private void func_interval(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() < 2) {
            return;
        }

        int[] values = new int[arguments.size()];
        for (int i = 0; i < arguments.size(); i++) {
            SQLExpr arg = arguments.get(i);
            if (arg instanceof SQLNumericLiteralExpr) {
                Number number = ((SQLNumericLiteralExpr) arg).getNumber();
                if (number instanceof Integer) {
                    values[i] = (Integer) number;
                } else {
                    return;
                }
            } else {
                return;
            }
        }

        SQLExpr replaceTo = null;
        int first = values[0];
        for (int i = 1; i < values.length; ++i) {
            if (values[i] == first) {
                replaceTo = new SQLIntegerExpr(i);
            } else if (values[i] > first) {
                replaceTo = new SQLIntegerExpr(i - 1);
                break;
            }
        }
        if (replaceTo == null) {
            replaceTo = new SQLIntegerExpr(values.length);
        }
        if (replaceInParent(x, replaceTo)) {
            optimizedCount++;
        }
    }

    private void func_abs(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 1) {
            return;
        }

        SQLExpr arg = arguments.get(0);
        if (arg instanceof SQLIntegerExpr) {
            Number num = ((SQLIntegerExpr) arg).getNumber();
            if (num instanceof Integer) {
                int intVal = ((Integer) num).intValue();
                int absVal = Math.abs(intVal);
                if (replaceInParent(x, new SQLIntegerExpr(absVal))) {
                    optimizedCount++;
                }
            } else if (num instanceof Long) {
                long intVal = ((Long) num).longValue();
                long absVal = Math.abs(intVal);
                if (replaceInParent(x, new SQLIntegerExpr(absVal))) {
                    optimizedCount++;
                }
            } else if (num instanceof BigInteger) {
                BigInteger absVal = ((BigInteger) num).abs();
                replaceInParent(x, new SQLIntegerExpr(absVal));
            }
        } else if (arg instanceof SQLNullExpr) {
            if (replaceInParent(x, new SQLNullExpr())) {
                optimizedCount++;
            }
        }
    }

    private void func_floor(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 1) {
            return;
        }

        SQLExpr arg = arguments.get(0);
        if (arg instanceof SQLIntegerExpr) {
            replaceInParent(x, arg.clone());
            return;
        } else if (arg instanceof SQLNumberExpr) {
            Number number = ((SQLNumberExpr) arg).getNumber();
            if (number instanceof BigDecimal) {
                long value = ((BigDecimal) number).setScale(0, BigDecimal.ROUND_FLOOR).longValue();
                replaceInParent(x, SQLIntegerExpr.ofIntOrLong(value));
            }
            return;
        } else if (arg instanceof SQLNullExpr) {
            if (replaceInParent(x, new SQLNullExpr())) {
                optimizedCount++;
            }
        }
    }

    private void func_nullif(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();

        if (arguments.size() != 2) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);

        boolean eq = false;
        if (arg0.equals(arg1)) {
            eq = true;
        }

        final SQLObject parent = x.getParent();
        if (parent instanceof SQLBetweenExpr) {
            return;
        }


        if (eq && replaceInParent(x, new SQLNullExpr())) {
            optimizedCount++;
            return;
        }

    }

    private void func_greatest(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 1) {
            if (replaceInParent(x, arguments.get(0).clone())) {
                optimizedCount++;
            }

            return;
        }

        Object[] values = new Object[arguments.size()];
        Class<?> clazz = null;
        for (int i = 0; i < arguments.size(); i++) {
            SQLExpr arg = arguments.get(i);
            if (arg instanceof SQLNullExpr) {
                if (replaceInParent(x, new SQLNullExpr())) {
                    optimizedCount++;
                }
                return;
            }
            Object value = ((SQLValuableExpr) arg).getValue();
            if (i == 0) {
                clazz = value.getClass();
            } else {
                if (value.getClass() != clazz) {
                    return;
                }
            }
            values[i] = value;
        }

        if (clazz == Integer.class) {
            int max = (Integer) values[0];
            for (int i = 1; i < values.length; ++i) {
                int intVal = ((Integer) values[i]).intValue();
                if (intVal > max) {
                    max = intVal;
                }
            }
            if (replaceInParent(x, new SQLIntegerExpr(max))) {
                optimizedCount++;
            }
            return;
        }
    }

    private void func_least(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() == 1) {
            if (replaceInParent(x, arguments.get(0).clone())) {
                optimizedCount++;
            }

            return;
        }

        Object[] values = new Object[arguments.size()];
        Class<?> clazz = null;
        for (int i = 0; i < arguments.size(); i++) {
            SQLExpr arg = arguments.get(i);
            if (arg instanceof SQLNullExpr) {
                continue;
            }

            Object value = ((SQLValuableExpr) arg).getValue();
            if (i == 0) {
                clazz = value.getClass();
            } else {
                if (value.getClass() != clazz) {
                    return;
                }
            }
            values[i] = value;
        }

        if (clazz == Integer.class) {
            Integer min = null;
            for (int i = 0; i < values.length; ++i) {
                Integer obj = ((Integer) values[i]);
                if (obj == null) {
                    continue;
                }

                if (min == null || obj.intValue() < min.intValue()) {
                    min = obj;
                }
            }

            if (min == null) {
                if (replaceInParent(x, new SQLNullExpr())) {
                    optimizedCount++;
                }
            } else if (replaceInParent(x, new SQLIntegerExpr(min))) {
                optimizedCount++;
            }
            return;
        }
    }

    private void func_period_add(SQLMethodInvokeExpr x) {
        final List<SQLExpr> arguments = x.getArguments();
        if (arguments.size() != 2) {
            return;
        }

        SQLExpr arg0 = arguments.get(0);
        SQLExpr arg1 = arguments.get(1);
        if (arg0 instanceof SQLIntegerExpr && arg1 instanceof SQLIntegerExpr) {
            int val = ((SQLIntegerExpr) arg0).getNumber().intValue();
            int delta = ((SQLIntegerExpr) arg1).getNumber().intValue();

            if (val == 0) {
                return;
            }

            if (val >= 100 && val < 10000) {
                val += 200000;
            }
            int year = val / 100;
            if (year == 0) {
                year = 2000;
            }
            int month = val % 100 - 1;

            Calendar calendar = timeZone != null ? Calendar.getInstance(timeZone) : Calendar.getInstance();
            calendar.set(Calendar.YEAR, year);
            calendar.set(Calendar.MONTH, month);
            calendar.add(Calendar.MONTH, delta);

            int result = calendar.get(Calendar.YEAR) * 100 + calendar.get(Calendar.MONTH) + 1;
            if (replaceInParent(x, new SQLIntegerExpr(result))) {
                optimizedCount++;
            }
        }
    }

    private void dateAdd(SQLMethodInvokeExpr x, SQLExpr arg0, SQLExpr arg1, boolean negative) {
        String text;
        boolean timestamp = false;
        if (arg0 instanceof SQLCharExpr) {
            text = ((SQLCharExpr) arg0).getText();
            if (text.length() > 10) {
                timestamp = true;
            }
        } else if (arg0 instanceof SQLDateExpr) {
            text = ((SQLDateExpr) arg0).getLiteral();
        } else {
            text = ((SQLTimestampExpr) arg0).getLiteral();
            timestamp = true;
        }

        SQLIntervalUnit unit = null;
        SQLExpr valueExpr = null;
        if (arg1 instanceof SQLIntegerExpr) {
            valueExpr = arg1;
            unit = SQLIntervalUnit.DAY;
        } else {
            SQLIntervalExpr interval = (SQLIntervalExpr) arg1;
            unit = interval.getUnit();
            valueExpr = interval.getValue();
        }

        Integer interval = null;
        if (valueExpr instanceof SQLIntegerExpr) {
            interval = ((SQLIntegerExpr) valueExpr).getNumber().intValue();
        } else if (valueExpr instanceof SQLCharExpr) {
            String intervalText = ((SQLCharExpr) valueExpr).getText();
            try {
                interval = Integer.parseInt(intervalText);
            } catch (NumberFormatException ex) {
                // skip
            }
        }

        if (interval == null) {
            return;
        }

        Date date = MySqlUtils.parseDate(text, timeZone);
        if (date != null
                && timeZone != null
                && unit != null) {
            int delta = interval;
            if (negative) {
                delta = -delta;
            }

            Calendar calendar = Calendar.getInstance(timeZone);
            calendar.setTime(date);
            switch (unit) {
                case YEAR:
                    calendar.add(Calendar.YEAR, delta);
                    break;
                case MONTH:
                    calendar.add(Calendar.MONTH, delta);
                    break;
                case DAY:
                    calendar.add(Calendar.DAY_OF_YEAR, delta);
                    break;
                case HOUR:
                    calendar.add(Calendar.HOUR_OF_DAY, delta);
                    break;
                case MINUTE:
                    calendar.add(Calendar.MINUTE, delta);
                    break;
                case SECOND:
                    calendar.add(Calendar.SECOND, delta);
                    break;
                default:
                    return;
            }

            DateFormat tzFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
            tzFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
            tzFormat.setTimeZone(timeZone);

            String result = tzFormat.format(calendar.getTime());

            SQLExpr replaceTo;
            if (result.endsWith(" 00:00:00") && !timestamp) {
                replaceTo = new SQLDateExpr(result.substring(0, result.length() - 9));
            } else {
                replaceTo = new SQLTimestampExpr(result);
            }
            if (replaceInParent(x, replaceTo)) {
                optimizedCount++;
            }

        }
    }

    static DateFormat dateFormatFromValue(String value) {
        return dateFormatFromValue(value, null);
    }

    static DateFormat dateFormatFromValue(String value, TimeZone timeZone) {
        if (value == null) {
            return null;
        }

        SimpleDateFormat df = null;
        if (value.length() == 19) {
            df = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
        }

        if (value.length() == 10) {
            if (value.endsWith("-00")) {
                return null;
            }

            df = new SimpleDateFormat("yyyy-MM-dd");
        }

        if (value.length() == 8) {
            df = new SimpleDateFormat("yyyyMMdd");
        }

        if (df != null && timeZone != null) {
            df.setTimeZone(timeZone);
        }

        return df;
    }

    static DateFormat mysqlDateFormat(String fmt, TimeZone timeZone) {
        if (fmt == null) {
            return null;
        }


        DateFormat df = MySqlUtils.toJavaFormat(fmt);
        if (df != null && timeZone != null) {
            df.setTimeZone(timeZone);
        }

        return df;
    }

    public void visitBinaryFunc(SQLBinaryOpExpr x) {
        SQLExpr left = x.getLeft();
        SQLExpr right = x.getRight();

        if (left instanceof SQLCastExpr && right instanceof SQLValuableExpr) {
            binaryOp_cast(x);
            return;
        }

        if (!(left instanceof SQLMethodInvokeExpr && right instanceof SQLLiteralExpr)) {
            return;
        }

        SQLMethodInvokeExpr func = (SQLMethodInvokeExpr) left;
        long methodNameHashCode64 = func.methodNameHashCode64();

        if (methodNameHashCode64 == FnvHash.Constants.DATE_FORMAT) {
            binaryOp_date_format(x);

        } else if (methodNameHashCode64 == FnvHash.Constants.DATE_DIFF
                || methodNameHashCode64 == FnvHash.Constants.DATEDIFF)
        {
            binaryOp_date_diff(x);

        } else if (methodNameHashCode64 == FnvHash.Constants.YEARMONTH) {
            binaryOp_yearmonth(x);
        } else if (methodNameHashCode64 == FnvHash.Constants.COALESCE) {
            binaryOp_coalesce(x);
        } else if (methodNameHashCode64 == FnvHash.Constants.ADDDATE
                || methodNameHashCode64 == FnvHash.Constants.DATE_ADD)
        {
            binaryOp_add_date(x);
        } else if (methodNameHashCode64 == FnvHash.Constants.SUBDATE) {
            binaryOp_sub_date(x);

        } else if (methodNameHashCode64 == FnvHash.Constants.FROM_UNIXTIME) {
            binaryOp_from_unixtime(x);

        } else if (methodNameHashCode64 == FnvHash.Constants.DATE) {
            binaryOp_date(x);
        }
    }

    public void binaryOp_cast(SQLBinaryOpExpr x) {
        SQLCastExpr cast = (SQLCastExpr) x.getLeft();
        SQLValuableExpr right = (SQLValuableExpr) x.getRight();

        final SQLExpr expr = cast.getExpr();
        final SQLDataType dataType = cast.getDataType();
        final Object value = right.getValue();
        final SQLDataType exprDataType = expr.computeDataType();

        if (exprDataType == null) {
            return;
        }
        if (!x.getOperator().isRelational()) {
            return;
        }

        SQLBinaryOperator op = x.getOperator();
        if ((op == Equality || op == NotEqual || op == LessThanOrGreater)) {
            if (value instanceof String
                    && dataType instanceof SQLCharacterDataType
                    && exprDataType.isInt()
            ) {
                try {
                    long longVal = Long.parseLong((String) value);
                    x.setRight(SQLIntegerExpr.ofIntOrLong(longVal));
                    x.setLeft(expr);
                } catch (NumberFormatException ex) {
                    // skip
                }
            } else if (value instanceof Long
                    && dataType.nameHashCode64() == FnvHash.Constants.BIGINT
                    && exprDataType instanceof SQLCharacterDataType
            ) {
                SQLCharExpr charExpr = new SQLCharExpr(value.toString());
                x.setRight(charExpr);
                x.setLeft(expr);

            } else if (value instanceof Long
                    && dataType instanceof SQLCharacterDataType
                    && exprDataType.nameHashCode64() == FnvHash.Constants.BIGINT
            ) {
                x.setLeft(expr);
            }
        }
    }

    public void visitBinaryOpBinaryOpExpr(SQLBinaryOpExpr x) {
        if (!(x.getLeft() instanceof SQLBinaryOpExpr)) {
            return;
        }

        if (!(x.getRight() instanceof SQLNumericLiteralExpr)) {
            return;
        }

        SQLBinaryOpExpr left = (SQLBinaryOpExpr) x.getLeft();
        final SQLExpr leftRight = left.getRight();
        if (!(leftRight instanceof SQLNumericLiteralExpr)) {
            return;
        }

        if (!x.getOperator().isRelational()) {
            return;
        }

        Number number = ((SQLNumericLiteralExpr) x.getRight()).getNumber();
        Number leftRightNumber = ((SQLNumericLiteralExpr) leftRight).getNumber();

        switch (left.getOperator()) {
            case DIV:
            case Divide:
                if (number instanceof Integer && leftRightNumber instanceof Integer) {
                    int intVal = number.intValue();
                    int leftRightVal = leftRightNumber.intValue();
                    long result = ((long) intVal) * ((long) leftRightVal);
                    x.setLeft(left.getLeft());
                    x.setRight(SQLIntegerExpr.ofIntOrLong(result));

                    visitBinaryFunc(x);
                    visitBinaryOpBinaryOpExpr(x);
                }
                break;
            case Multiply:
                if (BigDecimal.ONE.equals(leftRightNumber) || new BigDecimal("1.0").equals(leftRightNumber)) {
                    if (number instanceof Long) {
                        x.setLeft(left.getLeft());
                    }
                }
                break;
            default:
                break;
        }
    }

    public boolean visit(SQLBinaryOpExpr x) {
        if (isGroup(x)) {
            return super.visit(x);
        }

        x.getLeft().accept(this);
        x.getRight().accept(this);

        if (x.getLeft() instanceof SQLNullExpr || x.getRight() instanceof SQLNullExpr) {
            switch (x.getOperator()) {
                case Add:
                case Subtract:
                case Multiply:
                case DIV:
                case Divide:
                    if (replaceInParent(x, new SQLNullExpr())) {
                        return false;
                    }
                    break;
                default:
                    break;
            }
        }

        visitBinaryFunc(x);
        visitBinaryOpBinaryOpExpr(x);

        {
            SQLExpr left = x.getLeft();
            SQLExpr right = x.getRight();

            switch (x.getOperator()) {
                case BooleanAnd:
                case BooleanOr:
                case BooleanXor:
                    if (left instanceof SQLDateExpr || left instanceof SQLTimestampExpr) {
                        return false;
                    }
                    if (right instanceof SQLDateExpr || right instanceof SQLTimestampExpr) {
                        return false;
                    }

                    if (left instanceof SQLNumericLiteralExpr && right instanceof SQLBooleanExpr) {
                        boolean boolVal = ((SQLBooleanExpr) right).getBooleanValue();
                        x.setRight(new SQLIntegerExpr(boolVal ? 1 : 0));
                    } else if (left instanceof SQLBooleanExpr && right instanceof SQLNumericLiteralExpr) {
                        boolean boolVal = ((SQLBooleanExpr) left).getBooleanValue();
                        x.setLeft(new SQLIntegerExpr(boolVal ? 1 : 0));
                    } else  if (right instanceof SQLBooleanExpr && (left instanceof SQLMethodInvokeExpr || left instanceof SQLCharExpr || left instanceof SQLName)) {
                        return false;
                    } else  if (left instanceof SQLBooleanExpr && (right instanceof SQLMethodInvokeExpr || right instanceof SQLCharExpr || right instanceof SQLName)) {
                        return false;
                    } else  if (left instanceof SQLBooleanExpr
                            && right instanceof SQLBinaryOpExpr
                            && !(((SQLBinaryOpExpr) right).getOperator().isRelational() || ((SQLBinaryOpExpr) right).getOperator().isLogical())) {
                        return false;
                    } else  if (right instanceof SQLBooleanExpr
                            && left instanceof SQLBinaryOpExpr
                            && !(((SQLBinaryOpExpr) left).getOperator().isRelational() || ((SQLBinaryOpExpr) left).getOperator().isLogical())) {
                        return false;
                    }
                    break;
                case Like:
                case NotLike:
                case RLike:
                case NotRLike:
                    if (left instanceof SQLIntegerExpr) {
                        x.setLeft(new SQLCharExpr(((SQLIntegerExpr) left).getNumber().toString()));
                    }

                    if (right instanceof SQLIntegerExpr) {
                        x.setRight(new SQLCharExpr(((SQLIntegerExpr) right).getNumber().toString()));
                    }
                    break;
                case Multiply: {
                    SQLDataType leftDataType = left.computeDataType();
                    if (right instanceof SQLIntervalExpr) {
                        SQLExpr value = ((SQLIntervalExpr) right).getValue();

                        boolean one = false;
                        if (value instanceof SQLIntegerExpr
                                && ((SQLIntegerExpr) value).getNumber() instanceof Integer
                                && ((SQLIntegerExpr) value).getNumber().intValue() == 1) {
                            one = true;
                        } else if (value instanceof SQLCharExpr && ((SQLCharExpr) value).getText().equals("1")) {
                            one = true;
                        }

                        if (one && leftDataType != null && leftDataType.isInt()) {
                            SQLIntervalExpr newInterval = new SQLIntervalExpr(left.clone(), ((SQLIntervalExpr) right).getUnit());
                            if (replaceInParent(x, newInterval)) {
                                return false;
                            }
                        }
                    }

                    if (right instanceof SQLIntegerExpr
                            && ((SQLIntegerExpr) right).getNumber() instanceof Integer
                            && ((SQLIntegerExpr) right).getNumber().intValue() == 1) {
                        replaceInParent(x, left);
                        return false;
                    }
                }
                    break;
                case COLLATE:
                    if (left instanceof SQLCharExpr) {
                        replaceInParent(x, left);
                        return false;
                    }
                    break;
                default:
                    break;
            }
        }

        {
            SQLExpr left = x.getLeft();
            SQLExpr right = x.getRight();
            switch (x.getOperator()) {
                case Equality:
                    if (left instanceof SQLNullExpr && right instanceof SQLLiteralExpr) {
                        replaceInParent(x, new SQLNullExpr());
                    } else  if (left instanceof SQLLiteralExpr && right instanceof SQLNullExpr) {
                        replaceInParent(x, new SQLNullExpr());
                    } else if (left instanceof SQLLiteralExpr
                            && right instanceof SQLLiteralExpr
                            && left.equals(right)) {
                        replaceInParent(x, new SQLBooleanExpr(true));
                    } else if (left instanceof SQLLiteralExpr
                            && right instanceof SQLLiteralExpr
                            && left.getClass() == right.getClass()
                            && !left.equals(right)) {
                        replaceInParent(x, new SQLBooleanExpr(false));
                    } else if (left instanceof SQLIdentifierExpr
                            && right instanceof SQLIdentifierExpr
                            && left.equals(right)) {
                        SQLColumnDefinition column = ((SQLIdentifierExpr) left).getResolvedColumn();
                        if (column != null && column.containsNotNullConstaint()) {
                            replaceInParent(x, new SQLBooleanExpr(true));
                        }
                    } else if (left instanceof SQLPropertyExpr
                            && right instanceof SQLPropertyExpr
                            && left.equals(right)) {
                        SQLColumnDefinition column = ((SQLPropertyExpr) left).getResolvedColumn();
                        if (column != null && column.containsNotNullConstaint()) {
                            replaceInParent(x, new SQLBooleanExpr(true));
                        } else {
                            replaceInParent(x, new SQLBinaryOpExpr(left.clone(), SQLBinaryOperator.IsNot, new SQLNullExpr()));
                        }
                    } else if (left instanceof SQLCastExpr) {
                        SQLCastExpr cast = (SQLCastExpr) left;
                        SQLExpr castExpr = cast.getExpr();

                        boolean predicate = x.getParent() instanceof SQLBinaryOpExprGroup
                                || (x.getParent() instanceof SQLBinaryOpExpr
                                    && ((SQLBinaryOpExpr) x.getParent()).getOperator().isLogical())

                                ;

                        if (predicate
                                && castExpr instanceof SQLIdentifierExpr
                                && castExpr.computeDataType() != null
                                && castExpr.computeDataType().isString()
                                && right instanceof SQLIntegerExpr
                        ) {
                            replaceInParent(x,
                                new SQLBinaryOpExpr(castExpr.clone()
                                        , Equality
                                        , new SQLCharExpr(((SQLIntegerExpr) right).getNumber().toString()))
                            );
                        }
                    }
                    break;
                case LessThan:
                case GreaterThan:
                    if (left instanceof SQLName
                            && right instanceof SQLName && left.equals(right)) {
                        // skip
                        SQLColumnDefinition column = ((SQLName) left).getResolvedColumn();
                        if (column != null) {
                            if (column.isPrimaryKey()) {
                                replaceInParent(x, new SQLBooleanExpr(false));
                            }
                        }
                    }
                    break;
                case NotEqual:
                case LessThanOrGreater:
                    if (left instanceof SQLLiteralExpr
                            && right instanceof SQLLiteralExpr
                            && left.getClass() == right.getClass()) {
                        if (left.equals(right)) {
                            replaceInParent(x, new SQLBooleanExpr(false));
                        } else {
                            replaceInParent(x, new SQLBooleanExpr(true));
                        }
                    }
                    break;
                case BooleanOr:
                    if (left instanceof SQLBooleanExpr) {
                        if (((SQLBooleanExpr) left).getBooleanValue()) {
                            replaceInParent(x, left);
                        } else {
                            replaceInParent(x, right);
                        }
                    } else if (right instanceof SQLBooleanExpr) {
                        if (((SQLBooleanExpr) right).getBooleanValue()) {
                            replaceInParent(x, right);
                        } else {
                            replaceInParent(x, left);
                        }
                    }
                    break;
                case BooleanAnd:
                    if (left instanceof SQLBooleanExpr) {
                        if (((SQLBooleanExpr) left).getBooleanValue()) {
                            replaceInParent(x, right);
                        } else {
                            replaceInParent(x, left);
                        }
                    } else if (right instanceof SQLBooleanExpr) {
                        if (((SQLBooleanExpr) right).getBooleanValue()) {
                            replaceInParent(x, left);
                        } else {
                            replaceInParent(x, right);
                        }
                    }
                    break;
                case Like:
                    SQLColumnDefinition column = null;
                    if (left instanceof SQLIdentifierExpr) {
                        column = ((SQLIdentifierExpr) left).getResolvedColumn();
                    } else if (left instanceof SQLPropertyExpr) {
                        column = ((SQLPropertyExpr) left).getResolvedColumn();
                    }

                    if (column != null
                            && column.containsNotNullConstaint()
                            && right instanceof SQLCharExpr) {
                        boolean matchAll = false;
                        String pattern = ((SQLCharExpr) right).getText();
                        if (pattern.length() > 0) {
                            int count = 0;
                            for (int i = 0; i < pattern.length(); ++i) {
                                char ch = pattern.charAt(i);
                                if (ch == '%') {
                                    count++;
                                }
                            }
                            if (count == pattern.length()) {
                                matchAll = true;
                            }
                        }

                        if (matchAll) {
                            replaceInParent(x, new SQLBooleanExpr(true));
                        }
                    }
                    break;
                default:
                    break;
            }
        }

        final SQLBinaryOperator operator = x.getOperator();
        if ((operator == BooleanAnd || operator == BooleanOr)
                && x.getLeft() instanceof SQLBinaryOpExpr
                && ((SQLBinaryOpExpr) x.getLeft()).getOperator() == x.getOperator()) {
            SQLBinaryOpExprGroup group = new SQLBinaryOpExprGroup(operator, x.getDbType());
            group.add(x.getLeft());
            group.add(x.getRight());
            if (replaceInParent(x, group)) {
                optimizedCount++;
                group.accept(this);
                return false;
            }
        }

        if (x.getOperator() == BooleanAnd) {
            SQLExpr left = x.getLeft();
            if (left instanceof SQLBooleanExpr) {
                boolean booleanValue = ((SQLBooleanExpr) left).getBooleanValue();
                SQLExpr replaceTo;
                if (booleanValue) {
                    replaceTo = x.getRight();
                } else {
                    replaceTo = x.getLeft();
                }
                if (replaceInParent(x, replaceTo)) {
                    optimizedCount++;
                    return false;
                }
            }

            SQLExpr right = x.getRight();
            if (right instanceof SQLBooleanExpr) {
                boolean booleanValue = ((SQLBooleanExpr) right).getBooleanValue();
                SQLExpr replaceTo;
                if (booleanValue) {
                    replaceTo = x.getLeft();
                } else {
                    replaceTo = x.getRight();
                }
                if (replaceInParent(x, replaceTo)) {
                    optimizedCount++;
                    return false;
                }
            }
        }

        if (x.getOperator() == BooleanOr) {
            SQLExpr left = x.getLeft();
            if (left instanceof SQLBooleanExpr) {
                if (((SQLBooleanExpr) left).getBooleanValue()) {
                    if (replaceInParent(x, x.getRight())) {
                        optimizedCount++;
                        return false;
                    }
                } else {
                    if (replaceInParent(x, new SQLBooleanExpr(true))) {
                        optimizedCount++;
                        return false;
                    }
                }
            }

            SQLExpr right = x.getRight();
            if (right instanceof SQLBooleanExpr) {
                if (((SQLBooleanExpr) right).getBooleanValue()) {
                    if (replaceInParent(x, x.getLeft())) {
                        optimizedCount++;
                        return false;
                    }
                } else {
                    if (replaceInParent(x, new SQLBooleanExpr(true))) {
                        optimizedCount++;
                        return false;
                    }
                }
            }
        }

        if (operator == SQLBinaryOperator.Like && x.getRight() instanceof SQLCharExpr) {
            // fname like 'aBc' -> fname = 'aBc'
//            SQLCharExpr right = (SQLCharExpr) x.getRight();
//            String pattern = right.getText();
//            if (pattern.indexOf('%') == -1 && pattern.indexOf('_') == -1) {
//                x.setOperator(SQLBinaryOperator.Equality);
//            }
        } else if (operator == SQLBinaryOperator.NotLike && x.getRight() instanceof SQLCharExpr) {
            // fname note like 'aBc' -> fname <> 'aBc'
//            SQLCharExpr right = (SQLCharExpr) x.getRight();
//            String pattern = right.getText();
//            if (pattern.indexOf('%') == -1 && pattern.indexOf('_') == -1) {
//                x.setOperator(SQLBinaryOperator.LessThanOrGreater);
//            }
        }

        // col + 3 + 4 => col + 7
        if (isNameAndLiteral(x.getLeft()) && x.getRight() instanceof SQLLiteralExpr) {
            SQLBinaryOpExpr left = (SQLBinaryOpExpr) x.getLeft();
            SQLLiteralExpr right = (SQLLiteralExpr) x.getRight();

            if (left.getRight() instanceof SQLLiteralExpr) {

                SQLLiteralExpr leftRight = (SQLLiteralExpr) left.getRight();
                if (left.getOperator() == x.getOperator()) {
                    switch (left.getOperator()) {
                        case Add:
                        case Multiply:
                        case Concat:
                            x.setLeft(left.getLeft());
                            x.setRight(new SQLBinaryOpExpr(leftRight, left.getOperator(), right));
                            x.getRight().accept(this);
                            break;
                        case Subtract:
                            x.setLeft(left.getLeft());
                            x.setRight(new SQLBinaryOpExpr(leftRight, SQLBinaryOperator.Add, right));
                            x.getRight().accept(this);
                            break;
                        default:
                            break;
                    }
                }
            } else {
                SQLLiteralExpr leftLeft = (SQLLiteralExpr) left.getLeft();
                if (left.getOperator() == x.getOperator()) {
                    switch (left.getOperator()) {
                        case Add:
                        case Multiply:
                        case Concat:
                            x.setLeft(left.getRight());
                            x.setRight(new SQLBinaryOpExpr(leftLeft, left.getOperator(), right));
                            x.getRight().accept(this);
                            break;
                        default:
                            break;
                    }
                }
            }
        }

        // swap
        if (x.getLeft() instanceof SQLIntervalExpr
                && (x.getRight() instanceof SQLCharExpr || x.getRight() instanceof SQLDateExpr || x.getRight() instanceof SQLTimestampExpr))
        {
            SQLExpr left = x.getLeft();
            SQLExpr right = x.getRight();
            x.setRight(left);
            x.setLeft(right);
        }

        boolean charAddInterval = false;
        if (x.getLeft() instanceof SQLCharExpr && x.getRight() instanceof SQLIntervalExpr) {
            String text = ((SQLCharExpr) x.getLeft()).getText();
            Date date = MySqlUtils.parseDate(text, timeZone);
            if (date == null) {
                return false;
            }

            if (text.length() <= 10) {
                x.setLeft(new SQLDateExpr(date, timeZone));
            } else {
                x.setLeft(new SQLTimestampExpr(date, timeZone));
            }
            charAddInterval = true;
        }

        if (x.getLeft() instanceof SQLValuableExpr
                && x.getRight() instanceof SQLValuableExpr) {
            handValueValue(x);
            return false;
        }

        if ((x.getLeft() instanceof SQLDateExpr || x.getLeft() instanceof SQLTimestampExpr)
                && x.getRight() instanceof SQLIntervalExpr
                && (operator == SQLBinaryOperator.Add || operator == SQLBinaryOperator.Subtract)) {
            boolean datetime = false;

            SQLExpr left = x.getLeft();
            String text;
            if (left instanceof SQLDateExpr) {
                text = ((SQLDateExpr) left).getValue();
            } else {
                text = ((SQLTimestampExpr) left).getValue();
                datetime = true;
            }

            SQLIntervalExpr right = (SQLIntervalExpr) x.getRight();

            String intervalText;
            int intervalInt;
            if (right.getValue() instanceof SQLCharExpr) {
                intervalText = ((SQLCharExpr) right.getValue()).getText();
                try {
                    intervalInt = Integer.parseInt(intervalText);
                } catch (NumberFormatException ex) {
                    // skip
                    return false;
                }
            } else if (right.getValue() instanceof SQLIntegerExpr) {
                intervalInt = ((SQLIntegerExpr) right.getValue()).getNumber().intValue();
            } else {
                return false;
            }

            if (operator == SQLBinaryOperator.Subtract) {
                intervalInt = -intervalInt;
            }

            Date date = MySqlUtils.parseDate(text, timeZone);
            if (date == null) {
                return false;
            }
            Calendar calendar = Calendar.getInstance();
            calendar.setTime(date);

            boolean changed = SQLIntervalUnit.add(calendar, intervalInt, right.getUnit());
            if (!changed) {
                return false;
            }

            if (right.getUnit().isDateTime()) {
                datetime = true;
            }

            Date dateChanged = calendar.getTime();

            if (dateChanged != null) {
                SQLExpr replaceTo;
                if (charAddInterval) {
                    SimpleDateFormat dateFormat;
                    if (datetime) {
                        dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
                    } else {
                        dateFormat = new SimpleDateFormat("yyyy-MM-dd");
                    }
                    if (timeZone != null) {
                        dateFormat.setTimeZone(timeZone);
                    }

                    replaceTo = new SQLCharExpr(dateFormat.format(dateChanged));
                } else {
                    replaceTo = datetime ? new SQLTimestampExpr(dateChanged, timeZone) : new SQLDateExpr(dateChanged, timeZone);
                }

                if (replaceInParent(x, replaceTo)) {
                    optimizedCount++;
                    return false;
                }
            }
        }

        if (x.isLeftLiteralAndRightName()) {
            SQLName right = (SQLName) x.getRight();
            SQLLiteralExpr left = (SQLLiteralExpr) x.getLeft();

            switch (x.getOperator()) {
                case LessThan:
                    // 3 < col -> col > 3
                    x.setRight(left);
                    x.setLeft(right);
                    x.setOperator(SQLBinaryOperator.GreaterThan);
                    optimizedCount++;
                    break;
                case LessThanOrEqual:
                    // 3 < col -> col > 3
                    x.setRight(left);
                    x.setLeft(right);
                    x.setOperator(SQLBinaryOperator.GreaterThanOrEqual);
                    optimizedCount++;
                    break;
                case Equality:
                    // 3 = id -> id = 3
                    x.setRight(left);
                    x.setLeft(right);
                    break;
                case GreaterThanOrEqual:
                    // 3 >= id -> id <= 3
                    x.setRight(left);
                    x.setLeft(right);
                    x.setOperator(SQLBinaryOperator.LessThanOrEqual);
                    optimizedCount++;
                    break;
                case GreaterThan:
                    // 3 > id -> id < 3
                    x.setRight(left);
                    x.setLeft(right);
                    x.setOperator(SQLBinaryOperator.LessThan);
                    optimizedCount++;
                    break;
                default:
                    break;
            }
        }

        if (x.isLeftNameAndRightLiteral()) {
            SQLName name = (SQLName) x.getLeft();
            SQLLiteralExpr literal = (SQLLiteralExpr) x.getRight();

            // if (op == SQLBinaryOperator.Like)

            switch (operator) {
                case Like:
                case NotLike:
                    break;
                default:
                    handleNameLiteral(x, name, literal);
                    break;
            }
        }


        if (x.isLeftFunctionAndRightLiteral()) {
            SQLMethodInvokeExpr functionCall = (SQLMethodInvokeExpr) x.getLeft();
            SQLLiteralExpr literal = (SQLLiteralExpr) x.getRight();

            long nameHash = functionCall.methodNameHashCode64();
            SQLDataType dataType = functionCall.getResolvedReturnDataType();

            if (dataType != null) {
                final long dataTypeNameHash = dataType.nameHashCode64();
                if (dataTypeNameHash == FnvHash.Constants.BIGINT) {
                    if (literal instanceof SQLCharExpr) {
                        String text = ((SQLCharExpr) literal).getText();
                        try {
                            // id = '123' -> id = 123
                            long val = Long.parseLong(text);
                            x.setRight(new SQLIntegerExpr(val));
                            optimizedCount++;
                        } catch (NumberFormatException ex) {
                            // id = '123' -> false
                            if (replaceInParent(x, new SQLBooleanExpr(false))) {
                                optimizedCount++;
                                return false;
                            }
                        }
                    }
                }
            }
        }

        if (operator == BooleanAnd) {
            if (isLeftNameAndRightLiteral(x.getLeft()) && isLeftNameAndRightLiteral(x.getRight())) {
                SQLBinaryOpExpr left = (SQLBinaryOpExpr) x.getLeft();
                SQLBinaryOpExpr right = (SQLBinaryOpExpr) x.getRight();

                mergeTwiceOp(x, left, right);
            } else if (isBothName(x.getLeft()) && isLeftNameAndRightLiteral(x.getRight())) {
                SQLBinaryOpExpr left = (SQLBinaryOpExpr) x.getLeft();
                SQLBinaryOpExpr right = (SQLBinaryOpExpr) x.getRight();

                if (left.getOperator() == right.getOperator()) {
                    if (left.isBothName()) {
                        if (left.getLeft().equals(right.getLeft())) {
                            SQLBinaryOpExpr left2 = left.clone();
                            left2.setLeft(right.getRight().clone());
                            SQLBinaryOpExpr replaceTo = new SQLBinaryOpExpr(left, BooleanAnd, left2);
                            x.setLeft(replaceTo);
                            optimizedCount++;
                        } else if (left.getRight().equals(right.getLeft())) {
                            SQLBinaryOpExpr left2 = left.clone();
                            left2.setRight(right.getRight().clone());
                            SQLBinaryOpExpr replaceTo = new SQLBinaryOpExpr(left, BooleanAnd, left2);
                            x.setLeft(replaceTo);
                            optimizedCount++;
                        }
                    } else {
                        if (left.getLeft().equals(right.getLeft())) {
                            left.setLeft(right.getRight().clone());
                            optimizedCount++;
                        } else if (left.getRight().equals(right.getLeft())) {
                            left.setRight(right.getRight().clone());
                            optimizedCount++;
                        }
                    }
                }
            }
        } else if (operator == BooleanOr) {
            if (isLeftNameAndRightLiteral(x.getLeft()) && isLeftNameAndRightLiteral(x.getRight())) {
                SQLBinaryOpExpr left = (SQLBinaryOpExpr) x.getLeft();
                SQLBinaryOpExpr right = (SQLBinaryOpExpr) x.getRight();

                mergeTwiceOp(x, left, right);
            }
        }

        // a - 1 = 2 -> a = 3
        if (isLeftNameAndRightLiteral(x.getLeft()) && x.getRight() instanceof SQLLiteralExpr) {
            if (operator == Equality
                    || operator == SQLBinaryOperator.NotEqual
                    || operator == SQLBinaryOperator.LessThanOrGreater
                    || operator == SQLBinaryOperator.LessThanOrEqual
                    || operator == SQLBinaryOperator.LessThan
                    || operator == SQLBinaryOperator.GreaterThanOrEqual
                    || operator == SQLBinaryOperator.GreaterThan)
            {
                SQLBinaryOpExpr left = (SQLBinaryOpExpr) x.getLeft();
                final SQLExpr leftRight = left.getRight();
                final SQLExpr leftLeft = left.getLeft();
                final SQLExpr right = x.getRight();

                SQLExpr valuExpr = null;

                switch (left.getOperator()) {
                    case Add: {
                        if (leftRight instanceof SQLIntegerExpr
                                && right instanceof SQLIntegerExpr) {
                            int val = ((SQLIntegerExpr) right).getNumber().intValue()
                                    - ((SQLIntegerExpr) leftRight).getNumber().intValue();
                            x.setRight(new SQLIntegerExpr(val));
                            x.setLeft(leftLeft);
                            optimizedCount++;
                            return false;
                        } else if (leftRight instanceof SQLIntegerExpr
                                && right instanceof SQLNumberExpr) {
                            Number leftRightVal = ((SQLIntegerExpr) leftRight).getNumber();
                            Number rightVal = ((SQLNumberExpr) right).getNumber();


                            if (rightVal instanceof BigDecimal && leftRightVal instanceof Integer) {
                                final BigDecimal rightDecimal = (BigDecimal) rightVal;
                                final BigDecimal leftRightDecimal = new BigDecimal(leftRightVal.intValue());
                                valuExpr = new SQLNumberExpr(
                                        rightDecimal.subtract(leftRightDecimal));
                            }
                        }
                        break;
                    }
                    case Subtract: {
                        if (leftRight instanceof SQLIntegerExpr
                                && right instanceof SQLIntegerExpr) {
                            int val = ((SQLIntegerExpr) right).getNumber().intValue()
                                    + ((SQLIntegerExpr) leftRight).getNumber().intValue();
                            x.setRight(new SQLIntegerExpr(val));
                            x.setLeft(leftLeft);
                            optimizedCount++;
                            return false;
                        } else if (leftRight instanceof SQLIntegerExpr
                                && right instanceof SQLNumberExpr) {
                            Number leftRightVal = ((SQLIntegerExpr) leftRight).getNumber();
                            Number rightVal = ((SQLNumberExpr) right).getNumber();


                            if (rightVal instanceof BigDecimal && leftRightVal instanceof Integer) {
                                final BigDecimal rightDecimal = (BigDecimal) rightVal;
                                final BigDecimal leftRightDecimal = new BigDecimal(leftRightVal.intValue());
                                valuExpr = new SQLNumberExpr(
                                        rightDecimal.add(leftRightDecimal));
                            }
                        }
                        break;
                    }
                    default:
                        break;
                }

                if (valuExpr != null) {
                    x.setRight(valuExpr);
                    x.setLeft(leftLeft);
                    optimizedCount++;
                    return false;
                }
            }
        }

        if (operator == Equality
                && x.isLeftNameAndRightLiteral()
                && x.getParent() instanceof SQLSelectQueryBlock)
        {
            SQLSelectQueryBlock queryBlock = (SQLSelectQueryBlock) x.getParent();
            if (queryBlock.getLimit() == null
                    && queryBlock.getFrom() instanceof SQLJoinTableSource
                    && isBothName(((SQLJoinTableSource) queryBlock.getFrom()).getCondition())) {
                SQLJoinTableSource joinTableSource = (SQLJoinTableSource) queryBlock.getFrom();
                SQLBinaryOpExpr joinCondition = (SQLBinaryOpExpr) joinTableSource.getCondition();

                if (joinTableSource.getJoinType() == SQLJoinTableSource.JoinType.INNER_JOIN
                        && joinCondition.getOperator() == Equality
                        && joinCondition.getLeft().equals(x.getLeft())) {
                    SQLExpr to = SQLBinaryOpExpr.and(joinCondition.clone(),
                            x.clone(),
                            new SQLBinaryOpExpr(joinCondition.getRight().clone()
                                    , Equality
                                    , x.getRight().clone()));
                    if (replaceInParent(x, null)) {
                        joinTableSource.setCondition(to);
                        optimizedCount++;
                        return false;
                    }
                }
            }
        }

        return false;
    }


    public void binaryOp_date(SQLBinaryOpExpr x) {
        SQLMethodInvokeExpr func = (SQLMethodInvokeExpr) x.getLeft();
        if (func.getArguments().size() != 1) {
            return;
        }

        SQLExpr arg0 = func.getArguments().get(0);
        SQLExpr right = x.getRight();

        final SQLDataType dataType = arg0.computeDataType();
        if (dataType == null) {
            return;
        }

        if (dataType.nameHashCode64() != FnvHash.Constants.TIMESTAMP
            && dataType.nameHashCode64() != FnvHash.Constants.DATETIME) {
            return;
        }

        Date date = null;
        if (right instanceof SQLCharExpr) {
            String text = ((SQLCharExpr) right).getText();

            SimpleDateFormat dateFormat = null;
            if (text.length() == 10) {
                dateFormat = new SimpleDateFormat("yyyy-MM-dd");
            }

            if (dateFormat == null) {
                return;
            }

            if (timeZone != null) {
                dateFormat.setTimeZone(timeZone);
            }

            try {
                date = dateFormat.parse(text);
            } catch (ParseException e) {
                // skip
            }
        } else if (right instanceof SQLDateExpr) {
            date = ((SQLDateExpr) right).getDate(timeZone);
        }

        if (date == null) {
            return;
        }

        SQLExpr replaceTo = null;
        switch (x.getOperator()) {
            case Equality: {
                Date nextDate = nextDate(date);
                SQLBinaryOpExpr ge = new SQLBinaryOpExpr(arg0.clone(), GreaterThanOrEqual, new SQLTimestampExpr(date, timeZone));
                SQLBinaryOpExpr lt = new SQLBinaryOpExpr(arg0.clone(), LessThan, new SQLTimestampExpr(nextDate, timeZone));
                replaceTo = new SQLBinaryOpExpr(ge, BooleanAnd, lt);
                break;
            }
            case LessThanOrGreater:
            case NotEqual: {
                Date nextDate = nextDate(date);
                SQLBinaryOpExpr lt = new SQLBinaryOpExpr(arg0.clone(), LessThan, new SQLTimestampExpr(date, timeZone));
                SQLBinaryOpExpr ge = new SQLBinaryOpExpr(arg0.clone(), GreaterThanOrEqual, new SQLTimestampExpr(nextDate, timeZone));
                replaceTo = new SQLBinaryOpExpr(lt, BooleanOr, ge);
                break;
            }
            case LessThanOrEqual:
                Date nextDate = nextDate(date);
                replaceTo = new SQLBinaryOpExpr(arg0.clone(), LessThan, new SQLTimestampExpr(nextDate, timeZone));
                break;

            case LessThan:
            case GreaterThan:
            case GreaterThanOrEqual:
                replaceTo = new SQLBinaryOpExpr(arg0.clone(), x.getOperator(), new SQLTimestampExpr(date, timeZone));
                break;
            default:
                break;
        }

        if (replaceTo != null) {
            replaceInParent(x, replaceTo);
        }
    }

    private Date nextDate(Date date) {
        Calendar calendar = timeZone == null
                ? Calendar.getInstance()
                : Calendar.getInstance(timeZone);
        calendar.setTime(date);
        calendar.add(Calendar.DAY_OF_MONTH, 1);
        return calendar.getTime();
    }


    public void binaryOp_coalesce(SQLBinaryOpExpr x) {
        SQLMethodInvokeExpr func = (SQLMethodInvokeExpr) x.getLeft();
        if (func.getArguments().size() != 2) {
            return;
        }

        SQLExpr arg0 = func.getArguments().get(0);
        SQLExpr arg1 = func.getArguments().get(1);
        final SQLExpr right = x.getRight();
        if (!(arg0 instanceof SQLName)) {
            return;
        }
        if (!(arg1 instanceof SQLLiteralExpr)) {
            return;
        }
        if (!(right instanceof SQLLiteralExpr)) {
            return;
        }

        SQLExpr replaceTo = null;
        switch (x.getOperator()) {
            case NotEqual:
            case LessThanOrGreater:
                if (arg1.equals(right)) {
                    replaceTo = new SQLBinaryOpExpr(SQLBinaryOpExpr.isNotNull(arg0.clone())
                            , BooleanAnd
                            , new SQLBinaryOpExpr(arg0.clone(), x.getOperator(), right.clone())
                    );
                }
                break;
            default:
                break;
        }

        if (replaceTo != null) {
            replaceInParent(x, replaceTo);
        }
    }

    public void binaryOp_yearmonth(SQLBinaryOpExpr x) {
        SQLMethodInvokeExpr func = (SQLMethodInvokeExpr) x.getLeft();
        if (func.getArguments().size() != 1) {
            return;
        }

        SQLExpr arg0 = func.getArguments().get(0);
        SQLExpr right = x.getRight();

        String text;
        if (right instanceof SQLCharExpr) {
            text = ((SQLCharExpr) right).getText();
        } else {
            return;
        }

        Date date = null;
        SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM");
        try {
            date = dateFormat.parse(text);
        } catch (ParseException e) {
            return;
        }

        if (date == null) {
            return;
        }

        SQLDataType dataType;
        if (arg0 instanceof SQLName) {
            dataType = arg0.computeDataType();
        } else {
            return;
        }

        if (dataType == null) {
            return;
        }

        Calendar calendar = timeZone == null
                ? Calendar.getInstance()
                : Calendar.getInstance(timeZone);

        SQLBinaryOpExpr replaceTo = null;
        SQLExpr ts = null, next_month = null;
        if (dataType.nameHashCode64() == FnvHash.Constants.TIMESTAMP) {
            ts = new SQLTimestampExpr(date, timeZone);
            calendar.setTime(date);
            calendar.add(Calendar.MONTH, 1);
            next_month = new SQLTimestampExpr(calendar.getTime(), timeZone);
        } else if (dataType.nameHashCode64() == FnvHash.Constants.DATE) {
            ts = new SQLDateExpr(date, timeZone);
            calendar.setTime(date);
            calendar.add(Calendar.MONTH, 1);
            next_month = new SQLDateExpr(calendar.getTime(), timeZone);
        } else {
            return;
        }

        switch (x.getOperator()) {
            case Equality:
                SQLBinaryOpExpr gteq = new SQLBinaryOpExpr(arg0.clone(), SQLBinaryOperator.GreaterThanOrEqual, ts);
                SQLBinaryOpExpr le = new SQLBinaryOpExpr(arg0.clone(), SQLBinaryOperator.LessThan, next_month);
                replaceTo = new SQLBinaryOpExpr(gteq, BooleanAnd, le);
                break;
            case LessThan:
            case GreaterThanOrEqual:
                x.setLeft(arg0.clone());
                x.setRight(ts);
                optimizedCount++;
                break;
            case GreaterThan:
                replaceTo = new SQLBinaryOpExpr(arg0.clone(), SQLBinaryOperator.GreaterThanOrEqual, next_month);
                break;
            case LessThanOrEqual:
                replaceTo = new SQLBinaryOpExpr(arg0.clone(), SQLBinaryOperator.LessThan, next_month);
                break;
            default:
                break;
        }

        if (replaceTo != null && replaceInParent(x, replaceTo)) {
            optimizedCount++;
        }
    }

    public void binaryOp_date_diff(SQLBinaryOpExpr x) {
        SQLExpr right = x.getRight();

        SQLMethodInvokeExpr func = (SQLMethodInvokeExpr) x.getLeft();
        final List<SQLExpr> arguments = func.getArguments();
        if (arguments.size() < 2) {
            return;
        }

        SQLExpr arg0, arg1, arg2;

        if (arguments.size() == 3) {
            arg0 = arguments.get(0);
            arg1 = arguments.get(1);
            arg2 = arguments.get(2);
        } else if (arguments.size() == 2) {
            arg0 = new SQLCharExpr("DAY");
            arg1 = arguments.get(0);
            arg2 = arguments.get(1);
        } else {
            return;
        }

        Object value = null;
        if (right instanceof SQLValuableExpr) {
            value = ((SQLValuableExpr) right).getValue();
        }

        int intVal;
        if (value instanceof BigDecimal) {
            BigDecimal decimal = (BigDecimal) value;
            if (decimal.scale() == 0) {
                intVal = decimal.intValue();
            } else if (decimal.scale() == 1 && decimal.toString().endsWith(".0")) {
                intVal = decimal.intValue();
            } else {
                return;
            }
        } else if (value instanceof Integer) {
            intVal = ((Integer) value).intValue();
        } else if (value instanceof Long) {
            long longVal = ((Long) value).longValue();
            if (longVal >= Integer.MIN_VALUE && longVal <= Integer.MAX_VALUE) {
                intVal = (int) longVal;
            } else {
                return;
            }
        } else {
            return;
        }

        int field = -1;

        if (arg0 instanceof SQLCharExpr) {
            long fieldHash = FnvHash.hashCode64(((SQLCharExpr) arg0).getText());
            if (FnvHash.Constants.HOUR == fieldHash) {
                field = Calendar.HOUR_OF_DAY;
            } else if (FnvHash.Constants.MINUTE == fieldHash) {
                field = Calendar.MINUTE;
            } else if (FnvHash.Constants.SECOND == fieldHash) {
                field = Calendar.SECOND;
            } else if (FnvHash.Constants.DAY == fieldHash) {
                field = Calendar.DAY_OF_MONTH;
            } else if (FnvHash.Constants.MONTH == fieldHash) {
                field = Calendar.MONTH;
            } else if (FnvHash.Constants.YEAR == fieldHash) {
                field = Calendar.YEAR;
            }
        }

        if (field == -1) {
            return;
        }

        Calendar calendar = timeZone == null
                ? Calendar.getInstance()
                : Calendar.getInstance(timeZone);

        Date date = null;
        SQLDataType dataType = null;
        SQLExpr replaceTo = null;
        if (arg1 instanceof SQLName && (arg2 instanceof SQLTimestampExpr || arg2 instanceof SQLDateExpr)) {
            if (arg2 instanceof SQLTimestampExpr) {
                date = ((SQLTimestampExpr) arg2).getDate(timeZone);
            } else if (arg2 instanceof SQLDateExpr) {
                date = ((SQLDateExpr) arg2).getDate(timeZone);
            }
            dataType = arg1.computeDataType();
        } else if (arg2 instanceof SQLName && (arg1 instanceof SQLTimestampExpr || arg1 instanceof SQLDateExpr || arg1 instanceof SQLCharExpr)) {
            if (arg1 instanceof SQLTimestampExpr) {
                date = ((SQLTimestampExpr) arg1).getDate(timeZone);
            } else if (arg1 instanceof SQLDateExpr) {
                date = ((SQLDateExpr) arg1).getDate(timeZone);
            } else if (arg1 instanceof SQLCharExpr) {
                String text = ((SQLCharExpr) arg1).getText();
                date = MySqlUtils.parseDate(text, timeZone);
            }
            dataType = arg2.computeDataType();
        } else {
            return;
        }


        if (date == null) {
            return;
        }

        calendar.setTime(date);
        if (arg1 instanceof SQLName) {
            calendar.add(field, intVal);
        } else {
            calendar.add(field, -intVal);
        }

        if (dataType == null) {
            return;
        }

        if (arg1 instanceof SQLName) {
            if (dataType.nameHashCode64() == FnvHash.Constants.TIMESTAMP) {
                switch (x.getOperator()) {
                    case LessThan:
                    case LessThanOrEqual:
                    case GreaterThan:
                    case GreaterThanOrEqual: {
                        SQLTimestampExpr ts = new SQLTimestampExpr(calendar.getTime(), timeZone);
                        replaceTo = new SQLBinaryOpExpr(arg1.clone(), x.getOperator(), ts);
                    }
                    break;
                    default:
                        break;
                }
            }
        } else {
            if (dataType.nameHashCode64() == FnvHash.Constants.TIMESTAMP) {
                SQLTimestampExpr ts = new SQLTimestampExpr(calendar.getTime(), timeZone);
                SQLBinaryOperator op = null;
                switch (x.getOperator()) {
                    case LessThan:
                        op = SQLBinaryOperator.GreaterThan;
                        break;
                    case LessThanOrEqual:
                        op = SQLBinaryOperator.GreaterThanOrEqual;
                        break;
                    case GreaterThan:
                        op = SQLBinaryOperator.LessThan;
                        break;
                    case GreaterThanOrEqual:
                        op = SQLBinaryOperator.LessThanOrEqual;
                        break;
                    default:
                        break;
                }
                if (op != null) {
                    replaceTo = new SQLBinaryOpExpr(arg2.clone(), op, ts);
                }
            }
        }

        if (replaceTo != null && replaceInParent(x, replaceTo)) {
            optimizedCount++;
        }
    }

    public void binaryOp_date_format(SQLBinaryOpExpr x) {
        SQLMethodInvokeExpr func = (SQLMethodInvokeExpr) x.getLeft();
        if (func.getArguments().size() != 2) {
            return;
        }

        SQLExpr arg0 = func.getArguments().get(0);
        SQLExpr arg1 = func.getArguments().get(1);
        SQLLiteralExpr val = (SQLLiteralExpr) x.getRight();

        boolean timestampCastToDate = false;
        if (arg0 instanceof SQLCastExpr) {
            SQLCastExpr cast = (SQLCastExpr) arg0;
            final SQLExpr expr = cast.getExpr();
            if (expr == null) {
                return;
            }

            SQLDataType exprDataType = expr.computeDataType();
            if (exprDataType == null) {
                return;
            }

            if (expr instanceof SQLName
                    && exprDataType.nameHashCode64() == FnvHash.Constants.TIMESTAMP
                    && cast.dateTypeHashCode() == FnvHash.Constants.DATE) {
                arg0 = expr;
                timestampCastToDate = true;
            }
        }

        SQLDataType dataType = null;
        if (arg0 instanceof SQLName) {
            dataType = arg0.computeDataType();
        } else {
            return;
        }

        if (dataType == null) {
            return;
        }

        String fmt = null;
        DateFormat javaFormat = null;
        if (arg1 instanceof SQLCharExpr) {
            fmt = ((SQLCharExpr) arg1).getText();

            boolean containsYMD = fmt.contains("%Y") && fmt.contains("%m") && fmt.contains("%d");
            if (!containsYMD) {
                return;
            }

            javaFormat = MySqlUtils.toJavaFormat(fmt, timeZone);
        }

        if (javaFormat == null) {
            return;
        }

        if (timestampCastToDate) {
            if ((!"%Y%m%d".equals(fmt)) && !"%Y-%m-%d".equals(fmt)) {
                return;
            }
        }

        Date date = null;
        if (val instanceof SQLCharExpr) {
            try {
                date = javaFormat.parse(((SQLCharExpr) val).getText());
            } catch (ParseException e) {
                // skip
            }
        }
        if (date == null) {
            return;
        }

        final SQLBinaryOperator op = x.getOperator();
        if (dataType.nameHashCode64() == FnvHash.Constants.TIMESTAMP) {
            SQLTimestampExpr ts = new SQLTimestampExpr(date, timeZone);
            switch (op) {
                case Equality: {
                    if ("%Y%m%d".equals(fmt) || "%Y-%m-%d".equals(fmt)) {
                        SQLBinaryOpExpr gteq = new SQLBinaryOpExpr(arg0.clone(), SQLBinaryOperator.GreaterThanOrEqual, ts.clone());

                        ts.addDay(1);
                        SQLBinaryOpExpr le = new SQLBinaryOpExpr(arg0.clone(), SQLBinaryOperator.LessThan, ts);

                        SQLBinaryOpExpr replaceTo = new SQLBinaryOpExpr(gteq, BooleanAnd, le);
                        if (replaceInParent(x, replaceTo)) {
                            optimizedCount++;
                        }
                    }
                }
                break;
                case GreaterThanOrEqual:
                case LessThan: {
                    x.setLeft(arg0.clone());
                    x.setRight(ts);
                    optimizedCount++;
                }
                break;
                case GreaterThan:
                    if ("%Y%m%d".equals(fmt) || "%Y-%m-%d".equals(fmt)) {
                        ts.addDay(1);
                        x.setLeft(arg0.clone());
                        x.setRight(ts);
                        x.setOperator(SQLBinaryOperator.GreaterThanOrEqual);
                        optimizedCount++;
                    }
                    break;
                case LessThanOrEqual:
                    if ("%Y%m%d".equals(fmt) || "%Y-%m-%d".equals(fmt)) {
                        ts.addDay(1);
                        x.setLeft(arg0.clone());
                        x.setRight(ts);
                        x.setOperator(SQLBinaryOperator.LessThan);
                        optimizedCount++;
                    }
                    break;
                default:
                    break;
            }
        } else if (dataType.nameHashCode64() == FnvHash.Constants.DATE) {
            SQLDateExpr ts = new SQLDateExpr(date, timeZone);
            switch (op) {
                case Equality: {
                    if ("%Y%m%d".equals(fmt) || "%Y-%m-%d".equals(fmt)) {
                        SQLBinaryOpExpr gteq = new SQLBinaryOpExpr(arg0.clone(), SQLBinaryOperator.GreaterThanOrEqual, ts.clone());

                        ts.addDay(1);
                        SQLBinaryOpExpr le = new SQLBinaryOpExpr(arg0.clone(), SQLBinaryOperator.LessThan, ts);

                        SQLBinaryOpExpr replaceTo = new SQLBinaryOpExpr(gteq, BooleanAnd, le);
                        if (replaceInParent(x, replaceTo)) {
                            optimizedCount++;
                        }
                    }
                }
                break;
                case GreaterThanOrEqual:
                case LessThan: {
                    x.setLeft(arg0.clone());
                    x.setRight(ts);
                    optimizedCount++;
                }
                break;
                case GreaterThan:
                    if ("%Y%m%d".equals(fmt) || "%Y-%m-%d".equals(fmt)) {
                        ts.addDay(1);
                        x.setLeft(arg0.clone());
                        x.setRight(ts);
                        x.setOperator(SQLBinaryOperator.GreaterThanOrEqual);
                        optimizedCount++;
                    }
                    break;
                case LessThanOrEqual:
                    if ("%Y%m%d".equals(fmt) || "%Y-%m-%d".equals(fmt)) {
                        ts.addDay(1);
                        x.setLeft(arg0.clone());
                        x.setRight(ts);
                        x.setOperator(SQLBinaryOperator.LessThan);
                        optimizedCount++;
                    }
                    break;
                default:
                    break;
            }
        }
    }

    public void binaryOp_add_date(SQLBinaryOpExpr x) {
        SQLMethodInvokeExpr func = (SQLMethodInvokeExpr) x.getLeft();

        final List<SQLExpr> arguments = func.getArguments();
        if (arguments.size() == 2) {

            SQLExpr arg0 = func.getArguments().get(0);
            SQLExpr arg1 = func.getArguments().get(1);
            SQLLiteralExpr val = (SQLLiteralExpr) x.getRight();

            if (!(arg0 instanceof SQLName)) {
                return;
            }

            Date date = null;
            if (val instanceof SQLDateExpr) {
                date = ((SQLDateExpr) val).getDate(timeZone);
            } else if (val instanceof SQLCharExpr) {
                String chars = ((SQLCharExpr) val).getText();
                date = MySqlUtils.parseDate(chars, timeZone);
            }

            if (date == null) {
                return;
            }

            Calendar calendar = timeZone == null
                    ? Calendar.getInstance()
                    : Calendar.getInstance(timeZone);
            calendar.setTime(date);

            int delta = 0;
            Integer field = null;
            if (arg1 instanceof SQLIntegerExpr
                    && ((SQLIntegerExpr) arg1).getNumber() instanceof Integer)
            {
                field = Calendar.DAY_OF_YEAR;
                delta = ((SQLIntegerExpr) arg1).getNumber().intValue();
            }

            if (delta == 0 || field == null) {
                return;
            }
            calendar.add(field, -delta);

            SQLExpr replaceToValExpr;
            if (val instanceof SQLDateExpr) {
                replaceToValExpr = new SQLDateExpr(calendar.getTime(), timeZone);
            } else {
                replaceToValExpr = new SQLTimestampExpr(calendar.getTime(), timeZone);
            }

            x.setLeft(arg0);
            x.setRight(replaceToValExpr);

            return;
        }
    }

    public void binaryOp_from_unixtime(SQLBinaryOpExpr x) {
        SQLMethodInvokeExpr func = (SQLMethodInvokeExpr) x.getLeft();
        final List<SQLExpr> arguments = func.getArguments();
        final SQLExpr right = x.getRight();

        Date date = null;
        if (right instanceof SQLDateExpr) {
            date = ((SQLDateExpr) right).getDate(timeZone);
        } else if (right instanceof SQLCharExpr) {
            String chars = ((SQLCharExpr) right).getText();
            if (arguments.size() == 1) {
                date = MySqlUtils.parseDate(chars, timeZone);
            } else if (arguments.size() == 2 && arguments.get(1) instanceof SQLCharExpr) {
                String format = ((SQLCharExpr) arguments.get(1)).getText();
                if ("YYYY-MM-DD".equals(format) || "yyyy-MM-dd".equals(format)) {
                    format = "%Y-%m-%d";
                }

                if (format.indexOf("%Y") == -1 || format.indexOf("%m") == -1) {
                    return;
                }

                DateFormat dateFormat = MySqlUtils.toJavaFormat(format, timeZone);
                if (dateFormat == null) {
                    return;
                }
                try {
                    date = dateFormat.parse(chars);
                } catch (ParseException e) {
                    // skip
                }
            }
        }

        if (date == null) {
            return;
        }

        long unixtime = date.getTime() / 1000;
        SQLIntegerExpr unixtimeRight = SQLIntegerExpr.ofIntOrLong(unixtime);

        if (arguments.size() == 1) {
            SQLExpr arg0 = func.getArguments().get(0);
            x.setRight(unixtimeRight);
            x.setLeft(arg0);
            return;
        }

        if (arguments.size() == 2) {
            SQLExpr arg0 = func.getArguments().get(0);
            x.setRight(unixtimeRight);
            x.setLeft(arg0);
            return;
        }
    }

    public void binaryOp_to_char(SQLBinaryOpExpr x) {
        SQLMethodInvokeExpr func = (SQLMethodInvokeExpr) x.getLeft();
        final List<SQLExpr> arguments = func.getArguments();
        if (arguments.size() != 2) {
            return;
        }

        final SQLExpr right = x.getRight();
        if (!(right instanceof SQLCharExpr)) {
            return;
        }
        String rightText = ((SQLCharExpr) right).getText();

        SQLExpr arg0 = func.getArguments().get(0);
        SQLExpr arg1 = func.getArguments().get(1);

        if (!(arg1 instanceof SQLCharExpr)) {
            return;
        }
        String format = ((SQLCharExpr) arg1).getText();

        Date date = null;

        DateFormat dateFormat;
        if (format.equals("yyyymmdd")) {
            dateFormat = new SimpleDateFormat("yyyyMMdd");
            if (timeZone != null) {
                dateFormat.setTimeZone(timeZone);
            }
        } else {
            dateFormat = MySqlUtils.toJavaFormat(format, timeZone);
        }

        if (dateFormat == null) {
            return;
        }

        try {
            date = dateFormat.parse(rightText);
        } catch (ParseException e) {
            return;
        }
        if (date == null) {
            return;
        }

        x.setRight(new SQLDateExpr(date, timeZone));
        x.setLeft(arg0);

        visitBinaryFunc(x);
    }

    public void binaryOp_sub_date(SQLBinaryOpExpr x) {
        SQLMethodInvokeExpr func = (SQLMethodInvokeExpr) x.getLeft();

        final List<SQLExpr> arguments = func.getArguments();
        if (arguments.size() == 2) {

            SQLExpr arg0 = func.getArguments().get(0);
            SQLExpr arg1 = func.getArguments().get(1);
            SQLLiteralExpr val = (SQLLiteralExpr) x.getRight();

            if (!(arg0 instanceof SQLName)) {
                return;
            }

            Date date = null;
            if (val instanceof SQLDateExpr) {
                date = ((SQLDateExpr) val).getDate(timeZone);
            } else if (val instanceof SQLCharExpr) {
                String chars = ((SQLCharExpr) val).getText();
                date = MySqlUtils.parseDate(chars, timeZone);
            }

            if (date == null) {
                return;
            }

            Calendar calendar = timeZone == null
                    ? Calendar.getInstance()
                    : Calendar.getInstance(timeZone);
            calendar.setTime(date);

            int delta = 0;
            Integer field = null;
            if (arg1 instanceof SQLIntegerExpr
                    && ((SQLIntegerExpr) arg1).getNumber() instanceof Integer)
            {
                field = Calendar.DAY_OF_YEAR;
                delta = ((SQLIntegerExpr) arg1).getNumber().intValue();
            }

            if (delta == 0 || field == null) {
                return;
            }
            calendar.add(field, delta);

            SQLExpr replaceToValExpr;
            if (val instanceof SQLDateExpr) {
                replaceToValExpr = new SQLDateExpr(calendar.getTime(), timeZone);
            } else {
                replaceToValExpr = new SQLTimestampExpr(calendar.getTime(), timeZone);
            }

            x.setLeft(arg0);
            x.setRight(replaceToValExpr);

            return;
        }
    }

    private SQLExpr mergeTwiceOp(SQLExpr x, SQLBinaryOpExpr left, SQLBinaryOpExpr right) {
        final SQLBinaryOperator leftOp = left.getOperator();
        final SQLBinaryOperator rightOp = right.getOperator();
        final SQLExpr leftLeft = left.getLeft();
        final SQLExpr rightLeft = right.getLeft();
        final SQLExpr leftRight = left.getRight();
        final SQLExpr rightRight = right.getRight();

        final SQLBinaryOperator op;
        if (x instanceof SQLBinaryOpExprGroup) {
            op = ((SQLBinaryOpExprGroup) x).getOperator();
        } else if (x instanceof SQLBinaryOpExpr) {
            op = ((SQLBinaryOpExpr) x).getOperator();
        } else {
            return null;
        }

        SQLExpr replaceTo = null;

        if (leftLeft.equals(rightLeft)) {
            if (leftOp == SQLBinaryOperator.Is
                    && rightOp == SQLBinaryOperator.IsNot
                    && leftRight instanceof SQLNullExpr
                    && rightRight instanceof SQLNullExpr) {
                if (op == BooleanAnd) {
                    replaceTo = new SQLBooleanExpr(false);
                } else if (op == BooleanOr) {
                    replaceTo = new SQLBooleanExpr(true);
                }

                if (replaceTo != null && replaceInParent(x, replaceTo)) {
                    optimizedCount++;
                    return replaceTo;
                }
            }

            Object leftRightVal = null, rightRightVal = null;

            if (leftRight instanceof SQLNumericLiteralExpr) {
                leftRightVal = ((SQLNumericLiteralExpr) leftRight).getNumber();
            } else if (leftRight instanceof SQLTimestampExpr) {
                leftRightVal = ((SQLTimestampExpr) leftRight).getDate(timeZone);
            } else if (leftRight instanceof SQLDateExpr) {
                leftRightVal = ((SQLDateExpr) leftRight).getDate(timeZone);
            } else if (leftRight instanceof SQLCharExpr) {
                leftRightVal = ((SQLCharExpr) leftRight).getText();
            }

            if (rightRight instanceof SQLNumericLiteralExpr) {
                rightRightVal = ((SQLNumericLiteralExpr) rightRight).getNumber();
            } else if (rightRight instanceof SQLTimestampExpr) {
                rightRightVal = ((SQLTimestampExpr) rightRight).getDate(timeZone);
            } else if (rightRight instanceof SQLDateExpr) {
                rightRightVal = ((SQLDateExpr) rightRight).getDate(timeZone);
            } else if (rightRight instanceof SQLCharExpr) {
                rightRightVal = ((SQLCharExpr) rightRight).getText();
            }

            if (leftRightVal instanceof BigDecimal && rightRightVal instanceof Integer) {
                rightRightVal = new BigDecimal(((Integer) rightRightVal).intValue());
            } else if (leftRightVal instanceof Integer && rightRightVal instanceof BigDecimal) {
                leftRightVal = new BigDecimal(((Integer) leftRightVal).intValue());
            }

            if (leftRightVal == null
                    || rightRightVal == null
                    || leftRightVal.getClass() != rightRightVal.getClass() || !(rightRightVal instanceof Comparable)) {
                return null;
            }

            int compareResult = ((Comparable) leftRightVal).compareTo(rightRightVal);
            if (leftRightVal instanceof String && compareResult != 0 && ((String) leftRightVal).length() != ((String) rightRightVal).length()) {
                switch (left.getOperator()) {
                    case Equality:
                    case NotEqual:
                    case LessThanOrGreater:
                    case Like:
                    case NotLike:
                        break;
                    default:
                        return null;
                }
                switch (right.getOperator()) {
                    case Equality:
                    case NotEqual:
                    case LessThanOrGreater:
                    case Like:
                    case NotLike:
                        break;
                    default:
                        return null;
                }
            }

            if (op == BooleanAnd) {
                replaceTo = mergeTwiceOp_and(left, right, leftOp, rightOp, leftLeft, leftRight, compareResult, leftRightVal, rightRightVal);
            } else if (op == BooleanOr) {
                replaceTo = mergeTwiceOp_or(left, right, leftOp, rightOp, leftLeft, compareResult);
            }
        }

        if (replaceTo == null) {
            return null;
        }

        if (x instanceof SQLBinaryOpExprGroup) {
            SQLBinaryOpExprGroup group = (SQLBinaryOpExprGroup) x;
            if (group.getItems().size() >= 2) {
                group.getItems().remove(left);
                group.getItems().remove(right);

                if (group.getItems().size() == 0) {
                    if (!replaceInParent(x, replaceTo)) {
                        group.add(replaceTo);
                    }
                } else {
                    group.add(replaceTo);
                }
                optimizedCount++;

                replaceTo.accept(this);

                return replaceTo;
            }
        } else {
            if (replaceInParent(left, replaceTo)) {
                replaceInParent(right, null);
                optimizedCount++;
                return replaceTo;
            }
        }
//        if (replaceInParent(x, replaceTo)) {
//            optimizedCount++;
//        }

        return null;
    }

    private SQLExpr mergeTwiceOp_or(SQLBinaryOpExpr left
            , SQLBinaryOpExpr right
            , SQLBinaryOperator leftOp
            , SQLBinaryOperator rightOp
            , SQLExpr leftLeft
            , int compareResult)
    {
        SQLExpr replaceTo = null;
        switch (leftOp) {
            case Equality:
                switch (rightOp) {
                    case Equality:
                        if (compareResult == 0) {
                            replaceTo = left;
                        }
                        break;
                    case LessThanOrGreater:
                    case NotEqual:
                        if (compareResult == 0) {
                            replaceTo = SQLBinaryOpExpr.isNotNull(leftLeft.clone());
                        } else {
                            replaceTo = right;
                        }
                        break;
                    case GreaterThanOrEqual:
                        if (compareResult >= 0) {
                            replaceTo = right;
                        }
                        break;
                    case GreaterThan:
                        if (compareResult > 0) {
                            replaceTo = right;
                        } else if (compareResult == 0) {
                            replaceTo = right;
                            right.setOperator(GreaterThanOrEqual);
                        }
                        break;
                    case LessThanOrEqual:
                        if (compareResult <= 0) {
                            replaceTo = right;
                        }
                        break;
                    case LessThan:
                        if (compareResult < 0) {
                            replaceTo = right;
                        } else if (compareResult == 0) {
                            replaceTo = right;
                            right.setOperator(LessThanOrEqual);
                        }
                        break;
                    default:
                        break;
                }
                break;

            case LessThanOrGreater:
            case NotEqual:
                switch (rightOp) {
                    case Equality:
                        if (compareResult == 0) {
                            replaceTo = SQLBinaryOpExpr.isNotNull(leftLeft.clone());
                        } else {
                            replaceTo = left;
                        }
                        break;
                    case LessThanOrGreater:
                    case NotEqual:
                        if (compareResult == 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = SQLBinaryOpExpr.isNotNull(leftLeft.clone());
                        }
                        break;
                    case LessThan:
                        if (compareResult >= 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = SQLBinaryOpExpr.isNotNull(leftLeft.clone());
                        }
                        break;
                    case LessThanOrEqual:
                        if (compareResult > 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = SQLBinaryOpExpr.isNotNull(leftLeft.clone());
                        }
                        break;
                    case GreaterThan:
                        if (compareResult <= 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = SQLBinaryOpExpr.isNotNull(leftLeft.clone());
                        }
                        break;
                    case GreaterThanOrEqual:
                        if (compareResult < 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = SQLBinaryOpExpr.isNotNull(leftLeft.clone());
                        }
                        break;
                    default:
                        break;
                }
                break;

            case GreaterThan:
                switch (rightOp) {
                    case Equality:
                        if (compareResult < 0) {
                            replaceTo = left;
                        } else if (compareResult == 0) {
                            replaceTo = left;
                            left.setOperator(GreaterThanOrEqual);
                        }
                        break;
                    case NotEqual:
                    case LessThanOrGreater:
                        if (compareResult < 0) {
                            replaceTo = SQLBinaryOpExpr.isNotNull(leftLeft.clone());
                        } else {
                            replaceTo = right;
                        }
                        break;

                    case LessThan:
                        if (compareResult < 0) {
                            replaceTo = SQLBinaryOpExpr.isNotNull(leftLeft.clone());
                        } else if (compareResult == 0) {
                            replaceTo = left;
                            left.setOperator(LessThanOrGreater);
                        }
                        break;
                    case LessThanOrEqual:
                        if (compareResult <= 0) {
                            replaceTo = SQLBinaryOpExpr.isNotNull(leftLeft.clone());
                        }
                        break;

                    case GreaterThan:
                    case GreaterThanOrEqual:
                        if (compareResult < 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = right;
                        }
                        break;
                    default:
                        break;
                }
                break;

            case GreaterThanOrEqual:
                switch (rightOp) {
                    case Equality:
                        if (compareResult <= 0) {
                            replaceTo = left;
                        }
                        break;
                    case NotEqual:
                    case LessThanOrGreater:
                        if (compareResult <= 0) {
                            replaceTo = SQLBinaryOpExpr.isNotNull(leftLeft.clone());
                        } else {
                            replaceTo = right;
                        }
                        break;
                    case LessThan:
                    case LessThanOrEqual:
                        if (compareResult <= 0) {
                            replaceTo = SQLBinaryOpExpr.isNotNull(leftLeft.clone());
                        }
                        break;
                    case GreaterThan:
                    case GreaterThanOrEqual:
                        if (compareResult <= 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = right;
                        }
                        break;
                    default:
                        break;
                }
                break;


            case LessThan:
                switch (rightOp) {
                    case Equality:
                        if (compareResult > 0) {
                            replaceTo = left;
                        } else if (compareResult == 0) {
                            replaceTo = left;
                            left.setOperator(LessThanOrEqual);
                        }
                        break;
                    case NotEqual:
                    case LessThanOrGreater:
                        if (compareResult > 0) {
                            replaceTo = SQLBinaryOpExpr.isNotNull(leftLeft.clone());
                        } else {
                            replaceTo = right;
                        }
                        break;

                    case LessThan:
                    case LessThanOrEqual:
                        if (compareResult > 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = right;
                        }
                        break;

                    case GreaterThan:
                        if (compareResult > 0) {
                            replaceTo = SQLBinaryOpExpr.isNotNull(leftLeft.clone());
                        } else if (compareResult == 0) {
                            replaceTo = left;
                            left.setOperator(LessThanOrGreater);
                        }
                        break;

                    case GreaterThanOrEqual:
                        if (compareResult >= 0) {
                            replaceTo = SQLBinaryOpExpr.isNotNull(leftLeft.clone());
                        }
                        break;

                    default:
                        break;
                }
                break;

            case LessThanOrEqual:
                switch (rightOp) {
                    case Equality:
                        if (compareResult >= 0) {
                            replaceTo = left;
                        }
                        break;
                    case NotEqual:
                    case LessThanOrGreater:
                        if (compareResult >= 0) {
                            replaceTo = SQLBinaryOpExpr.isNotNull(leftLeft.clone());
                        } else {
                            replaceTo = right;
                        }
                        break;
                    case LessThan:
                    case LessThanOrEqual:
                        if (compareResult >= 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = right;
                        }
                        break;
                    case GreaterThan:
                    case GreaterThanOrEqual:
                        if (compareResult >= 0) {
                            replaceTo = SQLBinaryOpExpr.isNotNull(leftLeft.clone());
                        }
                        break;
                    default:
                        break;
                }
                break;
            default:
                break;
        }
        return replaceTo;
    }

    private SQLExpr mergeTwiceOp_and(SQLBinaryOpExpr left
            , SQLBinaryOpExpr right
            , SQLBinaryOperator leftOp
            , SQLBinaryOperator rightOp
            , SQLExpr leftLeft
            , SQLExpr leftRight
            , int compareResult
            , Object leftRightVal
            , Object rightRightVal)
    {

        SQLExpr replaceTo = null;
        switch (leftOp) {
            case LessThan:
                switch (rightOp) {
                    case Equality:
                        if (compareResult > 0) {
                            replaceTo = right;
                        } else {
                            replaceTo = new SQLBooleanExpr(false);
                        }
                        break;
                    case NotEqual:
                    case LessThanOrGreater:
                        if (compareResult <= 0) {
                            replaceTo = left;
                        }
                        break;
                    case LessThan:
                        if (compareResult < 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = right;
                        }
                        break;
                    case LessThanOrEqual:
                        if (compareResult <= 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = right;
                        }
                        break;
                    case GreaterThan:
                        if (compareResult <= 0) {
                            replaceTo = new SQLBooleanExpr(false);
                        } else {
                            SQLName leftName = (SQLName) left.getLeft();
                            final SQLColumnDefinition resolvedColumn = leftName.getResolvedColumn();
                            if (resolvedColumn != null
                                    && resolvedColumn.getDataType() != null
                                    && resolvedColumn.getDataType().isInt()
                                    && leftRightVal instanceof Integer
                                    && rightRightVal instanceof Integer) {
                                int leftRightValInt = ((Integer) leftRightVal).intValue();
                                int rightRightValInt = ((Integer) rightRightVal).intValue();

                                if (leftRightValInt - rightRightValInt == 2) {
                                    replaceTo = new SQLBinaryOpExpr(leftLeft.clone()
                                            , Equality
                                            , new SQLIntegerExpr(rightRightValInt + 1)
                                    );
                                }
                            }
                        }

                        break;
                    case GreaterThanOrEqual:
                        if (compareResult <= 0) {
                            replaceTo = new SQLBooleanExpr(false);
                        }
                        break;
                    default:
                        break;
                }
                break;

            case LessThanOrEqual:
                switch (rightOp) {
                    case Equality:
                        if (compareResult >= 0) {
                            replaceTo = right;
                        } else {
                            replaceTo = new SQLBooleanExpr(false);
                        }
                        break;
                    case NotEqual:
                    case LessThanOrGreater:
                        if (compareResult < 0) {
                            replaceTo = left;
                        } else if (compareResult == 0){
                            replaceTo = left;
                            left.setOperator(LessThan);
                        }
                        break;
                    case LessThan:
                        if (compareResult < 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = right;
                        }
                        break;
                    case LessThanOrEqual:
                        if (compareResult < 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = right;
                        }
                        break;
                    case GreaterThan:
                        if (compareResult <= 0) {
                            replaceTo = new SQLBooleanExpr(false);
                        }
                        break;
                    case GreaterThanOrEqual:
                        if (compareResult < 0) {
                            replaceTo = new SQLBooleanExpr(false);
                        } else if (compareResult == 0) {
                            replaceTo = new SQLBinaryOpExpr(leftLeft.clone(), Equality, leftRight.clone());
                        }
                        break;
                    default:
                        break;
                }
                break;

            case Equality:
                switch (rightOp) {
                    case Equality:
                        if (compareResult == 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = new SQLBooleanExpr(false);
                        }
                        break;
                    case NotEqual:
                    case LessThanOrGreater:
                        if (compareResult == 0) {
                            replaceTo = new SQLBooleanExpr(false);
                        } else {
                            replaceTo = left;
                        }
                        break;
                    case LessThan:
                        if (compareResult >= 0) {
                            replaceTo = new SQLBooleanExpr(false);
                        } else {
                            replaceTo = left;
                        }
                        break;
                    case LessThanOrEqual:
                        if (compareResult <= 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = new SQLBooleanExpr(false);
                        }
                        break;
                    case GreaterThan:
                        if (compareResult <= 0) {
                            replaceTo = new SQLBooleanExpr(false);
                        } else {
                            replaceTo = left;
                        }
                        break;
                    case GreaterThanOrEqual:
                        if (compareResult >= 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = new SQLBooleanExpr(false);
                        }
                        break;
                    case Like:
                        if (right.getRight() instanceof SQLCharExpr) {
                            String text = ((SQLCharExpr) right.getRight()).getText();
                            if (text.indexOf('%') == -1 && text.indexOf('_') == -1) {
                                if (compareResult == 0) {
                                    replaceTo = left;
                                } else {
                                    replaceTo = new SQLBooleanExpr(false);
                                }
                            }
                        }
                        break;
                    case NotLike:
                        if (right.getRight() instanceof SQLCharExpr) {
                            String text = ((SQLCharExpr) right.getRight()).getText();
                            if (text.indexOf('%') == -1 && text.indexOf('_') == -1) {
                                if (compareResult == 0) {
                                    replaceTo = new SQLBooleanExpr(false);
                                }
                            }
                        }
                        break;
                    default:
                        break;
                }
                break;
            case NotEqual:
            case LessThanOrGreater:
                switch (rightOp) {
                    case NotEqual:
                    case LessThanOrGreater:
                        if (compareResult == 0) {
                            replaceTo = left;
                        }
                        break;
                    case Equality:
                        if (compareResult == 0) {
                            replaceTo = new SQLBooleanExpr(false);
                        } else {
                            replaceTo = right;
                        }
                        break;
                    case LessThan:
                        if (compareResult >= 0) {
                            replaceTo = right;
                        }
                        break;
                    case LessThanOrEqual:
                        if (compareResult > 0) {
                            replaceTo = right;
                        } else if (compareResult == 0) {
                            replaceTo = right;
                            right.setOperator(LessThan);
                        }
                        break;
                    case GreaterThan:
                        if (compareResult <= 0) {
                            replaceTo = right;
                        }
                        break;
                    case GreaterThanOrEqual:
                        if (compareResult < 0) {
                            replaceTo = right;
                        } else if (compareResult == 0) {
                            replaceTo = right;
                            right.setOperator(GreaterThan);
                        }
                        break;
                    default:
                        break;
                }
                break;

            case GreaterThan:
                switch (rightOp) {
                    case LessThan:
                        if (compareResult >= 0) {
                            replaceTo = new SQLBooleanExpr(false);
                        } else {
                            if (left.getLeft() instanceof SQLCastExpr) {

                                SQLDataType sqlDataType = left.getLeft().computeDataType();

                                if (sqlDataType != null
                                        && sqlDataType.isInt()
                                        && leftRightVal instanceof Integer
                                        && rightRightVal instanceof Integer) {
                                    int leftRightValInt = ((Integer) leftRightVal).intValue();
                                    int rightRightValInt = ((Integer) rightRightVal).intValue();

                                    if (rightRightValInt - leftRightValInt == 2) {
                                        replaceTo = new SQLBinaryOpExpr(leftLeft.clone()
                                                , Equality
                                                , new SQLIntegerExpr(leftRightValInt + 1)
                                        );
                                    }
                                }
                            } else {
                                SQLName leftName = (SQLName) left.getLeft();
                                final SQLColumnDefinition resolvedColumn = leftName.getResolvedColumn();
                                if (resolvedColumn != null
                                        && resolvedColumn.getDataType() != null
                                        && resolvedColumn.getDataType().isInt()
                                        && leftRightVal instanceof Integer
                                        && rightRightVal instanceof Integer) {
                                    int leftRightValInt = ((Integer) leftRightVal).intValue();
                                    int rightRightValInt = ((Integer) rightRightVal).intValue();

                                    if (rightRightValInt - leftRightValInt == 2) {
                                        replaceTo = new SQLBinaryOpExpr(leftLeft.clone()
                                                , Equality
                                                , new SQLIntegerExpr(leftRightValInt + 1)
                                        );
                                    }
                                }
                            }
                        }
                        break;
                    case LessThanOrEqual:
                        if (compareResult > 0) {
                            replaceTo = new SQLBooleanExpr(false);
                        }
                        break;
                    case Equality:
                        if (compareResult >= 0) {
                            replaceTo = new SQLBooleanExpr(false);
                        }
                        break;
                    case NotEqual:
                    case LessThanOrGreater:
                        if (compareResult >= 0) {
                            replaceTo = left;
                        }
                        break;
                    case GreaterThan:
                        if (compareResult > 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = right;
                        }
                        break;
                    case GreaterThanOrEqual:
                        if (compareResult >= 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = right;
                        }
                        break;
                    default:
                        break;
                }
                break;

            case GreaterThanOrEqual:
                switch (rightOp) {
                    case GreaterThan:
                        if (compareResult > 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = right;
                        }
                        break;
                    case GreaterThanOrEqual:
                        if (compareResult > 0) {
                            replaceTo = left;
                        } else {
                            replaceTo = right;
                        }
                        break;
                    case LessThan:
                        if (compareResult >= 0) {
                            replaceTo = new SQLBooleanExpr(false);
                        }
                        break;
                    case LessThanOrEqual:
                        if (compareResult > 0) {
                            replaceTo = new SQLBooleanExpr(false);
                        } else if (compareResult == 0) {
                            replaceTo = new SQLBinaryOpExpr(leftLeft.clone(), Equality, leftRight.clone());
                        }
                        break;
                    case Equality:
                        if (compareResult <= 0) {
                            replaceTo = right;
                        } else {
                            replaceTo = new SQLBooleanExpr(false);
                        }
                        break;
                    case NotEqual:
                    case LessThanOrGreater:
                        if (compareResult > 0) {
                            replaceTo = left;
                        } else if (compareResult == 0){
                            replaceTo = left;
                            left.setOperator(GreaterThan);
                        }
                        break;
                    default:
                        break;
                }
                break;
            case Like:
                switch (rightOp) {
                    case Equality:
                        if (leftRight instanceof SQLCharExpr) {
                            String text = ((SQLCharExpr) leftRight).getText();
                            if (text.indexOf('%') == -1 && text.indexOf('_') == -1) {
                                if (compareResult == 0) {
                                    replaceTo = right;
                                } else {
                                    replaceTo = new SQLBooleanExpr(false);
                                }
                            }
                        }
                        break;
                    default:
                        break;
                }
                break;
            default:
                break;
        }
        return replaceTo;
    }


    private void handValueValue(SQLBinaryOpExpr x) {
        SQLBinaryOperator op = x.getOperator();
        SQLValuableExpr leftExpr = (SQLValuableExpr) x.getLeft();
        SQLValuableExpr rightExpr = (SQLValuableExpr) x.getRight();

        if (rightExpr instanceof SQLNullExpr) {
            final SQLObject parent = x.getParent();
            if ((op == SQLBinaryOperator.Is || op == SQLBinaryOperator.IsNot)
                    && parent instanceof SQLBinaryOpExpr) {
                SQLBinaryOperator parentOp = ((SQLBinaryOpExpr) parent).getOperator();
                switch (parentOp) {
                    case Like:
                    case NotLike:
                    case RLike:
                    case NotRLike:
                    case SoudsLike:
                        return;
                    default:
                        break;
                }
            }

            SQLExpr replaceTo = null;
            if (op == SQLBinaryOperator.Is) {
                if (leftExpr instanceof SQLNullExpr) {
                    replaceTo = new SQLBooleanExpr(true);
                } else {
                    replaceTo = new SQLBooleanExpr(false);
                }
            } else if (op == SQLBinaryOperator.IsNot) {
                if (leftExpr instanceof SQLNullExpr) {
                    replaceTo = new SQLBooleanExpr(false);
                } else {
                    replaceTo = new SQLBooleanExpr(true);
                }
            }

            if (replaceTo != null && replaceInParent(x, replaceTo)) {
                optimizedCount++;
            }
        }

        if (leftExpr instanceof SQLDateExpr
                && rightExpr instanceof SQLIntegerExpr) {
            SQLDateExpr dateExpr = (SQLDateExpr) leftExpr;
            Date date = dateExpr.getDate();
            if (date != null && (dbType == DbType.mysql || dbType == DbType.ads)) {
                String text = new SimpleDateFormat("yyyyMMdd").format(date);
                int dateIntValue = Integer.parseInt(text);
                x.setLeft(leftExpr = new SQLIntegerExpr(dateIntValue));
            }
        }

        if (leftExpr instanceof SQLTimestampExpr
                && rightExpr instanceof SQLIntegerExpr) {
            SQLTimestampExpr dateExpr = (SQLTimestampExpr) leftExpr;
            Date date = dateExpr.getDate(timeZone);
            if (date != null && (dbType == DbType.mysql || dbType == DbType.ads)) {
                String text = new SimpleDateFormat("yyyyMMddHHmmss").format(date);
                long dateIntValue = Long.parseLong(text);
                x.setLeft(leftExpr = new SQLIntegerExpr(dateIntValue));
            }
        }

        if (leftExpr instanceof SQLCharExpr && (rightExpr instanceof SQLTimestampExpr || rightExpr instanceof SQLDateExpr)) {
            SQLTimestampExpr tsExpr = new SQLTimestampExpr(((SQLCharExpr) leftExpr).getText());
            Date date = tsExpr.getDate(timeZone);
            if (date != null && (dbType == DbType.mysql || dbType == DbType.ads)) {
                String text = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss").format(date);
                SQLTimestampExpr replaceTo = new SQLTimestampExpr(text);
                x.setLeft(leftExpr = replaceTo);
            }
        }

        if (leftExpr instanceof SQLTimestampExpr
                && rightExpr instanceof SQLTimestampExpr) {
            {
                SQLTimestampExpr leftDateExpr = (SQLTimestampExpr) leftExpr;
                Date date = leftDateExpr.getDate(timeZone);
                if (date != null && (dbType == DbType.mysql || dbType == DbType.ads)) {
                    String text = new SimpleDateFormat("yyyyMMddHHmmss").format(date);
                    long dateIntValue = Long.parseLong(text);
                    x.setLeft(leftExpr = new SQLIntegerExpr(dateIntValue));
                }
            }
            {
                SQLTimestampExpr rightDateExpr = (SQLTimestampExpr) rightExpr;
                Date date = rightDateExpr.getDate(timeZone);
                if (date != null && (dbType == DbType.mysql || dbType == DbType.ads)) {
                    String text = new SimpleDateFormat("yyyyMMddHHmmss").format(date);
                    long dateIntValue = Long.parseLong(text);
                    x.setRight(rightExpr = new SQLIntegerExpr(dateIntValue));
                }
            }
        }

        Object leftVal = leftExpr.getValue();
        Object rightVal = rightExpr.getValue();

        if (leftVal instanceof Integer && rightVal instanceof String) {
            long rightValLong;
            try {
                rightValLong = Long.parseLong((String) rightVal);
            } catch (NumberFormatException ex) {
                return;
            }

            if (rightValLong >= Integer.MIN_VALUE && rightValLong <= Integer.MAX_VALUE) {
                rightVal = (int) rightValLong;
            } else {
                rightVal = rightValLong;
            }
        } else if (leftVal instanceof Integer && rightVal instanceof byte[] && rightExpr instanceof SQLHexExpr) {
            byte[] bytes = (byte[]) rightVal;
            long hex = 0;
            for (int i = 0; i < bytes.length; i++) {
                byte b = bytes[i];
                hex = hex * 16 + b;
            }
            if (hex >= Integer.MIN_VALUE && hex <= Integer.MAX_VALUE) {
                rightVal = (int) hex;
            } else {
                rightVal = hex;
            }
        }

        if (leftVal instanceof Integer && rightVal instanceof Integer) {
            int leftIntVal = ((Integer) leftVal).intValue();
            int rightIntVal = ((Integer) rightVal).intValue();

            SQLExpr replaceTo = null;
            switch (op) {
                case Add:
                    // 3 + 4 -> 7
                    replaceTo = SQLIntegerExpr.ofIntOrLong((long) leftIntVal + (long) rightIntVal);
                    break;
                case Subtract: {
                    // 4 - 3 -> 1
                    long result = (long) leftIntVal - (long) rightIntVal;
                    replaceTo = SQLIntegerExpr.ofIntOrLong((long) leftIntVal - (long) rightIntVal);
                    break;
                }
                case Multiply: {
                    // 4 * 3 -> 21
                    long result = ((long)leftIntVal) * (long) rightIntVal;
                    replaceTo = SQLIntegerExpr.ofIntOrLong(result);
                    break;
                }
                case Divide: {
                    // 4 * 3 -> 21
                    if (rightIntVal == 0) {
                        break;
                    }
                    int div = leftIntVal / rightIntVal;
                    int mod = leftIntVal % rightIntVal;
                    if (mod == 0) {
                        replaceTo = new SQLIntegerExpr(div);
                    }
                    break;
                }
                case Mod:
                case Modulus: {
                    if (rightIntVal == 0) {
                        break;
                    }
                    int mod = leftIntVal % rightIntVal;
                    replaceTo = new SQLIntegerExpr(mod);
                    break;
                }
                case GreaterThan: {
                    boolean result = leftIntVal > rightIntVal;
                    replaceTo = new SQLBooleanExpr(result);
                    break;
                }
                case GreaterThanOrEqual: {
                    boolean result = leftIntVal >= rightIntVal;
                    replaceTo = new SQLBooleanExpr(result);
                    break;
                }
                case LessThan: {
                    boolean result = leftIntVal < rightIntVal;
                    replaceTo = new SQLBooleanExpr(result);
                    break;
                }
                case LessThanOrEqual: {
                    boolean result = leftIntVal <= rightIntVal;
                    replaceTo = new SQLBooleanExpr(result);
                    break;
                }
                case Equality: {
                    boolean result = leftIntVal == rightIntVal;
                    replaceTo = new SQLBooleanExpr(result);
                    break;
                }
                case NotEqual: {
                    boolean result = leftIntVal != rightIntVal;
                    replaceTo = new SQLBooleanExpr(result);
                    break;
                }
                default:
                    break;
            }

            if (replaceTo != null) {
                replaceInParent(x, replaceTo);
                return;
            }
        }

        if (leftVal instanceof Integer && rightVal instanceof Long) {
            leftVal = new Long(((Integer) leftVal).intValue());
        } else if (leftVal instanceof Long && rightVal instanceof Integer) {
            rightVal = new Long(((Integer) rightVal).intValue());
        }

        if (leftVal instanceof Long && rightVal instanceof Long) {
            long lefLongVal = ((Long) leftVal).longValue();
            long rightLongVal = ((Long) rightVal).longValue();

            switch (op) {
                case Add: {
                    // 3 + 4 -> 7
                    SQLIntegerExpr replaceTo = SQLIntegerExpr.add(lefLongVal, rightLongVal);
                    if (replaceTo.getNumber() instanceof BigInteger) {
                        break;
                    }
                    if (replaceInParent(x, replaceTo)) {
                        optimizedCount++;
                        return;
                    }
                    break;
                }
                case Subtract:
                    // 4 - 3 -> 1
                    if (replaceInParent(x, new SQLIntegerExpr(lefLongVal - rightLongVal))) {
                        optimizedCount++;
                        return;
                    }
                    break;
                case Multiply: {
                    // 4 * 3 -> 21
                    if (lefLongVal < Integer.MIN_VALUE  || lefLongVal > Integer.MAX_VALUE) {
                        break;
                    }
                    if (rightLongVal < Integer.MIN_VALUE  || rightLongVal > Integer.MAX_VALUE) {
                        break;
                    }
                    long result = lefLongVal * rightLongVal;
                    if (result != 0 && replaceInParent(x, new SQLIntegerExpr(result))) {
                        optimizedCount++;
                        return;
                    }
                    break;
                }
                case Divide: {
                    // 4 * 3 -> 21
                    long result = lefLongVal / rightLongVal;
                    long mod = lefLongVal % rightLongVal;
                    if (mod == 0 && replaceInParent(x, new SQLIntegerExpr(result))) {
                        optimizedCount++;
                        return;
                    }

                    break;
                }
                case Mod:
                case Modulus: {
                    long result = lefLongVal % rightLongVal;
                    if (replaceInParent(x, new SQLIntegerExpr(result))) {
                        optimizedCount++;
                        return;
                    }
                    break;
                }
                case GreaterThan: {
                    boolean result = lefLongVal > rightLongVal;
                    if (replaceInParent(x, new SQLBooleanExpr(result))) {
                        optimizedCount++;
                        return;
                    }
                    break;
                }
                case GreaterThanOrEqual: {
                    boolean result = lefLongVal >= rightLongVal;
                    if (replaceInParent(x, new SQLBooleanExpr(result))) {
                        optimizedCount++;
                        return;
                    }
                    break;
                }
                case LessThan: {
                    boolean result = lefLongVal < rightLongVal;
                    if (replaceInParent(x, new SQLBooleanExpr(result))) {
                        optimizedCount++;
                        return;
                    }
                    break;
                }
                case LessThanOrEqual: {
                    boolean result = lefLongVal <= rightLongVal;
                    if (replaceInParent(x, new SQLBooleanExpr(result))) {
                        optimizedCount++;
                        return;
                    }
                    break;
                }
                default:
                    break;
            }
        }

        if (leftVal instanceof BigDecimal && (rightVal instanceof Long || rightVal instanceof Integer)) {
            rightVal = new BigDecimal(((Number) rightVal).longValue());
        } else if (rightExpr instanceof BigDecimal && (leftVal instanceof Long || leftVal instanceof Integer)) {
            leftVal = new BigDecimal(((Number) leftVal).longValue());
        }

        if (leftVal instanceof Double && rightVal instanceof Number) {
            rightVal = ((Number) rightVal).doubleValue();
        } else if (leftVal instanceof Float && rightVal instanceof Number) {
            leftVal = ((Float) leftVal).doubleValue();
            rightVal = ((Number) rightVal).doubleValue();
        } else if (leftVal instanceof Integer && rightVal instanceof BigDecimal) {
            leftVal = BigDecimal.valueOf(((Integer) leftVal).intValue());
        }

        if (leftVal instanceof BigDecimal && rightVal instanceof BigDecimal) {
            BigDecimal leftRealVal = (BigDecimal) leftVal;
            BigDecimal rightRealVal = (BigDecimal) rightVal;

            boolean decimalExpr = leftExpr instanceof SQLDecimalExpr || rightExpr instanceof SQLDecimalExpr;
            BigDecimal value = null;
            switch (op) {
                case Add:
                    value = leftRealVal.add(rightRealVal);
                    break;
                case Subtract:
                    value = leftRealVal.subtract(rightRealVal);
                    break;
                case Multiply:
                    if (!decimalExpr) {
                        value = leftRealVal.multiply(rightRealVal);
                    }
                    break;
                case Divide: {
                    if ((!decimalExpr) && !BigDecimal.ZERO.equals(rightRealVal)) {
                        try {
                            value = leftRealVal.divide(rightRealVal);
                            optimizedCount++;
                        } catch (Exception e) {
                            // ignore
                        }
                    }
                    break;
                }
                default:
                    break;
            }
            if (value != null) {
                if (decimalExpr) {
                    replaceInParent(x, new SQLDecimalExpr(value));
                } else {
                    replaceInParent(x, new SQLNumberExpr(value));
                }
            }
        } else if (leftVal instanceof Double && rightVal instanceof Double) {
            double leftDoubleVal = (Double) leftVal;
            double rightDoubleVal = (Double) rightVal;

            switch (op) {
                case Add:
                    if (replaceInParent(x, new SQLNumberExpr(leftDoubleVal + rightDoubleVal))) {
                        optimizedCount++;
                        return;
                    }
                    break;
                case Subtract:
                    if (replaceInParent(x,
                            new SQLNumberExpr(leftDoubleVal - rightDoubleVal))) {
                        optimizedCount++;
                        return;
                    }
                    break;
                case Multiply:
                    if (replaceInParent(x, new SQLNumberExpr(leftDoubleVal * rightDoubleVal))) {
                        optimizedCount++;
                        return;
                    }
                    break;
                case Divide: {
                    if (replaceInParent(x, new SQLNumberExpr(leftDoubleVal / rightDoubleVal))) {
                        optimizedCount++;
                        return;
                    }

                    break;
                }
                default:
                    break;
            }
        } else if (leftVal instanceof String && rightVal instanceof String) {

            int compareTo = ((String) leftVal).compareTo((String) rightVal);

            switch (op) {
                case Concat:
                    SQLCharExpr replaceTo = new SQLCharExpr((String) leftVal + (String) rightVal);
                    replaceInParent(x, replaceTo);
                    break;
                case GreaterThanOrEqual:
                    replaceInParent(x, new SQLBooleanExpr(compareTo > 0 || compareTo == 0));
                    break;
                case GreaterThan:
                    replaceInParent(x, new SQLBooleanExpr(compareTo > 0));
                    break;
                case LessThanOrEqual:
                    replaceInParent(x, new SQLBooleanExpr(compareTo < 0 || compareTo == 0));
                    break;
                case LessThan:
                    replaceInParent(x, new SQLBooleanExpr(compareTo < 0));
                    break;
                case Equality:
                    replaceInParent(x, new SQLBooleanExpr(compareTo == 0));
                default:
                    break;
            }
        } else if (leftVal instanceof String && rightVal instanceof Integer) {

            try {
                Integer leftInt = Integer.valueOf((String) leftVal);

                int compareTo = leftInt.compareTo((Integer) rightVal);

                switch (op) {
                    case GreaterThanOrEqual:
                        replaceInParent(x, new SQLBooleanExpr(compareTo > 0 || compareTo == 0));
                        break;
                    case GreaterThan:
                        replaceInParent(x, new SQLBooleanExpr(compareTo > 0));
                        break;
                    case LessThanOrEqual:
                        replaceInParent(x, new SQLBooleanExpr(compareTo < 0 || compareTo == 0));
                        break;
                    case LessThan:
                        replaceInParent(x, new SQLBooleanExpr(compareTo < 0));
                        break;
                    case Equality:
                        replaceInParent(x, new SQLBooleanExpr(compareTo == 0));
                        break;
                    default:
                        break;
                }

            } catch (Exception e) {
                replaceInParent(x, new SQLBooleanExpr(false));
            }
        }

    }



    /**
     * 1. a or b or true -> true
     * 2. a or b or true -> a or b
     * 3. a and b and true -> a and b
     * 4. a and b and false -> false
     */
    protected boolean visitGroupBoolean(SQLBinaryOpExprGroup x) {
        SQLBinaryOperator operator = x.getOperator();
        List<SQLExpr> items = x.getItems();

        if (operator == BooleanOr) {
            for (int i = items.size() - 1; i >= 0; i--) {
                SQLExpr expr = items.get(i);
                if (expr instanceof SQLBooleanExpr) {
                    final boolean value = ((SQLBooleanExpr) expr).getBooleanValue();
                    if (value) {
                        // a or b or true -> true
                        if (replaceInParent(x, new SQLBooleanExpr(true))) {
                            return true;
                        }
                    } else {
                        // a or b or true -> a or b
                        items.remove(i);
                        optimizedCount++;
                    }
                }
            }
        } else if (operator == BooleanAnd) {
            for (int i = items.size() - 1; i >= 0; i--) {
                SQLExpr expr = items.get(i);
                if (expr instanceof SQLBooleanExpr) {
                    final boolean value = ((SQLBooleanExpr) expr).getBooleanValue();
                    if (value) {
                        // a and b and true -> a and b
                        items.remove(i);
                        optimizedCount++;
                    } else {
                        // a and b and false -> false
                        if (replaceInParent(x, new SQLBooleanExpr(false))) {
                            return true;
                        }
                    }
                }
            }
        }

        return false;
    }

    public void endVisit(SQLBinaryOpExprGroup x) {
        visitGroupBoolean(x);
    }

    public boolean visit(SQLBinaryOpExprGroup x) {
        for (int i = 0; i < x.getItems().size(); i++) {
            SQLExpr item = x.getItems().get(i);
            item.accept(this);
        }

        SQLBinaryOperator operator = x.getOperator();
        {
            List<SQLExpr> items = x.getItems();
            for (int i = x.getItems().size() - 1; i >= 0; i--) {
                SQLExpr expr = items.get(i);
                if (items.indexOf(expr) < i) {
                    items.remove(i);
                    optimizedCount++;
                }
            }
        }

        if (x.getOperator() == BooleanOr){
            List<SQLExpr> items = x.getItems();
            for (int i = items.size() - 1; i >= 0; i--) {
                if (i == items.size() - 1) {
                    continue;
                }

                SQLExpr item = items.get(i);
                SQLExpr postItem = items.get(i + 1);

                if (item instanceof SQLBinaryOpExpr
                        && ((SQLBinaryOpExpr) item).isLeftNameAndRightLiteral()
                        && ((SQLBinaryOpExpr) item).getOperator() == Equality) {
                    SQLBinaryOpExpr binaryOpExpr = (SQLBinaryOpExpr) item;
                    SQLName leftName = (SQLName) binaryOpExpr.getLeft();
                    SQLLiteralExpr right = (SQLLiteralExpr) binaryOpExpr.getRight();

                    if (postItem instanceof SQLBinaryOpExpr
                            && ((SQLBinaryOpExpr) postItem).isLeftNameAndRightLiteral()) {

                        SQLBinaryOpExpr postBinaryOpExpr = (SQLBinaryOpExpr) postItem;
                        SQLName postLeft = (SQLName) postBinaryOpExpr.getLeft();
                        SQLLiteralExpr postRight = (SQLLiteralExpr) postBinaryOpExpr.getRight();

                        if (leftName.equals(postLeft)
                                && postBinaryOpExpr.getOperator() == Equality
                                && !(postRight instanceof SQLNullExpr)
                        ) {
                            SQLInListExpr in = new SQLInListExpr(leftName.clone());
                            in.addTarget(right.clone());
                            in.addTarget(postRight.clone());
                            in.setParent(x);
                            items.remove(i + 1);
                            items.set(i, in);
                        }
                    } else if (postItem instanceof SQLInListExpr
                            && leftName.equals(((SQLInListExpr) postItem).getExpr())
                            && !((SQLInListExpr) postItem).isNot()) {
                        SQLInListExpr in = (SQLInListExpr) postItem;
                        in.addTarget(0, right.clone());
                        items.remove(i);
                    }
                } else if (item instanceof SQLInListExpr && !((SQLInListExpr) item).isNot()) {
                    SQLInListExpr in = (SQLInListExpr) item;
                    SQLExpr expr = in.getExpr();

                    if (postItem instanceof SQLBinaryOpExpr
                            && ((SQLBinaryOpExpr) postItem).isLeftNameAndRightLiteral()
                            && ((SQLBinaryOpExpr) postItem).getOperator() == Equality) {
                        SQLBinaryOpExpr postBinaryOpExpr = (SQLBinaryOpExpr) postItem;
                        SQLName postLeft = (SQLName) postBinaryOpExpr.getLeft();
                        SQLLiteralExpr postRight = (SQLLiteralExpr) postBinaryOpExpr.getRight();

                        if (expr.equals(postLeft)) {
                            in.addTarget(postRight.clone());
                            items.remove(i + 1);
                        }
                    } else if (postItem instanceof SQLInListExpr
                            && expr.equals(((SQLInListExpr) postItem).getExpr())
                            && !((SQLInListExpr) postItem).isNot()) {
                        SQLInListExpr postIn = (SQLInListExpr) postItem;
                        for (SQLExpr postValue : postIn.getTargetList()) {
                            in.addTarget(postValue.clone());
                        }
                        items.remove(i + 1);
                    }
                }
            }
        }
        if (x.getItems().size() == 1) {
            SQLUtils.replaceInParent(x, x.getItems().get(0));
            return false;
        }

        if (visitGroupBoolean(x)) {
            return false;
        }
        List<SQLExpr> items = new ArrayList<SQLExpr>(x.getItems());
        Collections.sort(items, SQLExprComparor.instance);

//        List<SQLExpr> items1 = new ArrayList<SQLExpr>(x.getItems());
//        try {
//            Collections.sort(items1, new Comparator<SQLExpr>() {
//
//                @Override
//                public int compare(SQLExpr a, SQLExpr b) {
//                    if (a instanceof SQLBinaryOpExpr && b instanceof SQLBinaryOpExpr) {
//                        return compare((SQLBinaryOpExpr) a, (SQLBinaryOpExpr) b);
//                    }
//                    return 0;
//                }
//
//                public int compare(SQLBinaryOpExpr a, SQLBinaryOpExpr b) {
//                    if (isNameAndLiteral(a) && b.isBothName()) {
//                        return -1;
//                    }
//
//                    if (isNameAndLiteral(b) && a.isBothName()) {
//                        return 1;
//                    }
//
//                    if (isNameAndLiteral(a) && isNameAndLiteral(b)) {
//                        SQLName aName = getName(a);
//                        SQLName bName = getName(b);
//
//                        SQLLiteralExpr aLiteral = getLiterlal(a);
//                        SQLLiteralExpr bLiteral = getLiterlal(b);
//
//                        if (aName.equals(bName) && aLiteral.getClass() == bLiteral.getClass()) {
//
//                        }
//                    }
//
//                    return 0;
//                }
//
//                private SQLName getName(SQLBinaryOpExpr x) {
//                    SQLExpr left = x.getLeft();
//                    SQLExpr right = x.getRight();
//
//                    if (left instanceof SQLName) {
//                        return (SQLName) left;
//                    }
//
//                    if (right instanceof SQLName) {
//                        return (SQLName) right;
//                    }
//
//                    return null;
//                }
//
//                private SQLLiteralExpr getLiterlal(SQLBinaryOpExpr x) {
//                    SQLExpr left = x.getLeft();
//                    SQLExpr right = x.getRight();
//
//                    if (left instanceof SQLLiteralExpr) {
//                        return (SQLLiteralExpr) left;
//                    }
//
//                    if (right instanceof SQLLiteralExpr) {
//                        return (SQLLiteralExpr) right;
//                    }
//
//                    return null;
//                }
//
//
//            });
//        } catch (IllegalArgumentException e) {
//            return false;
//            // skip
//        }

        if (operator == BooleanAnd) {
            Map<SQLExpr, SQLExpr> constFields = new HashMap<SQLExpr, SQLExpr>();

            Map<SQLName, List<SQLBinaryOpExpr>> nameConditions = new HashMap<SQLName, List<SQLBinaryOpExpr>>();
            for (int i = 0; i < items.size(); i++) {
                SQLExpr item = items.get(i);

                if (item instanceof SQLBinaryOpExpr) {
                    SQLBinaryOpExpr binaryOpExpr = (SQLBinaryOpExpr) item;
                    SQLExpr left = binaryOpExpr.getLeft();
                    SQLExpr right = binaryOpExpr.getRight();
                    if (binaryOpExpr.getOperator() == Equality
                            && left instanceof SQLName
                            && right instanceof SQLValuableExpr) {
                        constFields.put(left, right);
                    }

                    if (isLeftNameAndRightLiteral(item)) {
                        List<SQLBinaryOpExpr> exprList = nameConditions.get(left);
                        if (exprList == null) {
                            exprList = new ArrayList<SQLBinaryOpExpr>();
                            nameConditions.put((SQLName) left, exprList);
                        }
                        exprList.add((SQLBinaryOpExpr) item);
                    }

                    if (right instanceof SQLLiteralExpr
                            && left instanceof SQLCastExpr
                            && ((SQLCastExpr) left).getExpr() instanceof SQLName
                    ) {
                        SQLName leftName = (SQLName) ((SQLCastExpr) left).getExpr();
                        List<SQLBinaryOpExpr> exprList = nameConditions.get(leftName);
                        if (exprList == null) {
                            exprList = new ArrayList<SQLBinaryOpExpr>();
                            nameConditions.put(leftName, exprList);
                        }
                        exprList.add((SQLBinaryOpExpr) item);
                    }
                }
            }

            for (int i = 0; i < items.size(); i++) {
                SQLExpr item = items.get(i);

                if (item instanceof SQLBinaryOpExpr) {
                    SQLBinaryOpExpr binaryOpExpr = (SQLBinaryOpExpr) item;
                    SQLExpr left = binaryOpExpr.getLeft();
                    SQLExpr right = binaryOpExpr.getRight();
                    if (left instanceof SQLName && right instanceof SQLName) {
                        SQLExpr leftLiteral = constFields.get(left);

                        final SQLObject parent = x.getParent();
                        if (parent instanceof SQLSelectQueryBlock
                                && !(((SQLSelectQueryBlock) parent).getFrom() instanceof SQLJoinTableSource)) {
                            if (leftLiteral != null) {
                                binaryOpExpr.setLeft(leftLiteral.clone());
                            }

                            SQLExpr rightLiteral = constFields.get(right);
                            if (rightLiteral != null) {
                                binaryOpExpr.setRight(rightLiteral.clone());
                            }
                        } else {
                            SQLBinaryOpExpr binaryOpExpr2 = null;
                            if (leftLiteral != null) {
                                binaryOpExpr2 = binaryOpExpr.clone();
                                binaryOpExpr2.setLeft(leftLiteral.clone());
                            }

                            SQLExpr rightLiteral = constFields.get(right);
                            if (rightLiteral != null) {
                                binaryOpExpr2 = binaryOpExpr.clone();
                                binaryOpExpr2.setRight(rightLiteral.clone());
                            }

                            if (items.contains(binaryOpExpr2)) {
                                binaryOpExpr2 = null;
                            }

                            if (binaryOpExpr2 != null) {
                                binaryOpExpr2.setParent(x);
                                x.getItems().add(binaryOpExpr2);
                                optimizedCount++;
                            }
                        }
                    }
                }
            }

            for (Map.Entry<SQLName, List<SQLBinaryOpExpr>> entry : nameConditions.entrySet()) {
                List<SQLBinaryOpExpr> exprList = entry.getValue();
                SQLBinaryOpExpr left = exprList.get(0);
                for (int i = 1; i < exprList.size(); i++) {
                    SQLBinaryOpExpr right = exprList.get(i);
                    SQLExpr replaceTo = mergeTwiceOp(x, left, right);
                    if (replaceTo == null) {
                        left = right;
                    } else if (replaceTo instanceof SQLBinaryOpExpr) {
                        left = (SQLBinaryOpExpr) replaceTo;
                    }
                }
            }
        } else if (operator == BooleanOr) {
            Map<SQLExpr, SQLExpr> constFields = new HashMap<SQLExpr, SQLExpr>();
            Map<SQLName, List<SQLBinaryOpExpr>> nameConditions = new HashMap<SQLName, List<SQLBinaryOpExpr>>();
            for (int i = 0; i < items.size(); i++) {
                SQLExpr item = items.get(i);

                if (item instanceof SQLBinaryOpExpr) {
                    SQLBinaryOpExpr binaryOpExpr = (SQLBinaryOpExpr) item;
                    SQLExpr left = binaryOpExpr.getLeft();
                    SQLExpr right = binaryOpExpr.getRight();
                    if (binaryOpExpr.getOperator() == Equality
                            && left instanceof SQLName
                            && right instanceof SQLValuableExpr) {
                        constFields.put(left, right);
                    }

                    if (isLeftNameAndRightLiteral(item)) {
                        List<SQLBinaryOpExpr> exprList = nameConditions.get(left);
                        if (exprList == null) {
                            exprList = new ArrayList<SQLBinaryOpExpr>();
                            nameConditions.put((SQLName) left, exprList);
                        }
                        exprList.add((SQLBinaryOpExpr) item);
                    }
                }
            }

            for (Map.Entry<SQLName, List<SQLBinaryOpExpr>> entry : nameConditions.entrySet()) {
                List<SQLBinaryOpExpr> exprList = entry.getValue();

                SQLBinaryOpExpr left = exprList.get(0);
                for (int i = 1; i < exprList.size(); i++) {
                    SQLBinaryOpExpr right = exprList.get(i);
                    SQLExpr replaceTo = mergeTwiceOp(x, left, right);
                    if (replaceTo == null) {
                        left = right;
                    } else if (replaceTo instanceof SQLBinaryOpExpr) {
                        left = (SQLBinaryOpExpr) replaceTo;
                    }
                }
            }
        }

        for (int i = 0; i < items.size(); i++) {
            SQLExpr item = items.get(i);
            item.accept(this);
        }

        if (x.getItems().size() == 0) {
            replaceInParent(x, null);
        }
        return false;
    }

    public boolean visit(SQLInListExpr x) {
        x.getExpr().accept(this);
        for (SQLExpr item : x.getTargetList()) {
            item.accept(this);
        }

        List<SQLExpr> targetList = x.getTargetList();
        Map<Object, Integer> indics = new HashMap<Object, Integer>();
        for (int i = targetList.size() - 1; i >= 0; i--) {
            SQLExpr itemExpr = targetList.get(i);
            indics.put(itemExpr, i);
        }

        for (int i = targetList.size() - 1; i >= 0; i--) {
            SQLExpr item = targetList.get(i);
            Integer idx = indics.get(item);
            if (idx == null) {
                continue;
            }

            if (idx.intValue() < i) {
                targetList.remove(i);
            }
        }

        SQLExpr expr = x.getExpr();


        if (targetList.size() == 1) {

            SQLExpr item0 = targetList.get(0);
            if (item0 instanceof SQLLiteralExpr) {
                if (expr instanceof SQLName) {
                    SQLName name = (SQLName) expr;
                    SQLDataType nameDataType = name.computeDataType();

                    if (item0 instanceof SQLLiteralExpr) {
                        SQLLiteralExpr literal = (SQLLiteralExpr) item0;
                        handleNameLiteral(x, name, literal);
                    }
                }
                if (expr instanceof SQLIdentifierExpr && ((SQLIdentifierExpr) expr).getResolvedColumn() != null) {
                    String typeName = ((SQLIdentifierExpr) expr).getResolvedColumn().getDataType().getName();
                    if ("multivalue".equalsIgnoreCase(typeName) || "varchar".equalsIgnoreCase(typeName)) {
                        return false;
                    }
                }
                if (expr instanceof SQLPropertyExpr && ((SQLPropertyExpr) expr).getResolvedColumn() != null) {
                    String typeName = ((SQLPropertyExpr) expr).getResolvedColumn().getDataType().getName();
                    if ("multivalue".equalsIgnoreCase(typeName) || "varchar".equalsIgnoreCase(typeName)) {
                        return false;
                    }
                }

                SQLBinaryOpExpr binaryOpExpr = new SQLBinaryOpExpr(
                        expr.clone()
                        , x.isNot() ? LessThanOrGreater : Equality
                        , item0.clone()
                );
                if (replaceInParent(x, binaryOpExpr)) {
                    binaryOpExpr.accept(this);
                    optimizedCount++;
                    return false;
                }
            }
        }

        // 1 in (1, 2) -> true
        if (expr instanceof SQLValuableExpr) {
            Object value = ((SQLValuableExpr) expr).getValue();
            if (value != null) {
                Class<?> valueClass = value.getClass();
                int typeMatchCount = 0;
                for (int i = 0; i < targetList.size(); i++) {
                    SQLExpr item = targetList.get(i);
                    if (item instanceof SQLValuableExpr) {
                        Object itemValue = ((SQLValuableExpr) item).getValue();
                        if (value.equals(itemValue)) {
                            SQLBooleanExpr trueExpr = new SQLBooleanExpr(!x.isNot());
                            if (replaceInParent(x, trueExpr)) {
                                optimizedCount++;
                                return false;
                            }
                        } else if (itemValue != null && itemValue.getClass() == valueClass) {
                            typeMatchCount++;
                        }
                    }
                }
                if (typeMatchCount == targetList.size()) {
                    SQLBooleanExpr falseExpr = new SQLBooleanExpr(x.isNot());
                    if (replaceInParent(x, falseExpr)) {
                        optimizedCount++;
                        return false;
                    }
                }
            }
        }

        if (expr instanceof SQLName) {
            SQLName name = (SQLName) expr;
            SQLDataType nameDataType = name.computeDataType();

            for (int i = targetList.size() - 1; i >= 0; i--) {
                SQLExpr item = targetList.get(i);

                if (nameDataType != null
                        && nameDataType.isInt()
                        && item instanceof SQLCharExpr
                        && ((SQLCharExpr) item).getText().length() == 0
                ) {
                    targetList.remove(i);
                    continue;
                }

                if (item instanceof SQLLiteralExpr) {
                    SQLLiteralExpr literal = (SQLLiteralExpr) item;
                    handleNameLiteral(x, name, literal);
                }
            }
        }

        if (targetList.size() > 0) {
            SQLExpr first = targetList.get(0);
            if (first instanceof SQLCharExpr || first instanceof SQLIdentifierExpr) {
                for (int i = 1; i < targetList.size(); i++) {
                    SQLExpr item = targetList.get(i);
                    if (item instanceof SQLHexExpr) {
                        SQLCharExpr charExpr = ((SQLHexExpr) item).toCharExpr();
                        if (charExpr != null) {
                            charExpr.setParent(x);
                            targetList.set(i, charExpr);
                        }
                    }
                }
            }
        }

        if (targetList.size() == 0) {
            SQLBooleanExpr falseExpr = new SQLBooleanExpr(x.isNot());
            if (replaceInParent(x, falseExpr)) {
                optimizedCount++;
                return false;
            }
            return false;
        }

        if (expr instanceof SQLMethodInvokeExpr) {

            SQLMethodInvokeExpr methodInvokeExpr = (SQLMethodInvokeExpr) expr;

            SQLDataType methodDataType = methodInvokeExpr.computeDataType();

            long hashCode64 = methodInvokeExpr.methodNameHashCode64();
            if (hashCode64 == FnvHash.Constants.CONCAT) {
                boolean allChars = true;
                for (SQLExpr item : targetList) {
                    if (!(item instanceof SQLCharExpr)) {
                        allChars = false;
                        break;
                    }
                }

                if (allChars) {
                    SQLMethodInvokeExpr concat = methodInvokeExpr;
                    for (SQLExpr item : concat.getArguments()) {
                        if (item instanceof SQLCharExpr) {
                            String text = ((SQLCharExpr) item).getText();
                            if (text == null || text.length() == 0) {
                                continue;
                            }

                            boolean contains = false;
                            for (SQLExpr target : targetList) {
                                String value = ((SQLCharExpr) target).getText();
                                if (value.indexOf(text) != -1) {
                                    contains = true;
                                    break;
                                }
                            }

                            if (!contains) {
                                replaceInParent(x, new SQLBooleanExpr(x.isNot()));
                                return false;
                            }
                        }
                    }
                }
            } else if (hashCode64 == FnvHash.Constants.STR_TO_DATE) {
                SQLExpr item = targetList.get(0);
                for (SQLExpr target : targetList) {
                    if (target instanceof SQLLiteralExpr) {
                        SQLLiteralExpr literal = (SQLLiteralExpr) target;
                        handleNameLiteral(x, expr, literal);
                    }
                }
            }
        }

        return false;
    }

    public boolean visit(SQLSelectGroupByClause x) {
        final List<SQLExpr> items = x.getItems();
        for (int i = 0; i < items.size(); i++) {
            SQLExpr item = items.get(i);
            if (item != null) {
                if (item instanceof SQLCurrentTimeExpr) {
                    SQLCurrentTimeExpr expr = (SQLCurrentTimeExpr) item;
                    replaceInParent(expr, new SQLIdentifierExpr(expr.toString()));
                } else {
                    item.accept(this);
                }
            }
        }

        if(x.getHaving() != null) {
            x.getHaving().accept(this);
        }

        return false;
    }

    public boolean visit(SQLSelectOrderByItem x) {
        if (x.getExpr() instanceof SQLCurrentTimeExpr) {
            SQLCurrentTimeExpr expr = (SQLCurrentTimeExpr)x.getExpr();
            replaceInParent(expr, new SQLIdentifierExpr(expr.toString()));
        } else {
            x.getExpr().accept(this);
        }
        return false;
    }

    public boolean visit(SQLBetweenExpr x) {
        x.getTestExpr().accept(this);
        x.getBeginExpr().accept(this);
        x.getEndExpr().accept(this);

        SQLExpr testExpr = x.getTestExpr();
        SQLExpr beginExpr = x.getBeginExpr();
        SQLExpr endExpr = x.getEndExpr();

        SQLDataType testType = testExpr.computeDataType();
        SQLDataType beginType = beginExpr.computeDataType();
        SQLDataType endType = endExpr.computeDataType();

        boolean isIntType = (beginType != null && beginType.isInt())
                || (endType != null && endType.isInt());

        if (isIntType && testType != null && testType.isString()) {
            boolean cast = true;
            if (testExpr instanceof SQLCharExpr) {
                String text = ((SQLCharExpr) testExpr).getText();
                try {
                    long l = Long.parseLong(text);
                    testExpr = SQLIntegerExpr.ofIntOrLong(l);
                    x.setTestExpr(testExpr);
                    testType = testExpr.computeDataType();
                    cast = false;
                } catch (NumberFormatException ex) {
                    // skip
                }
            }
            if (cast) {
                SQLCastExpr castExpr = new SQLCastExpr(testExpr.clone(), new SQLDataTypeImpl("BIGINT"));
                testExpr = castExpr;
                x.setTestExpr(testExpr);
                testType = testExpr.computeDataType();
            }
        }

        SQLExpr expr = x.getTestExpr();
        if (expr instanceof SQLName) {
            SQLName name = (SQLName) expr;

            if (x.getBeginExpr() instanceof SQLLiteralExpr) {
                handleNameLiteral(x, name, (SQLLiteralExpr) x.getBeginExpr());
            }

            if (x.getEndExpr() instanceof SQLLiteralExpr) {
                handleNameLiteral(x, name, (SQLLiteralExpr) x.getEndExpr());
            }
        } else if (expr instanceof SQLValuableExpr
                && x.getBeginExpr() instanceof SQLValuableExpr
                && x.getEndExpr() instanceof SQLValuableExpr) {
            Object value = ((SQLValuableExpr) expr).getValue();
            Object begin = ((SQLValuableExpr) x.getBeginExpr()).getValue();
            Object end = ((SQLValuableExpr) x.getEndExpr()).getValue();

            if (value instanceof Integer
                    && begin instanceof Integer
                    && end instanceof Integer) {
                int valueInt = ((Integer) value).intValue();
                int beginInt = ((Integer) begin).intValue();
                int endInt = ((Integer) end).intValue();
                boolean result = valueInt >= beginInt && valueInt <= endInt;
                if (x.isNot()) {
                    result = !result;
                }

                if (replaceInParent(x, new SQLBooleanExpr(result))) {
                    optimizedCount++;
                }
            }

            if (value instanceof String
                    && begin instanceof String
                    && end instanceof String) {
                String valueStr = (String) value;
                String beginStr = (String) begin;
                String endStr = (String) end;

                if (valueStr.length() != ((String) begin).length()
                        || valueStr.length() != ((String) endStr).length()) {
                    return false;
                }

                boolean result = valueStr.compareTo(beginStr) >= 0
                        && valueStr.compareTo(endStr) <= 0;

                if (x.isNot()) {
                    result = !result;
                }

                if (replaceInParent(x, new SQLBooleanExpr(result))) {
                    optimizedCount++;
                    return false;
                }
            }
        } else if (expr.computeDataType() != null) {
            if (x.getBeginExpr() instanceof SQLLiteralExpr) {
                handleNameLiteral(x, expr, (SQLLiteralExpr) x.getBeginExpr());
            }

            if (x.getEndExpr() instanceof SQLLiteralExpr) {
                handleNameLiteral(x, expr, (SQLLiteralExpr) x.getEndExpr());
            }
        }

        if (x.getBeginExpr() instanceof SQLValuableExpr && x.getEndExpr() instanceof SQLValuableExpr) {
            Object beginVal = ((SQLValuableExpr) x.getBeginExpr()).getValue();
            Object endVal = ((SQLValuableExpr) x.getEndExpr()).getValue();

            if (beginVal != null && endVal != null && beginVal.getClass() == endVal.getClass() && beginVal instanceof Comparable && beginVal.getClass() != String.class) {
                int result = ((Comparable) beginVal).compareTo(endVal);
                if (result > 0) {
                    replaceInParent(x, new SQLBooleanExpr(false));
                }
            }
        }

        return false;
    }

    public boolean visit(SQLExistsExpr x) {
        x.getSubQuery().accept(this);

        final SQLSelect select = x.getSubQuery();

        final SQLSelectQueryBlock queryBlock = select.getQueryBlock();
        if (queryBlock != null && queryBlock.getWhere() instanceof SQLBooleanExpr) {
            boolean value = ((SQLBooleanExpr) queryBlock.getWhere()).getBooleanValue();
            if (x.isNot() && !value) {
                replaceInParent(x, new SQLBooleanExpr(true));
            } else if (!x.isNot() && !value) {
                replaceInParent(x, new SQLBooleanExpr(false));
            }
        }
        return false;
    }

    public static boolean isBothName(SQLExpr x) {
        return x instanceof SQLBinaryOpExpr
                && ((SQLBinaryOpExpr) x).isBothName();
    }

    public static boolean isLeftNameAndRightLiteral(SQLExpr x) {
        return x instanceof SQLBinaryOpExpr
                && ((SQLBinaryOpExpr) x).isLeftNameAndRightLiteral();
    }

    public static boolean isNameAndLiteral(SQLExpr x) {
        return x instanceof SQLBinaryOpExpr
                && ((SQLBinaryOpExpr) x).isNameAndLiteral();
    }

    @Override
    public boolean visit(SQLIntervalExpr x) {
        x.getValue().accept(this);

        SQLExpr value = x.getValue();
        if (value instanceof SQLNumberExpr) {
            Number num = ((SQLNumberExpr) value).getNumber();
            if (num instanceof BigDecimal) {
                BigDecimal decimal = (BigDecimal) num;
                if (decimal.toString().endsWith("0")) {
                    BigDecimal truncatedVal = decimal.setScale(0, RoundingMode.FLOOR);
                    x.setValue(new SQLNumberExpr(truncatedVal));
                    optimizedCount++;
                }
            }
        }

        return false;
    }
}
